﻿note
	description: "[
		Implementation of {STS_RATIONAL_NUMBER}.
		
		Notice that, as stated at {STS_RATIONAL_NUMBER} description, that
		specification (and this implementation)) does not try to offer a
		high-grade numeric class. Its intent is just to introduce examples of
		how several numeric types and sets may receive a set-theoretic
		treatment.
		
		Ideally, {STS_RATIONAL_NUMBER} could inherit from
		{STS_HOMOGENEOUS_ORDERED_PAIR [STS_INTEGER_NUMBER, ...]}. However, for
		the sake of regularity, in such a case {STS_COMPLEX_NUMBER} should
		inherit from {STS_HOMOGENEOUS_ORDERED_PAIR [STS_REAL_NUMBER, ...]}. But
		then {STS_RATIONAL_NUMBER} would inherit twice from
		{STS_HOMOGENEOUS_ORDERED_PAIR}, and such a double inheritance would be
		valid only if both {STS_HOMOGENEOUS_ORDERED_PAIR} ancestors had the
		same actual generic parameters, which would not be the case. Such a
		mess does not occur here, since {COMPLEX_NUMBER} is not an ancestor of
		{RATIONAL_NUMBER}, and each one of these classes can inherit from the
		appropriate derivation of {STS_HOMOGENEOUS_ORDERED_PAIR}. Inheriting
		from {HOMOGENEOUS_ORDERED_PAIR} is not possible because the (renamed)
		fields `a' and `b' would change their expansion status, which is a
		forbidden construct. The same observations apply to the hypothesis of
		inheriting from {STS_COMPOSABLE_ORDERED_PAIR} and
		{COMPOSABLE_ORDERED_PAIR}.
		]"
	author: "Rosivaldo Fernandes Alves"
	date: "$Date$"
	revision: "$Revision$"

expanded class
	RATIONAL_NUMBER

inherit
	STS_RATIONAL_NUMBER
		rename
			p as numerator,
			q as denominator
		redefine
			default_create,
			sign,
			machine_epsilon,
			ulp,
			relative_epsilon,
			is_real,
			is_proper_real,
			is_imaginary,
			is_proper_imaginary,
			is_proper_complex,
			is_nan,
			is_negative_infinity,
			is_negative_zero,
			is_positive_zero,
			is_positive_infinity,
			is_infinite,
			is_finite,
			is_integer,
			real_is_invertible,
			out,
			real_three_way_comparison,
			real_min,
			real_max,
			min,
			max,
			multipliable,
			modulus,
			abs,
			argument,
			phase,
			real_quotient,
			real_reciprocal,
			real_inverse,
			value_is_infinite,
			value_is_finite,
			gcd,
			div,
			rem,
			converted_integer,
			integer_product_overflows
		end

	STS_HOMOGENEOUS_ORDERED_PAIR [STS_INTEGER_NUMBER, STS_INTEGER_EQUALITY]
		rename
			a as numerator,
			first as numerator,
			b as denominator,
			second as denominator,
			is_in as pair_is_in,
			is_not_in as pair_is_not_in,
			equals as pair_equals,
			unequals as pair_unequals,
			inverse as pair_inverse,
			composed_with_inverse_like as composable_composed
		redefine
			default_create,
			out
		select
			pair_inverse,
			composed
		end

	STS_COMPOSABLE_ORDERED_PAIR [
		STS_INTEGER_NUMBER, STS_INTEGER_NUMBER, STS_INTEGER_NUMBER,
		STS_INTEGER_EQUALITY, STS_INTEGER_EQUALITY, STS_INTEGER_EQUALITY
		]
		rename
			a as numerator,
			first as numerator,
			b as denominator,
			second as denominator,
			eq_w as Eq,
			eq_a as Eq,
			eq_b as Eq,
			is_in as pair_is_in,
			is_not_in as pair_is_not_in,
			equals as pair_equals,
			unequals as pair_unequals,
			inverse as composable_pair_inverse,
			composed as composable_composed,
			composed_with_homogeneous as composable_composed,
			composed_with_inverse_like as composable_composed
		redefine
			default_create,
			out
		end

	DEBUG_OUTPUT
		rename
			debug_output as out
		redefine
			default_create,
			out
		end

create
	default_create,
	make,
	make_from_expanded,
	make_from_rational,
	make_from_native_integer,
	make_from_native_integer_ref,
	make_from_integer,
	make_from_native_natural,
	make_from_native_natural_ref

convert
	make_from_rational ({STS_RATIONAL_NUMBER}),
	make_from_native_integer ({INTEGER}),
	make_from_native_integer_ref ({INTEGER_REF}),
	make_from_integer ({STS_INTEGER_NUMBER}),
	make_from_native_natural ({NATURAL}),
	make_from_native_natural_ref ({NATURAL_32_REF}),
	value: {REAL, REAL_REF, NUMERIC, COMPARABLE, HASHABLE},
	as_real: {REAL_NUMBER},
	as_complex: {COMPLEX_NUMBER},
	as_ordered_pair: {
		HOMOGENEOUS_ORDERED_PAIR [STS_INTEGER_NUMBER, STS_INTEGER_EQUALITY],
		HOMOGENEOUS_ORDERED_PAIR
			[STS_INTEGER_NUMBER, STS_NUMERIC_EQUALITY [STS_INTEGER_NUMBER]],

		ORDERED_PAIR [
			STS_INTEGER_NUMBER, STS_INTEGER_NUMBER,
			STS_INTEGER_EQUALITY, STS_INTEGER_EQUALITY
			],
		ORDERED_PAIR [
			STS_INTEGER_NUMBER, STS_INTEGER_NUMBER,
			STS_NUMERIC_EQUALITY [STS_INTEGER_NUMBER],
			STS_NUMERIC_EQUALITY [STS_INTEGER_NUMBER]
			]
		},
	as_tuple: {
		N_TUPLE [STS_INTEGER_NUMBER, STS_INTEGER_EQUALITY],
		N_TUPLE
			[STS_INTEGER_NUMBER, STS_NUMERIC_EQUALITY [STS_INTEGER_NUMBER]],

		STS_N_TUPLE [STS_INTEGER_NUMBER, STS_INTEGER_EQUALITY],
		STS_N_TUPLE
			[STS_INTEGER_NUMBER, STS_NUMERIC_EQUALITY [STS_INTEGER_NUMBER]]
		}

feature {NONE} -- Initialization

	default_create
			-- Create the rational number 0/1.
		do
			p := {INTEGER_NUMBER}.Zero
			q := One.p
		ensure then
			numerator: p ≍ Zero
			denominator: q ≍ One
		end

	make (num, den: STS_INTEGER_NUMBER)
			-- Create a rational number whose value `p'/`q' equals `num'/`den'
		require
			non_zero_denominator: den ≭ Zero
		do
			p := converted_integer (num)
			q := converted_integer (den)
				check
					non_zero_denominator: q ≭ Zero -- q ≍ den ≭ Zero
				end
			make_from_expanded (p, q)
		ensure
			numerator: p ≍ (num / den).p
			denominator: q ≍ (num / den).q
		end

	make_from_expanded (num, den: INTEGER_NUMBER)
			-- Create the rational number `num'/`den'.
			-- Useful to create rational numbers out of manifest constants.
		require
			non_zero_denominator: den ≭ Zero
		local
			l_gcd: like gcd
		do
			l_gcd := gcd (num, den)
				check
					non_zero_gcd: l_gcd ≭ Zero -- den /= 0
					num_good_divisor: num.divisible (l_gcd) -- l_gcd ≭ Zero
					den_good_divisor: den.divisible (l_gcd) -- l_gcd ≭ Zero
				end
			p := num // l_gcd
			q := den // l_gcd
			if
				q < Zero and
				p ≭ {INTEGER_NUMBER}.Min_value and
				q ≭ {INTEGER_NUMBER}.Min_value
			then
				p := - p
				q := - q
			end
		ensure
			numerator: p ≍ (num / den).p
			denominator: q ≍ (num / den).q
		end

	make_from_rational (pq: STS_RATIONAL_NUMBER)
			-- Create a rational number whose value `p'/`q' equals `pq'.`p'/`pq'.`q'
		do
			if attached {RATIONAL_NUMBER} pq as exp_pq then
					check
						non_zero_denominator: exp_pq.q ≭ Zero -- Class invariant
					end
				make_from_expanded (exp_pq.p, exp_pq.q)
			else
					check
						non_zero_denominator: -- {STS_RATIONAL_NUMBER} invariant
							pq.q ≭ Zero
					end
				make (pq.p, pq.q)
			end
		ensure
			numerator: p ≍ (pq.p / pq.q).p
			denominator: q ≍ (pq.p / pq.q).q
		end

	make_from_native_integer (v: like native_integer_anchor)
			-- Create a rational number out of value `v'.
		do
			create p.make (v)
			q := One.p
		ensure
			numerator: p.value = {INTEGER_NUMBER}.adjusted_value (v)
			denominator: q ≍ One
		end

	make_from_native_integer_ref (v_ref: like native_integer_ref_anchor)
			-- Create a rational number out of value `v_ref'.`item'.
		do
			make_from_native_integer (v_ref.item)
		ensure
			numerator: p.value = {INTEGER_NUMBER}.adjusted_value (v_ref.item)
			denominator: q ≍ One
		end

	make_from_integer (i: STS_INTEGER_NUMBER)
			-- Create a rational number with `numerator'.`value' = -- {INTEGER_NUMBER}.adjusted_value (`i'.`value'.) and `denominator' = 1
		do
			p := converted_integer (i)
			q := One.p
		ensure
			numerator: p.value = {INTEGER_NUMBER}.adjusted_value (i.value)
			imaginary: q ≍ One
		end

	make_from_native_natural (v: like native_natural_anchor)
			-- Create a rational number out of value `v'.
		do
			create p.make_from_native_natural (v)
			q := One.p
		ensure
			numerator: p.value =
				{INTEGER_NUMBER}.adjusted_value (v.as_integer_32)
			denominator: q ≍ One
		end

	make_from_native_natural_ref (v_ref: like native_natural_ref_anchor)
			-- Create a rational number out of value `v_ref'.`as_integer_32'.
		do
			make_from_native_natural (v_ref.item)
		ensure
			numerator: p.value =
				{INTEGER_NUMBER}.adjusted_value (v_ref.as_integer_32)
			denominator: q ≍ One
		end

		--	make_from_natural (n: STS_NATURAL)
		--			-- Create an integer number with value `n'.`integer_value'.
		--		do
		--			make_from_integer (n)
		--		ensure
		--			value: value = adjusted_value (n.integer_value)
		--		end

feature -- Primitive

	p: like Integer_anchor
			-- Numerator of current rational number

	numerator: like Integer_anchor
			-- <Precursor>
		do
			Result := p
		ensure then
			definition: Result ≍ p
		end

	q: like Integer_anchor
			-- Denominator of current rational number

	denominator: like Integer_anchor
			-- <Precursor>
		do
			Result := q
		ensure then
			definition: Result ≍ q
		end

	eq: STS_INTEGER_EQUALITY
			-- Equality for integer numbers
		once
			Result := {INTEGER_SET}.eq
		ensure then
			class
		end

	value: like native_real_anchor
			-- <Precursor>
			-- NOTICE: This implementation allows two different rational numbers
			-- to have the same `value', due to lack of precision of
			-- (floating-point) real numbers implementation.
		do
				check
					good_divisor: p.real_divisible (q) -- Class invariant: q /= 0
				end
			Result := p.real_quotient (q).value
		end

feature -- Access

	sign: like integer_anchor
			-- <Precursor>
		do
			Result := p.sign ⋅ q.sign -- TODO: Take native sign bit instead?
		end

	sign_bit: like integer_anchor -- TODO: Natural instead.
			-- <Precursor>
		do
				check
						-- value_sign_bit (value).as_integer_8 ∈ {0, 1}
					big_enough: {INTEGER_NUMBER}.Native_min_value ≤ value_sign_bit (value).as_integer_8
					small_enough: value_sign_bit (value).as_integer_8 ≤ {INTEGER_NUMBER}.Native_max_value
				end
			create Result.make_from_stored_value (value_sign_bit (value).as_integer_8)
		end

	zero: RATIONAL_NUMBER
			-- <Precursor>
		once
		ensure then
			class
		end

	one: RATIONAL_NUMBER
			-- <Precursor>
		once
			create Result.make_from_expanded ({INTEGER_NUMBER}.One, {INTEGER_NUMBER}.One)
		ensure then
			class
		end

	two: INTEGER_NUMBER
			-- <Precursor>
		once
			create Result.make (2)
		ensure then
			class
		end

	imaginary_unit: COMPLEX_NUMBER
			-- <Precursor>
		once
			Result := {COMPLEX_NUMBER}.I
		ensure then
			class
		end

	Machine_epsilon: REAL_NUMBER
			-- <Precursor>
		once
			Result := {REAL_NUMBER}.machine_epsilon
		ensure then
			class
		end

	ulp,
	relative_epsilon: like real_anchor
			-- <Precursor>
		do
			Result := as_real.ulp
		end

	previous_float: like real_anchor
			-- <Precursor>
		do
			Result := as_real.previous_float
		end

	next_float: like real_anchor
			-- <Precursor>
		do
			Result := as_real.next_float
		end

	Min_value: RATIONAL_NUMBER
			-- Minimum value representable by this implementation of rational
			-- numbers
		once
			create Result.make_from_expanded ({INTEGER_NUMBER}.Min_value, One.p)
		ensure
			class
			numerator: Result.p ≍ {INTEGER_NUMBER}.Min_value
			denominator: Result.q ≍ One
		end

	Max_value: RATIONAL_NUMBER
			-- Maximum value representable by this implementation of rational
			-- numbers
		once
			create Result.make_from_expanded ({INTEGER_NUMBER}.Min_value, - One.p)
		ensure
			class
			numerator: Result.p ≍ {INTEGER_NUMBER}.Min_value
			denominator: Result.q ≍ -One
		end

feature -- Quality

	is_real: BOOLEAN = True
			-- <Precursor>

	is_proper_real: BOOLEAN
			-- <Precursor>
		do
			Result := p ≭ Zero
		end

	is_imaginary: BOOLEAN
			-- <Precursor>
		do
			Result := p ≍ Zero
		end

	is_proper_imaginary: BOOLEAN = False
			-- <Precursor>

	is_proper_complex: BOOLEAN = False
			-- <Precursor>

	is_nan: BOOLEAN = False
			-- <Precursor>
			-- A two's-complement representation of integer numbers, like that
			-- used on `p' and `q', has no NaN; besides, current class invariant
			-- ensures that no rational representation is a NaN.

	is_negative_infinity: BOOLEAN = False
			-- <Precursor>
			-- A two's-complement representation of integer numbers, like that
			-- used on `p' and `q', has no infinity; besides, current class
			-- invariant ensures that no rational representation is an infinity.

	is_negative_zero: BOOLEAN = False
			-- <Precursor>
			-- A two's-complement representation of integer numbers, like that used on `p' and `q', has no negative zero; besides, current class invariant ensures
			-- that no rational representation is a negative zero.

	is_positive_zero: BOOLEAN
			-- <Precursor>
		do
			Result := p ≍ Zero
		end

	is_positive_infinity: BOOLEAN = False
			-- <Precursor>
			-- A two's-complement representation of integer numbers, like that
			-- used on `p' and `q', has no infinity; besides, current class
			-- invariant ensures that no rational representation is an
			-- infinity.

	is_infinite: BOOLEAN = False
			-- <Precursor>
			-- A two's-complement representation of integer numbers, like that
			-- used on `p' and `q', has no infinity; besides, current class
			-- invariant ensures that no rational representation is an
			-- infinity.

	is_finite: BOOLEAN = True
			-- <Precursor>
			-- A two's-complement representation of integer numbers, like that
			-- used on `p' and `q', has no infinity; besides, current class
			-- invariant ensures that no rational representation is an
			-- infinity.

	is_rational: BOOLEAN = True
			-- <Precursor>

	is_integer: BOOLEAN
			-- <Precursor>
		do
			Result := q.abs ≍ One
		end

	real_is_invertible: BOOLEAN = True
			-- <Precursor>
			-- An implementation akin to IEEE 754 inverts zero with no major
			-- consequence.

feature -- Output

	out: STRING
			-- <Precursor>
		do
			Result := p.out
			Result.append_character ('/')
				check
					argument_not_void:
						q.out /= Void -- {INTEGER_NUMBER}.out definition
				end
			Result.append (q.out)
		ensure then
			definition: Result ~ p.out + "/" + q.out
		end

feature -- Comparison

	real_three_way_comparison (x: STS_REAL_NUMBER): like Integer_anchor
			-- <Precursor>
		local
			xv, v: like Real_anchor.value
		do
			v := value
			xv := x.value
			if v < xv then
				Result := - One.p
			elseif xv < v then
				Result := One.p
			end
		end

	real_min (x: STS_REAL_NUMBER): like real_anchor
			-- <Precursor>
		do
			if real_is_less_equal (x) then
				create Result.make_from_real (Current) -- TODO: make_from_rational?
			else
--				Result := x -- TODO: Result := create {REAL_NUMBER}.make_from_real (x) instead of create Result.make_from_real (x)
				create Result.make_from_real (x)
			end
		end

	real_max (x: STS_REAL_NUMBER): like real_anchor
			-- <Precursor>
		do
			if real_is_greater_equal (x) then
				create Result.make_from_real (Current) -- TODO: make_from_rational?
			else
				create Result.make_from_real (x)
			end
		end

	min alias "∧" (pq: STS_RATIONAL_NUMBER): like Rational_anchor
			-- <Precursor>
		do
			if Current ≤ pq then
				Result := Current
			else
				create Result.make_from_rational (pq)
			end
		end

	max alias "∨" (pq: STS_RATIONAL_NUMBER): like Rational_anchor
			-- <Precursor>
		do
			if Current ≥ pq then
				Result := Current
			else
				create Result.make_from_rational (pq)
			end
		end

feature -- Relationship

	multipliable (other: STS_RATIONAL_NUMBER): BOOLEAN
			-- <Precursor>
			--| NOTICE: This refinement assumes that a product overflow that
			--| occasionly unequals zero will *consistently* produce a non-zero
			--| result.
		local
			l_gcd_1, l_gcd_2: like gcd
		do
			l_gcd_1 := gcd (other.p, q)
			l_gcd_2 := gcd (p, other.q)
				check
					good_divisor_1:  -- l_gcd_1 /= 0 ⇐ q /= 0
						q.divisible (l_gcd_1)
					good_divisor_2: -- l_gcd_2 /= 0 ⇐ other.q /= 0
						other.q.divisible (l_gcd_2)
				end
			Result := (q // l_gcd_1) ⋅ (other.q // l_gcd_2) ≭ Zero
		ensure then
			gcd_1: attached gcd (other.p, q) as gcd_1
			gcd_2: attached gcd (p, other.q) as gcd_2
			good_divisor_1: q.divisible (gcd_1)
			good_divisor_2: other.q.divisible (gcd_2)
			definition: Result = (q // gcd_1) ⋅ (other.q // gcd_2) ≭ Zero
		end

feature -- Operation

	pair_inverse:
		HOMOGENEOUS_ORDERED_PAIR [STS_INTEGER_NUMBER, STS_INTEGER_EQUALITY]
			-- The components of current rational number in inverted order, i.e.
			-- (`q`, `p`).
		do
			create Result.make (q, p)
		end

	composable_pair_inverse: COMPOSABLE_ORDERED_PAIR [
		STS_INTEGER_NUMBER, STS_INTEGER_NUMBER, STS_INTEGER_NUMBER,
		STS_INTEGER_EQUALITY, STS_INTEGER_EQUALITY, STS_INTEGER_EQUALITY
		]
			-- The components of current rational number in inverted order, i.e.
			-- (`q`, `p`).
		do
			create Result.make (q, p)
		end

	composed alias "∘" (
		ij: STS_ORDERED_PAIR [
			STS_INTEGER_NUMBER, STS_INTEGER_NUMBER,
			STS_INTEGER_EQUALITY, STS_INTEGER_EQUALITY
			]):
		HOMOGENEOUS_ORDERED_PAIR [STS_INTEGER_NUMBER, STS_INTEGER_EQUALITY]
			-- Composition of (`p', `q') and `ij`, i.e. the pair
			-- (`ij`.`first`, `q`).
		do
			create Result.make (ij.a, q)
		end

	composable_composed (
		ij: STS_ORDERED_PAIR [
			STS_INTEGER_NUMBER, STS_INTEGER_NUMBER,
			STS_INTEGER_EQUALITY, STS_INTEGER_EQUALITY
			]
		): COMPOSABLE_ORDERED_PAIR [
			STS_INTEGER_NUMBER, STS_INTEGER_NUMBER, STS_INTEGER_NUMBER,
			STS_INTEGER_EQUALITY, STS_INTEGER_EQUALITY, STS_INTEGER_EQUALITY
			]
			-- Composition of (`p', `q') and `ij`, i.e. the pair
			-- (`ij`.`first`, `q`).
		do
			create Result.make (ij.a, q)
		end

	argument,
	phase: like real_anchor
			-- <Precursor>
		do
			if sign_bit ≍ One then
				create Result.make (Pi)
			end
		end

	conjugate: like complex_anchor
			-- <Precursor>
		do
			Result := as_complex.conjugate
		end

	complex_plus (z: STS_COMPLEX_NUMBER): like complex_anchor
			-- <Precursor>
		do
			Result := as_complex + z
		end

	complex_minus (z: STS_COMPLEX_NUMBER): like complex_anchor
			-- <Precursor>
		do
			Result := as_complex - z
		end

	complex_product (z: STS_COMPLEX_NUMBER): like complex_anchor
			-- <Precursor>
		do
			Result := as_complex ⋅ z
		end

	complex_quotient (z: STS_COMPLEX_NUMBER): like complex_anchor
			-- <Precursor>
		do
				check
					good_divisor: as_complex.divisible (z) -- divisible (z)
				end
			Result := as_complex / z
		end

	real_quotient (x: STS_REAL_NUMBER): like Real_anchor
			-- <Precursor>
		do
				check
					ignore_precondition: value.divisible (x.value) implies True
				end
			create Result.make (value / x.value)
		ensure then
			definition: Result.value = adjusted_value (value / x.value)
		end

	real_reciprocal,
	real_inverse: like Real_anchor
			-- <Precursor>
		require else
			always_invertible: True
		do
			Result := as_real.reciprocal
		end

	modulus,
	abs: like rational_anchor
			-- <Precursor>
		do
			if p.sign ⋅ q.sign ≥ Zero then
				Result := Current
			elseif p ≍ {INTEGER_NUMBER}.Min_value then
					check
							-- Class invariant: q /= 0
						good_divisor: p.divisible (- q)
					end
				Result := p / -q
			elseif q ≍ {INTEGER_NUMBER}.Min_value then
					check
							-- q ≍ {INTEGER_NUMBER}.Min_value /= 0
						good_divisor: (- p).divisible (q)
					end
				Result := -p / q
			else
					check
							-- Class invariant: q /= 0
						good_divisor: (- p).divisible (q)
					end
				Result := -p / q
			end
		end

	opposite alias "-" alias "−": like rational_anchor
			-- <Precursor>
		do

			if q > Zero then
				if p ≭ {INTEGER_NUMBER}.Min_value then
						check
								-- p ≭ {INTEGER_NUMBER}.Min_value
							(- p).sign ≍ - p.sign
							minus_p_divisible_q: (- p).divisible (q) -- q > 0
						end
					create Result.make_from_expanded (- p, q)
				else
						check
								-- q ≭ {INTEGER_NUMBER}.Min_value ⇐ q > 0
							(- q).sign ≍ - q.sign
							p_divisible_minus_q: -- -q < 0 ⇐ q > 0
								p.divisible (- q)
						end
					create Result.make_from_expanded (p, - q)
				end
			else
				if p ≭ {INTEGER_NUMBER}.Min_value then
						check
								-- p ≭ {INTEGER_NUMBER}.Min_value
							(- p).sign ≍ - p.sign
							minus_p_divisible_q: -- Class invariant: q /= 0
								(- p).divisible (q)
						end
					create Result.make_from_expanded (- p, q)
				else
						check
								-- q ≭ {INTEGER_NUMBER}.Min_value ⇐
								-- 	p ≍ {INTEGER_NUMBER}.Min_value
							(- q).sign ≍ - q.sign

								-- -q /= 0 ⇐ Class invariant: q /= 0
							p_divisible_minus_q: p.divisible (- q)
						end
					create Result.make_from_expanded (p, - q)
				end
			end
		end

feature -- Conversion

	as_singleton: SET_OF_ORDERED_PAIRS [
		STS_INTEGER_NUMBER, STS_INTEGER_NUMBER,
		STS_INTEGER_EQUALITY, STS_INTEGER_EQUALITY
		]
			-- The ordered pair (`p', `q') alone into a singleton, i.e.
			-- {(`p', `q')}.
		do
--			create Result.make_singleton (Current) -- TODO: System crash! {SET_OF_ORDERED_PAIRS}.converted_pair fails on noting that `Current' is not an {ORDERED_PAIR}.
			create Result.make_singleton (as_ordered_pair)
		end

	as_tuple: N_TUPLE [STS_INTEGER_NUMBER, STS_INTEGER_EQUALITY]
			-- <Precursor>
		do
			Result := {INTEGER_UNIVERSE}.Unit.after (q).after (p)
		end

	as_ordered_pair:
		HOMOGENEOUS_ORDERED_PAIR [STS_INTEGER_NUMBER, STS_INTEGER_EQUALITY]
			-- Current rational number converted into an effective ordered pair
			-- of integer numbers
		do
			create Result.make (p, q)
		ensure
			first: Result.a ≍ p
			second: Result.b ≍ q
		end

	as_complex: like complex_anchor
			-- Current rational number represented as a complex number
		do
			create Result.make_from_real (Current) -- TODO: Change it?
		ensure
			real: Result.a ≍ Current
			imaginary: Result.b ≍ Zero
		end

	as_real: like real_anchor
			-- Current rational number represented as a real number
		do
				check
					good_divisor: q.real_divisible (q) -- `q' /= 0
				end
			Result := p.real_quotient (q)
		end

	truncated_to_integer: like integer_anchor
			-- <Precursor>
		do
				check
					good_divisor: p.divisible (q) -- Class invariant: q /= 0
				end
			Result := p // q
		end

feature -- Math

	value_is_infinite (v: like value): BOOLEAN
			-- <Precursor>
		do
			Result := {REAL_NUMBER}.value_is_infinite (v)
		end

	value_is_finite (v: like value): BOOLEAN
			-- <Precursor>
		do
			Result := {REAL_NUMBER}.value_is_finite (v)
		end

	hypotenuse (x, y: STS_REAL_NUMBER): like real_anchor
			-- <Precursor>
			--| TODO: DRY.
		do
			Result := {COMPLEX_NUMBER}.hypotenuse (x, y)
		ensure then
			class
		end

	arc_tangent_yx (y, x: STS_REAL_NUMBER): like real_anchor
			-- <Precursor>
		do
			Result := {COMPLEX_NUMBER}.arc_tangent_yx (y, x)
		ensure then
			class
		end

	splitted (a_q: STS_REAL_NUMBER): HOMOGENEOUS_ORDERED_PAIR [STS_REAL_NUMBER, STS_REAL_EQUALITY]
			-- <Precursor>
		do
			Result := as_real.splitted (a_q)
		end

	gcd (i, j: STS_INTEGER_NUMBER): like Integer_anchor
			-- <Precursor>
		do
			if j ≍ Zero then
				create Result.make_from_integer (i.abs)
			else
					check
						good_divisor: i.divisible (j) -- j /= 0
					end
				Result := gcd (j, i \\ j)
			end
		ensure then
			class
		end

	div (i, j: STS_INTEGER_NUMBER): like Integer_anchor
			-- <Precursor>
		do
			if (i \\ j) ≍ zero then
				create Result.make_from_integer (i // j)
			elseif i < zero then
				if j < zero then
					create Result.make_from_integer (i // j)
				else
						check
							j > zero -- i.divisible (j)
						end
					create Result.make_from_integer (i // j - one.p)
				end
			else
					check
						i > zero -- (i \\ j) /= 0
					end
				if j < zero then
					create Result.make_from_integer (i // j - one.p)
				else
						check
							j > zero -- i.divisible (j)
						end
					create Result.make_from_integer (i // j)
				end
			end
		ensure then
			class
		end

	rem (i, j: STS_INTEGER_NUMBER): like Integer_anchor
			-- <Precursor>
		local
			i_rem_j: like Integer_anchor
		do
			create i_rem_j.make_from_integer (i \\ j)
			if i_rem_j ≍ zero then
				Result := i_rem_j
			elseif i < zero then
				if j < zero then
					Result := i_rem_j
				else
						check
							j > zero -- i.divisible (j)
						end
					Result := i_rem_j + j
				end
			else
					check
						i > zero -- i_rem_j /= 0
					end
				if j < zero then
					Result := i_rem_j + j
				else
						check
							j > zero -- i.divisible (j)
						end
					Result := i_rem_j
				end
			end
		ensure then
			class
		end

feature -- Factory

	real_from_value (v: like real_anchor.value): like real_anchor
			-- <Precursor>
		do
			create Result.make (v)
		ensure then
			class
		end

	real_from_double_value
		(dv: like real_anchor.value.to_double): like real_anchor
			-- <Precursor>
		do
			create Result.make_from_native_double_real (dv)
		ensure then
			class
		end

	converted_integer (i: STS_INTEGER_NUMBER): like integer_anchor
			-- <Precursor>
		do
			if attached {like integer_anchor} i as exp_i then
				Result := exp_i
			else
				create Result.make_from_integer (i)
			end
		ensure then
			class
		end

feature -- Predicate

	integer_product_overflows (i, j: STS_INTEGER_NUMBER): BOOLEAN
			-- <Precursor>
		do
			Result := i < zero and j < zero and i.max_value_exists and then
				i < i.max_value // j or
				i < zero and j > zero and i.min_value_exists and then
				i < i.min_value // j or
				i > zero and j < - one and i.min_value_exists and then
				i > i.min_value // j or
				i > zero and j > zero and i.max_value_exists and then
				i > i.max_value // j
		ensure then
			class
		end

feature -- Implementation

	adjusted_value (v: like value): like value
			-- <Precursor>
		do
			Result := {REAL_NUMBER}.adjusted_value (v)
		ensure then
			class
			definition: Result = {REAL_NUMBER}.adjusted_value (v)
		end

feature -- Anchor

	complex_anchor: COMPLEX_NUMBER
			-- <Precursor>
		once
		ensure then
			class
		end

	real_anchor: REAL_NUMBER
			-- <Precursor>
		once
		ensure then
			class
		end

	rational_anchor: RATIONAL_NUMBER
			-- <Precursor>
		once
		ensure then
			class
		end

	integer_anchor: INTEGER_NUMBER
			-- <Precursor>
		once
		ensure then
			class
		end

	native_integer_anchor: INTEGER
			-- Anchor for a native integer number
		once
		ensure
			class
		end

	native_integer_ref_anchor: like native_integer_anchor.to_reference
			-- Anchor for a reference to a native integer number
		do
			create Result
		ensure
			class
		end

	native_natural_anchor: NATURAL
			-- Anchor for a native natural number
		once
		ensure
			class
		end

	native_natural_ref_anchor: like native_natural_anchor.to_reference
			-- Anchor for a reference to a native natural number
		do
			create Result
		ensure
			class
		end

invariant
	irreducible: gcd (p, q) ≍ One
	zero_real: Zero.identity.value = 0
	zero_imaginary: Zero.Zero.value = 0
	one_real: One.identity.value = 1
	one_imaginary: One.Zero.value = 0
	Two: Two.value = 2
	Imaginary_unit_real: Imaginary_unit.real ≍ Zero.identity
	Imaginary_unit_imaginary: Imaginary_unit.imaginary ≍ One.identity
	Machine_epsilon: Machine_epsilon ≍ One.relative_epsilon
	Machine_epsilon: Machine_epsilon = {REAL_NUMBER}.machine_epsilon
	Min_value_numerator: Min_value.p ≍ {INTEGER_NUMBER}.min_value
	Min_value_denominator: Min_value.q ≍ One
	Max_value_numerator: Max_value.p ≍ {INTEGER_NUMBER}.min_value
	Max_value_denominator: Max_value.q ≍ - One
	is_real_definition: is_real = zero.real_equals (zero)
	is_proper_imaginary_definition: is_proper_imaginary = (is_imaginary and zero.real_unequals (zero))
	is_proper_complex_definition: is_proper_complex = (not is_real and not is_imaginary)
	is_nan_definition: is_nan = value.is_nan
	total_order: is_nan implies Current.real_equals (Current)
	is_negative_infinity_definition: is_negative_infinity = value.is_negative_infinity
	is_negative_zero_definition: is_negative_zero = (Current.real_equals (zero) and sign_bit ≍ one)
	is_positive_infinity_definition: is_positive_infinity = value.is_positive_infinity
	is_infinite_definition: is_infinite = (identity.is_infinite or zero.is_infinite)
	is_infinite_real_definition: is_infinite = (is_negative_infinity or is_positive_infinity)
	is_finite_definition: is_finite = (identity.is_finite and zero.is_finite)
	is_finite_real_definition: is_finite = not (is_nan or is_infinite)

		-- is_rational
	is_rational_definition: is_rational = (is_real and identity.is_rational)
	splitted_one: attached splitted (one) as splitted_one
	a: attached splitted_one.a as a
	b: attached splitted_one.b as b
	is_rational_real_definition: is_rational = (
		(a.is_integer or zero.p.max_value_exists and then (a - one) ≍ zero.max_value) and
		(b.is_integer or zero.p.max_value_exists and then (b - one) ≍ zero.max_value)
		)

		-- real_is_invertible
	scale_log: attached value_logb ((identity.abs.real_max (zero.abs)).value) as scale_log
	int_scale_log: attached if value_is_finite (scale_log) then
			scale_log.truncated_to_integer
		else
			0
		end as int_scale_log
	scaled_a: attached (identity.value * 2 ^ (- int_scale_log)) as scaled_a
	scaled_b: attached (zero.value * 2 ^ (- int_scale_log)) as scaled_b
	denom: attached (scaled_a ^ 2 + scaled_b ^ 2) as denom
	real_is_invertible_quasi_definition: denom /= 0 implies real_is_invertible
	real_is_invertible_real_quasi_definition: Current.real_unequals (zero) implies real_is_invertible

note
	copyright: "Copyright (c) 2012-2022, Rosivaldo Fernandes Alves"
	license: "[
		Eiffel Forum License v2
		(see http://www.eiffel.com/licensing/forum.txt)
		]"
	source: "https://launchpad.net/set-theory"
end
