﻿note
	description: "[
		Implementation of {STS_REAL_NUMBER}
		
		This implementation treats real numbers as if they were encoded by a
		6-bit, base 2, IEEE 754-like, floating point number with the following
		parameters:
			sign bit: 1 bit;
			exponent width: 3 bits;
			significand (mantissa) precision: 3 bits (2 explicitly stored).
		Conceptual adapted value layout: s_eee_mm
		
		I'd like to thank Jeroen van der Zijp, whose paper
		"Fast Half Float Conversions" 
		(http://www.fox-toolkit.org/ftp/fasthalffloatconversion.pdf) gave me a
		clear understanding on how to represent, store and convert different
		sizes of IEEE 754-like binary floating-point numbers.
		
		Notice that, as stated at {STS_REAL_NUMBER} description, that
		specification (and this implementation)) does not try to offer a
		high-grade numeric class.
		Its intent is just to introduce examples of how several numeric types
		and sets may receive a set-theoretic treatment.
		]"
	author: "Rosivaldo Fernandes Alves"
	date: "$Date$"
	revision: "$Revision$"

expanded class
	REAL_NUMBER

inherit
	STS_REAL_NUMBER
		rename
			i as imaginary_unit
		redefine
			default_create,
			sign,
			Machine_epsilon,
			ulp, relative_epsilon,
			is_real,
			is_proper_real,
			is_imaginary,
			is_proper_imaginary,
			is_proper_complex,
			is_nan,
			is_negative_infinity,
			is_negative_zero,
			is_positive_zero,
			is_positive_infinity,
			is_finite,
			is_infinite,
			Is_invertible,
			out,
			equals,
			is_less,
			is_less_equal,
			three_way_comparison,
			min,
			max,
			modulus, abs,
			argument, phase,
			opposite,
			quotient,
			reciprocal, inverse,
			value_is_infinite,
			value_is_finite
		end

	DEBUG_OUTPUT
		rename
			debug_output as out
		redefine
			default_create,
			out
		end

create
	default_create,
	make,
	make_from_native_real_ref,
	make_from_native_double_real,
	make_from_native_double_real_ref,
	make_from_real,
	make_from_native_integer,
	make_from_native_integer_ref,
	make_from_integer,
	make_from_native_natural,
	make_from_native_natural_ref,
	make_abs,
	make_opposite

create {REAL_NUMBER, REAL_UNIVERSE}
	make_from_bit_pattern

convert
	make ({REAL}),
	make_from_native_real_ref ({REAL_REF}),
	make_from_native_double_real ({DOUBLE}),
	make_from_native_double_real_ref ({DOUBLE_REF}),
	make_from_real ({STS_REAL_NUMBER}),
	make_from_native_integer ({INTEGER}),
	make_from_native_integer_ref ({INTEGER_REF}),
	make_from_integer ({STS_INTEGER_NUMBER}),
	make_from_native_natural ({NATURAL}),
	make_from_native_natural_ref ({NATURAL_32_REF}),
	value: {REAL, REAL_REF, NUMERIC, COMPARABLE, HASHABLE},
	as_complex: {COMPLEX_NUMBER}

feature {NONE} -- Initialization

	default_create
			-- Create a real number with `value' = 0
		do
			value_bit_pattern := 0
		ensure then
			value: value = 0
		end

	make (v: like value)
			-- Create a real number with `value' = `adjusted_value' (`v').
		do
			value_bit_pattern := bit_pattern_from_native_real (v)
		ensure
			adjusted_value: value = adjusted_value (v)
		end

	make_from_bit_pattern (bp: like value_bit_pattern)
			-- Create a real number with `value' determined by `bp'
		require
			no_extra_bit: bp & (⊝ Value_bit_pattern_mask) = 0
		do
			value_bit_pattern := bp
		ensure
			value_bit_pattern: value_bit_pattern = bp
		end

	make_from_native_real_ref (v_ref: like native_real_ref_anchor)
			-- Create a real number with `value' = `adjusted_value' (`v_ref'.`item').
		do
			make (v_ref.item)
		ensure
			adjusted_value: value = adjusted_value (v_ref.item)
		end

	make_from_native_double_real (v: like native_double_real_anchor)
			-- Create a real number with `value' = `adjusted_value' (`v'.`truncated_to_real').
			--| TODO: Get rid of truncated_to_real.
		do
			make (v.truncated_to_real)
		ensure
			adjusted_value: value = adjusted_value (v.truncated_to_real)
		end

	make_from_native_double_real_ref (v_ref: like native_double_real_ref_anchor)
			-- Create a real number with `value' = `adjusted_value' (`v_ref'.`truncated_to_real').
			--| TODO: Get rid of truncated_to_real.
		do
			make_from_native_double_real (v_ref.item)
		ensure
			adjusted_value: value = adjusted_value (v_ref.truncated_to_real)
		end

	make_from_real (x: STS_REAL_NUMBER)
			-- Create a real number with `value' = `adjusted_value' (`x'.`value').
			--| Notice that `x' might not conform to `Current', so `x'.`value' might not be kept intact after adjusted for current real type.
		do
			if attached {REAL_NUMBER} x as exp_x then
				value_bit_pattern := exp_x.value_bit_pattern
			else
				make (x.value)
			end
		ensure
			adjusted_value: value = adjusted_value (x.value)
		end

	make_from_native_integer (v: like native_integer_anchor)
			-- Create a real number with `value' = `adjusted_value' (`v'.`to_real').
		do
			make (v)
		ensure
			adjusted_value: value = adjusted_value (v)
		end

	make_from_native_integer_ref (v_ref: like native_integer_ref_anchor)
			-- Create a real number with `value' =  `adjusted_value' (`v_ref'.`to_real').
		do
			make (v_ref.to_real)
		ensure
			adjusted_value: value = adjusted_value (v_ref.to_real)
		end

	make_from_integer (i: STS_INTEGER_NUMBER)
			-- Create a real number with `value' = `adjusted_value' (`i'.`value'.`to_real').
		do
			make_from_native_integer (i.value) -- TODO: convert
		ensure
			adjusted_value: value = adjusted_value (i.value) -- TODO: convert
		end

	make_from_native_natural (v: like native_natural_anchor)
			-- Create a real number with `value' = `adjusted_value' (`v').
		do
			make (v)
		ensure
			adjusted_value: value = adjusted_value (v)
		end

	make_from_native_natural_ref (v_ref: like native_natural_ref_anchor)
			-- Create a real number with `value' =  `adjusted_value' (`v_ref'.`item').
		do
			make (v_ref.item)
		ensure
			adjusted_value: value = adjusted_value (v_ref.item)
		end

	make_abs (x: REAL_NUMBER)
			-- Create a real number with value `x'.`value'.`abs'.
		do
			value_bit_pattern := x.value_bit_pattern & ⊝ ({NATURAL_8} 1 |<< (Exponent_width + Mantissa_width))
		ensure
			value: value = x.value.abs
		end

	make_opposite (x: REAL_NUMBER)
			-- Create a real number with value -`x'.`value'
		do
			value_bit_pattern := x.value_bit_pattern ⊕ ({NATURAL_8} 1 |<< (Exponent_width + Mantissa_width))
		ensure
			value: value = - x.value
		end

feature -- Primitive

	value: like native_real_anchor
			-- <Precursor>
		local
			m: like value_bit_pattern
			sgn: like value.sign
			exp: like Exponent_bias
		do
			sgn := if sign_bit_status = 0 then 1 else - 1 end
			exp := exponent_bit_pattern.as_integer_8 - Exponent_bias
			m := mantissa_bit_pattern
			if max_exponent < exp then
				if m = 0 then
					Result := sgn × {like value}.Positive_infinity
				else
					Result := {COMPLEX_NUMBER}.c_copysign ({like value}.Nan, sgn)
				end
			elseif min_normal_exponent ≤ exp then
					-- Normal range
				Result := (sgn × (2 ^ exp) × (1 + m / Mantissa_scale)).truncated_to_real -- TODO: Get rid of truncated_to_real.
			elseif 0 < m then
					-- Subnormal range
				Result := (sgn × (2 ^ min_normal_exponent) × m / Mantissa_scale).truncated_to_real -- TODO: Get rid of truncated_to_real.
			else
				Result := sgn × {REAL} 0.0
			end
		ensure then
			el_sgn: attached if sign_bit_status = 0 then 1 else - 1 end as el_sgn
			el_exp: attached (exponent_bit_pattern.as_integer_8 - Exponent_bias) as el_exp
			el_m: attached mantissa_bit_pattern as el_m

			when_nan: max_exponent < el_exp and el_m /= 0 implies Result.is_nan
			when_huge: max_exponent < el_exp and el_m = 0 implies Result = el_sgn × {like value}.Positive_infinity
			when_normal: min_normal_exponent ≤ el_exp and el_exp ≤ max_exponent implies Result = (el_sgn × (2 ^ el_exp) × (1 + el_m / Mantissa_scale)).truncated_to_real
			when_subnormal: el_exp < min_normal_exponent and 0 < el_m implies Result = (el_sgn × (2 ^ min_normal_exponent) × el_m / Mantissa_scale).truncated_to_real
			when_tiny: el_exp < min_normal_exponent and el_m = 0 implies Result = el_sgn × 0
		end

feature -- Access

	sign: like integer_anchor
			-- <Precursor>
		local
			e: like value_bit_pattern
			m: like value_bit_pattern
		do
			e := exponent_bit_pattern
			m := mantissa_bit_pattern
			if sign_bit_status = 0 then
				if e = max_exponent_bit_pattern then
					if m = 0 then
							check
								is_positive_infinity
							end
						create Result.make (1)
					else
							check
								is_nan
								value < 0
							end
						create Result.make (-1)
					end
				elseif e | m = 0 then
						check
							positive_zero: value = 0
							nothing_to_do: Result = 0
						end
				else
						check
							0 < value
						end
					create Result.make (1)
				end
			elseif e | m = 0 then
					check
						negative_zero: value = - 0.0
						nothing_to_do: Result = 0
					end
			else
					check
						value < 0
					end
				create Result.make (-1)
			end
		end

	sign_bit: like Integer_anchor -- TODO: Natural instead.
			-- <Precursor>
		do
				check
						-- sign_bit_status.as_integer_8 ∈ {0, 1}
					big_enough: {INTEGER_NUMBER}.Native_min_value ≤ sign_bit_status.as_integer_8
					small_enough: sign_bit_status.as_integer_8 ≤ {INTEGER_NUMBER}.Native_max_value
				end
			create Result.make_from_stored_value (sign_bit_status.as_integer_8)
		end

	zero: REAL_NUMBER
			-- <Precursor>
		once
		ensure then
			class
		end

	one: REAL_NUMBER
			-- <Precursor>
		once
			create Result.make (1)
		ensure then
			class
		end

	two: INTEGER_NUMBER
			-- <Precursor>
		once
			create Result.make (2)
		ensure then
			class
		end

	imaginary_unit: COMPLEX_NUMBER
			-- <Precursor>
		once
			Result := {COMPLEX_NUMBER}.I
		ensure then
			class
		end

	nan: REAL_NUMBER
			-- Representation of not a number (NaN)
		local
			e: like Exponent_mask
			m: like value_bit_pattern
		once
			e := Exponent_mask |<< Mantissa_width
			m := Mantissa_mask
				check
					no_extra_bit: ({NATURAL_8} 0 | e | m) & (⊝ Value_bit_pattern_mask) = 0
						-- 0 | (Exponent_mask |<< Mantissa_width) | Mantissa_mask < Value_bit_pattern_mask
				end
			create Result.make_from_bit_pattern ({NATURAL_8} 0 | e | m) -- TODO: Get rid of {NATURAL_8}?
		ensure
			class
			definition: Result.is_nan
		end

	negative_infinity: REAL_NUMBER
			-- Representation of negative infinity
		local
			s: like value_bit_pattern
			e: like Exponent_mask
		once
			s := {NATURAL_8} 1 |<< (Exponent_width + Mantissa_width) -- TODO: Get rid of {NATURAL_8}?
			e := Exponent_mask |<< Mantissa_width
				check
					no_extra_bit: (s | e | 0) & (⊝ Value_bit_pattern_mask) = 0
						-- (1 |<< (Exponent_width + Mantissa_width)) | (Exponent_mask |<< Mantissa_width) | 0 < Value_bit_pattern_mask
				end
			create Result.make_from_bit_pattern (s | e | 0)
		ensure
			class
			definition: Result.is_negative_infinity
		end

	positive_infinity: REAL_NUMBER
			-- Representation of positive infinity
		local
			e: like Exponent_mask
		once
			e := Exponent_mask |<< Mantissa_width
				check
					no_extra_bit: ({NATURAL_8} 0 | e | 0) & (⊝ Value_bit_pattern_mask) = 0
						-- (0 | (Exponent_mask |<< Mantissa_width) | 0) < Value_bit_pattern_mask
				end
			create Result.make_from_bit_pattern ({NATURAL_8} 0 | e | 0) -- TODO: Get rid of {NATURAL_8}?
		ensure
			class
			definition: Result.is_positive_infinity
		end

	min_value: REAL_NUMBER
			-- Minimum value representable by current kind of real number.
		once
			Result := - Max_value
		ensure
			class
			definition: Result ≍ - Max_value
		end

	max_value: REAL_NUMBER
			-- Maximum value representable by current kind of real number.
		local
			e: like Exponent_mask
			m: like value_bit_pattern
		once
			e := (max_exponent_bit_pattern - 1) |<< Mantissa_width
			m := Max_mantissa
				check
					no_extra_bit: ({NATURAL_8} 0 | e | m) & (⊝ Value_bit_pattern_mask) = 0
						-- 0 | ((max_exponent_bit_pattern - 1) |<< Mantissa_width) | Max_mantissa < Value_bit_pattern_mask
				end
			create Result.make_from_bit_pattern ({NATURAL_8} 0 | e | m) -- TODO: Get rid of {NATURAL_8}?
		ensure
			class
			definition: Result.value = (2 ^ max_exponent) × (1 + Max_mantissa / Mantissa_scale)
		end

	machine_epsilon: REAL_NUMBER
			-- <Precursor>
		local
			e: like Exponent_mask
		once
			e := (Exponent_bias.as_natural_8 - Mantissa_width) |<< Mantissa_width
				check
					no_extra_bit: ({NATURAL_8} 0 | e | 0) & (⊝ Value_bit_pattern_mask) = 0
						-- 0 | ((Exponent_bias - Mantissa_width) |<< Mantissa_width) | 0 < Value_bit_pattern_mask
				end
			create Result.make_from_bit_pattern ({NATURAL_8} 0 | e | 0) -- TODO: Get rid of {NATURAL_8}?
		ensure then
			class
		end

	epsilon: REAL_NUMBER
			-- Minimum normalized positive number representable by current kind of real number
		local
			e: like Exponent_mask
		once
			e := {NATURAL_8} 1 |<< Mantissa_width
				check
					no_extra_bit: ({NATURAL_8} 0 | e | 0) & (⊝ Value_bit_pattern_mask) = 0 -- 0 | (1 |<< Mantissa_width) | 0 < Value_bit_pattern_mask
				end
			create Result.make_from_bit_pattern ({NATURAL_8} 0 | e | 0)
		ensure
			class
			definition: Result.value = 2 ^ min_normal_exponent
		end

	ulp,
	relative_epsilon: like real_anchor
			-- <Precursor>
			--| TODO: Optimize it?
		do
			Result := abs
			if Max_value ≤ Result then
				Result := Result - Result.previous_float
			else
				Result := Result.next_float - Result
			end
		end

	previous_float: like real_anchor
			-- <Precursor>
		local
			e: like value_bit_pattern
			m: like value_bit_pattern
		do
			e := exponent_bit_pattern
			m := mantissa_bit_pattern
			if sign_bit_status = 0 then
				if e = max_exponent_bit_pattern then
					if m = 0 then
							check
								is_positive_infinity
							end
							-- Come down to `Max_value'
							check
								no_extra_bit: (value_bit_pattern - 1) & (⊝ Value_bit_pattern_mask) = 0
									-- value_bit_pattern - 1 < Value_bit_pattern_mask <== value_bit_pattern ≤ Value_bit_pattern_mask
							end
						create Result.make_from_bit_pattern (value_bit_pattern - 1)
					else
							check
								is_nan
							end
							-- Stay upon NaN.
							check
								no_extra_bit: value_bit_pattern & (⊝ Value_bit_pattern_mask) = 0 -- value_bit_pattern ≤ Value_bit_pattern_mask
							end
						create Result.make_from_bit_pattern (value_bit_pattern)
					end
				elseif e | m = 0 then
						check
							Current ≍ Zero
						end
						-- Flip sign and go to the tiniest negative value.
						check
							no_extra_bit: (({NATURAL_8} 1 |<< (Exponent_width + Mantissa_width)) | 0 | 1) & (⊝ Value_bit_pattern_mask) = 0
								-- ((1 |<< (Exponent_width + Mantissa_width)) | 0 | 1) < Value_bit_pattern_mask
						end
					create Result.make_from_bit_pattern (({NATURAL_8} 1 |<< (Exponent_width + Mantissa_width)) | 0 | 1)
				else
						check
							Zero < Current
						end
						-- Come one `ulp' down.
						check
							no_extra_bit: (value_bit_pattern - 1) & (⊝ Value_bit_pattern_mask) = 0
									-- value_bit_pattern - 1 < Value_bit_pattern_mask <== value_bit_pattern ≤ Value_bit_pattern_mask
						end
					create Result.make_from_bit_pattern (value_bit_pattern - 1)
				end
			else
				if e = max_exponent_bit_pattern then
						check
							is_nan or is_negative_infinity
						end
						-- Nothing changes.
					create Result.make_from_bit_pattern (value_bit_pattern)
				else
						check
							is_finite
							Current ≤ Zero
						end
						-- Come one `ulp' down.
					create Result.make_from_bit_pattern (value_bit_pattern + 1) -- The segative sign inverts the order.
				end
			end
		ensure then
			when_positive_infinity: is_positive_infinity implies Result ≍ Max_value

			el_exp: attached {like value} value_logb (value) as el_exp
			el_normal_v: attached {like value} (value / (2 ^ el_exp).truncated_to_real) as el_normal_v
			el_subnormal_v: attached {like value} (value / (2 ^ min_normal_exponent).truncated_to_real) as el_subnormal_v

			when_normal_positive_same_exponent: 0 < value and min_normal_exponent ≤ el_exp and (el_normal_v - 1) × Mantissa_scale > 0 implies
				Result.value = (2 ^ el_exp).truncated_to_real × (1 + ((el_normal_v - 1) × Mantissa_scale - 1) / Mantissa_scale)
			when_normal_positive_new_exponent_still_normal: 0 < value and min_normal_exponent < el_exp and (el_normal_v - 1) × Mantissa_scale = 0 implies
				Result.value = (2 ^ (el_exp - 1)).truncated_to_real × (1 + (Mantissa_scale - 1) / Mantissa_scale)
			when_normal_positive_new_subnormal_exponent: 0 < value and min_normal_exponent = el_exp and (el_normal_v - 1) × Mantissa_scale = 0 implies
				Result.value = (2 ^ el_exp).truncated_to_real × (Mantissa_scale - 1) / Mantissa_scale
			when_subnormal_positive: 0 < value and el_exp < min_normal_exponent implies
				Result.value = (2 ^ min_normal_exponent).truncated_to_real × (el_subnormal_v × Mantissa_scale - 1) / Mantissa_scale
			when_non_positive: Current ≤ Zero implies Result ≍ - abs.next_float
		end

	next_float: like real_anchor
			-- <Precursor>
		local
			e: like value_bit_pattern
			m: like value_bit_pattern
		do
			e := exponent_bit_pattern
			m := mantissa_bit_pattern
			if sign_bit_status = 0 then
				if e = max_exponent_bit_pattern then
						check
							is_nan or is_positive_infinity
						end
						-- Nothing changes.
						check
							no_extra_bit: value_bit_pattern & (⊝ Value_bit_pattern_mask) = 0 -- value_bit_pattern ≤ Value_bit_pattern_mask
						end
					create Result.make_from_bit_pattern (value_bit_pattern)
				else
						check
							is_finite
							Zero ≤ Current
						end
						-- Go one `ulp' up.
						check
							no_extra_bit: (value_bit_pattern + 1) & (⊝ Value_bit_pattern_mask) = 0
								-- value_bit_pattern + 1 < Value_bit_pattern_mask <== is_finite
						end
					create Result.make_from_bit_pattern (value_bit_pattern + 1)
				end
			else
				if e = max_exponent_bit_pattern then
					if m = 0 then
							check
								is_negative_infinity
							end
							-- Go up to `Min_value'.
							check
								no_extra_bit: (value_bit_pattern - 1) & (⊝ Value_bit_pattern_mask) = 0
									-- value_bit_pattern - 1 < Value_bit_pattern_mask <== value_bit_pattern ≤ Value_bit_pattern_mask
							end
						create Result.make_from_bit_pattern (value_bit_pattern - 1) -- The segative sign inverts the order.
					else
							check
								is_nan
							end
							-- Stay upon NaN.
							check
								no_extra_bit: value_bit_pattern & (⊝ Value_bit_pattern_mask) = 0 -- value_bit_pattern ≤ Value_bit_pattern_mask
							end
						create Result.make_from_bit_pattern (value_bit_pattern)
					end
				elseif e | m = 0 then
						check
							negative_zero: Current ≍ Zero
						end
						-- Flip sign and go to the tiniest positive value.
						check
							no_extra_bit: ({NATURAL_8} 0 | 0 | 1) & (⊝ Value_bit_pattern_mask) = 0 -- By definition
						end
					create Result.make_from_bit_pattern ({NATURAL_8} 0 | 0 | 1) -- TODO: Get rid of {NATURAL_8}?
				else
						check
							is_finite
							Current < Zero
						end
						-- Go one `ulp' up.
						check
							no_extra_bit: (value_bit_pattern - 1) & (⊝ Value_bit_pattern_mask) = 0
								-- value_bit_pattern - 1 < Value_bit_pattern_mask <== is_finite
						end
					create Result.make_from_bit_pattern (value_bit_pattern - 1) -- The segative sign inverts the order.
				end
			end
		ensure then
			when_nan: is_nan implies Result.is_nan
			when_negative: Current < Zero implies Result ≍ - abs.previous_float
			when_sero: Current ≍ Zero implies Result.value = Epsilon.value / Mantissa_scale

			el_exp: attached {like value} value_logb (value) as el_exp

			when_sub_normal: Zero < Current and el_exp < min_normal_exponent implies
				Result.value = (2 ^ min_normal_exponent).truncated_to_real × ((value / (2 ^ min_normal_exponent).truncated_to_real) × Mantissa_scale + 1) / Mantissa_scale -- TODO: × and * have different precedences.
			when_normal: Zero < Current and min_normal_exponent ≤ el_exp and Current < Max_value implies
				Result.value = (2 ^ el_exp).truncated_to_real × (1 + ((value / (2 ^ el_exp).truncated_to_real - 1) × Mantissa_scale + 1) / Mantissa_scale)
			when_huge: Max_value ≤ Current implies Result.is_positive_infinity
		end

feature -- Quality

	is_real: BOOLEAN = True
			-- <Precursor>

	is_proper_real: BOOLEAN
			-- <Precursor>
		do
			Result := value_bit_pattern & ((Exponent_mask |<< Mantissa_width) | Mantissa_mask) /= 0
		end

	is_imaginary: BOOLEAN
			-- <Precursor>
		do
			Result := value_bit_pattern & ((Exponent_mask |<< Mantissa_width) | Mantissa_mask) = 0
		end

	is_proper_imaginary: BOOLEAN = False
			-- <Precursor>

	is_proper_complex: BOOLEAN = False
			-- <Precursor>

	is_nan: BOOLEAN
			-- <Precursor>
		do
			Result := exponent_bit_pattern = max_exponent_bit_pattern and mantissa_bit_pattern /= 0
		end

	is_negative_infinity: BOOLEAN
			-- <Precursor>
		do
			Result := mantissa_bit_pattern = 0 and exponent_bit_pattern = max_exponent_bit_pattern and sign_bit_status = 1
		end

	is_negative_zero: BOOLEAN
			-- <Precursor>
		do
			Result := value_bit_pattern = {NATURAL_8} 1 |<< (Exponent_width + Mantissa_width)
		end

	is_positive_zero: BOOLEAN
			-- <Precursor>
		do
			Result := value_bit_pattern = 0
		end

	is_positive_infinity: BOOLEAN
			-- <Precursor>
		do
			Result := mantissa_bit_pattern = 0 and exponent_bit_pattern = max_exponent_bit_pattern and sign_bit_status = 0
		end

	is_finite: BOOLEAN
			-- <Precursor>
		do
			Result := exponent_bit_pattern < max_exponent_bit_pattern
		end

	is_infinite: BOOLEAN
			-- <Precursor>
		do
			Result := exponent_bit_pattern = max_exponent_bit_pattern and mantissa_bit_pattern = 0
		end

	is_invertible: BOOLEAN = True
			-- <Precursor>
			-- An implementation akin to IEEE 754 inverts zero with no major consequence.

	is_rational: BOOLEAN
			-- <Precursor>
		local
			m: like value_bit_pattern
			sgn: like value.sign
			exp: like Exponent_bias
			num, den: like value
		do
			sgn := if sign_bit_status = 0 then 1 else - 1 end
			exp := exponent_bit_pattern.as_integer_8 - Exponent_bias
			m := mantissa_bit_pattern
			if exp ≤ max_exponent then
				if min_normal_exponent ≤ exp then
						-- Normal range
					num := (sgn × (2 ^ exp) × (1 + m / Mantissa_scale)).truncated_to_real -- TODO: Get rid of truncated_to_real.
				elseif 0 < m then
						-- Subnormal range
					num := (sgn × (2 ^ min_normal_exponent) × m / Mantissa_scale).truncated_to_real -- TODO: Get rid of truncated_to_real.
				end
				from
					den := 1
				until
					num = num.floor_real_32
				loop
					num := num * 2
					den := den * 2
				end
					check
						0 < den -- den = 2 ^ k, k ≥ 0.
					end
				Result := num.abs ≤ {INTEGER_NUMBER}.Native_min_value.abs and den ≤ {INTEGER_NUMBER}.Native_min_value.abs
			end
		end

feature -- Output

	out: STRING
			-- <Precursor>
		do
			Result := value.out
		ensure then
			definition: Result ~ value.out
		end

feature -- Comparison

	equals alias "≍" (x: STS_REAL_NUMBER): BOOLEAN
			-- <Precursor>
		do
			if Current ~ x then
				Result := True
			elseif attached {REAL_NUMBER} x as exp_x then
				if
						-- +0 = -0
					value_bit_pattern & ((Exponent_mask |<< Mantissa_width) | Mantissa_mask) = 0 and
					exp_x.value_bit_pattern & ((Exponent_mask |<< Mantissa_width) | Mantissa_mask) = 0
				then
					Result := True
				elseif is_nan then
					Result := exp_x.is_nan
				end
			else
				Result := Precursor {STS_REAL_NUMBER}(x)
			end
		end

	is_less alias "<" (x: STS_REAL_NUMBER): BOOLEAN
			-- <Precursor>
			--| Notice that the code is big, but after discarding assertions and removing dead code, hopefully it is faster than comparing the calculated values.
			--| TODO: Test it.
		local
			sb, x_sb: like sign_bit_status
		do
			if attached {REAL_NUMBER} x as exp_x then
				if exp_x.is_nan then
						check
							cant_be_less_than_nan: exp_x.value ≤ value
							nothing_to_do: not Result
						end
				elseif is_nan then
						check
							nan_is_minimum: value < exp_x.value
						end
					Result := True
				else
					sb := sign_bit_status
					x_sb := exp_x.sign_bit_status
					if sb | x_sb = 0 then
							check
								current_is_non_negative: 0 ≤ value
								x_is_non_negative: 0 ≤ exp_x.value
							end
						Result := value_bit_pattern < exp_x.value_bit_pattern
					elseif sb < x_sb then
							check
								x_is_non_positive: exp_x.value ≤ 0
								current_is_non_negative: 0 ≤ value
								nothing_to_do: not Result
							end
					elseif sb > x_sb then
							check
								current_is_non_positive: value ≤ 0
								x_is_non_negative: 0 ≤ exp_x.value
							end
						if
								-- +0 = -0
							value_bit_pattern & ((Exponent_mask |<< Mantissa_width) | Mantissa_mask) /= 0 or
							exp_x.value_bit_pattern & ((Exponent_mask |<< Mantissa_width) | Mantissa_mask) /= 0
						then
								check
									negative_below_non_negative: value < 0 implies 0 ≤ exp_x.value
									zero_below_positive: value = 0 implies 0 < exp_x.value
								end
							Result := True
						else
								check
									current_is_zero: value = 0
									x_is_zero: exp_x.value = 0
									nothing_to_do: not Result
								end
						end
					else
							check
								current_is_non_positive: sb = 1
								x_is_non_positive: x_sb = 1
							end
						Result := exp_x.value_bit_pattern < value_bit_pattern
					end
				end
			else
				Result := Precursor {STS_REAL_NUMBER}(x)
			end
		end

	is_less_equal alias "<=" alias "≤" (x: STS_REAL_NUMBER): BOOLEAN
			-- <Precursor>
			--| Please see the observation at `is_less' header.
		local
			sb, x_sb: like sign_bit_status
		do
			if attached {REAL_NUMBER} x as exp_x then
				if exp_x.is_nan then
						check
							cant_be_less_than_nan: exp_x.value ≤ value
						end
					Result := is_nan
				elseif is_nan then
						check
							nan_is_minimum: value < exp_x.value
						end
					Result := True
				else
					sb := sign_bit_status
					x_sb := exp_x.sign_bit_status
					if sb | x_sb = 0 then
							check
								current_is_non_negative: 0 ≤ value
								x_is_non_negative: 0 ≤ exp_x.value
							end
						Result := value_bit_pattern ≤ exp_x.value_bit_pattern
					elseif sb < x_sb then
							check
								x_is_non_positive: exp_x.value ≤ 0
								current_is_non_negative: 0 ≤ value
							end
							-- +0 = -0
						Result := value_bit_pattern & ((Exponent_mask |<< Mantissa_width) | Mantissa_mask) = 0 and
							exp_x.value_bit_pattern & ((Exponent_mask |<< Mantissa_width) | Mantissa_mask) = 0
					elseif sb > x_sb then
							check
								current_is_non_positive: value ≤ 0
								x_is_non_negative: 0 ≤ exp_x.value
							end
						Result := True
					else
							check
								current_is_non_positive: sb = 1
								x_is_non_positive: x_sb = 1
							end
						Result := exp_x.value_bit_pattern ≤ value_bit_pattern
					end
				end
			else
				Result := value ≤ x.value
			end
		end

	three_way_comparison alias "⋚" (x: STS_REAL_NUMBER): like integer_anchor
			-- <Precursor>
		local
			v: like Integer_anchor.value
		do
			v := value ⋚ x.value
				check
					not_too_small: {INTEGER_NUMBER}.Native_min_value ≤ v.as_integer_8 -- {INTEGER_NUMBER}.Native_min_value ≤ -1
					not_too_big: v.as_integer_8 ≤ {INTEGER_NUMBER}.Native_max_value -- 1 ≤ {INTEGER_NUMBER}.Native_max_value
				end
			create Result.make_from_stored_value (v.as_integer_8)
		end

	min alias "∧" (x: STS_REAL_NUMBER): like real_anchor
			-- <Precursor>
		do
			if Current ≤ x then
				Result := Current -- TODO: No-op when called from {STT_REAL_NUMBER}!!!
			else
--				Result := x -- TODO: Result := create {REAL_NUMBER}.make_from_real (x) instead of create Result.make_from_real (x)
				create Result.make_from_real (x)
			end
		end

	max alias "∨" (x: STS_REAL_NUMBER): like real_anchor
			-- <Precursor>
		do
			if Current ≥ x then
				Result := Current
			else
				create Result.make_from_real (x)
			end
		end

feature -- Operation

	argument,
	phase: like real_anchor
			-- <Precursor>
		do
			if is_nan then
				Result := Current
			elseif sign_bit_status = 1 then -- Negative zero or below
				create Result.make (Pi)
			end
		end

	conjugate: like complex_anchor
			-- <Precursor>
		do
			create Result.make_from_expanded (Current, - Zero)
		end

	complex_plus (z: STS_COMPLEX_NUMBER): like complex_anchor
			-- <Precursor>
		do
			Result := as_complex + z
		end

	complex_minus (z: STS_COMPLEX_NUMBER): like complex_anchor
			-- <Precursor>
		do
			Result := as_complex - z
		end

	complex_product (z: STS_COMPLEX_NUMBER): like complex_anchor
			-- <Precursor>
		do
			Result := as_complex ⋅ z
		end

	complex_quotient (z: STS_COMPLEX_NUMBER): like complex_anchor
			-- <Precursor>
		do
				check
					good_divisor: as_complex.divisible (z) -- divisible (z)
				end
			Result := as_complex / z
		end

	modulus,
	abs: like real_anchor
			-- Distance from current real number to the origin of the real number
			-- line
		do
			create Result.make_abs (Current)
		end

	opposite alias "-" alias "−": like real_anchor
			-- <Precursor>
		do
			create Result.make_opposite (Current)
		end

	quotient alias "/" alias "÷" (x: STS_REAL_NUMBER): like real_anchor
			-- <Precursor>
		do
				check
					good_divisor_ignored: value.divisible (x.value) implies True
				end
			create Result.make (value / x.value)
		ensure then
			nan_by_zero: is_nan and x ≍ Zero implies Result.is_nan
			negative_by_negative_zero: Nan < Current and Current < Zero and x.is_negative_zero implies Result.is_positive_infinity
			positive_by_negative_zero: Zero < Current and x.is_negative_zero implies Result.is_negative_infinity
			zero_by_zero: Current ≍ Zero and x ≍ Zero implies Result.is_nan
			negative_by_positive_zero: Nan < Current and Current < Zero and x.is_positive_zero implies Result.is_negative_infinity
			positive_by_positive_zero: Zero < Current and x.is_positive_zero implies Result.is_positive_infinity
		end

	reciprocal,
	inverse: like real_anchor
			-- <Precursor>
		require else
			always_divisible: True
		do
				check
					good_divisor_ignored: (1).to_real.divisible (value) implies
						True
				end
			create Result.make (1 / value)
		end

feature -- Conversion

	as_complex: like complex_anchor
			-- Current real number represented as a complex number
		do
			create Result.make_from_real (Current)
		ensure
			real: Result.a ≍ Current
			imaginary: Result.b ≍ Zero
		end

	to_rational: like rational_anchor
			-- <Precursor>
		local
			m: like value_bit_pattern
			sgn: like value.sign
			exp: like Exponent_bias
			num, den: like value
		do
			sgn := if sign_bit_status = 0 then 1 else - 1 end
			exp := exponent_bit_pattern.as_integer_8 - Exponent_bias
			m := mantissa_bit_pattern
			if min_normal_exponent ≤ exp then
					-- Normal range
				num := (sgn × (2 ^ exp) × (1 + m / Mantissa_scale)).truncated_to_real -- TODO: Get rid of truncated_to_real.
			elseif 0 < m then
					-- Subnormal range
				num := (sgn × (2 ^ min_normal_exponent) × m / Mantissa_scale).truncated_to_real -- TODO: Get rid of truncated_to_real.
			end
			from
				den := 1
			until
				num = num.floor_real_32
			loop
				num := num * 2
				den := den * 2
			end
				check
						-- Precondition: is_rational
					num.abs ≤ {INTEGER_NUMBER}.Native_min_value.abs
					den ≤ {INTEGER_NUMBER}.Native_min_value.abs
				end
			if num = {INTEGER_NUMBER}.Native_min_value.abs or den = {INTEGER_NUMBER}.Native_min_value.abs then
					-- Otherwise, e.g. pq = 8/q gets changed upon creation to -8/q = -pq.
				num := - num
				den := - den
			end
				check
					non_zero_denominator: den.truncated_to_integer ≭ Zero -- den = 2 ^ k, k ≥ 0.
				end
			create Result.make_from_expanded (num.truncated_to_integer, den.truncated_to_integer)
		end

	truncated_to_integer: like integer_anchor
			-- <Precursor>
		do
			create Result.make (value.truncated_to_integer)
		end

feature -- Math

	value_is_infinite (v: like real_anchor.value): BOOLEAN
			-- <Precursor>
		external
			"C macro signature (float): int use <math.h>"
		alias
			"isinf"
		end

	value_is_finite (v: like real_anchor.value): BOOLEAN
			-- <Precursor>
		external
			"C macro signature (float): int use <math.h>"
		alias
			"isfinite"
		end

	hypotenuse (x, y: STS_REAL_NUMBER): like real_anchor
			-- <Precursor>
			--| TODO: DRY.
		do
			Result := {COMPLEX_NUMBER}.hypotenuse (x, y)
		ensure then
			class
		end

	arc_tangent_yx (y, x: STS_REAL_NUMBER): like real_anchor
			-- <Precursor>
		do
			Result := {COMPLEX_NUMBER}.arc_tangent_yx (y, x)
		ensure then
			class
		end

	splitted (q: STS_REAL_NUMBER): HOMOGENEOUS_ORDERED_PAIR [STS_REAL_NUMBER, STS_REAL_EQUALITY]
			-- <Precursor>
		local
			num, den: like value
		do
			from
				num := value
				den := q.value
			until
				num = num.floor_real_32
			loop
				num := num * 2
				den := den * 2
			end
			create Result.make (create {REAL_NUMBER}.make (num), create {REAL_NUMBER}.make (den))
		end

feature -- Factory

	real_from_value (v: like real_anchor.value): like real_anchor
			-- <Precursor>
		do
			create Result.make (v)
		ensure then
			class
		end

	real_from_double_value (dv: like real_anchor.value.to_double): like real_anchor
			-- <Precursor>
		do
			create Result.make_from_native_double_real (dv)
		ensure then
			class
		end

feature -- Implementation

	adjusted_value (v: like value): like value
			-- <Precursor>
		local
			exp: like value
		do
			exp := value_logb (v)
			if max_exponent < exp then
				Result := v.sign × {like value}.positive_infinity
			elseif min_normal_exponent ≤ exp then
				Result := v × (2 ^ (Mantissa_width - exp)).truncated_to_real
				Result := Result.truncated_to_integer
				Result := Result / Mantissa_scale
				Result := (2 ^ exp).truncated_to_real × Result
			elseif min_subnormal_exponent ≤ exp then
				Result := v × (2 ^ (Mantissa_width.as_integer_8 - min_normal_exponent)).truncated_to_real
				Result := Result.truncated_to_integer
				Result := Result / Mantissa_scale
				Result := (2 ^ min_normal_exponent).truncated_to_real × Result
			else
					check
						no_surprising_nan: not (v.is_negative_infinity or v.is_positive_infinity) -- Otherwise Greatest_exponent < exp = +Infinity (see above.)
						nan_is_included: v.is_nan implies (0 × v).is_nan
					end
				Result := 0 × v
			end
		ensure then
			class
			el_exp: attached {like value} value_logb (v) as el_exp
			when_huge: max_exponent < el_exp implies Result = v.sign × {like value}.Positive_infinity
			when_normal: min_normal_exponent ≤ el_exp and el_exp <= max_exponent implies
				Result = (2 ^ el_exp).truncated_to_real × (v × (2 ^ (Mantissa_width - el_exp)).truncated_to_real).truncated_to_integer / Mantissa_scale
			when_subnormal: min_subnormal_exponent ≤ el_exp and el_exp < min_normal_exponent implies Result =
				(2 ^ min_normal_exponent).truncated_to_real × (v × (2 ^ (Mantissa_width.as_integer_8 - min_normal_exponent)).truncated_to_real).truncated_to_integer / Mantissa_scale
			when_tiny: el_exp < min_subnormal_exponent implies Result = 0 × v
		end

feature -- Anchor

	complex_anchor: COMPLEX_NUMBER
			-- <Precursor>
		once
		ensure then
			class
		end

	real_anchor: REAL_NUMBER
			-- <Precursor>
		once
		ensure then
			class
		end

	rational_anchor: RATIONAL_NUMBER
			-- <Precursor>
		once
		ensure then
			class
		end

	integer_anchor: INTEGER_NUMBER
			-- <Precursor>
		once
		ensure then
			class
		end

	native_real_ref_anchor: like value.to_reference
			-- Anchor for a reference to the native representation of the value of current real number
		do
			create Result
		ensure
			class
		end

	native_double_real_anchor: like value.to_double
			-- Anchor for a native real value of double precision
		do
		ensure
			class
		end

	native_double_real_ref_anchor: like native_double_real_anchor.to_reference
			-- Anchor for a reference to a native real value of double precision
		do
			create Result
		ensure
			class
		end

	native_integer_anchor: INTEGER
			-- Anchor for a native integer number
		once
		ensure
			class
		end

	native_integer_ref_anchor: like native_integer_anchor.to_reference
			-- Anchor for a reference to a native integer number
		do
			create Result
		ensure
			class
		end

	native_natural_anchor: NATURAL
			-- Anchor for a native natural number
		once
		ensure
			class
		end

	native_natural_ref_anchor: like native_natural_anchor.to_reference
			-- Anchor for a reference to a native natural number
		do
			create Result
		ensure
			class
		end

feature {REAL_NUMBER, REAL_UNIVERSE} -- Implementation

	value_bit_pattern: NATURAL_8
			-- Bit pattern of the `value' of current real number

feature {REAL_NUMBER} -- Implementation

	sign_bit_status: like value_bit_pattern
			-- Status of the sign bit of `value_bit_pattern'
			-- Notice that, unlike `sign', it has only two possible values (0 or 1), and it may be non zero for, e.g. negative zero.
		do
			Result := value_bit_pattern |>> (Exponent_width + Mantissa_width)
			Result := Result & 1
		ensure
			definition: Result = (value_bit_pattern |>> (Exponent_width + Mantissa_width)) & 1
		end

	exponent_bit_pattern: like value_bit_pattern
			-- Bit pattern that represents the exponent, on base 2, of current real number.
		do
			Result := value_bit_pattern |>> Mantissa_width
			Result := Result & Exponent_mask
		ensure
			definition: Result = (value_bit_pattern |>> Mantissa_width) & Exponent_mask
		end

	mantissa_bit_pattern: like value_bit_pattern
			-- Bit pattern that represents the mantissa, on base 2, of current real number.
		do
			Result := value_bit_pattern & Mantissa_mask
		ensure
			definition: Result = value_bit_pattern & Mantissa_mask
		end

feature {NONE} -- Implementation

	bit_pattern_from_native_real (v: like value): like value_bit_pattern
			-- `v' converted to a bit pattern that represents the value of a real number like current
		local
			exp: like value
			s: like value_bit_pattern
			e: like Exponent_mask
			m: like Mantissa_mask
		do
			s := value_sign_bit (v).as_natural_8
			exp := value_logb (v)
			if max_exponent < exp then
				e := max_exponent_bit_pattern
					check
						m = 0
					end
			elseif min_normal_exponent ≤ exp then
				e := (exp.truncated_to_integer + Exponent_bias).as_natural_8 -- TODO: Get rid of as_natural_8?
				m := (v.abs × (2 ^ (Mantissa_width - exp))).truncated_to_integer.as_natural_8
				m := m & Mantissa_mask
			elseif min_subnormal_exponent <= exp then
				e := 0
				m := (v.abs × (2 ^ (Mantissa_width.as_integer_8 - min_normal_exponent))).truncated_to_integer.as_natural_8 -- TODO: get rid of as_integer_8 and as_natural_8?
			elseif {like value}.Nan < exp then
				e := 0
				m := 0
			else
				e := max_exponent_bit_pattern
				m := Max_mantissa
			end
			s := s |<< (Exponent_width + Mantissa_width)
			e := e |<< Mantissa_width
			Result := s | e | m
		ensure
			class
			el_s: attached value_sign_bit (v) as el_s
			el_exp: attached value_logb (v) as el_exp

			when_huge: max_exponent < el_exp implies Result =
				(el_s |<< (Exponent_width + Mantissa_width)) |
				(max_exponent_bit_pattern |<< Mantissa_width) |
				0
			when_normal: min_normal_exponent ≤ el_exp and el_exp ≤ max_exponent implies Result =
				(el_s |<< (Exponent_width + Mantissa_width)) |
				((el_exp.truncated_to_integer + Exponent_bias).as_natural_8 |<< Mantissa_width) |
				((v.abs × (2 ^ (Mantissa_width - el_exp))).truncated_to_integer.as_natural_8 & Mantissa_mask)
			when_subnormal: min_subnormal_exponent <= el_exp and el_exp < min_normal_exponent implies Result =
				(el_s |<< (Exponent_width + Mantissa_width)) |
				(0 |<< Mantissa_width) |
				(v.abs × (2 ^ (Mantissa_width.as_integer_8 - min_normal_exponent))).truncated_to_integer.as_natural_8
			when_above_nan: {like value}.Nan < el_exp and el_exp < min_subnormal_exponent implies Result =
				(el_s |<< (Exponent_width + Mantissa_width)) |
				(0 |<< Mantissa_width) |
				0
			when_nan: el_exp.is_nan implies Result =
				(el_s |<< (Exponent_width + Mantissa_width)) |
				(max_exponent_bit_pattern |<< Mantissa_width) |
				Max_mantissa
		end

feature {REAL_UNIVERSE} -- Implementation

	exponent_width: NATURAL_8 = 3
			-- How many bits a real number uses to represent its exponent

	exponent_mask,
			-- Binary mask for the exponent of a real number
	max_exponent_bit_pattern: like exponent_bit_pattern = 0b111 -- 2 ^ `exponent_width' - 1
			-- Maximum biased value for the exponent of a real number

	exponent_bias,
			-- Bias value for the exponent of a real number
	max_exponent: like exponent_bit_pattern.as_integer_8 = 0b011 -- `max_exponent_bit_pattern' // 2
			-- Maximum unbiased value for the exponent of a real number
			--| TODO: Get rid of as_integer_8

	min_normal_exponent: like max_exponent = - 2 -- 1 - `exponent_bias'
			-- Minimum unbiased value for the exponent of a normal real number

	min_subnormal_exponent: like max_exponent = - 4 -- `min_normal_exponent' - `mantissa_width'
			-- Minimum unbiased value for the exponent of a subnormal real number

feature {REAL_UNIVERSE, INTEGER_NUMBER} -- Implementation

	mantissa_width: NATURAL_8 = 2
			-- Number of (explicit) bits used to store the mantissa of a real number

feature {REAL_UNIVERSE} -- Implementation

	mantissa_mask,
			-- Binary mask for the mantissa of a real number
	max_mantissa: like value_bit_pattern = 0b11 -- 2 ^ `mantissa_width' - 1
			-- Maximum value for the mantissa of a real number

	mantissa_scale: like Mantissa_mask = 0b100 -- 2 ^ `mantissa_width'
			-- Reference against wich the units of the mantissa of a real number are measured

	value_bit_pattern_width: NATURAL_8 = 6 -- 1 + exponent_width + mantissa_width
			-- Number of bits used to store the (codified) value of a real number
			--| TODO: Test.

	value_bit_pattern_mask: like value_bit_pattern = 0b1_111_11 -- 2 ^ `value_bit_pattern_width' - 1
			-- Binary mask for the (codified) value of a real number
			--| TODO: Test.

invariant
	zero_real: Zero.identity.value = 0
	zero_imaginary: Zero.Zero.value = 0
	one_real: One.identity.value = 1
	one_imaginary: One.Zero.value = 0
	Two: Two.value = 2
	Imaginary_unit_real: Imaginary_unit.real ≍ Zero.identity
	Imaginary_unit_imaginary: Imaginary_unit.imaginary ≍ One.identity
	Nan: Nan.is_nan
	Negative_infinity: Negative_infinity.is_negative_infinity
	Positive_infinity: Positive_infinity.is_positive_infinity
--	Min_value: Min_value ≍ - Max_value -- TODO: Invariant failure!
	at_least_min_value: not is_nan and not is_negative_infinity implies Min_value ≤ Current
--	Max_value: Max_value.value = (2 ^ max_exponent) × (1 + Max_mantissa / Mantissa_scale) -- TODO: Invariant failure!
--	at_most_max_value: not is_positive_infinity implies Current ≤ Max_value -- TODO: Invariant failure!
--	Machine_epsilon: Machine_epsilon ≍ One.relative_epsilon -- TODO: Invariant failure!
	Epsilon: Epsilon.value = 2 ^ min_normal_exponent
	is_real: is_real = Zero ≍ Zero
	is_proper_imaginary: is_proper_imaginary = (is_imaginary and (Zero ≭ Zero))
	is_proper_complex: is_proper_complex = (not is_real and not is_imaginary)

		-- is_invertible
	scale_log: attached value_logb ((identity.abs ∨ zero.abs).value) as scale_log
	int_scale_log: attached if value_is_finite (scale_log) then
			scale_log.truncated_to_integer
		else
			0
		end as int_scale_log
	scaled_a: attached (identity.value * 2 ^ (- int_scale_log)) as scaled_a
	scaled_b: attached (zero.value * 2 ^ (- int_scale_log)) as scaled_b
	denom: attached (scaled_a ^ 2 + scaled_b ^ 2) as denom
	quasi_definition: denom /= 0 implies Is_invertible
	is_invertible_real_quasi_definition: Current ≭ zero implies Is_invertible

	Exponent_mask: Exponent_mask = (2 ^ Exponent_width - 1).truncated_to_integer
	Greatest_biased_exponent: max_exponent_bit_pattern = Exponent_mask
	Exponent_bias: Exponent_bias = max_exponent_bit_pattern.as_integer_8 // 2
	Greatest_exponent: max_exponent = Exponent_bias
	Least_normal_exponent: min_normal_exponent = 1 - Exponent_bias
	Least_subnormal_exponent: min_subnormal_exponent = min_normal_exponent - Mantissa_width.as_integer_8
	Mantissa_mask: Mantissa_mask = (2 ^ Mantissa_width - 1).truncated_to_integer
	Greatest_mantissa: Max_mantissa = Mantissa_mask
	Mantissa_scale: Mantissa_scale = (2 ^ Mantissa_width).truncated_to_integer
	Value_bit_pattern_width: value_bit_pattern_width = 1 + Exponent_width + Mantissa_width
	Value_bit_pattern_mask: Value_bit_pattern_mask = (2 ^ Value_bit_pattern_width - 1).truncated_to_integer
	nan_bit_pattern: is_nan implies exponent_bit_pattern = max_exponent_bit_pattern and mantissa_bit_pattern /= 0
	zero_bit_pattern: Current ≍ Zero implies exponent_bit_pattern = 0 and mantissa_bit_pattern = 0
	ordinary_bit_pattern: not is_nan and Current ≭ Zero implies value_bit_pattern = bit_pattern_from_native_real (value)

note
	copyright: "Copyright (c) 2012-2022, Rosivaldo Fernandes Alves"
	license: "[
		Eiffel Forum License v2
		(see http://www.eiffel.com/licensing/forum.txt)
		]"
	source: "https://launchpad.net/set-theory"
end
