﻿note
	description: "Implementation of {STS_N_TUPLE}"
	author: "Rosivaldo Fernandes Alves"
	date: "$Date: Sat 2018-12-01 19:25:58 -0300 $"
	revision: "$Revision$"

class
	N_TUPLE [A, EQ -> STS_EQUALITY [A] create default_create end]

inherit
	STS_N_TUPLE [A, EQ]
		redefine
			has,
			term,
			terms,
			front,
			last,
			n,
			out,
			equals,
			exists,
			for_all,
			left_trimmed,
			right_trimmed,
			suffix,
			replacing,
			proper_right_reduced,
			indexed_term
		end

	DEBUG_OUTPUT
		rename
			debug_output as out
		redefine
			out
		end

create
	make_unit,
	make_after,
	make_n,
	make_from_tuple

convert
	make_from_tuple ({
		STS_N_TUPLE [A, EQ],
		STS_TRANSFORMABLE_TUPLE [A, A, EQ, EQ]
		})

feature {NONE} -- Initialization

	make_unit
			-- Create an empty n-tuple, i.e. ().
		do
			tail := Current
		ensure
			is_unit: is_unit
		end

	make_after (a: A; t: STS_N_TUPLE [A, EQ])
			-- Create an n-tuple whose `first' term and `tail' are, respectively, `a' and `t'.
		do
			create tail.make_unit
			tail := tail.converted (t)
			create first_storage.put (converted_term (a))
		ensure
			is_not_unit: not is_unit
			first: eq (first, a)
			tail: tail ≍ t
		end

	make_n (a_n: NATURAL; a: A)
			-- Create an n-tuple with `a_n' terms equal to `a'.
		local
			l_a: like term
			i: NATURAL
		do
			if a_n = 0 then
				make_unit
			else
				from
					l_a := converted_term (a)
					create tail.make_unit
					i := a_n - 1
				invariant
					n: tail.n = a_n - 1 - i
					repeated_terms: tail.n > 0 implies tail.terms.is_singleton and tail.terms ∋ l_a
				until
					i = 0
				loop
					tail := tail.after (l_a)
					i := i - 1
				variant
					i: natural_as_integer (i)
				end
				create first_storage.put (l_a)
			end
		ensure
			n: n = a_n
			repeated_terms: n > 0 implies terms.is_singleton and terms ∋ a
		end

	make_from_tuple (t: STS_N_TUPLE [A, EQ])
			-- Create an n-tuple equal to `t'.
		do
			if t.is_unit then
				make_unit
			else
				make_after (t.first, t.tail)
			end
		ensure
			same_tuple: Current ≍ t
		end

feature -- Primitive

	is_unit: BOOLEAN
			-- <Precursor>
		do
			Result := not attached first_storage
		ensure then
			definition: Result = not attached first_storage
		end

	first: like term_anchor
			-- <Precursor>
		do
			check
				attached_stored_first: attached first_storage -- not `is_unit'
			then
				Result := first_storage.item
			end
		ensure then
			attached_stored_first: attached first_storage -- not `is_unit'
			definition: Result = first_storage.item
		end

	tail: like trimmed_tuple_anchor
			-- <Precursor>

	eq: EQ
			-- <Precursor>
		do
			create Result
		ensure then
			class
		end

feature -- Membership

	has (a: A): BOOLEAN
			-- <Precursor>
		local
			t: like trimmed_tuple_anchor
			l_eq: EQ
		do
			from
				t := Current
				l_eq := eq -- Better to avoid creating `eq' over and over at each comparison
			invariant
				not_yet: right_trimmed (t.n).terms ∌ a
			until
				t.is_unit or else l_eq (a, t.first)
			loop
				t := t.tail
			variant
				n: natural_as_integer (t.n)
			end
			Result := not t.is_unit
		end

feature -- Access

	indices: TRANSFORMABLE_SET [NATURAL, A, STS_OBJECT_EQUALITY [NATURAL], EQ]
			-- <Precursor>
		local
			i: NATURAL
			t: like trimmed_tuple_anchor
		do
			from
				i := 0
				create Result.make_empty
				t := Current
			invariant
				building_indices: # Result = n - t.n
				positive: Result |∀ agent index_is_positive
				small_enough: Result |∀ agent index_is_less_equal (?, n - t.n)
			until
				t.is_unit
			loop
				i := i + 1
					check
						does_not_have: Result ∌ i -- Result ≍ {1..(i-1)}
					end
				Result := Result.extended (i)
				t := t.tail
			variant
				n: natural_as_integer (t.n)
			end
		end

	term alias "[]" (i: NATURAL): like term_anchor
			-- <Precursor>
		local
			j: NATURAL
			t: like trimmed_tuple_anchor
		do
			from
				j := i
				t := Current
			invariant
				is_not_unit: not t.is_unit -- t.n = j ⊇ 1
				looking_up: eq (Current [i], t.tail.indexed_left_reduced (t.first, 2, agent indexed_term (j, ?, ?, ?)))
			until
				j = 1
			loop
				j := j - 1
				t := t.tail
			variant
				j: natural_as_integer (j)
			end
				check
					is_not_unit: not t.is_unit -- loop invariant
				end
			Result := t.first
		end

	terms: like terms_anchor
			-- <Precursor>
		local
			t: like trimmed_tuple_anchor
		do
			from
				create Result.make_empty
				t := Current
			invariant
				positive: right_trimmed (t.n).indices |∀ agent index_is_positive -- by definition
				small_enough: right_trimmed (t.n).indices |∀ agent index_is_less_equal (?, n) -- t.n <= n
				building_up: Result ≍ (right_trimmed (t.n).indices ↦ agent term)
			until
				t.is_unit
			loop
				Result := Result & t.first
				t := t.tail
			variant
				n: natural_as_integer (t.n)
			end
		end

	front: like trimmed_tuple_anchor
			-- <Precursor>
		do
			Result := reversed.tail.reversed
		end

	last: like term_anchor
			-- <Precursor>
		local
			t: like trimmed_tuple_anchor
		do
			from
				t := Current
			invariant
				positive: 0 < t.n -- not `t'.`is_unit'
				small_enough: t.n <= t.n -- by definition
				looking_up: eq (last, t [t.n])
			until
				t.tail.is_unit
			loop
				t := t.tail
			variant
				tail_n: natural_as_integer (t.tail.n)
			end
				check
					is_not_unit: not t.is_unit -- 0 < t.n
				end
			Result := t.first
		end

feature -- Construction

	before (a: A): like extended_tuple_anchor
			-- <Precursor>
		do
			Result := reversed.after (a).reversed
		end

	after (a: A): like extended_tuple_anchor
			-- <Precursor>
		do
			create Result.make_after (a, Current)
		end

feature -- Measurement

	n: NATURAL
			-- <Precursor>
		local
			t: like trimmed_tuple_anchor
		do
			from
				t := Current
			invariant
				counting_up: Result = term_to_index.tuple_indexed_left_reduction (right_trimmed (t.n), 0, 1, agent just_index)
			until
				t.is_unit
			loop
				Result := Result + 1
				t := t.tail
			variant
				n: natural_as_integer (t.n)
			end
		end

feature -- Output

	out: STRING
			-- <Precursor>
		do
			Result := "(" + terms_out + ")"
		ensure then
			definition: Result ~ "(" + terms_out + ")"
		end

	terms_out: STRING
			-- Terms of current n-tuple, represented as a list with no left or right delimiters.
		local
			t: like trimmed_tuple_anchor
		do
			if is_unit then
				Result := ""
			else
				from
					Result := term_out (first)
					t := tail
				invariant
					building_up: Result ~ {TRANSFORMER [A, STRING, EQ, STS_OBJECT_EQUALITY [STRING]]}.tuple_indexed_left_reduction (right_trimmed (t.n), "", 1, agent appending_term_out)
				until
					t.is_unit
				loop
					Result.append_character (',')
						check
							other_arguments_not_void: term_out (t.first) /= Void -- `term_out' definition
						end
					Result.append (term_out (t.first))
					t := t.tail
				variant
					n: natural_as_integer (t.n)
				end
			end
		ensure
			definition: Result ~ {TRANSFORMER [A, STRING, EQ, STS_OBJECT_EQUALITY [STRING]]}.tuple_indexed_left_reduction (Current, "", 1, agent appending_term_out)
		end

	term_out (a: A): STRING
			-- Terse printable representation of `a'
		do
			if attached a then
					check
						other_not_void: a.out /= Void -- {ANY}.out definition
					end
				create Result.make_from_separate (a.out)
			else
				Result := "Void"
			end
		ensure
			class
			definition: if attached a then
					Result ~ a.out
				else
					Result ~ "Void"
				end
		end

feature -- Comparison

	equals alias "≍" (t: STS_N_TUPLE [A, EQ]): BOOLEAN
			-- <Precursor>
		local
			t1: like trimmed_tuple_anchor
			t2: STS_N_TUPLE [A, EQ]
			l_eq: EQ
		do
			if Current = t or Current ~ t then
				Result := True
			elseif n = t.n then
				from
					t1 := Current
					t2 := t
					l_eq := eq -- Better to avoid creating `eq' over and over at each comparison
				invariant
					so_far_so_good: right_trimmed (t1.n).for_all (agent (t.right_trimmed (t2.n)).terms_are_equal) and
						t.right_trimmed (t2.n).for_all (agent (right_trimmed (t1.n)).terms_are_equal)
				until
					(t1.is_unit or t2.is_unit) or else not l_eq (t1.first, t2.first)
				loop
					t1 := t1.tail
					t2 := t2.tail
				variant
					n: natural_as_integer (t1.n)
				end
				Result := t1.is_unit and t2.is_unit
			end
		end

feature -- Quantifier

	exists (p: PREDICATE [NATURAL, A]): BOOLEAN
			-- <Precursor>
		local
			i: NATURAL
			t: like trimmed_tuple_anchor
		do
			from
				i := 1
				t := Current
			invariant
				not_yet: right_trimmed (t.n).indices |∄ agent predicate_holds_to_term (p, ?)
			until
				t.is_unit or else p (i, t.first)
			loop
				i := i + 1
				t := t.tail
			variant
				n: natural_as_integer (t.n)
			end
			Result := not t.is_unit
		end

	for_all (p: PREDICATE [NATURAL, A]): BOOLEAN
			-- <Precursor>
		local
			i: NATURAL
			t: like trimmed_tuple_anchor
		do
			from
				i := 1
				t := Current
			invariant
				so_far_so_good: right_trimmed (t.n).indices |∀ agent predicate_holds_to_term (p, ?)
			until
				t.is_unit or else not p (i, t.first)
			loop
				i := i + 1
				t := t.tail
			variant
				n: natural_as_integer (t.n)
			end
			Result := t.is_unit
		end

feature -- Operation

	left_trimmed (m: NATURAL): like trimmed_tuple_anchor
			-- <Precursor>
		local
			i: NATURAL
		do
			from
				Result := Current
				i := m
			invariant
				trimming_down: Result ≍ suffix (n - (m - i).min (n))
			until
				Result.is_unit or i = 0
			loop
				Result := Result.tail
				i := i - 1
			variant
				i: natural_as_integer (i)
			end
		end

	right_trimmed (m: NATURAL): like trimmed_tuple_anchor
			-- <Precursor>
		do
			Result := reversed.left_trimmed (m)
			Result := Result.reversed
		end

	prefix (m: NATURAL): like trimmed_tuple_anchor
			-- <Precursor>
		local
			l_n: NATURAL
		do
			l_n := n
			Result := right_trimmed (l_n - m.min (l_n))
		end

	suffix (m: NATURAL): like trimmed_tuple_anchor
			-- <Precursor>
		local
			l_n: NATURAL
		do
			l_n := n
			Result := left_trimmed (l_n - m.min (l_n))
		end

	reversed: like tuple_anchor
			-- <Precursor>
		local
			t: N_TUPLE [A, EQ]
		do
			from
				Result := unit
				t := Current
			invariant
				n: Result.n = right_trimmed (t.n).n
				terms: Result.for_all (agent (right_trimmed (t.n)).reversed_terms_are_equal)
			until
				t.is_unit
			loop
				Result := Result.after (t.first)
				t := t.tail
			variant
				n: natural_as_integer (t.n)
			end
		end

	replacing (i: NATURAL; a: A): like tuple_anchor
			-- <Precursor>
		local
			before_i, t: like trimmed_tuple_anchor
			j: NATURAL
		do
			from
				before_i := unit
				t := Current
				j := i
			invariant
				t_n: t.n = n - (i - j)
				building_before_i_up: before_i.reversed ≍ right_trimmed (t.n)
			until
				j = 1
			loop
					check
						is_not_unit: not t.is_unit -- t.n = n - (i - j) > 0
					end
				before_i := before_i.after (t.first)
				t := t.tail
				j := j - 1
			variant
				t_n: natural_as_integer (t.n)
			end
				check
					is_not_unit: not t.is_unit -- t.n = n - (i - j) > 0
				end
			if eq (a, t.first) then
				Result := Current
			else
				from
					Result := t.tail.after (a)
					t := before_i
				invariant
					building_up: Result ≍ (prefix (i - 1).left_trimmed (t.n).before (a) + suffix (n - i))
				until
					t.is_unit
				loop
					Result := Result.after (t.first)
					t := t.tail
				variant
					n: natural_as_integer (t.n)
				end
			end
		end

	concatenated alias "+" (t: STS_N_TUPLE [A, EQ]): like extended_tuple_anchor
			-- <Precursor>
		local
			reversed_prefix: like tuple_anchor
		do
			from
				Result := converted (t)
				reversed_prefix := reversed
			invariant
				n: Result.n = (n - reversed_prefix.n) + t.n
				prefix: Result.prefix (n - reversed_prefix.n) ≍ reversed.prefix (n - reversed_prefix.n).reversed
				suffix: Result.suffix (t.n) ≍ t
			until
				reversed_prefix.is_unit
			loop
				Result := Result.after (reversed_prefix.first)
				reversed_prefix := reversed_prefix.tail
			variant
				n: natural_as_integer (reversed_prefix.n)
			end
		end

feature -- Transformation

	proper_right_reduced (f: FUNCTION [A, A, A]): A
			-- <Precursor>
		do
				check
					is_not_unit: not reversed.is_unit -- not is_unit
				end
			Result := reversed.proper_left_reduced (
				agent  (acc: A; x: A; ia_f: FUNCTION [A, A, A]): A
					do
						Result := ia_f (x, acc)
					end (?, ?, f)
				)
		end

feature -- Conversion

	natural_as_integer (a_n: NATURAL): INTEGER_64
			-- `a_n' converted into an integer value
			--| todo: make it more flexible.
			--| todo: DRY: {SET}.natural_as_integer
		do
			Result := a_n.as_integer_64
		ensure
			class
			definition: Result = a_n.as_integer_64
		end

	as_singleton: SET [STS_N_TUPLE [A, EQ], STS_N_TUPLE_EQUALITY [A, EQ]]
			-- <Precursor>
		do
			create Result.make_singleton (Current)
		end

	first_tail: ORDERED_PAIR [A, STS_N_TUPLE [A, EQ], EQ, STS_N_TUPLE_EQUALITY [A, EQ]]
			-- <Precursor>
		do
			create Result.make (first, tail)
		end

	front_last: ORDERED_PAIR [STS_N_TUPLE [A, EQ], A, STS_N_TUPLE_EQUALITY [A, EQ], EQ]
			-- <Precursor>
		do
			create Result.make (front, last)
		end

feature -- Factory

	unit: like trimmed_tuple_anchor
			-- <Precursor>
		do
			from
				Result := Current
			invariant
				looking_up: unit = Result.unit
			until
				Result.is_unit
			loop
				Result := Result.tail
			variant
				n: natural_as_integer (Result.n)
			end
		end

	converted (t: STS_N_TUPLE [A, EQ]): like tuple_anchor
			-- `t' converted to an n-tuple like current one
		local
			l_t: STS_N_TUPLE [A, EQ]
		do
			if attached {like tuple_anchor} t as ct then
				Result := ct
			else
				from
					Result := unit
					l_t := t
				invariant
					building_up: Result ≍ t.right_trimmed (l_t.n).reversed
				until
					l_t.is_unit
				loop
					Result := Result.after (l_t.first)
					l_t := l_t.tail
				variant
					n: natural_as_integer (l_t.n)
				end
				Result := Result.reversed
			end
		ensure
			definition: Result ≍ t
		end

	converted_term (a: A): like term_anchor
			-- `a' converted to a term like `term'
			-- TODO: Test it.
			-- TODO: DRY ({SET}.converted_element)
		do
			Result := a
		ensure
			definition: eq (Result, a)
		end

	converted_terms (s: STS_SET [A, EQ]): like terms_anchor
			-- <Precursor>
		do
			if attached {like terms_anchor} s as ct then
				Result := ct
			else
				create Result.make_empty
					check
						is_disjoint: Result.is_disjoint (s) -- Result.is_empty
					end
				Result := Result.batch_extended (s)
			end
		ensure then
			class
		end

feature -- Reduction

	indexed_term (i: NATURAL; acc: A; j: NATURAL_32; x_j: A): A
			-- <Precursor>
		do
			if j = i then
				Result := x_j
			else
				Result := acc
			end
		ensure then
			class
		end

	appending_term_out (acc: STRING; i: NATURAL; x: A): STRING
			-- `term_out' (`x') appended, as a new list entry, to `acc'.
		do
			if i = 1 then
				Result := term_out (x)
			else
				Result := acc + "," + term_out (x)
			end
		ensure
			class
			first: i = 1 implies Result ~ term_out (x)
			tail: i > 1 implies Result ~ acc + "," + term_out (x)
		end

feature -- Transformer

	term_to_term: TRANSFORMER [A, A, EQ, EQ]
			-- <Precursor>
		do
			create Result
		ensure then
			class
		end

	term_to_index: TRANSFORMER [A, NATURAL, EQ, STS_OBJECT_EQUALITY [NATURAL]]
			-- <Precursor>
		do
			create Result
		ensure then
			class
		end

feature -- Anchor

	term_anchor: A
			-- Anchor for objects like `term'
		require
			is_not_unit: not is_unit
		do
			Result := first
		end

	tuple_anchor,
	trimmed_tuple_anchor,
	extended_tuple_anchor: N_TUPLE [A, EQ]
			-- <Precursor>
		do
			Result := Current
		end

	terms_anchor: SET [A, EQ]
			-- <Precursor>
		do
			Result := terms
		end

feature {NONE} -- Implementation

	first_storage: detachable CELL [like term_anchor]
			-- Storage for `first'
		note
			option: stable
		attribute
		end

note
	copyright: "Copyright (c) 2012-2021, Rosivaldo Fernandes Alves"
	license: "[
		Eiffel Forum License v2
		(see http://www.eiffel.com/licensing/forum.txt)
		]"
	source: "https://launchpad.net/set-theory"

end
