﻿note
	description: "Implementation of {STS_SET}"
	author: "Rosivaldo Fernandes Alves"
	date: "$Date: Sat 2018-12-01 19:25:58 -0300 $"
	revision: "$Revision$"

class
	SET [A, EQ -> STS_EQUALITY [A] create default_create end]

inherit
	STS_SET [A, EQ]
		redefine
			has,
			with,
			without,
			cardinality,
			out,
			equals,
			is_subset,
			is_strict_subset,
			is_trivial_subset,
			is_proper_subset,
			is_disjoint,
			exists,
			exists_unique,
			exists_pair,
			exists_distinct_pair,
			for_all,
			for_all_pairs,
			for_all_distinct_pairs,
			trivial_subsets,
			complemented,
			intersected,
			subtracted,
			subtracted_symmetricaly,
			diagonal,
			crossed_tuples,
			equality_holds
		end

	DEBUG_OUTPUT
		rename
			debug_output as out
		redefine
			out
		end

create
	make_empty,
	make_singleton,
	make_extended,
	make_from_set

convert
	make_from_set ({
		STS_SET [A, EQ],
		STS_TRANSFORMABLE_SET [A, A, EQ, EQ]
		})

feature {NONE} -- Initialization

	make_empty
			-- Create an empty set, i.e. {} or ∅.
		do
			others := Current
		ensure
			is_empty: is_empty
		end

	make_singleton (a: A)
			-- Create a singleton in the form {`a'}.
		do
			create others.make_empty
			create any_storage.put (converted_element (a))
		ensure
			is_not_empty: not is_empty
			any: eq (any, a)
			others: others.is_empty
		end

	make_extended (a: A; s: STS_SET [A, EQ])
			-- Create a set whose `any' element and `others' are, respectively, `a' and `s'.
		require
			does_not_have: s ∌ a
		do
			create others.make_empty
			others := others.converted (s)
			create any_storage.put (converted_element (a))
		ensure
			is_not_empty: not is_empty
			any: eq (any, a)
			others: others ≍ s
		end

	make_from_set (s: STS_SET [A, EQ])
			-- Create a set equal to `s'.
		do
			if s.is_empty then
				make_empty
			else
					check
						does_not_have: s.others ∌ s.any -- {STS_SET} invariant
					end
				make_extended (s.any, s.others)
			end
		ensure
			same_set: Current ≍ s
		end

feature -- Primitive

	is_empty: BOOLEAN
			-- <Precursor>
		do
			Result := not attached any_storage
		ensure then
			definition: Result = not attached any_storage
		end

	any: like any_anchor
			-- <Precursor>
		do
			check
				attached_stored_any: attached any_storage -- not `is_empty'
			then
				Result := any_storage.item
			end
		ensure then
			attached_stored_any: attached any_storage -- not `is_empty'
--			definition: Result = any_storage.item -- TODO: False within REAL_UNIVERSE.
			definition: Result ~ any_storage.item
		end

	others: like subset_anchor
			-- <Precursor>

	eq: EQ
			-- <Precursor>
		do
			create Result
		ensure then
			class
		end

feature -- Membership

	has alias "∋" (a: A): BOOLEAN
			-- <Precursor>
			--| This feature is called very frequently. It is wortwhile to avoid agent indirection and recreating `eq' over and over at
			--| each comparison.
		local
			s: like subset_anchor
			l_eq: EQ
		do
			from
				s := Current
				l_eq := eq
			invariant
				not_found_yet: (Current ∖ s) |∄ agent equality_holds (a, ?)
			until
				s.is_empty or else l_eq (a, s.any)
			loop
				s := s.others
			variant
				cardinality: natural_as_integer (# s)
			end
			Result := not s.is_empty
		end

feature -- Construction

	with alias "&" (a: A): like superset_anchor
			-- <Precursor>
		do
			if Current ∋ a then
				Result := Current
			else
				Result := extended (a)
			end
		end

	without alias "/" (a: A): like subset_anchor
			-- <Precursor>
		local
			s: like subset_anchor
			l_eq: like eq
		do
			from
				Result := o
				s := Current
				l_eq := eq -- Better to avoid recreating `eq' over and over.
			invariant
				building_up: Result ≍ (Current ∖ s)
			until
				s.is_empty or else l_eq (s.any, a)
			loop
					check
						does_not_have: Result ∌ s.any -- Result ≍ (Current ∖ s)
					end
				Result := Result.extended (s.any)
				s := s.others
			variant
				cardinality: natural_as_integer (# s)
			end
			if s.is_empty then
				Result := Current
			else
					check
						unwanted_element_found: l_eq (s.any, a) -- Previous loop exit
					end
				s := s.others
				if not s.is_empty then
						check
							is_disjoint: s.is_disjoint (Result) -- Result ≍ ((Current ∖ s) / a)
						end
					Result := s.batch_extended (Result)
				end
			end
		end

	extended (a: A): like superset_anchor
			-- Current set extended with `a'
		require
			does_not_have: Current ∌ a
		do
			create Result.make_extended (a, Current)
		ensure
			is_not_empty: not Result.is_empty
			any: eq (Result.any, a)
			others: Result.others ≍ Current
		end

	batch_extended (s: STS_SET [A, EQ]): like superset_anchor
			-- Current set extended with every element in `s'
		require
			is_disjoint: is_disjoint (s)
		local
			l_s: STS_SET [A, EQ]
		do
			from
				Result := Current
				l_s := s
			invariant
				building_prefixed_s_up: Result.as_tuple.prefix (# (s ∖ l_s)).terms ≍ (s ∖ l_s)
				suffixed_current: Result.as_tuple.suffix (# Current).terms ≍ Current
			until
				l_s.is_empty
			loop
					check
						does_not_have: Result ∌ l_s.any -- Result ≍ (Current ∪ (s ∖ l_s))
					end
				Result := Result.extended (l_s.any)
				l_s := l_s.others
			variant
				cardinality: natural_as_integer (# l_s)
			end
		ensure
			prefixed_s: Result.as_tuple.prefix (# s).terms ≍ s
			suffixed_current: Result.as_tuple.suffix (# Current).terms ≍ Current
		end

feature -- Measurement

	cardinality alias "#": like natural_anchor
			-- <Precursor>
		local
			s: like subset_anchor
		do
			from
				s := Current
			invariant
				counting_up: Result = element_to_natural.set_reduction (Current ∖ s, 0, agent natural_successor)
			until
				s.is_empty
			loop
				Result := Result + 1
				s := s.others
			variant
				cardinality: natural_as_integer (# s)
			end
		end

feature -- Output

	out: STRING
			-- <Precursor>
		do
			Result := "{" + as_tuple.terms_out + "}"
		ensure then
			definition: Result ~ "{" + as_tuple.terms_out + "}"
		end

feature -- Comparison

	equals alias "≍" (s: STS_SET [A, EQ]): BOOLEAN
			-- <Precursor>
			--| This feature is called very frequently. It is wortwhile to avoid agent indirection and recreating `eq' over and over at
			--| each comparison.
		do
			Result := Current = s or Current ~ s or # Current = # s and then subset_relation_holds (Current, s)
		end

	is_subset alias "⊆" (s: STS_SET [A, EQ]): BOOLEAN
			-- <Precursor>
			--| This feature is called very frequently. It is wortwhile to avoid agent indirection and recreating `eq' over
			--| and over at each comparison.
		do
			Result := Current = s or Current ~ s or # Current ≤ # s and then subset_relation_holds (Current, s)
		end

	is_strict_subset alias "⊂" (s: STS_SET [A, EQ]): BOOLEAN
			-- <Precursor>
			--| This feature is called somewhat frequently. It is wortwhile to avoid agent indirection and recreating `eq' over and over
			--| at each comparison.
		do
			Result := Current /= s and Current /~ s and # Current < # s and then subset_relation_holds (Current, s)
		end

	is_trivial_subset (s: STS_SET [A, EQ]): BOOLEAN
			-- <Precursor>
		do
			Result := is_empty or Current ≍ s
		end

	is_proper_subset (s: STS_SET [A, EQ]): BOOLEAN
			-- <Precursor>
		do
			Result := not is_empty and Current ⊂ s
		end

	is_disjoint (s: STS_SET [A, EQ]): BOOLEAN
			-- <Precursor>
		local
			s1, s2: STS_SET [A, EQ]
			l_eq: EQ
		do
			Result := True
			if not is_empty and not s.is_empty then
				from
					l_eq := eq -- Better to avoid recreating `eq' over and over.
					s1 := Current
				invariant
					definition: Result = ((Current ∖ s1) ∩ s) ≍ o
				until
					not Result or s1.is_empty
				loop
					from
						s2 := s
					invariant
						not_found_yet: (s ∖ s2) ∌ s1.any
					until
						s2.is_empty or else l_eq (s1.any, s2.any)
					loop
						s2 := s2.others
					variant
						cardinality_2: natural_as_integer (# s2)
					end
					Result := s2.is_empty
					s1 := s1.others
				variant
					cardinality_1: natural_as_integer (# s1)
				end
			end
		end

feature -- Quantifier

	exists alias "|∃" (p: PREDICATE [A]): BOOLEAN
			-- <Precursor>
		local
			s: SET [A, EQ]
		do
			from
				s := Current
			invariant
				not_yet: not element_to_boolean.set_reduction (Current ∖ s, False, agent disjunction (?, p, ?))
			until
				s.is_empty or else p (s.any)
			loop
				s := s.others
			variant
				cardinality: natural_as_integer (# s)
			end
			Result := not s.is_empty
		end

	exists_unique alias "|∃!" (p: PREDICATE [A]): BOOLEAN
			-- <Precursor>
		local
			s: SET [A, EQ]
		do
			from
				s := Current
			invariant
				not_yet: ((Current ∖ s) | p) ≍ o
			until
				s.is_empty or else p (s.any)
			loop
				s := s.others
			variant
				cardinality: natural_as_integer (# s)
			end
			if not s.is_empty then
					check
						found: p (s.any) -- loop termination
					end
				Result := s.others |∄ p
			end
		end

	exists_pair (p: PREDICATE [A, A]): BOOLEAN
			-- <Precursor>
		local
			s1, s2: like subset_anchor
		do
			from
				s1 := Current
			invariant
				definition: Result = ((Current ∖ s1).crossed (Current)).exist_xy (p)
			until
				Result or s1.is_empty
			loop
				from
					s2 := Current
				invariant
					not_yet: (singleton (s1.any) × (Current ∖ s2)).do_not_exist_xy (p)
				until
					s2.is_empty or else p (s1.any, s2.any)
				loop
					s2 := s2.others
				variant
					cardinality_2: natural_as_integer (# s2)
				end
				Result := not s2.is_empty
				s1 := s1.others
			variant
				cardinality_1: natural_as_integer (# s1)
			end
		end

	exists_distinct_pair (p: PREDICATE [A, A]): BOOLEAN
			-- <Precursor>
		local
			s1, s2: like subset_anchor
			l_eq: like eq
		do
			from
				s1 := Current
				l_eq := eq -- Better to avoid recreating `eq' over and over.
			invariant
				definition: Result = ((Current ∖ s1) × Current ∖ ∆ (Current ∖ s1)).exist_xy (p)
			until
				Result or s1.is_empty
			loop
				from
					s2 := Current
				invariant
					not_yet: (singleton (s1.any) × (Current ∖ s2) ∖ ∆ (Current ∖ s2)).do_not_exist_xy (p)
				until
					s2.is_empty or else not l_eq (s1.any, s2.any) and p (s1.any, s2.any)
				loop
					s2 := s2.others
				variant
					cardinality_2: natural_as_integer (# s2)
				end
				Result := not s2.is_empty
				s1 := s1.others
			variant
				cardinality_1: natural_as_integer (# s1)
			end
		end

	for_all alias "|∀" (p: PREDICATE [A]): BOOLEAN
			-- <Precursor>
		local
			s: like subset_anchor
		do
			from
				s := Current
			invariant
				so_far_so_good: element_to_boolean.set_reduction (Current ∖ s, True, agent conjunction (?, p, ?))
			until
				s.is_empty or else not p (s.any)
			loop
				s := s.others
			variant
				cardinality: natural_as_integer (# s)
			end
			Result := s.is_empty
		end

	for_all_pairs (p: PREDICATE [A, A]): BOOLEAN
			-- <Precursor>
		local
			s1, s2: like subset_anchor
		do
			from
				Result := True
				s1 := Current
			invariant
				definition: Result = ((Current ∖ s1) × Current).for_all_xy (p)
			until
				not Result or s1.is_empty
			loop
				from
					s2 := Current
				invariant
					so_far_so_good: (singleton (s1.any) × (Current ∖ s2)).for_all_xy (p)
				until
					s2.is_empty or else not p (s1.any, s2.any)
				loop
					s2 := s2.others
				variant
					cardinality_2: natural_as_integer (# s2)
				end
				Result := s2.is_empty
				s1 := s1.others
			variant
				cardinality_1: natural_as_integer (# s1)
			end
		end

	for_all_distinct_pairs (p: PREDICATE [A, A]): BOOLEAN
			-- <Precursor>
		local
			s1, s2: like subset_anchor
			l_eq: like eq
		do
			from
				Result := True
				s1 := Current
				l_eq := eq -- Better to avoid recreating `eq' over and over.
			invariant
				definition: Result = ((Current ∖ s1) × Current ∖ ∆ (Current ∖ s1)).for_all_xy (p)
			until
				not Result or s1.is_empty
			loop
				from
					s2 := Current
				invariant
					so_far_so_good: (singleton (s1.any) × (Current ∖ s2) ∖ ∆ (Current ∖ s2)).for_all_xy (p)
				until
					s2.is_empty or else not l_eq (s1.any, s2.any) and not p (s1.any, s2.any)
				loop
					s2 := s2.others
				variant
					cardinality_2: natural_as_integer (# s2)
				end
				Result := s2.is_empty
				s1 := s1.others
			variant
				cardinality_1: natural_as_integer (# s1)
			end
		end

feature -- Operation

	subsets,
	powerset: SUBSET_FAMILY [A, EQ]
			-- <Precursor>
		local
			s: SET [A, EQ]
			sf: SUBSET_FAMILY [A, EQ]
		do
			from
				create Result.make_empty (Current)
					check
						not_in_family: o ∉ Result -- Result.is_empty
						already_a_subset: o ⊆ Result.superset -- By definition
					end
				Result := Result.extended (o)
				s := Current
			invariant
				building_up: Result ≍ {like element_to_sets}.set_reduction
					(Current ∖ s, o.as_singleton, agent sets_plus_same_sets_with_element)
				superset: Result.superset ≍ Current
			until
				s.is_empty
			loop
				from
					sf := Result
				invariant
					preceding: attached {like element_to_sets}.set_reduction
						(Current ∖ s, o.as_singleton, agent sets_plus_same_sets_with_element) as preceding
					succeeding: Result ≍ (
						preceding ∪ {like element_to_sets}.set_reduction
							(singleton (s.any), preceding ∖ sf, agent sets_plus_same_sets_with_element)
						)
					same_superset: Result.superset ≍ Current
				until
					sf.is_empty
				loop
						check
							does_not_have: sf.any ∌ s.any -- ∀ subset: sf ¦ (subset ⊆ (Current ∖ s)) ⇐ Loops invariants
							not_in_family: -- ∀ subset: Result ¦ (subset ∈ (preceding ∖ sf)) ⇐ `succeeding' loop invariant.
								sf.any.extended (s.any) ∉ Result
							already_a_subset: -- sf.superset = Result.superset = Current and s.any ∈ Current
								sf.any.extended (s.any) ⊆ Result.superset
						end
					Result := Result.extended (sf.any.extended (s.any))
					sf := sf.others
				variant
					sf_cardinality: natural_as_integer (# sf)
				end
				s := s.others
			variant
				s_cardinality: natural_as_integer (# s)
			end
		end

--	subsets,
--	powerset: SUBSET_FAMILY [A, EQ]
--			-- <Precursor>
--		local
--			s: SET [A, EQ]
--			sf: SUBSET_FAMILY [A, EQ]
--			sets_plus_same_sets_without_element: FUNCTION [
--				STS_SET [STS_SET [A, EQ], STS_SET_EQUALITY [A, EQ]], A,
--				STS_SET [STS_SET [A, EQ], STS_SET_EQUALITY [A, EQ]]
--				]
--		do
--			sets_plus_same_sets_without_element := agent (acc: STS_SET [STS_SET [A, EQ], STS_SET_EQUALITY [A, EQ]]; x: A): STS_SET [STS_SET [A, EQ], STS_SET_EQUALITY [A, EQ]]
--						-- Union of `acc' with `acc' mapped by {STS_SET [A, EQ]}.without (`x')
--				do
--					Result := acc ∪ (acc ↦ agent {STS_SET [A, EQ]}.without (x))
--				end
--			from
--				Result := as_singleton
--				s := Current
--			invariant
--				building_up: Result ≍ {like element_to_sets}.set_reduction (Current ∖ s, as_singleton, sets_plus_same_sets_without_element)
--				superset: Result.superset ≍ Current
--			until
--				s.is_empty
--			loop
--				from
--					sf := Result
--				invariant
--					preceding: attached {like element_to_sets}.set_reduction (Current ∖ s, as_singleton, sets_plus_same_sets_without_element) as preceding
--					succeeding: Result ≍ (preceding ∪ {like element_to_sets}.set_reduction (singleton (s.any), preceding ∖ sf, sets_plus_same_sets_without_element))
--					same_superset: Result.superset ≍ Current
--				until
--					sf.is_empty
--				loop
--						check
--								-- loop invariants imply that only (preceding ∖ sf) was processed so far.
--							does_not_have: Result ∌ (sf.any / s.any)
--							already_a_subset: (sf.any / s.any) ⊆ Result.superset -- sf.superset = Result.superset
--						end
--					Result := Result.extended (sf.any / s.any)
--					sf := sf.others
--				variant
--					sf_cardinality: natural_as_integer (# sf)
--				end
--				s := s.others
--			variant
--				s_cardinality: natural_as_integer (# s)
--			end
--		end

	trivial_subsets: like subsets.subfamily_anchor
			-- <Precursor>
		do
			Result := as_singleton
			if not is_empty then
					check
						does_not_have: Result ∌ o -- Result = {Current} /= {o}
						already_a_subset: o ⊆ Result.superset -- o.is_empty
					end
				Result := Result.extended (o)
			end
		end

	filtered alias "|" (p: PREDICATE [A]): like subset_anchor
			-- <Precursor>
		local
			s, previous_result, last_segment: like subset_anchor
		do
			from
				Result := o
				s := Current
			invariant
				every_compliant_element: (Current ∖ s) |∀ agent (Current ∖ s).iff (p, agent Result.has, ?)
				nothing_else: Result ⊆ (Current ∖ s)
			until
				s.is_empty
			loop
				if p (s.any) then
					from
						previous_result := Result
							check
								does_not_have: Result ∌ s.any -- Result ⊆ (Current ∖ s)
							end
						Result := Result.extended (s.any)
						last_segment := s
						s := s.others
					invariant
						previous_every_compliant_element: (Current ∖ last_segment) |∀ agent (Current ∖ last_segment).iff (p, agent previous_result.has, ?)
						previous_nothing_else: previous_result ⊆ (Current ∖ last_segment)
						compliant_segment: (last_segment ∖ s) |∀ p
						previous_result_is_disjoint: previous_result.is_disjoint (last_segment ∖ s) -- previous_result ⊆ (Current ∖ last_segment)
						building_up: Result ≍ previous_result.batch_extended (last_segment ∖ s)
					until
						s.is_empty or else not p (s.any)
					loop
							check
								loop_does_not_have: Result ∌ s.any
									-- previous_result.is_disjoint (last_segment ∖ s)
									-- Result ≍ previous_result.batch_extended (last_segment ∖ s)
							end
						Result := Result.extended (s.any)
						s := s.others
					variant
						inner_cardinality: natural_as_integer (# s)
					end
					if s.is_empty then
							check
								last_segment_is_disjoint: last_segment.is_disjoint (previous_result) -- previous_result.is_disjoint (last_segment ∖ s)
							end
							-- (last_segment ∖ s) |∀ p)
						Result := last_segment.batch_extended (previous_result)
					else
						s := s.others
					end
				else
					s := s.others
				end
			variant
				outer_cardinality: natural_as_integer (# s)
			end
		end

	complemented alias "∁" (s: STS_SET [A, EQ]): like set_anchor
			-- <Precursor>
		do
			if is_empty then
				Result := converted (s)
			else
				Result := Precursor {STS_SET}(s)
			end
		end

	united alias "∪" (s: STS_SET [A, EQ]): like superset_anchor
			-- <Precursor>
		local
			diff: STS_SET [A, EQ]
		do
			diff := s ∖ Current
				check
					is_disjoint: is_disjoint (diff) -- diff = s ∖ Current
				end
			Result := batch_extended (diff)
		end

	intersected alias "∩" (s: STS_SET [A, EQ]): like subset_anchor
			-- <Precursor>
		do
			if s.is_empty then
				Result := o
			else
				Result := Precursor {STS_SET}(s)
			end
		end

	subtracted alias "∖" (s: STS_SET [A, EQ]): like subset_anchor
			-- <Precursor>
		do
			if s.is_empty then
				Result := Current
			else
				Result := Precursor {STS_SET}(s)
			end
		end

	subtracted_symmetricaly alias "⊖" (s: STS_SET [A, EQ]): like set_anchor
			-- <Precursor>
		do
				check
					is_disjoint: (Current ∖ s).is_disjoint (s ∖ Current) -- (Current ∖ s) ⊆ Current and Current.is_disjoint (s ∖ Current)
				end
			Result := (Current ∖ s).batch_extended (s ∖ Current)
		end

	diagonal alias "∆": like diagonal_anchor
			-- <Precursor>
		local
			s: like subset_anchor
		do
			from
				Result := element_to_element.o
				s := Current
			invariant
				building_up: Result ≍ ((× (Current ∖ s)).filtered_xy (agent equality_holds))
			until
				s.is_empty
			loop
					check
						does_not_have: Result ∌ pair (s.any, s.any) -- Result ≍ (²(Current ∖ s) ｜ˣʸ agent equality_holds)
					end
				Result := Result.extended (pair (s.any, s.any))
				s := s.others
			variant
				cardinality: natural_as_integer (# s)
			end
		end

	crossed_tuples (ts: STS_SET [STS_N_TUPLE [A, EQ], STS_N_TUPLE_EQUALITY [A, EQ]]): like raised
			-- <Precursor>
		local
			s: like subset_anchor
		do
			create Result.make_empty
			if not ts.is_empty then
				from
					s := Current
				invariant
					building_up: Result ≍ tuple_extensor.set_map (element_to_tuple.cartesian_product (Current ∖ s, ts), agent second_after_first)
				until
					s.is_empty
				loop
						check
							is_disjoint: Result.is_disjoint (singleton_crossed_tuples (s.any, ts))
								-- Result ≍ tuple_extensor.set_map (
								-- 	element_to_tuple.cartesian_product (Current ∖ s, ts),
								--	agent second_after_first
								--	)
						end
					Result := Result.batch_extended (singleton_crossed_tuples (s.any, ts))
					s := s.others
				variant
					cardinality: natural_as_integer (# s)
				end
			end
		end

	singleton_crossed_tuples (x: A; ts: STS_SET [STS_N_TUPLE [A, EQ], STS_N_TUPLE_EQUALITY [A, EQ]]): like raised
			-- {`x'} "crossed" with `ts', i.e. set with every tuple in `ts' preceded by `x'.
		local
			l_ts: STS_SET [STS_N_TUPLE [A, EQ], STS_N_TUPLE_EQUALITY [A, EQ]]
		do
			from
				create Result.make_empty
				l_ts := ts
			invariant
				building_up: Result ≍ ((ts ∖ l_ts) ↦ agent {STS_N_TUPLE [A, EQ]}.after (x))
			until
				l_ts.is_empty
			loop
					check
						does_not_have: Result ∌ l_ts.any.after (x) -- Result ≍ ((ts ∖ l_ts) ↦ agent {STS_N_TUPLE [A, EQ]}.after (x))
					end
				Result := Result.extended (l_ts.any.after (x))
				l_ts := l_ts.others
			variant
				cardinality: (# l_ts).as_integer_64
			end
		ensure
			class
			definition: Result ≍ (ts ↦ agent {STS_N_TUPLE [A, EQ]}.after(x))
		end

feature -- Conversion

	natural_as_integer (n: NATURAL): INTEGER_64
			-- `n' converted into an integer value
			-- todo: make it more flexible.
		do
			Result := n.as_integer_64
		ensure
			class
			definition: Result = n.as_integer_64
		end

	as_singleton: like subsets.subfamily_anchor
			-- <Precursor>
		do
			create Result.make_singleton (Current)
		end

	n_times (n: NATURAL): SET_TUPLE [A, EQ]
			-- <Precursor>
		do
			create Result.make_n (n, Current)
		end

	as_tuple: like n_tuple_anchor
			-- <Precursor>
		local
			s: like subset_anchor
		do
			from
				Result := unit
				s := Current
			invariant
					-- TODO: VOID target when `as_tuple' gets redefined as a once routine and calls this one as a precursor.
--				building_up: Result.reversed ≍ as_tuple.right_trimmed (# s)

					-- NOTICE: This invariant is highly dependent upon `subtracted' implementation.
				building_up: not s.is_empty ⇒ Result ≍ element_to_tuple.set_reduction (Current ∖ s, unit, agent {STS_N_TUPLE [A, EQ]}.before)
				built: s.is_empty ⇒ Result ≍ element_to_tuple.set_reduction (Current ∖ s, unit, agent {STS_N_TUPLE [A, EQ]}.after)
			until
				s.is_empty
			loop
				Result := Result.after (s.any)
				s := s.others
			variant
				cardinality: natural_as_integer (# s)
			end
			Result := Result.reversed
		end

feature -- Factory

	o,
	empty_set: like subset_anchor
			-- <Precursor>
			--| Notice that this implementation aims to prevent the creation of redundant empty sets, so preserving system space at the cost of some time.
		do
			from
				Result := Current
			invariant
				looking_up: o = Result.o
			until
				Result.is_empty
			loop
				Result := Result.others
			variant
				cardinality: natural_as_integer (# Result)
			end
		end

	singleton (a: A): like set_anchor
			-- <Precursor>
		do
				check
					does_not_have: o ∌ a -- o.is_empty
				end
			Result := o.extended (a)
		end

	converted (s: STS_SET [A, EQ]): like set_anchor
			-- <Precursor>
		do
			if attached {like set_anchor} s as cs then
				Result := cs
			else
					check
						is_disjoint: o.is_disjoint (s) -- o.is_empty
					end
				Result := o.batch_extended (s)
			end
		end

	converted_element (a: A): like any_anchor
			-- `a' converted to an element like `any'
		do
			Result := a
		ensure
			definition: eq (Result, a)
		end

	pair (a, b: A): HOMOGENEOUS_ORDERED_PAIR [A, EQ]
			-- Ordered pair in the form (`a', `b')
		do
			create Result.make (a, b)
		ensure
			class
			first: Result.eq (Result.a, a)
			second: Result.eq (Result.b, b)
		end

	unit: like n_tuple_anchor
			-- <Precursor>
		do
			create Result.make_unit
		ensure then
			class
		end

feature -- Predicate

	equality_holds (x1, x2: A): BOOLEAN
			-- <Precursor>
		do
			Result := eq (x1, x2)
		ensure then
			class
		end

	subset_relation_holds (s1, s2: STS_SET [A, EQ]): BOOLEAN
			-- Does the relation `s1' ⊆ `s1' hold?
			--| Intended to help avoiding the indirection of agents and the repeated creation of `eq' at each element comparison.
		local
			l_eq: EQ
			l_s1, l_s2: STS_SET [A, EQ]
			a: A
		do
			from
				Result := True
				l_eq := eq
				l_s1 := s1
			invariant
				definition: Result = ((s1 ∖ l_s1) ⊆ s2)
			until
				not Result or l_s1.is_empty
			loop
				from
					a := l_s1.any
					Result := False
					l_s2 := s2
				invariant
					not_found_yet: (s2 ∖ l_s2) ∌ a
				until
					l_s2.is_empty or else l_eq (a, l_s2.any)
				loop
					l_s2 := l_s2.others
				variant
					cardinality_2: natural_as_integer (# l_s2)
				end
				Result := not l_s2.is_empty
				l_s1 := l_s1.others
			variant
				cardinality_1: natural_as_integer (# l_s1)
			end
		ensure
			class
			definition: Result = (s1 ⊆ s2)
		end

feature -- Transformer

	element_to_element: TRANSFORMER [A, A, EQ, EQ]
			-- <Precursor>
		do
			create Result
		ensure then
			class
		end

	element_to_natural: TRANSFORMER [A, NATURAL, EQ, STS_OBJECT_EQUALITY [NATURAL]]
			-- <Precursor>
		do
			create Result
		ensure then
			class
		end

	element_to_boolean: TRANSFORMER [A, BOOLEAN, EQ, STS_OBJECT_EQUALITY [BOOLEAN]]
			-- <Precursor>
		do
			create Result
		ensure then
			class
		end

	element_to_sets: TRANSFORMER [
		A, STS_SET [STS_SET [A, EQ], STS_SET_EQUALITY [A, EQ]],
		EQ, STS_SET_EQUALITY [STS_SET [A, EQ], STS_SET_EQUALITY [A, EQ]]
		]
			-- <Precursor>
		do
			create Result
		ensure then
			class
		end

	element_to_tuple: TRANSFORMER [A, STS_N_TUPLE [A, EQ], EQ, STS_N_TUPLE_EQUALITY [A, EQ]]
			-- <Precursor>
		do
			create Result
		ensure then
			class
		end

	tuple_extensor: TRANSFORMER [
		STS_ORDERED_PAIR [A, STS_N_TUPLE [A, EQ], EQ, STS_N_TUPLE_EQUALITY [A, EQ]],
		STS_N_TUPLE [A, EQ],
		STS_ORDERED_PAIR_EQUALITY [A, STS_N_TUPLE [A, EQ], EQ, STS_N_TUPLE_EQUALITY [A, EQ]],
		STS_N_TUPLE_EQUALITY [A, EQ]
		]
			-- <Precursor>
		do
			create Result
		ensure then
			class
		end

feature -- Anchor

	any_anchor: A
			-- Anchor for objects like `any'
		require
			is_not_empty: not is_empty
		do
			Result := any
		end

	set_anchor,
	subset_anchor,
	superset_anchor: SET [A, EQ]
			-- <Precursor>
		do
			Result := Current
		end

	product_anchor,
	square_anchor,
	diagonal_anchor: SET_OF_ORDERED_PAIRS [A, A, EQ, EQ]
			-- <Precursor>
		do
			Result := × o
		end

	n_tuple_anchor: N_TUPLE [A, EQ]
			-- <Precursor>
		do
			Result := unit
		end

feature {NONE} -- Implementation

	any_storage: detachable CELL [like any_anchor]
			-- Storage for `any'
		note
			option: stable
		attribute
		end

note
	copyright: "Copyright (c) 2012-2022, Rosivaldo Fernandes Alves"
	license: "[
		Eiffel Forum License v2
		(see http://www.eiffel.com/licensing/forum.txt)
		]"
	source: "https://launchpad.net/set-theory"

end
