﻿note
	description: "[
		Object that checks whether the properties verified within set theory hold
		for an implementation of {STS_COMPLEX_NUMBER}
		
		Notice that many straightforward algebraic properties of complex numbers
		break due to intricacies of a floating point implementation of their real
		and imaginary parts. Such breaks are documented as (many!) exceptions
		caught and handled when appropriate.
		]"
	author: "Rosivaldo Fernandes Alves"
	date: "$Date:$"
	revision: "$Revision$"

deferred class
	COMPLEX_NUMBER_PROPERTIES

inherit
	ELEMENT_PROPERTIES
		rename
			is_in_ok as element_is_in_ok,
			is_not_in_ok as element_is_not_in_ok
		end

feature -- Access

	zero: like complex_anchor
			-- The complex number 0 + 0i
		deferred
		ensure
			real: Result.a.value = 0
			imaginary: Result.b.value = 0
		end

	one: like complex_anchor
			-- The complex number 1 + 0i
		deferred
		ensure
			real: Result.a.value = 1
			imaginary: Result.b.value = 0
		end

	two: like integer_anchor -- TODO: Natural instead
			-- The integer number 2
			--| Necessary for some properties
		deferred
		ensure
			value: Result.value = 2
		end

--	three: like integer_anchor
--			-- The natural number 3
--		deferred
--		ensure
--			value: Result.value = 3
--		end

	machine_epsilon: like real_anchor
			-- The difference between 1 and the least value greater than 1 that is representable by a real number like `real_anchor'.
		do
			Result := one.a.relative_epsilon
		ensure
			definition: Result ≍ one.a.relative_epsilon
		end

	i,
	imaginary_unit: like complex_anchor
			-- The imaginary unit, i.e. the complex number 0 + 1i.
		deferred
		ensure
			real: Result.a ≍ zero.a
			imaginary: Result.b ≍ one.a
		end

	current_complex_universe: like complex_universe_anchor
			-- The current complex "Universe", i.e. set with every complex number currently in system memory.
			-- Notice that this "universe" may change from a call to another.
		deferred
		end

	current_real_universe: like real_universe_anchor
			-- The current real "Universe", i.e. set with every real number currently in system memory.
			-- Notice that this "universe" may change from a call to another.
		deferred
		end

feature -- Properties (Primitive)

	a_ok (z: STS_COMPLEX_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_COMPLEX_NUMBER}.a?
		local
			broken_definition: BOOLEAN
		do
			if not broken_definition then
				check
						good_divisor: (z + z.conjugate).divisible (two) -- two.is_invertible
					definition:  z.a.complex_equals ((z + z.conjugate) / two)
				end
			end
			Result := True
		rescue
			print (z) print ('%N')
				check
					same_good_divisor: (z + z.conjugate).divisible (two) -- two.is_invertible
				end
			if not broken_definition and z.a.complex_unequals ((z + z.conjugate) / two) then
				check
						a_good_divisor: (z.a + z.a).divisible (two) -- two.is_invertible
						b_good_divisor: (z.b - z.b).divisible (two)
					unfolded_definition_a: ((z + z.conjugate) / two).a ≍ if (z + z.conjugate) ≍ zero then
							zero.a
						elseif (z + z.conjugate).is_proper_imaginary then
							zero.a
						elseif (z + z.conjugate).is_proper_real then
							(z.a + z.a) / two
						else
							(z.a + z.a) / two
						end
					unfolded_definition_b: ((z + z.conjugate) / two).b ≍ if (z + z.conjugate) ≍ zero then
							zero.a
						elseif (z + z.conjugate).is_proper_imaginary then
							(z.b - z.b) / two
						elseif (z + z.conjugate).is_proper_real then
							zero.a
						else
							(z.b - z.b) / two
						end
					broken_definition_cause: z.a.is_finite and (z.a + z.a).is_infinite or z.b.is_nan or z.b.is_infinite
					broken_definition_a: z.a.is_finite and (z.a + z.a).is_infinite implies ((z + z.conjugate) / two).a.is_infinite
					broken_definition_b: z.b.is_nan or z.b.is_infinite implies ((z + z.conjugate) / two).b.is_nan
				end
			end
			broken_definition := True
			retry
		end

	real_ok (z: STS_COMPLEX_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_COMPLEX_NUMBER}.real?
		local
			broken_definition: BOOLEAN
		do
			if not broken_definition then
				check
						good_divisor: (z + z.conjugate).divisible (two) -- two.is_invertible
					definition:  z.real.complex_equals ((z + z.conjugate) / two)
				end
			end
			Result := True
		rescue
			print (z) print ('%N')
				check
					same_good_divisor: (z + z.conjugate).divisible (two) -- two.is_invertible
				end
			if not broken_definition and z.real.complex_unequals ((z + z.conjugate) / two) then
				check
						real_good_divisor: (z.real + z.real).divisible (two) -- two.is_invertible
						imaginary_good_divisor: (z.imaginary - z.imaginary).divisible (two)
					unfolded_definition_real: ((z + z.conjugate) / two).real ≍ if (z + z.conjugate) ≍ zero then
							zero.a
						elseif (z + z.conjugate).is_proper_imaginary then
							zero.a
						elseif (z + z.conjugate).is_proper_real then
							(z.real + z.real) / two
						else
							(z.real + z.real) / two
						end
					unfolded_definition_imaginary: ((z + z.conjugate) / two).imaginary ≍ if (z + z.conjugate) ≍ zero then
							zero.a
						elseif (z + z.conjugate).is_proper_imaginary then
							(z.imaginary - z.imaginary) / two
						elseif (z + z.conjugate).is_proper_real then
							zero.a
						else
							(z.imaginary - z.imaginary) / two
						end
					broken_definition_cause: z.real.is_finite and (z.real + z.real).is_infinite or z.imaginary.is_nan or z.imaginary.is_infinite
					broken_definition_real: z.real.is_finite and (z.real + z.real).is_infinite implies ((z + z.conjugate) / two).real.is_infinite
					broken_definition_imaginary: z.imaginary.is_nan or z.imaginary.is_infinite implies ((z + z.conjugate) / two).imaginary.is_nan
				end
			end
			broken_definition := True
			retry
		end

	b_ok (z: STS_COMPLEX_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_COMPLEX_NUMBER}.b?
		local
			broken_definition: BOOLEAN
		do
			if not broken_definition then
				check
						good_divisor: (z - z.conjugate).divisible (i ⋅ two) -- (i ⋅ two).is_invertible
					definition: z.b.complex_equals ((z - z.conjugate) / (i ⋅ two))
				end
			end
			Result := True
		rescue
				check
					same_good_divisor: (z - z.conjugate).divisible (i ⋅ two) -- (i ⋅ two).is_invertible
				end
			if not broken_definition and z.b.complex_unequals ((z - z.conjugate) / (i ⋅ two)) then
				check
						a_good_divisor: (z.b + z.b).divisible (two) -- two.is_invertible
						b_good_divisor: (z.a - z.a).divisible (two)
					unfolded_definition_a: ((z - z.conjugate) / (i ⋅ two)).a ≍ if (z - z.conjugate) ≍ zero then
							zero.a
						elseif (z - z.conjugate).is_proper_imaginary then
							(z.b + z.b) / two
						elseif (z - z.conjugate).is_proper_real then
							zero.a
						else
							(z.b + z.b) / two
						end
					unfolded_definition_b: ((z - z.conjugate) / (i ⋅ two)).b ≍ if (z - z.conjugate) ≍ zero then
							zero.a
						elseif (z - z.conjugate).is_proper_imaginary then
							zero.a
						elseif (z - z.conjugate).is_proper_real then
							(z.a - z.a) / two
						else
							(z.a - z.a) / two
						end
					broken_definition_cause: z.b.is_finite and (z.b + z.b).is_infinite or z.a.is_nan or z.a.is_infinite
					broken_definition_a: z.b.is_finite and (z.b + z.b).is_infinite implies ((z - z.conjugate) / (i ⋅ two)).a.is_infinite
					broken_definition_b: z.a.is_nan or z.a.is_infinite implies ((z - z.conjugate) / (i ⋅ two)).b.is_nan
				end
			end
			broken_definition := True
			retry
		end

	imaginary_ok (z: STS_COMPLEX_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_COMPLEX_NUMBER}.imaginary?
		local
			broken_definition: BOOLEAN
		do
			if not broken_definition then
				check
						good_divisor: (z - z.conjugate).divisible (i ⋅ two) -- (i ⋅ two).is_invertible
					definition: z.imaginary.complex_equals ((z - z.conjugate) / (i ⋅ two))
				end
			end
			Result := True
		rescue
				check
					same_good_divisor: (z - z.conjugate).divisible (i ⋅ two) -- (i ⋅ two).is_invertible
				end
			if not broken_definition and z.imaginary.complex_unequals ((z - z.conjugate) / (i ⋅ two)) then
				check
						real_good_divisor: (z.imaginary + z.imaginary).divisible (two) -- two.is_invertible
						imaginary_good_divisor: (z.real - z.real).divisible (two)
					unfolded_definition_real: ((z - z.conjugate) / (i ⋅ two)).real ≍ if (z - z.conjugate) ≍ zero then
							zero.a
						elseif (z - z.conjugate).is_proper_imaginary then
							(z.imaginary + z.imaginary) / two
						elseif (z - z.conjugate).is_proper_real then
							zero.a
						else
							(z.imaginary + z.imaginary) / two
						end
					unfolded_definition_imaginary: ((z - z.conjugate) / (i ⋅ two)).imaginary ≍ if (z - z.conjugate) ≍ zero then
							zero.a
						elseif (z - z.conjugate).is_proper_imaginary then
							zero.a
						elseif (z - z.conjugate).is_proper_real then
							(z.real - z.real) / two
						else
							(z.real - z.real) / two
						end
					broken_definition_cause: z.imaginary.is_finite and (z.imaginary + z.imaginary).is_infinite or z.real.is_nan or z.real.is_infinite
					broken_definition_real: z.imaginary.is_finite and (z.imaginary + z.imaginary).is_infinite implies
						((z - z.conjugate) / (i ⋅ two)).real.is_infinite
					broken_definition_imaginary: z.real.is_nan or z.real.is_infinite implies ((z - z.conjugate) / (i ⋅ two)).imaginary.is_nan
				end
			end
			broken_definition := True
			retry
		end

feature -- Properties (Membership)

	is_in_ok (z: STS_COMPLEX_NUMBER; s: STS_SET [STS_COMPLEX_NUMBER, STS_EQUALITY [STS_COMPLEX_NUMBER]]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_COMPLEX_NUMBER}.is_in?
		do
			check
				require_non_emptiness: z ∈ s implies not s.is_empty
			end
			Result := True
		end

	is_not_in_ok (z: STS_COMPLEX_NUMBER; s: STS_SET [STS_COMPLEX_NUMBER, STS_EQUALITY [STS_COMPLEX_NUMBER]]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_COMPLEX_NUMBER}.is_not_in?
		do
			check
				definition: z ∉ s = s ∌ z
				sufficient_emptiness: s.is_empty implies z ∉ s
			end
			Result := True
		end

feature -- Properties (Access)

	zero_ok (z1, z2: STS_COMPLEX_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_COMPLEX_NUMBER}.zero?
		local
			broken_left_absorption, broken_right_absorption, broken_dividend_absorption: BOOLEAN
		do
			if not (broken_left_absorption or broken_right_absorption or broken_dividend_absorption) then
				check
					real: z1.zero.is_real
					imaginary: z1.zero.is_imaginary
					not_proper_real: not z1.zero.is_proper_real
					not_proper_imaginary: not z1.zero.is_proper_imaginary
					not_proper_complex: not z1.zero.is_proper_complex
					equal_to_opposite: z1.zero ≍ - z1.zero
					neutral_left_term: (z1.zero + z2) ≍ z2
					neutral_right_term: (z1 + z2.zero) ≍ z1
					quasi_neutral_minuend: (z1.zero - z2) ≍ - z2
					neutral_subtrahend: (z1 - z2.zero) ≍ z1
					absorbing_left_factor: (z1.zero - z2) ≍ z1.zero
				end
			end
			if not (broken_right_absorption or broken_dividend_absorption) then
				check
					absorbing_right_factor: (z1 ⋅ z2.zero) ≍ z2.zero
				end
			end
			if not broken_dividend_absorption then
				check
					absorbing_dividend: z1.zero.divisible (z2) implies (z1.zero / z2) ≍ z1.zero
				end
			end
			Result := True
		rescue
			print (z1) print (", ")
			print (z2) print ('%N')
			if not broken_left_absorption and (z1.zero ⋅ z2) ≭ z1.zero then
				check
					absorbing_left_factor_unfolded_definition_a: (z1.zero ⋅ z2).a ≍ if z2 ≍ zero then
							zero.a
						elseif z2.is_proper_imaginary then
							- (zero.a ⋅ z2.b)
						elseif z2.is_proper_real then
							zero.a ⋅ z2.a
						else
							zero.a * z2.a - zero.a * z2.b
						end
					absorbing_left_factor_unfolded_definition_b: (z1.zero ⋅ z2).b ≍ if z2 ≍zero then
							zero.a
						elseif z2.is_proper_imaginary then
							zero.a
						elseif z2.is_proper_real then
							zero.a
						else
							zero.a * z2.b + zero.a * z2.a
						end
					broken_left_absorption:
						z2.is_proper_imaginary and not z2.b.is_finite or
						z2.is_proper_real and not z2.a.is_finite or
						z2.is_proper_complex and not (z2.a.is_finite and z2.b.is_finite)
					broken_left_absorption_a: z2.is_proper_imaginary and not z2.b.is_finite or
						z2.is_proper_real and not z2.a.is_finite or
						z2.is_proper_complex and not (z2.a.is_finite and z2.b.is_finite) implies
						(z1.zero ⋅ z2).a.is_nan
					broken_left_absorption_b: z2.is_proper_complex and not (z2.a.is_finite and z2.b.is_finite) implies (z1.zero ⋅ z2).b.is_nan
				end
				broken_left_absorption := True
				retry
			elseif not broken_right_absorption and (z1 ⋅ z2.zero) ≭ z2.zero then
				check
					absorbing_right_factor_unfolded_definition_a: (z1 ⋅ z2.zero).a ≍ if z1 ≍ zero then
							zero.a
						elseif z1.is_proper_imaginary then
							- z1.b ⋅ zero.a
						elseif z1.is_proper_real then
							z1.a ⋅ zero.a
						else
							z1.a * zero.a - z1.b * zero.a
						end
					absorbing_right_factor_unfolded_definition_b: (z1 ⋅ z2.zero).b ≍ if z1 ≍ zero then
							zero.a
						elseif z1.is_proper_imaginary then
							zero.a
						elseif z1.is_proper_real then
							zero.a
						else
							z1.a * zero.a + z1.b * zero.a
						end
					broken_right_absorption:
						z1.is_proper_imaginary and not z1.b.is_finite or
						z1.is_proper_real and not z1.a.is_finite or
						z1.is_proper_complex and not (z1.a.is_finite and z1.b.is_finite)
					broken_right_absorption_a: z1.is_proper_imaginary and not z1.b.is_finite or
						z1.is_proper_real and not z1.a.is_finite or
						z1.is_proper_complex and not (z1.a.is_finite and z1.b.is_finite) implies
						(z1 ⋅ z2.zero).a.is_nan
					broken_right_absorption_b: z1.is_proper_complex and not (z1.a.is_finite and z1.b.is_finite) implies (z1 ⋅ z2.zero).b.is_nan
				end
				broken_right_absorption := True
				retry
			elseif not broken_dividend_absorption and z1.zero.divisible (z2) and (z1.zero / z2) ≭ z1.zero then
				check
						zero_divisible_z2b: z2.is_proper_imaginary implies zero.a.divisible (z2.b) -- z2.b /= 0 ⇐ z2.is_proper_imaginary
						zero_divisible_z2a: z2.is_proper_real implies zero.a.divisible (z2.a) -- z2.a /= 0 ⇐ z2.is_proper_real
					absorbing_dividend_unfolded_definition_a: (z1.zero / z2).a ≍ if z2.is_proper_imaginary then
							(zero.a / z2.b)
						elseif z2.is_proper_real then
							(zero.a / z2.a)
						else
							proper_complex_quotient_implementation (zero, z2).a
						end
					absorbing_dividend_unfolded_definition_b: (z1.zero / z2).b ≍ if z2.is_proper_imaginary then
							zero.a
						elseif z2.is_proper_real then
							zero.a
						else
							proper_complex_quotient_implementation (zero, z2).b
						end
					broken_dividend_absorption: z2.is_proper_imaginary and z2.b.is_nan or
						z2.is_proper_real and z2.a.is_nan or
						z2.is_proper_complex and (z2.a.is_nan or z2.b.is_nan) or
						z2 ≍ zero
					broken_dividend_absorption_a:
						z2.is_proper_imaginary and z2.b.is_nan or
						z2.is_proper_real and z2.a.is_nan or
						z2.is_proper_complex and (z2.a.is_nan or
						z2.b.is_nan) or z2 ≍ zero implies
						(z1.zero / z2).a.is_nan
					broken_dividend_absorption_b: z2.b.is_proper_complex and (z1.a.is_nan or z1.b.is_nan) or z2 ≍ zero implies
						(z1.zero / z2).b.is_nan
				end
				broken_dividend_absorption := True
				retry -- TODO: goto?
			end
		end

	one_ok (z1, z2: STS_COMPLEX_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_COMPLEX_NUMBER}.one?
		do
			check
				real: z1.one.is_real
				not_imaginary: not z1.one.is_imaginary
				proper_real: z1.one.is_proper_real
				not_proper_imaginary: not z1.one.is_proper_imaginary
				not_proper_complex: not z1.one.is_proper_complex
				invertible: z1.one.is_invertible
				good_divisor: z1.divisible (z2.one)
				equal_to_inverse: z1.one.inverse ≍ z1.one
				neutral_left_factor: (z1.one ⋅ z2) ≍ z2
				neutral_right_factor: (z1 ⋅ z2.one) ≍ z1
					good_divisor: z1.divisible (z2.one) -- z2.one.real_is_invertible
				neutral_divisor: (z1 / z2.one) ≍ z1
			end
			Result := True
		rescue
			print (z1) print (", ")
			print (z2) print ('%N')
		end

	two_ok (z: STS_COMPLEX_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_COMPLEX_NUMBER}.two?
		do
			check
				definition: z.two ≍ (one.a.truncated_to_integer + one.a.truncated_to_integer)
				is_real: z.two.is_real
				is_not_imaginary: not z.two.is_imaginary
				is_proper_real: z.two.is_proper_real
				is_not_proper_imaginary: not z.two.is_proper_imaginary
				is_not_proper_complex: not z.two.is_proper_complex
			end
			Result := True
		end

	i_ok (z: STS_COMPLEX_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_COMPLEX_NUMBER}.i?
		do
			check
				definition: z.i ≍ (i⋅one + zero)
				is_not_real: not z.i.is_real
				is_imaginary: z.i.is_imaginary
				is_not_proper_real: not z.i.is_proper_real
				is_proper_imaginary: z.i.is_proper_imaginary
				is_not_proper_complex: not z.i.is_proper_complex
				negative_square: z.i ⋅ z.i ≍ - one
				negated_cube: z.i ⋅ z.i ⋅ z.i ≍ - i
				real_unit: z.i ⋅ z.i ⋅ z.i ⋅ z.i ≍ one
				fist_cubic_root: z.i ≍ (-i ⋅ -i ⋅ -i)
				product: z ⋅ z.i ≍ (i⋅z.a - z.b)
				quotient: (z / z.i) ≍ (-i⋅z.a + z.b)
			end
			Result := True
		end

	imaginary_unit_ok (z: STS_COMPLEX_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_COMPLEX_NUMBER}.imaginary_unit?
		do
			check
				definition: z.imaginary_unit ≍ (imaginary_unit⋅one + zero)
				is_not_real: not z.imaginary_unit.is_real
				is_imaginary: z.imaginary_unit.is_imaginary
				is_not_proper_real: not z.imaginary_unit.is_proper_real
				is_proper_imaginary: z.imaginary_unit.is_proper_imaginary
				is_not_proper_complex: not z.imaginary_unit.is_proper_complex
				negative_square: z.imaginary_unit ⋅ z.imaginary_unit ≍ - one
				negated_cube: z.imaginary_unit ⋅ z.imaginary_unit ⋅ z.imaginary_unit ≍ - imaginary_unit
				real_unit: z.imaginary_unit ⋅ z.imaginary_unit ⋅ z.imaginary_unit ⋅ z.imaginary_unit ≍ one
				fist_cubic_root: z.imaginary_unit ≍ (-imaginary_unit ⋅ -imaginary_unit ⋅ -imaginary_unit)
				product: z ⋅ z.imaginary_unit ≍ (imaginary_unit⋅z.a - z.b)
				quotient: (z / z.imaginary_unit) ≍ (-imaginary_unit⋅z.a + z.b)
			end
			Result := True
		end

feature -- Properties (Quality)

	is_real_ok (z: STS_COMPLEX_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for
			-- {STS_COMPLEX_NUMBER}.is_real?
		local
			broken_definition: BOOLEAN
		do
			if not broken_definition then
				check
					definition: z.is_real = z ≍ z.conjugate
				end
			end
			check
				may_be_imaginary: z.is_real and z ≍ zero implies z.is_imaginary
				is_not_proper_imaginary: z.is_real implies
					not z.is_proper_imaginary
				is_not_proper_complex: z.is_real implies not z.is_proper_complex
			end
			Result := True
		rescue
			print (z) print ('%N')
			if not broken_definition and z.is_real /= z ≍ z.conjugate then
				check
					conjugate_a: z.conjugate.a ≍ z.a
					conjugate_b: z.conjugate.b ≍ -z.b
					broken_definition_cause: z.b.is_nan
					consequence: -z.b ≍ z.b and not z.is_real
				end
			end
			broken_definition := True
			retry
		end

	is_proper_real_ok (z: STS_COMPLEX_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for
			-- {STS_COMPLEX_NUMBER}.is_proper_real?
		do
			check
				definition: z.is_proper_real = (z.is_real and not z.is_imaginary)
				is_not_imaginary: z.is_proper_real implies not z.is_imaginary
				is_not_proper_imaginary: z.is_proper_real implies
					not z.is_proper_imaginary
				is_not_proper_complex: z.is_proper_real implies
					not z.is_proper_complex
			end
			Result := True
		end

	is_imaginary_ok (z: STS_COMPLEX_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_COMPLEX_NUMBER}.is_imaginary?
		local
			broken_definition: BOOLEAN
		do
			if not broken_definition then
				check
					definition: z.is_imaginary = z ≍ -z.conjugate
				end
			end
			check
				may_be_real: z.is_imaginary and z ≍ zero implies z.is_real
				is_not_proper_real: z.is_imaginary implies not z.is_proper_real
				is_not_proper_complex: z.is_imaginary implies not z.is_proper_complex
			end
			Result := True
		rescue
			print (z) print ('%N')
			if not broken_definition and z.is_imaginary /= z ≍ -z.conjugate then
				check
					conjugate_a: -z.conjugate.a ≍ -z.a
					conjugate_b: -z.conjugate.b ≍ z.b
					broken_definition_cause: z.a.is_nan
					consequence: -z.a ≍ z.a and not z.is_imaginary
				end
				broken_definition := True
				retry
			end
		end

	is_proper_imaginary_ok (z: STS_COMPLEX_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for
			-- {STS_COMPLEX_NUMBER}.is_proper_imaginary?
		do
			check
				definition: z.is_proper_imaginary =
					(z.is_imaginary and not z.is_real)
				is_not_real: z.is_proper_imaginary implies not z.is_real
				is_not_proper_real: z.is_proper_imaginary implies
					not z.is_proper_real
				is_not_proper_complex: z.is_proper_imaginary implies
					not z.is_proper_complex
			end
			Result := True
		end

	is_proper_complex_ok (z: STS_COMPLEX_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for
			-- {STS_COMPLEX_NUMBER}.is_proper_complex?
		do
			check
				is_not_proper_real: z.is_proper_complex implies
					not z.is_proper_real
				is_not_proper_imaginary: z.is_proper_complex implies
					not z.is_proper_imaginary
			end
			Result := True
		end

	is_infinite_ok (z1, z2: STS_COMPLEX_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_COMPLEX_NUMBER}.is_infinite?
		do
			check
				absorbing_left_term: z1.is_infinite and z2.is_finite implies (z1 + z2).is_infinite
				absorbing_right_term: z1.is_finite and z2.is_infinite implies (z1 + z2).is_infinite
				absorbing_minuend: z1.is_infinite and z2.is_finite implies (z1 - z2).is_infinite
				absorbing_subtrahend: z1.is_finite and z2.is_infinite implies (z1 - z2).is_infinite
				absorbing_left_factor: z1.is_infinite and (z2.is_finite and z2 ≭ zero or z2.is_infinite) implies (z1 ⋅ z2).is_infinite
				absorbing_right_factor: (z1.is_finite and z1 ≭ zero or z1.is_infinite) and z2.is_infinite implies (z1 ⋅ z2).is_infinite
				absorbing_dividend: z1.is_infinite and z2.is_finite and z1.divisible (z2) implies (z1 / z2).is_infinite
					good_divisor: z2.is_infinite implies z1.divisible (z2) -- z2.is_invertible
				nullifying_divisor: z1.is_finite and z2.is_infinite implies (z1 / z2) ≍ zero
			end
			Result := True
		end

	is_rational_ok (z: STS_COMPLEX_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_COMPLEX_NUMBER}.is_rational?
		do
			check
				is_real: z.is_rational implies z.is_real
				is_not_proper_imaginary: z.is_rational implies not z.is_proper_imaginary
				is_not_proper_complex: z.is_rational implies not z.is_proper_complex
				is_not_infinite: z.is_rational implies not z.is_infinite
				is_finite: z.is_rational implies z.is_finite
			end
			Result := True
		end

	is_integer_ok (z: STS_COMPLEX_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_COMPLEX_NUMBER}.is_integer?
		do
			check
				is_real: z.is_integer implies z.is_real
				is_not_proper_imaginary: z.is_integer implies not z.is_proper_imaginary
				is_not_proper_complex: z.is_integer implies not z.is_proper_complex
				is_not_infinite: z.is_integer implies not z.is_infinite
				is_finite: z.is_integer implies z.is_finite
				is_rational: z.is_integer implies z.is_rational
			end
			Result := True
		end

feature -- Properties (Comparison)

	equals_ok (z1, z2, z3: STS_COMPLEX_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_COMPLEX_NUMBER}.equals?
		do
			check
				r: attached current_real_universe as r
				c: attached current_complex_universe as c
				definition: z1 ≍ z2 = (× r).for_all_xy (
					agent (x, y: STS_REAL_NUMBER; ia_z1, ia_z2: STS_COMPLEX_NUMBER): BOOLEAN
						do
							Result := (x ≍ ia_z1.a) = (x ≍ ia_z2.a) and (y ≍ ia_z1.b) = (y ≍ ia_z2.b)
						end (?, ?, z1, z2)
					)
				by_membership: (2 ^ # c) ≤ max_subsets implies
					z1 ≍ z2 = c.powerset |∀ agent (s: STS_SET [STS_COMPLEX_NUMBER, STS_COMPLEX_EQUALITY]; ia_z1, ia_z2: STS_COMPLEX_NUMBER): BOOLEAN
						do
							Result := ia_z1 ∈ s = ia_z2 ∈ s
						end (?, z1, z2)
				reflexive: z1 ≍ z1
				symmetric: z1 ≍ z2 implies z2 ≍ z1
				transitive: z1 ≍ z2 and z2 ≍ z3 implies z1 ≍ z3
				euclidian: z1 ≍ z3 and z2 ≍ z3 implies z1 ≍ z2
			end
			Result := True
		end

	unequals_ok (z1, z2: STS_COMPLEX_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_COMPLEX_NUMBER}.unequals?
		do
			check
				r: attached current_real_universe as r
				c: attached current_complex_universe as c
				definition: z1 ≭ z2 = (z1.a ≭ z2.a or z1.b ≭ z2.b)
				by_existence: z1 ≭ z2 = (× r).exist_xy (
					agent (x, y: STS_REAL_NUMBER; ia_z1, ia_z2: STS_COMPLEX_NUMBER): BOOLEAN
						do
							Result := (x ≍ ia_z1.a) /= (x ≍ ia_z2.a) or (y ≍ ia_z1.b) /= (y ≍ ia_z2.b)
						end (?, ?, z1, z2)
					)
				by_membership: (2 ^ # c) ≤ max_subsets implies z1 ≭ z2 =
					c.powerset |∃ agent (s: STS_SET [STS_COMPLEX_NUMBER, STS_COMPLEX_EQUALITY]; ia_z1, ia_z2: STS_COMPLEX_NUMBER): BOOLEAN
						do
							Result := ia_z1 ∈ s /= ia_z2 ∈ s
						end (?, z1, z2)
				irreflexive: not (z1 ≭ z1)
			end
			Result := True
		end

feature -- Properties (Operation)

	modulus_ok (z: STS_COMPLEX_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for
			-- {STS_COMPLEX_NUMBER}.modulus?
		local
			broken_order: BOOLEAN
		do
			if not broken_order then
				check
					non_negative: z.modulus ≥ zero.a
				end
			end
			Result := True
		rescue
			if not broken_order and z.modulus < zero.a then
				check
					broken_order: z.a.is_nan and not z.b.is_infinite or
						not z.a.is_infinite and z.b.is_nan
					is_nan: z.modulus.is_nan
				end
				broken_order := True
				retry
			end
		end

	abs_ok (z: STS_COMPLEX_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_COMPLEX_NUMBER}.abs?
		local
			broken_order: BOOLEAN
		do
			if not broken_order then
				check
					non_negative: z.abs >= zero.a
				end
			end
			Result := True
		rescue
			if not broken_order and z.abs < zero.a then
				check
					broken_order: z.a.is_nan and not z.b.is_infinite or not z.a.is_infinite and z.b.is_nan
					is_nan: z.abs.is_nan
				end
				broken_order := True
				retry
			end
		end

	argument_ok (z: STS_COMPLEX_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_COMPLEX_NUMBER}.argument?
		local
			l_broken_argument: BOOLEAN
		do
			if not l_broken_argument then
				check
					at_least_minus_pi: - Pi ≤ z.argument.value
					at_most_pi: z.argument.value ≤ Pi
				end
			end
			Result := True
		rescue
			if not l_broken_argument and z.argument.value < - Pi then
				check
					broken_argument: z.a.is_nan or z.b.is_nan
					is_nan: z.argument.is_nan
				end
				l_broken_argument := True
				retry
			end
		end

	phase_ok (z: STS_COMPLEX_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_COMPLEX_NUMBER}.phase?
		local
			l_broken_phase: BOOLEAN
		do
			if not l_broken_phase then
				check
					at_least_minus_pi: - Pi ≤ z.phase.value
					at_most_pi: z.phase.value ≤ Pi
				end
			end
			Result := True
		rescue
			if not l_broken_phase and z.phase.value < - Pi then
				check
					broken_phase: z.a.is_nan or z.b.is_nan
					is_nan: z.phase.is_nan
				end
				l_broken_phase := True
				retry
			end
		end

	conjugate_ok (z1, z2: STS_COMPLEX_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_COMPLEX_NUMBER}.conjugate?
	do
			check
				involutive: z1.conjugate.conjugate ≍ z1
				same_modulus: z1.conjugate.abs ≍ z1.abs
				distributes_over_sum: (z1 + z2).conjugate ≍ (z1.conjugate + z2.conjugate)
				distributes_over_subtraction: (z1 - z2).conjugate ≍ (z1.conjugate - z2.conjugate)
				distributes_over_product: (z1 ⋅ z2).conjugate ≍ (z1.conjugate ⋅ z2.conjugate)
					good_divisor: z1.divisible (z2) implies z1.conjugate.divisible (z2.conjugate) -- `conjugate' doesn't change `abs'.
				distributes_over_division: z1.divisible (z2) implies (z1 / z2).conjugate ≍ (z1.conjugate / z2.conjugate)
			end
			Result := True
		rescue
			print (z1) print (", ")
			print (z2) print ('%N')
		end

	plus_ok (z1, z2: STS_COMPLEX_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_COMPLEX_NUMBER}.plus?
		do
			check
				neutral_left_term: (zero + z1) ≍ z1
				neutral_right_term: (z1 + zero) ≍ z1
				commutative: (z1 + z2) ≍ (z2 + z1)
			end
			Result := True
		rescue
			print (z1) print (", ")
			print (z2) print ('%N')
		end

	identity_ok (z: STS_COMPLEX_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_COMPLEX_NUMBER}.identity?
		do
			check
				same_number: + z ≍ z
				as_left_term: + z ≍ (z + zero)
				as_right_term: + z ≍ (zero + z)
				as_left_factor: + z ≍ (z ⋅ one)
				as_right_factor: + z ≍ (one ⋅ z)
					good_divisor: z.divisible (one) -- one.real_is_invertible
				as_quotient: + z ≍ (z / one)
			end
			Result := True
		rescue
			print (z) print ('%N')
		end

	minus_ok (z1, z2: STS_COMPLEX_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_COMPLEX_NUMBER}.minus?
		do
			check
				quasi_neutral_minuend: (zero - z1) ≍ - z1
				neutral_subtrahend: (z1 - zero) ≍ z1
				quasi_commutative: (z1 - z2).abs ≍ (z2 - z1).abs
			end
			Result := True
		rescue
			print (z1) print (", ")
			print (z2) print ('%N')
		end

	opposite_ok (z: STS_COMPLEX_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_COMPLEX_NUMBER}.opposite?
		local
			broken_left_opposition, broken_right_opposition: BOOLEAN
		do
			if not (broken_left_opposition or broken_right_opposition) then
				check
					neutral_term: - zero ≍ zero
					opposite_left_term: ((- z) + z) ≍ zero
				end
			end
			if not (broken_right_opposition) then
				check
					opposite_right_term: (z + (- z)) ≍ zero
				end
			end
			check
				involutive: - - z ≍ z
				same_modulus: (- z).abs ≍ z.abs
				as_subtrahend: - z ≍ (zero - z)
				as_left_factor: - z ≍ (z ⋅ (- one))
				as_right_factor: - z ≍ ((- one) ⋅ z)
					divisible: z.divisible (- one) -- (- one).is_invertible
				as_quotient: - z ≍ (z / (- one))
			end
			Result := True
		rescue
			if not broken_left_opposition and ((- z) + z) ≭ zero then
				check
					opposite_left_term_unfolded_definition_a: ((- z) + z).a ≍ (- z.a + z.a)
					opposite_left_term_unfolded_definition_b: ((- z) + z).b ≍ (- z.b + z.b)
					not_finite: not z.a.is_finite or not z.b.is_finite
					broken_left_opposition_a: not z.a.is_finite implies ((- z) + z).a.is_nan
					broken_left_opposition_b: not z.b.is_finite implies ((- z) + z).b.is_nan
				end
				broken_left_opposition := True
				retry
			elseif not broken_right_opposition and (z + (- z)) ≭ zero then
				check
					opposite_right_term_unfolded_definition_a: (z + (- z)).a ≍ (z.a - z.a)
					opposite_right_term_unfolded_definition_b: (z + (- z)).b ≍ (z.b - z.b)
					not_finite: not z.a.is_finite or not z.b.is_finite
					broken_right_opposition_a: not z.a.is_finite implies (z + (- z)).a.is_nan
					broken_right_opposition_b: not z.b.is_finite implies (z + (- z)).b.is_nan
				end
				broken_right_opposition := True
				retry
			end
		end

	product_ok (z1, z2, z3: STS_COMPLEX_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for
			-- {STS_COMPLEX_NUMBER}.product?
		local
			broken_left_absorption,
			broken_right_absorption,
			broken_morphism: BOOLEAN
		do
			if
				not (
					broken_left_absorption or
					broken_right_absorption or
					broken_morphism
					)
			then
				check
					absorbing_left_factor: (zero ⋅ z1) ≍ zero
				end
			end
			if not (broken_right_absorption or broken_morphism) then
				check
					absorbing_right_factor: (z1 ⋅ zero) ≍ zero
				end
			end
			if not broken_morphism then
				check
					neutral_left_factor: (one ⋅ z1) ≍ z1
					neutral_right_factor: (z1 ⋅ one) ≍ z1
					commutative: (z1 ⋅ z2) ≍ (z2 ⋅ z1)
					quasi_homomorphic: z1.abs ≤ z2.abs implies
						(z1 ⋅ z3).abs ≤ (z2 ⋅ z3).abs
				end
			end
			Result := True
		rescue
			print (z1) print (", ")
			print (z2) print (", ")
			print (z3) print ('%N')
			if
				not broken_left_absorption and (zero ⋅ z1) ≭ zero
			then
				check
					left_absorption_unfolded_definition_a:
						(zero ⋅ z1).a ≍ if z1 ≍ zero then
							zero.a
						elseif z1.is_proper_imaginary then
							- (zero.a ⋅ z1.b)
						elseif z1.is_proper_real then
							zero.a ⋅ z1.a
						else
							zero.a * z1.a - zero.a * z1.b
						end
					left_absorption_unfolded_definition_b:
						(zero ⋅ z1).b ≍ if z1 ≍ zero then
							zero.a
						elseif z1.is_proper_imaginary then
							zero.a
						elseif z1.is_proper_real then
							zero.a
						else
							zero.a * z1.b + zero.a * z1.a
						end
					broken_left_absorption:
						not (z1.a.is_finite and z1.b.is_finite)
					broken_left_absorption_a: (
						z1.is_proper_imaginary and not z1.b.is_finite or
						z1.is_proper_real and not z1.a.is_finite or
						z1.is_proper_complex and not (
							z1.a.is_finite and z1.b.is_finite
							)
						) =
						(zero ⋅ z1).a.is_nan
					broken_left_absorption_a: (
						z1.is_proper_complex and
						not (z1.a.is_finite and z1.b.is_finite)
						) = (zero ⋅ z1).b.is_nan
				end
				broken_left_absorption := True
				retry
			elseif not broken_right_absorption and (z1 ⋅ zero) ≭ zero then
				check
					right_absorption_unfolded_definition_a:
						(z1 ⋅ zero).a ≍ if z1 ≍ zero then
							zero.a
						elseif z1.is_proper_imaginary then
							- z1.b ⋅ zero.a
						elseif z1.is_proper_real then
							z1.a ⋅ zero.a
						else
					 		z1.a * zero.a - z1.b * zero.a
						end
					right_absorption_unfolded_definition_b:
						(z1 ⋅ zero).b ≍ if z1 ≍ zero then
							zero.a
						elseif z1.is_proper_imaginary then
							zero.a
						elseif z1.is_proper_real then
							zero.a
						else
							z1.a * zero.a + z1.b * zero.a
						end
					broken_right_absorption:
						not (z1.a.is_finite and z1.b.is_finite)
					broken_right_absorption_a: (
						z1.is_proper_imaginary and not z1.b.is_finite or
						z1.is_proper_real and not z1.a.is_finite or
						z1.is_proper_complex and not (
							z1.a.is_finite and z1.b.is_finite
							)
						) =
						(z1 ⋅ zero).a.is_nan
					broken_right_absorption_b: (z1.is_proper_complex and not (z1.a.is_finite and z1.b.is_finite)) = (z1 * zero).b.is_nan
				end
				broken_right_absorption := True
				retry
			elseif
				not broken_morphism and
				z1.abs ≤ z2.abs and
				(z1 ⋅ z3).abs > (z2 ⋅ z3).abs
			then
				check
					unfolded_morphism_definition_z1z3_abs:
						(z1 * z3).abs ≍ if z1 ≍ zero and z3 ≍ zero then
							z1.hypotenuse (zero.a, zero.a)
						elseif z1 ≍ zero and z3.is_proper_imaginary then
							z1.hypotenuse (- z1.b ⋅ z3.b, zero.a)
						elseif z1 ≍ zero and z3.is_proper_real then
							z1.hypotenuse (z1.a ⋅ z3.a, zero.a)
						elseif z1 ≍ zero and z3.is_proper_complex then
							z1.hypotenuse (
								z1.a * z3.a - z1.b * z3.b,
								z1.a * z3.b + z1.b * z3.a
								)
						elseif z1.is_proper_imaginary and z3 ≍ zero then
							z1.hypotenuse (- z1.b * z3.b, zero.a)
						elseif
							z1.is_proper_imaginary and z3.is_proper_imaginary
						then
							z1.hypotenuse (- z1.b ⋅ z3.b, zero.a)
						elseif z1.is_proper_imaginary and z3.is_proper_real then
							z1.hypotenuse (zero.a, z1.b ⋅ z3.a)
						elseif
							z1.is_proper_imaginary and z3.is_proper_complex
						then
							z1.hypotenuse (- z1.b ⋅ z3.b, z1.b ⋅ z3.a)
						elseif z1.is_proper_real and z3 ≍ zero then
							z1.hypotenuse (z1.a ⋅ z3.a, zero.a)
						elseif z1.is_proper_real and z3.is_proper_imaginary then
							z1.hypotenuse (zero.a, z1.a ⋅ z3.b)
						elseif z1.is_proper_real and z3.is_proper_real then
							z1.hypotenuse (z1.a ⋅ z3.a, zero.a)
						elseif z1.is_proper_real and z3.is_proper_complex then
							z1.hypotenuse (z1.a ⋅ z3.a, z1.a ⋅ z3.b)
						elseif z1.is_proper_complex and z3 ≍ zero then
							z1.hypotenuse (
								z1.a * z3.a - z1.b * z3.b,
								z1.a * z3.b + z1.b * z3.a
								)
						elseif
							z1.is_proper_complex and z3.is_proper_imaginary
						then
							z1.hypotenuse (- z1.b ⋅ z3.b, z1.a ⋅ z3.b)
						elseif z1.is_proper_complex and z3.is_proper_real then
							z1.hypotenuse (z1.a ⋅ z3.a, z1.b ⋅ z3.a)
						else
							z1.hypotenuse (
								proper_complex_product_implementation (z1, z3).a,
								proper_complex_product_implementation (z1, z3).b
								)
						end
					unfolded_morphism_definition_z2z3_abs:
						(z2 * z3).abs ≍ if z2 ≍ zero and z3 ≍ zero then
							z2.hypotenuse (zero.a, zero.a)
						elseif z2 ≍ zero and z3.is_proper_imaginary then
							z2.hypotenuse (- z2.b ⋅ z3.b, zero.a)
						elseif z2 ≍ zero and z3.is_proper_real then
							z2.hypotenuse (z2.a ⋅ z3.a, zero.a)
						elseif z2 ≍ zero and z3.is_proper_complex then
							z2.hypotenuse (
								z2.a * z3.a - z2.b * z3.b,
								z2.a * z3.b + z2.b * z3.a
								)
						elseif z2.is_proper_imaginary and z3 ≍ zero then
							z2.hypotenuse (- z2.b ⋅ z3.b, zero.a)
						elseif
							z2.is_proper_imaginary and z3.is_proper_imaginary
						then
							z2.hypotenuse (- z2.b ⋅ z3.b, zero.a)
						elseif z2.is_proper_imaginary and z3.is_proper_real then
							z2.hypotenuse (zero.a, z2.b ⋅ z3.a)
						elseif
							z2.is_proper_imaginary and z3.is_proper_complex
						then
							z2.hypotenuse (- z2.b ⋅ z3.b, z2.b ⋅ z3.a)
						elseif z2.is_proper_real and z3 ≍ zero then
							z2.hypotenuse (z2.a ⋅ z3.a, zero.a)
						elseif z2.is_proper_real and z3.is_proper_imaginary then
							z2.hypotenuse (zero.a, z2.a ⋅ z3.b)
						elseif z2.is_proper_real and z3.is_proper_real then
							z2.hypotenuse (z2.a ⋅ z3.a, zero.a)
						elseif z2.is_proper_real and z3.is_proper_complex then
							z2.hypotenuse (z2.a ⋅ z3.a, z2.a ⋅ z3.b)
						elseif z2.is_proper_complex and z3 ≍ zero then
							z2.hypotenuse (
								z2.a * z3.a - z2.b * z3.b,
								z2.a * z3.b + z2.b * z3.a
								)
						elseif
							z2.is_proper_complex and z3.is_proper_imaginary
						then
							z2.hypotenuse (- z2.b ⋅ z3.b, z2.a ⋅ z3.b)
						elseif z2.is_proper_complex and z3.is_proper_real then
							z2.hypotenuse (z2.a ⋅ z3.a, z2.b ⋅ z3.a)
						else
							z2.hypotenuse (
								proper_complex_product_implementation (z2, z3).a,
								proper_complex_product_implementation (z2, z3).b
								)
						end
					broken_morphism_cause:
						((z1 ⋅ z3).abs - (z2 ⋅ z3).abs) ≤ (z1 ⋅ z3).abs.ulp
						or
						z2 ≍ zero and
						z3.is_proper_imaginary and
						not z3.b.is_finite
						or
						z2 ≍ zero and
						z3.is_proper_real and
						not z3.a.is_finite
						or
						z2 ≍ zero and
						z3.is_proper_complex and
						not (z3.a.is_finite and z3.b.is_finite)
						or
						z2.is_proper_imaginary and
						z3 ≍ zero and
						not z2.b.is_finite
						or
						z2.is_proper_imaginary and
						z3.is_proper_imaginary and
						(z2.b.is_nan or z3.b.is_nan)
						or
						z2.is_proper_imaginary and
						z3.is_proper_real and
						(z2.b.is_nan or z3.a.is_nan)
						or
						z2.is_proper_imaginary and
						z3.is_proper_complex and
						(z2.b.is_nan or z3.a.is_nan or z3.b.is_nan)
						or
						z2.is_proper_real and
						z3 ≍ zero and
						not z2.a.is_finite
						or
						z2.is_proper_real and
						z3.is_proper_imaginary and
						(z2.a.is_nan or z3.b.is_nan)
						or
						z2.is_proper_real and
						z3.is_proper_real and
						(z2.a.is_nan or z3.a.is_nan)
						or
						z2.is_proper_real and
						z3.is_proper_complex and
						(z2.a.is_nan or z3.a.is_nan or z3.b.is_nan)
						or
						z2.is_proper_complex and
						z3 ≍ zero and
						not (z2.a.is_finite and z2.b.is_finite)
						or
						z2.is_proper_complex and
						z3.is_proper_imaginary and
						(z2.a.is_nan or z2.b.is_nan or z3.b.is_nan)
						or
						z2.is_proper_complex and
						z3.is_proper_real and
						(z2.a.is_nan or z2.b.is_nan or z3.a.is_nan)
						or
						z2.is_proper_complex and
						z3.is_proper_complex and (
							proper_complex_product_implementation
								(z2, z3).a.is_nan or
							proper_complex_product_implementation
								(z2, z3).b.is_nan
							)
						or
						(z1.a.is_nan xor z1.b.is_nan) and
						not z1.is_infinite and
						(z1 ⋅ z3).is_infinite and
						not (z2 ⋅ z3).is_infinite
						or
						z1.is_infinite and
						z2.is_finite and
						z2.abs.is_infinite
					z1z3_abs_is_infinite: (z1.a.is_nan xor z1.b.is_nan) and
						not z1.is_infinite and
						(z1 ⋅ z3).is_infinite and
						not (z2 ⋅ z3).is_infinite implies
						(z1 ⋅ z3).abs.is_infinite
					z2z3_abs_is_finite: z1.is_infinite and
						z2.is_finite and
						z2.abs.is_infinite implies
						(z2 ⋅ z3).abs.is_finite
					z2z3_abs_is_nan: z2 ≍ zero and
						z3.is_proper_imaginary and
						not z3.b.is_finite
						or
						z2 ≍ zero and
						z3.is_proper_real and
						not z3.a.is_finite
						or
						z2 ≍ zero and
						z3.is_proper_complex and
						not (z3.a.is_finite and z3.b.is_finite)
						or
						z2.is_proper_imaginary and
						z3 ≍ zero and
						not z2.b.is_finite
						or
						z2.is_proper_imaginary and
						z3.is_proper_imaginary and
						(z2.b.is_nan or z3.b.is_nan)
						or
						z2.is_proper_imaginary and
						z3.is_proper_real and
						(z2.b.is_nan or z3.a.is_nan)
						or
						z2.is_proper_imaginary and
						z3.is_proper_complex and
						(z2.b.is_nan or z3.a.is_nan or z3.b.is_nan)
						or
						z2.is_proper_real and
						z3 ≍ zero and
						not z2.a.is_finite
						or
						z2.is_proper_real and
						z3.is_proper_imaginary and
						(z2.a.is_nan or z3.b.is_nan)
						or
						z2.is_proper_real and
						z3.is_proper_real and
						(z2.a.is_nan or z3.a.is_nan)
						or
						z2.is_proper_real and
						z3.is_proper_complex and
						(z2.a.is_nan or z3.a.is_nan or z3.b.is_nan)
						or
						z2.is_proper_complex and
						z3 ≍ zero and
						not (z2.a.is_finite and z2.b.is_finite)
						or
						z2.is_proper_complex and
						z3.is_proper_imaginary and
						(z2.a.is_nan or z2.b.is_nan or z3.b.is_nan)
						or
						z2.is_proper_complex and
						z3.is_proper_real and
						(z2.a.is_nan or z2.b.is_nan or z3.a.is_nan)
						or
						z2.is_proper_complex and
						z3.is_proper_complex and
						(
							proper_complex_product_implementation (z2, z3).a.
								is_nan or
							proper_complex_product_implementation (z2, z3).b.
								is_nan
							) implies
						(z2 ⋅ z3).abs.is_nan
				end
				broken_morphism := True
				retry
			end
		end

	quotient_ok (z1, z2, z3: STS_COMPLEX_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for
			-- {STS_COMPLEX_NUMBER}.quotient?
		local
			broken_absorption, broken_morphism: BOOLEAN
		do
			if not (broken_absorption or broken_morphism) then
				check
					absorbing_dividend: zero.divisible (z1) implies (zero / z1) ≍ zero
				end
			end
			if not broken_morphism then
				check
						good_divisor: -- one.real_is_invertible
							z1.divisible (one)
					neutral_divisor: (z1 / one) ≍ z1
					quasi_oppositely_homomorphic: z3.divisible (z1) and
						z3.divisible (z2) and
						z1.abs ≤ z2.abs implies
						(z3 / z1).abs ≥ (z3 / z2).abs
				end
			end
			Result := True
		rescue
			print (z1) print (", ")
			print (z2) print (", ")
			print (z3) print ('%N')
			if
				not broken_absorption and
				zero.divisible (z1) and then
				(zero / z1) ≭ zero
			then
				check
						zero_divisible_z1a:
								-- z1.b /= 0 ⇐ z1.is_proper_imaginary
							z1.is_proper_imaginary implies
							zero.a.divisible (z1.b)
						zero_divisible_z1a:
								-- z1.a /= 0 ⇐ z1.is_proper_real
							z1.is_proper_real implies
							zero.a.divisible (z1.a)
					absorption_unfolded_definition_a:
						(zero / z1).a ≍
							if z1.is_proper_imaginary then
								(zero.a / z1.b)
							elseif z1.is_proper_real then
								(zero.a / z1.a)
							else
								proper_complex_quotient_implementation
									(zero, z1).a
							end
					absorption_unfolded_definition_b:
						(zero / z1).b ≍
							if z1.is_proper_imaginary then
								zero.a
							elseif z1.is_proper_real then
								zero.a
							else
								proper_complex_quotient_implementation
									(zero, z1).b
							end
					broken_absorption_cause: z1.a.is_nan or
						z1.b.is_nan or
						z1 ≍ zero
					broken_absorption_a:
						z1.is_proper_imaginary and z1.b.is_nan or
						z1.is_proper_complex and (z1.a.is_nan or z1.b.is_nan) or
						z1 ≍ zero implies
						(zero / z1).a.is_nan
					broken_absorption_b:
						z1.is_proper_complex and (z1.a.is_nan or z1.b.is_nan) or
						z1 ≍ zero implies
						(zero / z1).b.is_nan
				end
				broken_absorption := True
				retry
			elseif
				not broken_morphism and
				z3.divisible (z1) and
				z3.divisible (z2) and
				z1.abs ≤ z2.abs and then
				(z3 / z1).abs < (z3 / z2).abs
			then
				check
--						zero_divisible_z1b:
--								-- z1.b /= 0 ⇐ z1.is_proper_imaginary
--							z1.is_proper_imaginary implies
--							zero.real_divisible (z1.b)
--						zero_divisible_z1a:
--								-- z1.a /= 0 ⇐ z1.is_proper_real
--							z1.is_proper_real implies zero.real_divisible (z1.a)
--						z3b_divisible_z1b: z1.is_proper_imaginary implies
--							z3.b.divisible (z1.b)
--						z3b_divisible_z1a: z1.is_proper_real implies
--							z3.b.divisible (z1.a)
--						z3a_divisible_z1b: z1.is_proper_imaginary implies
--							(- z3.a).divisible (z1.b)
--						z3a_divisible_z1a: z1.is_proper_real implies
--							z3.a.divisible (z1.a)

--						zero_divisible_z2b:
--								-- z2.b /= 0 ⇐ z2.is_proper_imaginary
--							z2.is_proper_imaginary implies
--							zero.real_divisible (z2.b)
--						zero_divisible_z2a:
--								-- z2.a /= 0 ⇐ z2.is_proper_real
--							z2.is_proper_real implies
--							zero.real_divisible (z2.a)
--						z3b_divisible_z2b: z2.is_proper_imaginary implies
--							z3.b.divisible (z2.b)
--						z3b_divisible_z2a: z2.is_proper_real implies
--							z3.b.divisible (z2.a)
--						z3a_divisible_z2b: z2.is_proper_imaginary implies
--							(- z3.a).divisible (z2.b)
--						z3a_divisible_z2a: z2.is_proper_real implies
--							z3.a.divisible (z2.a)
--					unfolded_morphism_definition_z3_by_z1_abs: (z3 / z1).abs ≍
--						if z3 ≍ zero and z1.is_proper_imaginary then
--							z1.hypotenuse (zero.real_quotient (z1.b), zero)
--						elseif z3 ≍ zero and z1.is_proper_real then
--							z1.hypotenuse (zero.real_quotient (z1.a), zero)
--						elseif
--							z3.is_proper_imaginary and z1.is_proper_imaginary
--						then
--							z1.hypotenuse (z3.b / z1.b, zero)
--						elseif z3.is_proper_imaginary and z1.is_proper_real then
--							z1.hypotenuse (zero, z3.b / z1.a)
--						elseif z3.is_proper_real and z1.is_proper_imaginary then
--							z1.hypotenuse (zero, - z3.a / z1.b)
--						elseif z3.is_proper_real and z1.is_proper_real then
--							z1.hypotenuse (z3.a / z1.a, zero)
--						elseif
--							z3.is_proper_complex and z1.is_proper_imaginary
--						then
--							z1.hypotenuse (z3.b / z1.b, - z3.a / z1.b)
--						elseif z3.is_proper_complex and z1.is_proper_real then
--							z1.hypotenuse (z3.a / z1.a, z3.b / z1.a)
--						else
--							z1.hypotenuse (
--								proper_complex_quotient_implementation
--									(z3, z1).a,
--								proper_complex_quotient_implementation (z3, z1).b
--								)
--						end
--					unfolded_morphism_definition_z3_by_z2_abs: (z3 / z2).abs ≍
--						if z3 ≍ zero and z2.is_proper_imaginary then
--							z2.hypotenuse (zero.real_quotient (z2.b), zero)
--						elseif z3 ≍ zero and z2.is_proper_real then
--							z2.hypotenuse (zero.real_quotient (z2.a), zero)
--						elseif
--							z3.is_proper_imaginary and z2.is_proper_imaginary
--						then
--							z2.hypotenuse (z3.b / z2.b, zero)
--						elseif z3.is_proper_imaginary and z2.is_proper_real then
--							z2.hypotenuse (zero, z3.b / z2.a)
--						elseif z3.is_proper_real and z2.is_proper_imaginary then
--							z2.hypotenuse (zero, - z3.a / z2.b)
--						elseif z3.is_proper_real and z2.is_proper_real then
--							z2.hypotenuse (z3.a / z2.a, zero)
--						elseif
--							z3.is_proper_complex and z2.is_proper_imaginary
--						then
--							z2.hypotenuse (z3.b / z2.b, - z3.a / z2.b)
--						elseif z3.is_proper_complex and z2.is_proper_real then
--							z2.hypotenuse (z3.a / z2.a, z3.b / z2.a)
--						else
--							z2.hypotenuse (
--								proper_complex_quotient_implementation
--									(z3, z2).a,
--								proper_complex_quotient_implementation (z3, z2).b
--								)
--						end
					broken_morphism_cause:
--						z1.is_infinite and
--						z2.is_finite and
--						z2.abs.is_infinite
--						or
--						z3 ≍ zero and
--						z1.is_proper_imaginary and
--						z1.b.is_nan
--						or
--						z3 ≍ zero and
--						z1.is_proper_real and
--						z1.a.is_nan or
--						z3.is_proper_imaginary and
--						z1.is_proper_imaginary and
--						(z1.b.is_nan or z3.b.is_infinite and z1.b.is_infinite)
--						or
--						z3.is_proper_imaginary and
--						z1.is_proper_real and
--						(z1.a.is_nan or z3.b.is_infinite and z1.a.is_infinite)
--						or
--						z3.is_proper_real and
--						z1.is_proper_imaginary and
--						(z1.b.is_nan or z3.a.is_infinite and z1.b.is_infinite)
--						or
--						z3.is_proper_real and
--						z1.is_proper_real and
--						(z1.a.is_nan or z3.a.is_infinite and z1.a.is_infinite)
--						or
--						z3.is_proper_complex and
--						z1.is_proper_imaginary and (
--							z1.b.is_nan or
--							(z3.a.is_infinite or z3.b.is_infinite) and
--							z1.b.is_infinite
--							)
--						or
--						z3.is_proper_complex and
--						z1.is_proper_real and (
--							z1.a.is_nan or
--							(z3.a.is_infinite or z3.b.is_infinite) and
--							z1.a.is_infinite
--							)
--						or
--						z1.is_proper_complex and (
--							proper_complex_quotient_implementation
--								(z3, z1).a.is_nan or
--							proper_complex_quotient_implementation
--								(z3, z1).b.is_nan
--							)
--						or
--							-- For the case of an implementation that regards
--							-- zero invertible.
--						z3 ≍ zero and
--						z1 ≍ zero
--						or
							-- Indeed (z3 / z1).abs should not be less than (z3 / z2).abs, but it is out of precision loss.
						((z3 / z2).abs - (z3 / z1).abs) ≤ (z3 / z2).abs.ulp
--					z3_by_z2_abs_is_not_zero: z1.is_infinite and
--						z2.is_finite and
--						z2.abs.is_infinite implies
--						(z3 / z2).abs > zero
--					z3_by_z1_abs_is_nan: z3 ≍ zero and
--						z1.is_proper_imaginary and
--						z1.b.is_nan
--						or
--						z3 ≍ zero and
--						z1.is_proper_real and
--						z1.a.is_nan
--						or
--						z3.is_proper_imaginary and
--						z1.is_proper_imaginary and
--						(z1.b.is_nan or z3.b.is_infinite and z1.b.is_infinite)
--						or
--						z3.is_proper_imaginary and
--						z1.is_proper_real and
--						(z1.a.is_nan or z3.b.is_infinite and z1.a.is_infinite)
--						or
--						z3.is_proper_real and
--						z1.is_proper_imaginary and
--						(z1.b.is_nan or z3.a.is_infinite and z1.b.is_infinite)
--						or
--						z3.is_proper_real and
--						z1.is_proper_real and
--						(z1.a.is_nan or z3.a.is_infinite and z1.a.is_infinite)
--						or
--						z3.is_proper_complex and
--						z1.is_proper_imaginary and (
--							z1.b.is_nan or
--							(z3.a.is_infinite or z3.b.is_infinite) and
--							z1.b.is_infinite
--							)
--						or
--						z3.is_proper_complex and
--						z1.is_proper_real and (
--							z1.a.is_nan or
--							(z3.a.is_infinite or z3.b.is_infinite) and
--							z1.a.is_infinite
--							)
--						or
--						z1.is_proper_complex and (
--							proper_complex_quotient_implementation
--								(z3, z1).a.is_nan or
--							proper_complex_quotient_implementation
--								(z3, z1).b.is_nan
--							)
--						or
--						z3 ≍ zero and
--						z1 ≍ zero implies
--						(z3 / z1).abs.is_nan
--					when_z1_abs_equals_z2_abs: z1.abs ≍ z2.abs implies
--						((z3 / z2).abs - (z3 / z1).abs) ≤ (z3 / z2).abs.ulp
--						or
--						z1.abs.is_infinite and
--						((z1.a.value ^ 2 + z1.b.value ^ 2) ^ (1/2)).is_nan
				end
				broken_morphism := True
				retry
			end
		end

	reciprocal_ok (z1, z2: STS_COMPLEX_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_COMPLEX_NUMBER}.reciprocal?
		local
			broken_complex_reciprocal_a, broken_complex_reciprocal_b, broken_morphism: BOOLEAN
		do
			if z1.is_invertible then
				if not (broken_complex_reciprocal_a or broken_complex_reciprocal_b or broken_morphism) then
					check
						neutral_factor: one.reciprocal ≍ one
							one_divisible_z1b: z1.is_proper_imaginary implies one.a.divisible (z1.b) -- z1.b /= 0 ⇐ z1.is_proper_imaginary
							one_divisible_z1a: z1.is_proper_real implies one.a.divisible (z1.a) -- z1.a /= 0 ⇐ z1.is_proper_real
							z1a_divisible_norm: z1.is_proper_complex implies z1.a.divisible (z1.a ^ two + z1.b ^ two) -- z1.is_invertible
						imaginary_reciprocal_a: z1.is_proper_imaginary implies z1.reciprocal.a ≍ zero.a
						imaginary_reciprocal_b: z1.is_proper_imaginary implies z1.reciprocal.b ≍ - (one.a / z1.b)
						real_reciprocal_a: z1.is_proper_real implies z1.reciprocal.a ≍ (one.a / z1.a)
						real_reciprocal_b: z1.is_proper_real implies z1.reciprocal.b ≍ zero.a
						complex_reciprocal_a: z1.is_proper_complex implies z1.reciprocal.a ≍ (z1.a / (z1.a ^ two + z1.b ^ two))
					end
				end
				if not (broken_complex_reciprocal_b or broken_morphism) then
					check
							z1b_divisible_norm: z1.is_proper_complex implies (- z1.b).divisible (z1.a ^ two + z1.b ^ two) -- z1.is_invertible
						complex_reciprocal_b: z1.is_proper_complex implies z1.reciprocal.b ≍ (- z1.b / (z1.a ^ two + z1.b ^ two))
					end
				end
				if not broken_morphism then
					check
						quasi_oppositely_homomorphic: z2.is_invertible and z1.abs ≤ z2.abs implies z1.reciprocal.abs >= z2.reciprocal.abs
					end
				end
			end
			Result := True
		rescue
			print (z1) print ("; ")
			print (z2) print ('%N')
			if
				not broken_complex_reciprocal_a and (z1.is_invertible and z1.a.divisible (z1.a ^ two + z1.b ^ two)) and then
					z1.reciprocal.a ≭ (z1.a / (z1.a ^ two + z1.b ^ two))
			then
				check
					approximate_complex_reciprocal_a: (z1.reciprocal.a - (z1.a / (z1.a ^ two + z1.b ^ two))).abs ≤
						((machine_epsilon ∨ z1.reciprocal.a.relative_epsilon ∨ (z1.a / (z1.a ^ two + z1.b ^ two)).relative_epsilon) ⋅ Two)
				end
				broken_complex_reciprocal_a := True
				retry
			elseif
				not broken_complex_reciprocal_b and (z1.is_invertible and (- z1.b).divisible (z1.a ^ two + z1.b ^ two)) and then
					z1.reciprocal.b ≭ (- z1.b / (z1.a ^ two + z1.b ^ two))
			then
				check
					approximate_complex_reciprocal_b: (z1.reciprocal.b - (- z1.b / (z1.a ^ two + z1.b ^ two))).abs ≤
						((machine_epsilon ∨ z1.reciprocal.b.relative_epsilon ∨ (- z1.b / (z1.a ^ two + z1.b ^ two)).relative_epsilon) ⋅ Two)
				end
				broken_complex_reciprocal_b := True
				retry
			elseif not broken_morphism and (z1.is_invertible and z2.is_invertible) and then (z1.abs ≤ z2.abs and z1.reciprocal.abs < z2.reciprocal.abs) then
				check
						one_divisible_z1b: z1.is_proper_imaginary implies one.a.divisible (z1.b) -- z1.b /= 0 ⇐ z1.is_proper_imaginary
						one_divisible_z1a: z1.is_proper_real implies one.a.divisible (z1.a) -- z1.a /= 0 ⇐ z1.is_proper_real
						one_divisible_z2b: z2.is_proper_imaginary implies one.a.divisible (z2.b) -- z2.b /= 0 ⇐ z2.is_proper_imaginary
						one_divisible_z2a: z2.is_proper_real implies one.a.divisible (z2.a) -- z2.a /= 0 ⇐ z2.is_proper_real
						one_divisible_z1: not (z1.is_proper_imaginary or z1.is_proper_real) implies one.divisible (z1) -- z1.is_invertible
						one_divisible_z2: not (z2.is_proper_imaginary or z2.is_proper_real) implies one.divisible (z2) -- z2.is_invertible
					unfolded_morphism_definition_z1_reciprocal_abs: z1.reciprocal.abs ≍ if z1.is_proper_imaginary then
							z1.hypotenuse (zero.a, - one.a / z1.b)
						elseif z1.is_proper_real then
							z1.hypotenuse (one.a / z1.a, zero.a)
						else
							z1.hypotenuse (proper_complex_quotient_implementation (one, z1).a, proper_complex_quotient_implementation (one, z1).b)
						end
					unfolded_morphism_definition_z2_reciprocal_abs: z2.reciprocal.abs ≍ if z2.is_proper_imaginary then
							z2.hypotenuse (zero.a, - one.a / z2.b)
						elseif z2.is_proper_real then
							z2.hypotenuse (one.a / z2.a, zero.a)
						else
							z2.hypotenuse (proper_complex_quotient_implementation (one, z2).a, proper_complex_quotient_implementation (one, z2).b)
						end
					broken_morphism: z1.is_proper_imaginary and z1.b.is_nan or
						z1.is_proper_real and z1.a.is_nan or
						z1.is_proper_complex and (z1.a.is_nan or z1.b.is_nan) or
						z1.abs ≍ z2.abs -- Indeed z1.abs should be above z2.abs, but it is not out of precision loss.
					when_z1_reciprocal_abs_is_nan: z1.abs ≭ z2.abs implies z1.reciprocal.abs.is_nan
					when_z1_abs_equals_z2_abs: z1.abs ≍ z2.abs implies (z1.a.value ^ 2 + z1.b.value ^ 2) ^ (1/2) > (z2.a.value ^ 2 + z2.b.value ^ 2) ^ (1/2)
				end
				broken_morphism := True
				retry
			end
		end

	inverse_ok (z1, z2: STS_COMPLEX_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_COMPLEX_NUMBER}.inverse?
		local
			broken_complex_inverse_a, broken_complex_inverse_b, broken_morphism: BOOLEAN
		do
			if z1.is_invertible then
				if not (broken_complex_inverse_a or broken_complex_inverse_b or broken_morphism) then
					check
						neutral_factor: one.inverse ≍ one
							one_divisible_z1b: z1.is_proper_imaginary implies one.a.divisible (z1.b) -- z1.b /= 0 ⇐ z1.is_proper_imaginary
							one_divisible_z1a: z1.is_proper_real implies one.a.divisible (z1.a) -- z1.a /= 0 ⇐ z1.is_proper_real
							z1a_divisible_norm: z1.is_proper_complex implies z1.a.divisible (z1.a ^ two + z1.b ^ two) -- z1.is_invertible
						imaginary_inverse_a: z1.is_proper_imaginary implies z1.inverse.a ≍ zero.a
						imaginary_inverse_b: z1.is_proper_imaginary implies z1.inverse.b ≍ - (one.a / z1.b)
						real_inverse_a: z1.is_proper_real implies z1.inverse.a ≍ (one.a / z1.a)
						real_inverse_b: z1.is_proper_real implies z1.inverse.b ≍ zero.a
						complex_inverse_a: z1.is_proper_complex implies z1.inverse.a ≍ (z1.a / (z1.a ^ two + z1.b ^ two))
					end
				end
				if not (broken_complex_inverse_b or broken_morphism) then
					check
							z1b_divisible_norm: z1.is_proper_complex implies (- z1.b).divisible (z1.a ^ two + z1.b ^ two) -- z1.is_invertible
						complex_inverse_b: z1.is_proper_complex implies z1.inverse.b ≍ (- z1.b / (z1.a ^ two + z1.b ^ two))
					end
				end
				if not broken_morphism then
					check
						quasi_oppositely_homomorphic: z2.is_invertible and z1.abs ≤ z2.abs implies z1.inverse.abs >= z2.inverse.abs
					end
				end
			end
			Result := True
		rescue
			print (z1) print ("; ")
			print (z2) print ('%N')
			if
				not broken_complex_inverse_a and (z1.is_invertible and z1.a.divisible (z1.a ^ two + z1.b ^ two)) and then
					z1.inverse.a ≭ (z1.a / (z1.a ^ two + z1.b ^ two))
			then
				check
					approximate_complex_inverse_a: (z1.inverse.a - (z1.a / (z1.a ^ two + z1.b ^ two))).abs ≤
						((machine_epsilon ∨ z1.inverse.a.relative_epsilon ∨ (z1.a / (z1.a ^ two + z1.b ^ two)).relative_epsilon) ⋅ Two)
				end
				broken_complex_inverse_a := True
				retry
			elseif
				not broken_complex_inverse_b and (z1.is_invertible and (- z1.b).divisible (z1.a ^ two + z1.b ^ two)) and then
					z1.inverse.b ≭ (- z1.b / (z1.a ^ two + z1.b ^ two))
			then
				check
					approximate_complex_inverse_b: (z1.inverse.b - (- z1.b / (z1.a ^ two + z1.b ^ two))).abs ≤
						((machine_epsilon ∨ z1.inverse.b.relative_epsilon ∨ (- z1.b / (z1.a ^ two + z1.b ^ two)).relative_epsilon) ⋅ Two)
				end
				broken_complex_inverse_b := True
				retry
			elseif not broken_morphism and (z1.is_invertible and z2.is_invertible) and then (z1.abs ≤ z2.abs and z1.inverse.abs < z2.inverse.abs) then
				check
						one_divisible_z1b: z1.is_proper_imaginary implies one.a.divisible (z1.b) -- z1.b /= 0 ⇐ z1.is_proper_imaginary
						one_divisible_z1a: z1.is_proper_real implies one.a.divisible (z1.a) -- z1.a /= 0 ⇐ z1.is_proper_real
						one_divisible_z2b: z2.is_proper_imaginary implies one.a.divisible (z2.b) -- z2.b /= 0 ⇐ z2.is_proper_imaginary
						one_divisible_z2a: z2.is_proper_real implies one.a.divisible (z2.a) -- z2.a /= 0 ⇐ z2.is_proper_real
						one_divisible_z1: not (z1.is_proper_imaginary or z1.is_proper_real) implies one.divisible (z1) -- z1.is_invertible
						one_divisible_z2: not (z2.is_proper_imaginary or z2.is_proper_real) implies one.divisible (z2) -- z2.is_invertible
					unfolded_morphism_definition_z1_inverse_abs: z1.inverse.abs ≍ if z1.is_proper_imaginary then
							z1.hypotenuse (zero.a, - (one.a / z1.b))
						elseif z1.is_proper_real then
							z1.hypotenuse (one.a / z1.a, zero.a)
						else
							z1.hypotenuse (proper_complex_quotient_implementation (one, z1).a, proper_complex_quotient_implementation (one, z1).b)
						end
					unfolded_morphism_definition_z2_inverse_abs: z2.inverse.abs ≍ if z2.is_proper_imaginary then
							z2.hypotenuse (zero.a, - one.a / z2.b)
						elseif z2.is_proper_real then
							z2.hypotenuse (one.a / z2.a, zero.a)
						else
							z2.hypotenuse (proper_complex_quotient_implementation (one, z2).a, proper_complex_quotient_implementation (one, z2).b)
						end
					broken_morphism: z1.is_proper_imaginary and z1.b.is_nan or
						z1.is_proper_real and z1.a.is_nan or
						z1.is_proper_complex and (z1.a.is_nan or z1.b.is_nan) or
						z1.abs ≍ z2.abs -- Indeed z1.abs should be above z2.abs, but it is not out of precision loss.
					when_z1_inverse_abs_is_nan: z1.abs ≭ z2.abs implies z1.inverse.abs.is_nan
					when_z1_abs_equals_z2_abs: z1.abs ≍ z2.abs implies (z1.a.value ^ 2 + z1.b.value ^ 2) ^ (1/2) > (z2.a.value ^ 2 + z2.b.value ^ 2) ^ (1/2)
				end
				broken_morphism := True
				retry
			end
		end

feature -- Math

	Pi: like real_anchor.value = 3.1415926535897932384626433832795029
			-- Value of the constant π

feature -- Conversion

	natural_as_integer (n: NATURAL): INTEGER_64
			-- `n' converted into an integer value
			-- TODO: DRY
		do
			Result := n.as_integer_64
		ensure
			class
			definition: Result = n.as_integer_64
		end

feature -- Anchor

	complex_anchor: STS_COMPLEX_NUMBER
			-- Anchor for complex numbers
		deferred
		end

	real_anchor: STS_REAL_NUMBER
			-- Anchor for real numbers
		deferred
		end

	integer_anchor: STS_INTEGER_NUMBER
			-- Anchor for integer numbers
		deferred
		end

	real_math_anchor: SINGLE_MATH
			-- Anchor for accessing basic mathematical operations on native real numbers
		once
			create Result
		ensure
			class
		end

	complex_universe_anchor: STS_COMPLEX_SET
			-- Anchor for `current_complex_universe'
		deferred
		end

	real_universe_anchor: STS_REAL_SET
			-- Anchor for `current_real_universe'
		deferred
		end

feature {NONE} -- Implementation

	proper_complex_product_implementation (z1, z2: STS_COMPLEX_NUMBER): like complex_anchor
			-- Product of `z1' and `z2' as done by an implementation of {STS_COMPLEX_NUMBER}
		deferred
		ensure
			ref_a: attached {like real_anchor} Result.real_from_value (z1.a.value * z2.a.value - z1.b.value * z2.b.value) as ref_a
			ref_b: attached {like real_anchor} Result.real_from_value (z1.a.value * z2.b.value + z1.b.value * z2.a.value) as ref_b
			prod_a: (Result.a - ref_a).abs ≤ (machine_epsilon ∨ ref_a.relative_epsilon)
			prod_b: (Result.b - ref_b).abs ≤ (machine_epsilon ∨ ref_b.relative_epsilon)

			infinite_left_factor: z1.is_infinite and (z2.is_finite and z2 ≭ zero or z2.is_infinite) implies Result.is_infinite
			infinite_right_factor: (z1.is_finite and z1 ≭ zero or z1.is_infinite) and z2.is_infinite implies Result.is_infinite
		end

	proper_complex_quotient_implementation (z1, z2: STS_COMPLEX_NUMBER): like complex_anchor
			-- Quotient of `z1' by `z2' as done by an implementation of {STS_COMPLEX_NUMBER}
		require
			good_divisor: z1.divisible (z2)
		deferred
		ensure
			scale_log: attached Result.value_logb ((z2.a.abs ∨ z2.b.abs).value) as scale_log
			int_scale_log: attached if Result.value_is_finite (scale_log) then scale_log.truncated_to_integer else 0 end as int_scale_log
			scaled_c: attached (z2.a.value * 2 ^ (- int_scale_log)) as scaled_c
			scaled_d: attached (z2.b.value * 2 ^ (- int_scale_log)) as scaled_d
			denom: attached (scaled_c ^ 2 + scaled_d ^ 2) as denom
			invertible_denom: z2.is_proper_complex implies denom /= 0 -- z2.a /= 0 and z2.b /= 0 ⇐ z2.is_proper_complex

				-- (a+bi) / (c + di) = (ac+bd)/(c^2+d^2)+(bc-ad)i/(c^2+d^2)				
			complex_divisor_a_good_divisor: z2.is_proper_complex implies
				(z1.a.value * scaled_c + z1.b.value * scaled_d).divisible (denom) -- denom /= 0 ⇐ z2.is_proper_complex
			complex_divisor_b_good_divisor: z2.is_proper_complex implies (z1.b.value * scaled_c - z1.a.value * scaled_d).divisible (denom)
			complex_divisor_quot_a: attached {STS_REAL_NUMBER} Result.real_from_double_value
				(((z1.a.value * scaled_c + z1.b.value * scaled_d) / denom) * 2 ^ (- int_scale_log)) as complex_by_complex_quot_a
			complex_divisor_quot_b: attached {STS_REAL_NUMBER} Result.real_from_double_value
				(((z1.b.value * scaled_c - z1.a.value * scaled_d) / denom) * 2 ^ (- int_scale_log)) as complex_by_complex_quot_b
			complex_divisor_a: z2.is_proper_complex implies
				(Result.a - complex_by_complex_quot_a).abs ≤ (machine_epsilon ∨ complex_by_complex_quot_a.relative_epsilon)
			complex_divisor_b: z2.is_proper_complex implies
				(Result.b - complex_by_complex_quot_b).abs ≤ (machine_epsilon ∨ complex_by_complex_quot_b.relative_epsilon)

			infinite_dividend: z1.is_infinite and z2.is_finite implies Result.is_infinite
			infinite_divisor: z1.is_finite and z2.is_infinite implies Result ≍ zero
			zero_divisor: (z1.is_finite and z1 ≭ zero or z1.is_infinite) and z2 ≍ zero implies Result.is_infinite -- Indeed the precondition implies z2 /= 0.
		end

	max_subsets: NATURAL
			-- Maximum number of subsets to be checked within property statements
		deferred
		end

invariant
	Pi: Pi = {like real_math_anchor}.Pi.truncated_to_real -- TODO: Get rid of `truncated_to_real'.

note
	copyright: "Copyright (c) 2012-2022, Rosivaldo Fernandes Alves"
	license: "[
		Eiffel Forum License v2
		(see http://www.eiffel.com/licensing/forum.txt)
		]"
	source: "https://launchpad.net/set-theory"

end
