﻿note
	description: "[
		Object that checks whether the properties verified within set theory hold for an implementation of {STS_INTEGER_NUMBER}
		]"
	author: "Rosivaldo Fernandes Alves"
	date: "$Date: Sat 2018-11-10 18:52:01 -0300 $"
	revision: "$Revision$"

deferred class
	INTEGER_NUMBER_PROPERTIES

inherit
	RATIONAL_NUMBER_PROPERTIES
		rename
--			zero as rational_zero, -- TODO: Such a renaming should be needed only in descendants.
--			one as rational_one, -- TODO: Such a renaming should be needed only in descendants.
			is_in_ok as rational_is_in_ok,
			is_not_in_ok as rational_is_not_in_ok,
			zero_ok as rational_zero_ok,
			one_ok as rational_one_ok,
			is_integer_ok as rational_is_integer_ok,
			equals_ok as rational_equals_ok,
			unequals_ok as rational_unequals_ok,
			is_less_ok as rational_is_less_ok,
			is_less_equal_ok as rational_is_less_equal_ok,
			is_greater_ok as rational_is_greater_ok,
			is_greater_equal_ok as rational_is_greater_equal_ok,
			min_ok as rational_min_ok,
			max_ok as rational_max_ok,
			multipliable_ok as rnp_rational_multipliable_ok,
			modulus_ok as rational_modulus_ok,
			abs_ok as rational_abs_ok,
			plus_ok as rational_plus_ok,
			identity_ok as rational_identity_ok,
			minus_ok as rational_minus_ok,
			opposite_ok as rational_opposite_ok,
			product_ok as rational_product_ok,
			quotient_ok as rational_quotient_ok,
			power_ok as real_power_ok
		end

feature -- Access

	zero: like integer_anchor
			-- The integer number 0
		deferred
--		ensure
--			definition: Result.value = 0
		end

	one: like integer_anchor
			-- The integer number 1
		deferred
--		ensure
--			definition: Result.value = 1
		end

	current_integer_universe: like integer_universe_anchor
			-- The current integer "Universe", i.e. set with every integer
			-- number currently in system memory.
			-- Notice that this "universe" may change from a call to another.
		deferred
		end

feature -- Properties (Membership)

	is_in_ok (i: STS_INTEGER_NUMBER; s: STS_SET [STS_INTEGER_NUMBER, STS_EQUALITY [STS_INTEGER_NUMBER]]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_INTEGER_NUMBER}.is_in?
		do
			check
				require_non_emptiness: i ∈ s implies not s.is_empty
			end
			Result := True
		end

	is_not_in_ok (i: STS_INTEGER_NUMBER; s: STS_SET [STS_INTEGER_NUMBER, STS_EQUALITY [STS_INTEGER_NUMBER]]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_INTEGER_NUMBER}.is_not_in?
		do
			check
				definition: i ∉ s = s ∌ i
				sufficient_emptiness: s.is_empty implies i ∉ s
			end
			Result := True
		end

feature -- Properties (Access)

	zero_ok (i, j: STS_INTEGER_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_INTEGER_NUMBER}.zero?
		do
			if real_zero_ok (i, j) then
				check
					neutral_left_term: (i.zero + j) ≍ j
					neutral_right_term: (j + i.zero) ≍ j
					quasi_neutral_minuend: (i.zero - j) ≍ - j
					neutral_subtrahend: (j - i.zero) ≍ j
					left_absorbing_factor: (i.zero * j) ≍ i.zero
					right_absorbing_factor: (i * j.zero) ≍ j.zero
					left_absorbing_dividend: i.zero.divisible (j) implies (i.zero // j) ≍ i.zero
					negative_exponent: j < zero implies (i.zero ^ j).is_positive_infinity
					to_power_zero: (i.zero ^ zero) ≍ one
					absorbing_base: zero < j implies (i.zero ^ j) ≍ i.zero
					zero_exponent: (i ^ j.zero) ≍ one
				end
				Result := True
			end
		end

	one_ok (i, j: STS_INTEGER_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_INTEGER_NUMBER}.one?
		do
			if real_one_ok (i, j) then
				check
					neutral_left_factor: (i.one * j) ≍ j
					neutral_right_factor: (j * i.one) ≍ j
--					quasi_neutral_dividend: i.one.divisible (j) implies - one ≤ i.one // j and i.one // j ≤ one
						good_divisor: j.divisible (i.one) -- i.one /= 0
					neutral_divisor: (j // i.one) ≍ j
					neutral_exponent: (i ^ j.one) ≍ i
					absorbing_base: (i.one ^ j) ≍ i.one
				end
				Result := True
			end
		end

feature -- Properties (Quality)

	is_finite_ok (i: STS_INTEGER_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_INTEGER_NUMBER}.is_finite?
		do
			check
				always_finite: i.is_finite
			end
			Result := True
		end

	is_integer_ok (i: STS_INTEGER_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for
			-- {STS_INTEGER_NUMBER}.is_integer?
		do
			if rational_is_integer_ok (i) then
				check
					is_integer: i.is_integer
				end
				Result := True
			end
		end

feature -- Properties (Comparison)

	equals_ok (i, j, k: STS_INTEGER_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_INTEGER_NUMBER}.equals?
		do
			check
				z: attached current_integer_universe as z
				definition: (2 ^ # z) ≤ max_subsets implies i ≍ j = z.powerset |∀
					agent (s: STS_SET [STS_INTEGER_NUMBER, STS_NUMERIC_EQUALITY [STS_INTEGER_NUMBER]]; ia_i, ia_j: STS_INTEGER_NUMBER): BOOLEAN
						do
							Result := ia_i ∈ s = ia_j ∈ s
						end (?, i, j)
				by_quantification: i ≍ j = (z |∀ agent (l: STS_INTEGER_NUMBER; ia_i, ia_j: STS_INTEGER_NUMBER): BOOLEAN
					do
						Result := (l.value = ia_i.value) = (l.value = ia_j.value)
					end (?, i, j))
				complex_definition: i ≍ j = i.complex_equals (j)
				real_definition: i ≍ j = i.real_equals (j)
				reflexive: i ≍ i
				symmetric: i ≍ j implies j ≍ i
				transitive: i ≍ j and j ≍ k implies i ≍ k
				euclidian: i ≍ k and j ≍ k implies i ≍ j
			end
			Result := True
		end

	unequals_ok (i, j: STS_INTEGER_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for
			-- {STS_INTEGER_NUMBER}.unequals?
		do
			check
				z: attached current_integer_universe as z
				definition: i ≭ j = (i.value /= j.value)
				by_existence: i ≭ j = (
					z |∃ agent (
						k: STS_INTEGER_NUMBER
						ia_i, ia_j: STS_INTEGER_NUMBER
						): BOOLEAN
						do
							Result := (k.value = ia_i.value) /=
								(k.value = ia_j.value)
						end (?, i, j)
					)
				by_membership: (2 ^ # z) ≤ max_subsets implies i ≭ j =
					z.powerset |∃ agent (
						s: STS_SET
							[STS_INTEGER_NUMBER, STS_INTEGER_EQUALITY]
						ia_i, ia_j: STS_INTEGER_NUMBER
						): BOOLEAN
						do
							Result := ia_i ∈ s /= ia_j ∈ s
						end (?, i, j)
				complex_definition: i ≭ j = i.complex_unequals (j)
				real_definition: i ≭ j = i.real_unequals (j)
				irreflexive: not (i ≭ i)
			end
			Result := True
		end

	is_less_ok (i, j, k: STS_INTEGER_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_INTEGER_NUMBER}.is_less?
		do
			check
				definition: i < j = i.real_is_less (j)
				irreflexive: not (i < i)
				transitive: i < j and j < k implies i < k
			end
			Result := True
		end

	is_less_equal_ok (i, j, k: STS_INTEGER_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_INTEGER_NUMBER}.is_less_equal?
		do
			check
				definition: i ≤ j = i.real_is_less_equal (j)
				reflexive: i ≤ i
				transitive: i ≤ j and j ≤ k implies i ≤ k
				antisymmetric: i ≤ j and j ≤ i implies i ≍ j
			end
			Result := True
		end

	is_greater_ok (i, j, k: STS_INTEGER_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_INTEGER_NUMBER}.is_greater?
		do
			check
				definition: i > j = i.real_is_greater (j)
				irreflexive: not (i > i)
				transitive: i > j and j > k implies i > k
			end
			Result := True
		end

	is_greater_equal_ok (i, j, k: STS_INTEGER_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_INTEGER_NUMBER}.is_greater_equal?
		do
			check
				definition: i ≥ j = i.real_is_greater_equal (j)
				reflexive: i ≥ i
				transitive: i ≥ j and j ≥ k implies i ≥ k
				antisymmetric: i ≥ j and j ≥ i implies i ≍ j
			end
			Result := True
		end

	min_ok (i, j, k: STS_INTEGER_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_INTEGER_NUMBER}.min?
		do
			check
				definition: i.real_min (j) ≍ (i ∧ j)
				idempotent: (i ∧ i) ≍ i
				commutative: (i ∧ j) ≍ (j ∧ i)
				associative: ((i ∧ j) ∧ k) ≍ (i ∧ (j ∧ k))
			end
			Result := True
		end

	max_ok (i, j, k: STS_INTEGER_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for
			-- {STS_INTEGER_NUMBER}.max?
		do
			check
				definition: i.real_max (j) ≍ (i ∨ j)
				idempotent: (i ∨ i) ≍ i
				commutative: (i ∨ j) ≍ (j ∨ i)
				associative: ((i ∨ j) ∨ k) ≍ (i ∨ (j ∨ k))
			end
			Result := True
		end

feature -- Properties (Relationship)

	rational_multipliable_ok (i: STS_INTEGER_NUMBER; pq: STS_RATIONAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_INTEGER_NUMBER}.rational_multipliable?
		do
			if rnp_rational_multipliable_ok (i, pq) then
				check
					always_multipliable: i.rational_multipliable (pq)
				end
				Result := True
			end
		end

	divides_ok (i, j, k: STS_INTEGER_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for
			-- {STS_INTEGER_NUMBER}.divides?
		local
			broken_definition, broken_sum, broken_difference: BOOLEAN
		do
			if not (broken_definition or broken_sum or broken_difference) then
				check
					z: attached current_integer_universe as z
					definition: (i | j) = (
						i ≭ zero and (
							z |∃ agent
								(ia_i, ia_j, m: STS_INTEGER_NUMBER): BOOLEAN
								do
									Result := ia_j ≍ (m ⋅ ia_i)
								end (i, j, ?)
							)
						)
				end
			end
			if not (broken_sum or broken_difference) then
				check
					transitive: i | j and j | k implies i | k
					quasi_antisymmetric: i | j and j | i implies i.abs ≍ j.abs
					sum_divisor: i | j and i | k implies i | (j + k)
				end
			end
			if not broken_difference then
				check
					difference_divisor: i | j and i | k implies i | (j - k)
	--				euclid_lemma: i | (j ⋅ k) and gcd (i, j) ≍ one implies i | k -- TODO: Inherit from rational!
				end
			end
			Result := True
		rescue
			print (i) print (", ")
			print (j) print (", ")
			print (k) print ('%N')
			if
				not broken_definition and {EXCEPTIONS}.tag_name ~ "definition"
			then
				check
					z: attached current_integer_universe as z
					broken_definition: not (i | j) and z |∃ agent
						(ia_i, ia_j, m: STS_INTEGER_NUMBER): BOOLEAN
						do
								-- NOTICE: It assumes consistent results upon
								-- overflows.
							Result := product_overflows (m, ia_i) and
								ia_j ≍ (m ⋅ ia_i)
						end (i, j, ?)
				end
				broken_definition := True
				retry
			elseif not broken_sum and i | j and i | k and not (i | (j + k)) then
				check
					broken_sum: sum_overflows (j, k)
				end
				broken_sum := True
				retry
			elseif
				not broken_difference and i | j and i | k and not (i | (j - k))
			then
				check
					broken_difference: difference_overflows (j, k)
				end
				broken_difference := True
				retry
			end
		end

feature -- Properties (Operation)

	modulus_ok (i: STS_INTEGER_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_INTEGER_NUMBER}.modulus?
		local
			mod_is_neg, inherited_ok: BOOLEAN
		do
			if not mod_is_neg then
				inherited_ok := rational_modulus_ok (i)
				if inherited_ok then
					check
						non_negative: i.modulus ≥ zero
					end
				end
			end
			if inherited_ok then
				check
					when_negative: i < zero implies i.modulus ≍ - i
					when_non_negative: i ≥ zero implies i.modulus ≍ i
				end
				Result := True
			end
		rescue
			if not mod_is_neg and i.modulus < zero then
				check
					min_value: i.min_value_exists and i ≍ i.min_value
				end
				mod_is_neg := True
				retry
			end
		end

	abs_ok (i: STS_INTEGER_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_INTEGER_NUMBER}.abs?
		local
			abs_is_neg, inherited_ok: BOOLEAN
		do
			if not abs_is_neg then
				inherited_ok := rational_abs_ok (i)
				if inherited_ok then
					check
						non_negative: i.abs ≥ zero
					end
				end
			end
			if inherited_ok then
				check
					when_negative: i < zero implies i.abs ≍ - i
					when_non_negative: i ≥ zero implies i.abs ≍ i
				end
				Result := True
			end
		rescue
			if not abs_is_neg and i.abs < zero then
				check
					min_value: i.min_value_exists and i ≍ i.min_value
				end
				abs_is_neg := True
				retry
			end
		end

	plus_ok (i, j, k: STS_INTEGER_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_INTEGER_NUMBER}.plus?
		local
			broken_definition, broken_morphism, inherited_ok: BOOLEAN
		do
			if not (broken_definition or broken_morphism) then
				inherited_ok := real_plus_ok (i, j, k)
				if inherited_ok then
					check
						complex_definition: (i + j).complex_equals (i.complex_plus (j))
						real_definition: (i + j).real_equals (i.real_plus (j))
					end
				end
			end
			if inherited_ok and not broken_morphism then
				check
					neutral_left_term: (zero + i) ≍ i
					neutral_right_term: (i + zero) ≍ i
					commutative: (i + j) ≍ (j + i)
					associative: ((i + j) + k) ≍ (i + (j + k))
					homomorphic: i ≤ j implies (i + k) ≤ (j + k)
					isomorphic: i < j implies i + k < j + k
				end
			end
			Result := inherited_ok
		rescue
			if not broken_definition and (i + j).complex_unequals (i.complex_plus (j)) then
				check
					broken_definition: sum_overflows (i, j)
				end
				broken_definition := True
				retry
			elseif not broken_morphism and i ≤ j and (i + k) > (j + k) then
				check
					broken_morphism: sum_overflows (i, k) xor sum_overflows (j, k)
				end
				broken_morphism := True
				retry
			end
		end

	identity_ok (i: STS_INTEGER_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_INTEGER_NUMBER}.identity?
		do
			if rational_identity_ok (i) then -- TODO: The second i will be removed.
				check
					as_left_term: + i ≍ (i + zero)
					as_right_term: + i ≍ (zero + i)
					as_left_factor: + i ≍ (i ⋅ one)
					as_right_factor: + i ≍ (one ⋅ i)
						good_divisor: i.divisible (one) -- one.is_invertible
					as_quotient: + i ≍ (i // one)
				end
				Result := True
			end
		end

	minus_ok (i, j, k: STS_INTEGER_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_INTEGER_NUMBER}.minus?
		local
			broken_definition, broken_morphism, inherited_ok: BOOLEAN
		do
			if not (broken_definition or broken_morphism) then
				inherited_ok := real_minus_ok (i, j, k)
				if inherited_ok then
					check
						complex_definition: (i - j).complex_equals (i.complex_minus (j))
						real_definition: (i - j).real_equals (i.real_minus (j))
					end
				end
			end
			if inherited_ok and not broken_morphism then
				check
					quasi_neutral_minuend: (zero - i) ≍ - i
					neutral_subtrahend: (i - zero) ≍ i
					quasi_commutative: (i - j).abs ≍ (j - i).abs
					oppositely_homomorphic: i ≤ j implies (k - i) ≥ (k - j)
					oppositely_isomorphic: i < j implies k - i > k - j
				end
			end
			Result := inherited_ok
		rescue
			print (i) print (", ")
			print (j) print (", ")
			print (k) print ('%N')
			if not broken_definition and (i - j).complex_unequals (i.complex_minus (j)) then
				check
					broken_definition: difference_overflows (i, j)
				end
				broken_definition := True
				retry
			elseif not broken_morphism and i ≤ j and (k - i) < (k - j) then
				check
					broken_morphism: difference_overflows (k, i) xor difference_overflows (k, j)
				end
				broken_morphism := True
				retry
			end
		end

	opposite_ok (i: STS_INTEGER_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for
			-- {STS_INTEGER_NUMBER}.opposite?
		local
			broken_sign, inherited_ok: BOOLEAN
		do
			if not broken_sign then
				inherited_ok := rational_opposite_ok (i)
				if inherited_ok then
					check
						inverted_sign: (- i).sign ≍ - i.sign
					end
				end
			end
			if inherited_ok then
				check
					opposite_left_term: ((- i) + i) ≍ zero
					opposite_right_term: (i + (- i)) ≍ zero
					as_subtrahend: - i ≍ (zero - i)
					as_left_factor: - i ≍ (i ⋅ (- one))
					as_right_factor: - i ≍ ((- one) ⋅ i)
						divisible: i.divisible (- one) -- (- one).is_invertible
					as_quotient: - i ≍ (i // (- one))
				end
				Result := True
			end
		rescue
			if not broken_sign and (- i).sign ≭ - i.sign then
				check
					broken_sign: i.min_value_exists and then i ≍ i.min_value
				end
				broken_sign := True
				retry
			end
		end

	product_ok (i, j, k: STS_INTEGER_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for
			-- {STS_INTEGER_NUMBER}.product?
		local
			broken_definition, broken_homomorphism, broken_isomorphism: BOOLEAN
		do
			if
				not (
					broken_definition or
					broken_homomorphism or
					broken_isomorphism
					)
			then
				check
					complex_definition: (i ⋅ j).complex_equals (i.complex_product (j))
					real_definition: (i ⋅ j).real_equals (i.real_product (j))
				end
			end
			if not (broken_homomorphism or broken_isomorphism) then
				check
					neutral_left_factor: one ⋅ i ≍ i
					neutral_right_factor: i ⋅ one ≍ i
					absorbing_left_factor: zero ⋅ i ≍ zero
					absorbing_right_factor: i ⋅ zero ≍ zero
					commutative: (i ⋅ j) ≍ (j ⋅ i)
					quasi_homomorphic: i.abs ≤ j.abs implies
						(i ⋅ k).abs ≤ (j ⋅ k).abs
				end
			end
			if broken_isomorphism then
				check
					quasi_isomorphic: i ≭ zero and
						i.abs < j.abs and
						k ≭ zero implies
						(i ⋅ k).abs < (j ⋅ k).abs
				end
			end
			Result := True
		rescue
			print (i) print (", ")
			print (j) print (", ")
			print (k) print ('%N')
			if not broken_definition and (i ⋅ j).complex_unequals (i.complex_product (j)) then
				check
					broken_definition: product_overflows (i, j)
				end
				broken_definition := True
				retry
			elseif not broken_homomorphism and i.abs ≤ j.abs and (i ⋅ k).abs > (j ⋅ k).abs then
				check
					broken_homomorphism: product_overflows (i, k) or product_overflows (j, k) or k.min_value_exists and then (j ⋅ k).abs ≍ k.min_value
				end
				broken_homomorphism := True
				retry
			elseif not broken_isomorphism and i ≭ zero and i.abs < j.abs and k ≭ zero and (i ⋅ k).abs ≥ (j ⋅ k).abs then
				check
					broken_isomorphism: product_overflows (i, k) or product_overflows (j, k) or k.min_value_exists and then (j ⋅ k).abs ≍ k.min_value
				end
				broken_isomorphism := True
				retry
			end
		end

	quotient_ok (i, j, k: STS_INTEGER_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for
			-- {STS_INTEGER_NUMBER}.quotient?
		local
			broken_morphism: BOOLEAN
		do
			if not broken_morphism then
				check
					absorbing_dividend: zero.divisible (j) implies (zero / j) ≍ zero

						good_divisor: -- 1 is an invertible, neutral factor.
							i.divisible (one)
					neutral_divisor: (i / one) ≍ i
					multiplicative_inverse: one.divisible (j) implies (one / j) ≍ j.inverse

					quasi_oppositely_homomorphic: i.abs ≤ j.abs and
						k.divisible (i) and k.divisible (j) implies
						(k / i).abs ≥ (k / j).abs
				end
			end
			Result := True
		rescue
			print (i) print (", ")
			print (j) print (", ")
			print (k) print ('%N')
			if
				i.abs ≤ j.abs and k.divisible (i) and k.divisible (j) and then (k / i).abs < (k / j).abs
			then
				check
					broken_morphism: i.abs < zero or (k / i).abs < zero -- Two's complement minimum value.
				end
				broken_morphism := True
				retry
			end
		end

	integer_quotient_ok (i, j, k: STS_INTEGER_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for
			-- {STS_INTEGER_NUMBER}.integer_quotient?
		local
			broken_morphism: BOOLEAN
		do
			if not broken_morphism then
				check
					absorbing_dividend: i ≍ zero and i.divisible (j) implies
						(i // j) ≍ zero
						good_divisor: i.divisible (one) -- one /= 0
					neutral_divisor: (i // one) ≍ i
					quasi_oppositely_homomorphic: i.abs ≤ j.abs and
						k.divisible (i) and k.divisible (j) implies
						(k // i).abs ≥ (k // j).abs
				end
			end
			Result := True
		rescue
			print (i) print (", ")
			print (j) print (", ")
			print (k) print ('%N')
			if
				not broken_morphism and
				i.abs ≤ j.abs and k.divisible (i) and
				k.divisible (j) and then
				(k // i).abs < (k // j).abs
			then
				check
					broken_morphism:
							-- Possible when `i' or `k' is the minimumm
							-- two's-complement representable integer.
						i.abs < zero or
							-- Possible when `k' is the minimum
							-- two's-complement representable integer and
							-- `i'.abs = 1.
						(k // i).abs < zero
				end
				broken_morphism := True
				retry
			end
		end

	integer_remainder_ok (i, j: STS_INTEGER_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_INTEGER_NUMBER}.integer_remainder?
		local
			broken_upper_bound: BOOLEAN
		do
			if not broken_upper_bound then
				check
					absorbing_dividend: i ≍ zero and i.divisible (j) implies (i \\ j) ≍ zero
						good_divisor: i.divisible (one) -- one /= 0
					neutral_divisor: (i \\ one) ≍ zero
					upper_bounded: i.divisible (j) implies (i \\ j).abs < j.abs
				end
			end
			Result := True
		rescue
			if not broken_upper_bound and i.divisible (j) and (i \\ j).abs ≥ j.abs then
				check
					broken_upper_bound: j.min_value_exists and then j ≍ j.min_value
						good_divisor: i.divisible (j) -- j ≍ j.min_value /= 0
					still_upper_bounded: (i \\ j).abs ≤ (j + one).abs
				end
				broken_upper_bound := True
				retry
			end
		end

	power_ok (i: STS_INTEGER_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_INTEGER_NUMBER}.power?
		do
			if real_power_ok (i) then
				check
					base_zero: zero < i implies (zero ^ i) ≍ zero
					base_one: (one ^ i) ≍ one
					exponent_zero: (i ^ zero) ≍ one
					exponent_one: (i ^ one) ≍ i
				end
				Result := True
			end
		end

feature -- Anchor

	integer_universe_anchor: STS_SET [STS_INTEGER_NUMBER, STS_NUMERIC_EQUALITY [STS_INTEGER_NUMBER]]
			-- Anchor for `current_integer_universe'
			-- TODO: Change set type.
		deferred
		end

note
	copyright: "Copyright (c) 2012-2022, Rosivaldo Fernandes Alves"
	license: "[
		Eiffel Forum License v2
		(see http://www.eiffel.com/licensing/forum.txt)
		]"
	source: "https://launchpad.net/set-theory"

end
