﻿note
	description: "[
		Object that checks whether the properties verified within set theory
		hold for an implementation of {STS_RATIONAL_NUMBER}
		]"
	author: "Rosivaldo Fernandes Alves"
	date: "$Date$"
	revision: "$Revision$"

deferred class
	RATIONAL_NUMBER_PROPERTIES

inherit
	REAL_NUMBER_PROPERTIES
		rename
			i as imaginary_unit,
			is_in_ok as real_is_in_ok,
			is_not_in_ok as real_is_not_in_ok,
			zero_ok as real_zero_ok,
			one_ok as real_one_ok,
			is_nan_ok as real_is_nan_ok,
			is_negative_infinity_ok as real_is_negative_infinity_ok,
			is_positive_infinity_ok as real_is_positive_infinity_ok,
			is_infinite_ok as real_is_infinite_ok,
			is_rational_ok as real_is_rational_ok,
			is_integer_ok as real_is_integer_ok,
			equals_ok as real_equals_ok,
			unequals_ok as real_unequals_ok,
			is_less_ok as real_is_less_ok,
			is_less_equal_ok as real_is_less_equal_ok,
			is_greater_ok as real_is_greater_ok,
			is_greater_equal_ok as real_is_greater_equal_ok,
			min_ok as real_min_ok,
			max_ok as real_max_ok,
			plus_ok as real_plus_ok,
			identity_ok as real_identity_ok,
			minus_ok as real_minus_ok,
			opposite_ok as real_opposite_ok,
			product_ok as real_product_ok,
			quotient_ok as real_quotient_ok,
			reciprocal_ok as real_reciprocal_ok,
			inverse_ok as real_inverse_ok
		end

feature -- Access

	zero: like rational_anchor
			-- The rational number 0/1
		deferred
		end

	one: like rational_anchor
			-- The rational number 1/1
		deferred
		end

--	current_rational_universe: like rational_universe_anchor
--			-- The current RATIONAL "Universe", i.e. set with every RATIONAL number currently in system memory.
--			-- Notice that this "universe" may change from a call to another.
--		deferred
--		end

feature -- Properties (Membership)

	is_in_ok (pq: STS_RATIONAL_NUMBER; s: STS_SET [STS_RATIONAL_NUMBER, STS_EQUALITY [STS_RATIONAL_NUMBER]]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_RATIONAL_NUMBER}.is_in?
		do
			check
				require_non_emptiness: pq ∈ s implies not s.is_empty
			end
			Result := True
		end

	is_not_in_ok (pq: STS_RATIONAL_NUMBER; s: STS_SET [STS_RATIONAL_NUMBER, STS_EQUALITY [STS_RATIONAL_NUMBER]]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_RATIONAL_NUMBER}.is_not_in?
		do
			check
				definition: pq ∉ s = s ∌ pq
				sufficient_emptiness: s.is_empty implies pq ∉ s
			end
			Result := True
		end

feature -- Properties (Access)

	zero_ok (pq_1, pq_2: STS_RATIONAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for
			-- {STS_RATIONAL_NUMBER}.zero?
		do
			if real_zero_ok (pq_1, pq_2) then
				check
					neutral_left_term: (pq_1.zero + pq_2) ≍ pq_2
					neutral_right_term: (pq_1 + pq_2.zero) ≍ pq_1
					quasi_neutral_minuend: (pq_1.zero - pq_2) ≍ - pq_2
					neutral_subtrahend: (pq_1 - pq_2.zero) ≍ pq_1

					absorbing_left_factor: (pq_1.zero ⋅ pq_2) ≍ pq_1.zero

						right_good_factor:
								-- Denominator equal to 1 is a neutral factor.
							pq_1.multipliable (pq_2.zero)
					absorbing_right_factor: (pq_1 ⋅ pq_2.zero) ≍ pq_2.zero

					absorbing_dividend: pq_1.zero.divisible (pq_2) implies (pq_1.zero / pq_2) ≍ pq_1.zero
					negative_exponent: pq_2 < pq_2.zero implies
						(pq_1.zero ^ pq_2).is_positive_infinity
					to_power_zero: (pq_1.zero ^ zero) ≍ one
					absorbing_base: pq_2 > pq_1.zero implies
						(pq_1.zero ^ pq_2) ≍ pq_1.zero
					zero_exponent: (pq_1 ^ pq_2.zero) ≍ one
				end
				Result := True
			end
		end

	one_ok (pq_1, pq_2: STS_RATIONAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_RATIONAL_NUMBER}.one?
		do
			if real_one_ok (pq_1, pq_2) then
				check
					neutral_left_factor: (pq_1.one ⋅ pq_2) ≍ pq_2

						right_good_factor: -- 1 is a neutral factor.
							pq_1.multipliable (pq_2.one)
					neutral_right_factor: (pq_1 ⋅ pq_2.one) ≍
						pq_1

						good_divisor: -- 1 is an invertible, neutral factor.
							pq_1.divisible (pq_2.one)
					neutral_divisor: (pq_1 / pq_2.one) ≍ pq_1

					neutral_exponent: (pq_1 ^ pq_2.one) ≍ pq_1
					absorbing_base: (pq_1.one ^ pq_2) ≍ pq_1.one
				end
				Result := True
			end
		end

feature -- Properties (Quality)

	is_nan_ok (pq: STS_RATIONAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for
			-- {STS_RATIONAL_NUMBER}.is_nan?
		do
			check
				never_a_nan: not pq.is_nan
			end
			Result := True
		end

	is_negative_infinity_ok (pq: STS_RATIONAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for
			-- {STS_RATIONAL_NUMBER}.is_negative_infinity?
		do
			check
				never_a_negative_infinity: not pq.is_negative_infinity
			end
			Result := True
		end

	is_positive_infinity_ok (pq: STS_RATIONAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for
			-- {STS_RATIONAL_NUMBER}.is_positive_infinity?
		do
			check
				never_a_positive_infinity: not pq.is_positive_infinity
			end
			Result := True
		end

	is_infinite_ok (pq: STS_RATIONAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for
			-- {STS_RATIONAL_NUMBER}.is_infinite?
		do
			check
				never_is_infinite: not pq.is_infinite
			end
			Result := True
		end

	is_rational_ok (pq: STS_RATIONAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_RATIONAL_NUMBER}.is_rational?
		do
			if real_is_rational_ok (pq) then
				check
					is_rational: pq.is_rational -- By definition.
				end
				Result := True
			end
		end

	is_integer_ok (pq: STS_RATIONAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_RATIONAL_NUMBER}.is_integer?
		do
			check
				stricter_integerness: pq.is_integer implies pq.real_is_integer
			end
			Result := True
		end

feature -- Properties (Comparison)

	equals_ok (pq_1, pq_2, pq_3: STS_RATIONAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_RATIONAL_NUMBER}.equals?
		do
			check
--					q: attached current_rational_universe as q
--					definition: (2 ^ # q) ≤ max_subsets implies pq_1 ≍ pq_2 = q.powerset |∀
--						agent (s: STS_SET [STS_RATIONAL_NUMBER, STS_RATIONAL_EQUALITY]; ia_x, ia_y: STS_RATIONAL_NUMBER): BOOLEAN
--							do
--								Result := ia_x ∈ s = ia_y ∈ s
--							end (?, pq_1, pq_2)
--					by_quantification: pq_1 ≍ pq_2 = (q |∀ agent (w: STS_RATIONAL_NUMBER; ia_x, ia_y: STS_RATIONAL_NUMBER): BOOLEAN
--						do
--							Result := (w.value = ia_x.value) = (w.value = ia_y.value)
--						end (?, pq_1, pq_2))
				complex_consistent: pq_1 ≍ pq_2 implies pq_1.complex_equals (pq_2)
				real_consistent: pq_1 ≍ pq_2 implies pq_1.real_equals (pq_2)
				reflexive: pq_1 ≍ pq_1
				symmetric: pq_1 ≍ pq_2 implies pq_2 ≍ pq_1
				transitive: pq_1 ≍ pq_2 and pq_2 ≍ pq_3 implies pq_1 ≍ pq_3
				euclidian: pq_1 ≍ pq_3 and pq_2 ≍ pq_3 implies pq_1 ≍ pq_2
			end
			Result := True
		end

	unequals_ok (pq_1, pq_2: STS_RATIONAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_RATIONAL_NUMBER}.unequals?
		do
			check
--				q: attached current_rational_universe as q
--				by_existence: pq_1 ≭ pq_2 = (q |∃ agent (k: STS_RATIONAL_NUMBER; ia_pq_1, ia_pq_2: STS_RATIONAL_NUMBER): BOOLEAN
--					do
--						Result := (k.value = ia_pq_1.value) /= (k.value = ia_pq_2.value)
--					end (?, pq_1, pq_2))
--				by_membership: (2 ^ # q) ≤ max_subsets implies pq_1 ≭ pq_2 = q.powerset |∃
--					agent (s: STS_SET [STS_RATIONAL_NUMBER, STS_NUMERIC_EQUALITY [STS_RATIONAL_NUMBER]]; ia_pq_1, ia_pq_2: STS_RATIONAL_NUMBER): BOOLEAN
--						do
--							Result := ia_pq_1 ∈ s /= ia_pq_2 ∈ s
--						end (?, pq_1, pq_2)
				complex_consistent: pq_1.complex_unequals (pq_2) implies pq_1 ≭ pq_2
				real_consistent: pq_1.real_unequals (pq_2) implies pq_1 ≭ pq_2
				irreflexive: not (pq_1 ≭ pq_1)
			end
			Result := True
		end

	is_less_ok (pq_1, pq_2, pq_3: STS_RATIONAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for
			-- {STS_RATIONAL_NUMBER}.is_less?
		do
			check
				consistent: pq_1.real_is_less (pq_2) implies pq_1 < pq_2
				irreflexive: not (pq_1 < pq_1)
				transitive: pq_1 < pq_2 and pq_2 < pq_3 implies pq_1 < pq_3
			end
			Result := True
		rescue
			print (pq_1) print (", ")
			print (pq_2) print (", ")
			print (pq_3) print ('%N')
		end

	is_less_equal_ok (pq_1, pq_2, pq_3: STS_RATIONAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_RATIONAL_NUMBER}.is_less_equal?
		do
			check
				consistent: pq_1 ≤ pq_2 implies pq_1.real_is_less_equal (pq_2)
				reflexive: pq_1 ≤ pq_1
				transitive: pq_1 ≤ pq_2 and pq_2 ≤ pq_3 implies pq_1 ≤ pq_3
				antisymmetric: pq_1 ≤ pq_2 and pq_2 ≤ pq_1 implies pq_1 ≍ pq_2
			end
			Result := True
		end

	is_greater_ok (pq_1, pq_2, pq_3: STS_RATIONAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_RATIONAL_NUMBER}.is_greater?
		do
			check
				consistent: pq_1.real_is_greater (pq_2) implies pq_1 > pq_2
				irreflexive: not (pq_1 > pq_1)
				transitive: pq_1 > pq_2 and pq_2 > pq_3 implies pq_1 > pq_3
			end
			Result := True
		end

	is_greater_equal_ok (pq_1, pq_2, pq_3: STS_RATIONAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_RATIONAL_NUMBER}.is_greater_equal?
		do
			check
				consistent: pq_1 ≥ pq_2 implies pq_1.real_is_greater_equal (pq_2)
				reflexive: pq_1 ≥ pq_1
				transitive: pq_1 ≥ pq_2 and pq_2 ≥ pq_3 implies pq_1 ≥ pq_3
				antisymmetric: pq_1 ≥ pq_2 and pq_2 ≥ pq_1 implies pq_1 ≍ pq_2
			end
			Result := True
		end

	min_ok (pq_1, pq_2, pq_3: STS_RATIONAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theorpq_2 hold for {STS_RATIONAL_NUMBER}.min?
		do
			check
				consistent: pq_1.real_min (pq_2) ≍ (pq_1 ∧ pq_2)
				idempotent: (pq_1 ∧ pq_1) ≍ pq_1
				commutative: (pq_1 ∧ pq_2) ≍ (pq_2 ∧ pq_1)
				associative: ((pq_1 ∧ pq_2) ∧ pq_3) ≍ (pq_1 ∧ (pq_2 ∧ pq_3))
			end
			Result := True
		end

	max_ok (pq_1, pq_2, pq_3: STS_RATIONAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theorpq_2 hold for {STS_RATIONAL_NUMBER}.max?
		do
			check
				consistent: pq_1.real_max (pq_2) ≍ (pq_1 ∨ pq_2)
				idempotent: (pq_1 ∨ pq_1) ≍ pq_1
				commutative: (pq_1 ∨ pq_2) ≍ (pq_2 ∨ pq_1)
				associative: ((pq_1 ∨ pq_2) ∨ pq_3) ≍ (pq_1 ∨ (pq_2 ∨ pq_3))
			end
			Result := True
		end

feature -- Properties (Relationship)

	multipliable_ok (pq_1, pq_2: STS_RATIONAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for
			-- {STS_RATIONAL_NUMBER}.multipliable?
		do
			check
				gcd_1: attached gcd (pq_2.p, pq_1.q) as gcd_1
				gcd_2: attached gcd (pq_1.p, pq_2.q) as gcd_2
				good_divisor_1: -- gcd_1 /= 0 ⇐ pq_1.q /= 0
					pq_1.q.divisible (gcd_1)
				good_divisor_2: -- gcd_2 /= 0 ⇐ pq_2.q /= 0
					pq_2.q.divisible (gcd_2)
				unexpected_zero_product:
						-- Which is possible only upon an overflow.
					(pq_1.q // gcd_1) ⋅ (pq_2.q // gcd_2) ≍ zero.p implies
					not pq_1.multipliable (pq_2)
			end
			Result := True
		end

feature -- Properties (Operation)

	plus_ok (pq_1, pq_2, pq_3: STS_RATIONAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for
			-- {STS_RATIONAL_NUMBER}.plus?
		local
			broken_associativity, broken_morphism: BOOLEAN
		do
			if not (broken_associativity or broken_morphism) then
				check
					neutral_left_term: (zero + pq_1) ≍ pq_1
					neutral_right_term: (pq_1 + zero) ≍ pq_1
					commutative: (pq_1 + pq_2) ≍ (pq_2 + pq_1)
					associative:
						((pq_1 + pq_2) + pq_3) ≍ (pq_1 + (pq_2 + pq_3))
				end
			end
			if not broken_morphism then
				check
					homomorphic:
						pq_1 ≤ pq_2 implies (pq_1 + pq_3) ≤ (pq_2 + pq_3)
					isomorphic: pq_1 < pq_2 implies pq_1 + pq_3 < pq_2 + pq_3
				end
			end
			Result := True
		rescue
			print (pq_1) print (", ")
			print (pq_2) print (", ")
			print (pq_3) print ('%N')
			if
				not broken_associativity and
				((pq_1 + pq_2) + pq_3) ≭ (pq_1 + (pq_2 + pq_3))
			then
				check
					gcd_12: attached gcd (pq_1.q, pq_2.q) as gcd_12
						non_zero_gcd_12: gcd_12 ≭ zero.p -- pq_1.q /= 0 and pq_2.q /= 0
						good_divisor_12: ( -- (pq_2.q // gcd_12) /= 0
							pq_1.p * (pq_2.q // gcd_12) + pq_2.p * (pq_1.q // gcd_12)
							).divisible (pq_1.q * (pq_2.q // gcd_12))
					sum_12: attached ((pq_1.p * (pq_2.q // gcd_12) + pq_2.p * (pq_1.q // gcd_12)) / (pq_1.q * (pq_2.q // gcd_12))) as sum_12
					gcd_12_3: attached gcd (sum_12.q, pq_3.q) as gcd_12_3
						non_zero_gcd_12_3: gcd_12_3 ≭ zero.p -- sum_12.q /= 0 and pq_3.q /= 0
						good_divisor_12_3: ( -- (pq_3.q // gcd_12_3) /= 0
							sum_12.p * (pq_3.q // gcd_12_3) + pq_3.p * (sum_12.q // gcd_12_3)
							).divisible (sum_12.q * (pq_3.q // gcd_12_3))
					sum_12_3: ((pq_1 + pq_2) + pq_3) ≍
						((sum_12.p * (pq_3.q // gcd_12_3) + pq_3.p * (sum_12.q // gcd_12_3)) / (sum_12.q * (pq_3.q // gcd_12_3)))

					gcd_23: attached gcd (pq_2.q, pq_3.q) as gcd_23
						non_zero_gcd_23: gcd_23 ≭ zero.p -- pq_2.q /= 0 and pq_3.q /= 0
						good_divisor_23: ( -- (pq_3.q // gcd_23) /= 0
							pq_2.p * (pq_3.q // gcd_23) + pq_3.p * (pq_2.q // gcd_23)
							).divisible (pq_2.q * (pq_3.q // gcd_23))
					sum_23: attached ((pq_2.p * (pq_3.q // gcd_23) + pq_3.p * (pq_2.q // gcd_23)) / (pq_2.q * (pq_3.q // gcd_23))) as sum_23
					gcd_1_23: attached gcd (pq_1.q, sum_23.q) as gcd_1_23
						non_zero_gcd_1_23: gcd_1_23 ≭ zero.p -- pq_1.q /= 0 and sum_23.q /= 0
						good_divisor_12_3: ( -- (sum_23.q // gcd_1_23) /= 0
							pq_1.p * (sum_23.q // gcd_1_23) + sum_23.p * (pq_1.q // gcd_1_23)
							).divisible (pq_1.q * (sum_23.q // gcd_1_23))
					sum_1_23: (pq_1 + (pq_2 + pq_3)) ≍
						((pq_1.p * (sum_23.q // gcd_1_23) + sum_23.p * (pq_1.q // gcd_1_23)) / (pq_1.q * (sum_23.q // gcd_1_23)))

					broken_associativity: product_overflows (pq_1.p, pq_2.q // gcd_12) or
						product_overflows (pq_2.p, pq_1.q // gcd_12) or
						sum_overflows (pq_1.p * (pq_2.q // gcd_12), pq_2.p * (pq_1.q // gcd_12)) or
						product_overflows (pq_1.q, pq_2.q // gcd_12) or
						product_overflows (sum_12.p, pq_3.q // gcd_12_3) or
						product_overflows (pq_3.p, sum_12.q // gcd_12_3) or
						sum_overflows (sum_12.p * (pq_3.q // gcd_12_3), pq_3.p * (sum_12.q // gcd_12_3)) or
						product_overflows (sum_12.q, pq_3.q // gcd_12_3) or

						product_overflows (pq_2.p, pq_3.q // gcd_23) or
						product_overflows (pq_3.p, pq_2.q // gcd_23) or
						sum_overflows (pq_2.p * (pq_3.q // gcd_23), pq_3.p * (pq_2.q // gcd_23)) or
						product_overflows (pq_2.q, pq_3.q // gcd_23) or
						product_overflows (pq_1.p, sum_23.q // gcd_1_23) or
						product_overflows (sum_23.p, pq_1.q // gcd_1_23) or
						sum_overflows (pq_1.p * (sum_23.q // gcd_1_23), sum_23.p * (pq_1.q // gcd_1_23)) or
						product_overflows (pq_1.q, sum_23.q // gcd_1_23)

				end
				broken_associativity := True
				retry
			elseif
				not broken_morphism and
				pq_1 ≤ pq_2 and (pq_1 + pq_3) > (pq_2 + pq_3)
			then
				check
					gcd_13: attached gcd (pq_1.q, pq_3.q) as gcd_13
						non_zero_gcd_13: gcd_13 ≭ zero.p -- pq_1.q /= 0 and pq_3.q /= 0
						good_divisor_13: ( -- (pq_3.q // gcd_13) /= 0
							pq_1.p * (pq_3.q // gcd_13) + pq_3.p * (pq_1.q // gcd_13)
							).divisible (pq_1.q * (pq_3.q // gcd_13))
					sum_13: (pq_1 + pq_3) ≍ ((pq_1.p * (pq_3.q // gcd_13) + pq_3.p * (pq_1.q // gcd_13)) / (pq_1.q * (pq_3.q // gcd_13)))

					gcd_23: attached gcd (pq_2.q, pq_3.q) as gcd_23
						non_zero_gcd_23: gcd_23 ≭ zero.p -- pq_2.q /= 0 and pq_3.q /= 0
						good_divisor_23: ( -- (pq_3.q // gcd_23) /= 0
							pq_2.p * (pq_3.q // gcd_23) + pq_3.p * (pq_2.q // gcd_23)
							).divisible (pq_2.q * (pq_3.q // gcd_23))
					sum_23: (pq_2 + pq_3) ≍ ((pq_2.p * (pq_3.q // gcd_23) + pq_3.p * (pq_2.q // gcd_23)) / (pq_2.q * (pq_3.q // gcd_23)))

					broken_morphism: product_overflows (pq_1.p, pq_3.q // gcd_13) or
						product_overflows (pq_3.p, pq_1.q // gcd_13) or
						sum_overflows (pq_1.p * (pq_3.q // gcd_13), pq_3.p * (pq_1.q // gcd_13)) or
						product_overflows (pq_1.q, pq_3.q // gcd_13) or

						product_overflows (pq_2.p, pq_3.q // gcd_23) or
						product_overflows (pq_3.p, pq_2.q // gcd_23) or
						sum_overflows (pq_2.p * (pq_3.q // gcd_23), pq_3.p * (pq_2.q // gcd_23)) or
						product_overflows (pq_2.q, pq_3.q // gcd_23)
				end
				broken_morphism := True
				retry
			end
		end

	identity_ok (pq: STS_RATIONAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for
			-- {STS_RATIONAL_NUMBER}.identity?
		do
			if
				real_identity_ok (pq, pq) and -- TODO: The second pq will be removed.
				to_rational_ok (pq)
			then
				check
					as_left_term: + pq ≍ (pq + zero)
					as_right_term: + pq ≍ (zero + pq)

						left_good_factor: -- 1 is a neutral factor.
							pq.multipliable (one)
					as_left_factor: + pq ≍ (pq ⋅ one)

					as_right_factor: + pq ≍ (one ⋅ pq)

						good_divisor: -- 1 is an invertible, neutral factor.
							pq.divisible (one)
					as_quotient: + pq ≍ (pq / one)
				end
				Result := True
			end
		end

	minus_ok (pq_1, pq_2, pq_3: STS_RATIONAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for
			-- {STS_RATIONAL_NUMBER}.minus?
		local
			broken_homomorphism, broken_isomorphism: BOOLEAN
		do
			if not (broken_homomorphism or broken_isomorphism) then
				check
					quasi_neutral_minuend: (zero - pq_1) ≍ -pq_1
					neutral_subtrahend: (pq_1 - zero) ≍ pq_1
					quasi_commutative: (pq_1 - pq_2).abs ≍ (pq_2 - pq_1).abs
					oppositely_homomorphic: pq_1 ≤ pq_2 implies (pq_3 - pq_1) ≥ (pq_3 - pq_2)
				end
			end
			if not broken_isomorphism then
				check
					oppositely_isomorphic: pq_1 < pq_2 implies pq_3 - pq_1 > pq_3 - pq_2
				end
			end
			Result := True
		rescue
			print (pq_1) print (", ")
			print (pq_2) print (", ")
			print (pq_3) print ('%N')
			if
				not broken_homomorphism and
				pq_1 ≤ pq_2 and (pq_3 - pq_1) < (pq_3 - pq_2)
			then
				check
					gcd_31: attached gcd (pq_3.q, pq_1.q) as gcd_31
						non_zero_gcd_31: gcd_31 ≭ zero.p -- pq_3.q /= 0 and pq_1.q /= 0
						good_divisor_31: ( -- (pq_1.q // gcd_31) /= 0
							pq_3.p * (pq_1.q // gcd_31) - pq_1.p * (pq_3.q // gcd_31)
							).divisible (pq_3.q * (pq_1.q // gcd_31))
					diff_31: (pq_3 - pq_1) ≍ (
						(pq_3.p * (pq_1.q // gcd_31) - pq_1.p * (pq_3.q // gcd_31)) /
						(pq_3.q * (pq_1.q // gcd_31))
						)

					gcd_32: attached gcd (pq_3.q, pq_2.q) as gcd_32
						non_zero_gcd_32: gcd_32 ≭ zero.p -- pq_3.q /= 0 and pq_2.q /= 0
						good_divisor_32: ( -- (pq_2.q // gcd_32) /= 0
							pq_3.p * (pq_2.q // gcd_32) - pq_2.p * (pq_3.q // gcd_32)
							).divisible (pq_3.q * (pq_2.q // gcd_32))
					diff_32: (pq_3 - pq_2) ≍ (
						(pq_3.p * (pq_2.q // gcd_32) - pq_2.p * (pq_3.q // gcd_32)) /
						(pq_3.q * (pq_2.q // gcd_32))
						)

					broken_homomorphism: product_overflows (pq_3.p, pq_1.q // gcd_31) or
						product_overflows (pq_1.p, pq_3.q // gcd_31) or
						difference_overflows (pq_3.p * (pq_1.q // gcd_31), pq_1.p * (pq_3.q // gcd_31)) or
						product_overflows (pq_3.q, pq_1.q // gcd_31) or

						product_overflows (pq_3.p, pq_2.q // gcd_32) or
						product_overflows (pq_2.p, pq_3.q // gcd_32) or
						difference_overflows (pq_3.p * (pq_2.q // gcd_32), pq_2.p * (pq_3.q // gcd_32)) or
						product_overflows (pq_3.q, pq_2.q // gcd_32)
				end
				broken_homomorphism := True
				retry
			elseif
				not broken_isomorphism and
				pq_1 < pq_2 and (pq_3 - pq_1) ≤ (pq_3 - pq_2)
			then
				check
					gcd_31: attached gcd (pq_3.q, pq_1.q) as gcd_31
						non_zero_gcd_31: gcd_31 ≭ zero.p -- pq_3.q /= 0 and pq_1.q /= 0
						good_divisor_31: ( -- (pq_1.q // gcd_31) /= 0
							pq_3.p * (pq_1.q // gcd_31) - pq_1.p * (pq_3.q // gcd_31)
							).divisible (pq_3.q * (pq_1.q // gcd_31))
					diff_31: (pq_3 - pq_1) ≍ (
						(pq_3.p * (pq_1.q // gcd_31) - pq_1.p * (pq_3.q // gcd_31)) /
						(pq_3.q * (pq_1.q // gcd_31))
						)

					gcd_32: attached gcd (pq_3.q, pq_2.q) as gcd_32
						non_zero_gcd_32: gcd_32 ≭ zero.p -- pq_3.q /= 0 and pq_2.q /= 0
						good_divisor_32: ( -- (pq_2.q // gcd_32) /= 0
							pq_3.p * (pq_2.q // gcd_32) - pq_2.p * (pq_3.q // gcd_32)
							).divisible (pq_3.q * (pq_2.q // gcd_32))
					diff_32: (pq_3 - pq_2) ≍ (
						(pq_3.p * (pq_2.q // gcd_32) - pq_2.p * (pq_3.q // gcd_32)) /
						(pq_3.q * (pq_2.q // gcd_32))
						)

					broken_isomorphism: product_overflows (pq_3.p, pq_1.q // gcd_31) or
						product_overflows (pq_1.p, pq_3.q // gcd_31) or
						difference_overflows (pq_3.p * (pq_1.q // gcd_31), pq_1.p * (pq_3.q // gcd_31)) or
						product_overflows (pq_3.q, pq_1.q // gcd_31) or

						product_overflows (pq_3.p, pq_2.q // gcd_32) or
						product_overflows (pq_2.p, pq_3.q // gcd_32) or
						difference_overflows (pq_3.p * (pq_2.q // gcd_32), pq_2.p * (pq_3.q // gcd_32)) or
						product_overflows (pq_3.q, pq_2.q // gcd_32)
				end
				broken_isomorphism := True
				retry
			end
		end

	opposite_ok (pq: STS_RATIONAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for
			-- {STS_RATIONAL_NUMBER}.opposite?
		do
			check
				inverted_sign: (- pq).sign ≍ - pq.sign
				opposite_left_term: ((- pq) + pq) ≍ zero
				opposite_right_term: (pq + (- pq)) ≍ zero
				as_subtrahend: - pq ≍ (zero - pq)
				as_left_factor: pq.multipliable (- one) implies
					- pq ≍ (pq ⋅ (- one))
				as_right_factor: (- one).multipliable (pq) implies - pq ≍ ((- one) ⋅ pq)

					divisible: -- -1 is an invertible, quasi-neutral factor.
						pq.divisible (- one)
				as_quotient: - pq ≍ (pq / (- one))
			end
			Result := True
		end

	product_ok (pq_1, pq_2, pq_3: STS_RATIONAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for
			-- {STS_RATIONAL_NUMBER}.product?
		local
			broken_homomorphism: BOOLEAN
		do
			if not broken_homomorphism then
				check
					neutral_left_factor: (one ⋅ pq_1) ≍ pq_1

						right_good_factor: -- 1 is a neutral factor.
							pq_1.multipliable (one)
					neutral_right_factor: pq_1 ⋅ one ≍ pq_1

					absorbing_left_factor: (zero ⋅ pq_1) ≍
						zero

						absorbing_right_good_factor:
								-- Denominator equal to 1 is a neutral factor.
							pq_1.multipliable (zero)
					absorbing_right_factor: pq_1 ⋅ zero ≍
						zero
					commutative: pq_1.multipliable (pq_2) implies
						pq_2.multipliable (pq_1) -- Commutativity
						and then (pq_1 ⋅ pq_2) ≍ (pq_2 ⋅ pq_1)

					quasi_homomorphic: pq_1.abs ≤ pq_2.abs and
						pq_1.multipliable (pq_3) and
						pq_2.multipliable (pq_3) implies
						(pq_1 ⋅ pq_3).abs ≤ (pq_2 ⋅ pq_3).abs
					quasi_isomorphic: pq_1 ≭ zero.p and
						pq_1.abs < pq_2.abs and
						pq_3 ≭ zero.p and
						pq_1.multipliable (pq_3) and
						pq_2.multipliable (pq_3) implies
						(pq_1 ⋅ pq_3).abs < (pq_2 ⋅ pq_3).abs
				end
			end
			Result := True
		rescue
			print (pq_1) print (", ")
			print (pq_2) print (", ")
			print (pq_3) print ('%N')
			if
				not broken_homomorphism and
				pq_1.abs ≤ pq_2.abs and (
						(
							pq_1.multipliable (pq_3) and
							pq_2.multipliable (pq_3)
							) and then
					(pq_1 ⋅ pq_3).abs > (pq_2 ⋅ pq_3).abs
					)
			then
				check
					gcd_1: attached gcd (pq_1.p, pq_3.q) as gcd_1
					gcd_2: attached gcd (pq_3.p, pq_1.q) as gcd_2
					non_zero_gcd_1: gcd_1 ≭ zero.p -- pq_3.q /= 0
					non_zero_gcd_2: gcd_2 ≭ zero.p -- pq_1.q /= 0
					good_divisor_13: -- pq_1.q /= 0 and pq_3.q /= 0
						((pq_1.p // gcd_1) * (pq_3.p // gcd_2)).divisible
							((pq_1.q // gcd_2) * (pq_3.q // gcd_1))
					pq_1_pq_3_unfolded_numerator: (pq_1 * pq_3).p ≍
						(
							((pq_1.p // gcd_1) * (pq_3.p // gcd_2)) /
							((pq_1.q // gcd_2) * (pq_3.q // gcd_1))
							).p
					pq_1_pq_3_unfolded_denominator: (pq_1 * pq_3).q ≍
						(
							((pq_1.p // gcd_1) * (pq_3.p // gcd_2)) /
							((pq_1.q // gcd_2) * (pq_3.q // gcd_1))
							).q

					gcd_3: attached gcd (pq_2.p, pq_3.q) as gcd_3
					gcd_4: attached gcd (pq_3.p, pq_2.q) as gcd_4
					non_zero_gcd_3: gcd_3 ≭ zero.p -- pq_3.q /= 0
					non_zero_gcd_4: gcd_4 ≭ zero.p -- pq_2.q /= 0
					good_divisor_23: -- pq_2.q /= 0 and pq_3.q /= 0
						((pq_2.p // gcd_3) * (pq_3.p // gcd_4)).divisible
							((pq_2.q // gcd_4) * (pq_3.q // gcd_3))
					pq_2_pq_3_unfolded_numerator: (pq_2 * pq_3).p ≍
						(
							((pq_2.p // gcd_3) * (pq_3.p // gcd_4)) /
							((pq_2.q // gcd_4) * (pq_3.q // gcd_3))
							).p
					pq_2_pq_3_unfolded_denominator: (pq_2 * pq_3).q ≍
						(
							((pq_2.p // gcd_3) * (pq_3.p // gcd_4)) /
							((pq_2.q // gcd_4) * (pq_3.q // gcd_3))
							).q

					broken_homomorphism:
						pq_1.abs < zero.p -- An integer overflow might yield a negative rational "absolute" value.
							or
							-- An implementation might allow some overflows that
							-- do not yield a denominator equal to 0.
						product_overflows
							(pq_1.p // gcd_1, pq_3.p // gcd_2) or
						product_overflows
							(pq_1.q // gcd_2, pq_3.q // gcd_1) or
						product_overflows
							(pq_2.p // gcd_3, pq_3.p // gcd_4) or
						product_overflows
							(pq_2.q // gcd_4, pq_3.q // gcd_3)
				end
				broken_homomorphism := True
				retry
			end
		end

	quotient_ok (pq_1, pq_2, pq_3: STS_RATIONAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for
			-- {STS_RATIONAL_NUMBER}.quotient?
		local
			broken_morphism: BOOLEAN
		do
			if not broken_morphism then
				check
					absorbing_dividend: zero.divisible (pq_2) implies (zero / pq_2) ≍ zero

						good_divisor: -- 1 is an invertible, neutral factor.
							pq_1.divisible (one)
					neutral_divisor: (pq_1 / one) ≍ pq_1

					quasi_oppositely_homomorphic: pq_1.abs ≤ pq_2.abs and
						pq_3.divisible (pq_1) and pq_3.divisible (pq_2) implies
						(pq_3 / pq_1).abs ≥ (pq_3 / pq_2).abs
				end
			end
			Result := True
		rescue
			print (pq_1) print (", ")
			print (pq_2) print (", ")
			print (pq_3) print ('%N')
			if
				not broken_morphism and
				pq_1.abs ≤ pq_2.abs and
				pq_3.divisible (pq_1) and
				pq_3.divisible (pq_2) and then
				(pq_3 / pq_1).abs < (pq_3 / pq_2).abs
			then
				check
					pq_1_is_invertible: pq_1.is_invertible -- pq_3.divisible (pq_1)
					pq_3_p_divisible_gcd_1: pq_3.p.divisible (gcd (pq_3.p, pq_1.inverse.q)) -- gcd (pq_3.p, pq_1.inverse.q) /= 0 ⇦ pq_1.inverse.q /= 0
					pq_1_inverse_p_divisible_gcd_2: pq_1.inverse.p.divisible (gcd (pq_1.inverse.p, pq_3.q)) -- gcd (pq_1.inverse.p, pq_3.q) /= 0 ⇦ pq_3.q /= 0
					pq_3_q_divisible_gcd_2: pq_3.q.divisible (gcd (pq_1.inverse.p, pq_3.q))
					pq_1_inverse_q_divisible_gcd_1: pq_1.inverse.q.divisible (gcd (pq_3.p, pq_1.inverse.q))
					unfolded_pq_3_by_pq_1_good_divisor:
							-- ((pq_3.q // gcd (pq_1.inverse.p, pq_3.q)) ⋅ (pq_1.inverse.q // gcd (pq_3.p, pq_1.inverse.q))) /= 0 ⇦
							-- ⇦ pq_3.q /= 0 and pq_1.inverse.q /= 0 and pq_3.divisible (pq_1)
						((pq_3.p // gcd (pq_3.p, pq_1.inverse.q)) ⋅ (pq_1.inverse.p // gcd (pq_1.inverse.p, pq_3.q))).divisible
							((pq_3.q // gcd (pq_1.inverse.p, pq_3.q)) ⋅ (pq_1.inverse.q // gcd (pq_3.p, pq_1.inverse.q)))
					unfolded_pq_3_by_pq_1: (pq_3 / pq_1) ≍ (
						((pq_3.p // gcd (pq_3.p, pq_1.inverse.q)) ⋅ (pq_1.inverse.p // gcd (pq_1.inverse.p, pq_3.q))) /
						((pq_3.q // gcd (pq_1.inverse.p, pq_3.q)) ⋅ (pq_1.inverse.q // gcd (pq_3.p, pq_1.inverse.q)))
						)

					pq_2_is_invertible: pq_2.is_invertible -- pq_3.divisible (pq_2)
					pq_3_p_divisible_gcd_1: pq_3.p.divisible (gcd (pq_3.p, pq_2.inverse.q)) -- gcd (pq_3.p, pq_2.inverse.q) /= 0 ⇦ pq_2.inverse.q /= 0
					pq_2_inverse_p_divisible_gcd_2: pq_2.inverse.p.divisible (gcd (pq_2.inverse.p, pq_3.q)) -- gcd (pq_2.inverse.p, pq_3.q) /= 0 ⇦ pq_3.q /= 0
					pq_3_q_divisible_gcd_2: pq_3.q.divisible (gcd (pq_2.inverse.p, pq_3.q))
					pq_2_inverse_q_divisible_gcd_1: pq_2.inverse.q.divisible (gcd (pq_3.p, pq_2.inverse.q))
					unfolded_pq_3_by_pq_2_good_divisor:
							-- ((pq_3.q // gcd (pq_2.inverse.p, pq_3.q)) ⋅ (pq_2.inverse.q // gcd (pq_3.p, pq_2.inverse.q))) /= 0 ⇦
							-- ⇦ pq_3.q /= 0 and pq_2.inverse.q /= 0 and pq_3.divisible (pq_2)
						((pq_3.p // gcd (pq_3.p, pq_2.inverse.q)) ⋅ (pq_2.inverse.p // gcd (pq_2.inverse.p, pq_3.q))).divisible
							((pq_3.q // gcd (pq_2.inverse.p, pq_3.q)) ⋅ (pq_2.inverse.q // gcd (pq_3.p, pq_2.inverse.q)))
					unfolded_pq_3_by_pq_2: (pq_3 / pq_2) ≍ (
						((pq_3.p // gcd (pq_3.p, pq_2.inverse.q)) ⋅ (pq_2.inverse.p // gcd (pq_2.inverse.p, pq_3.q))) /
						((pq_3.q // gcd (pq_2.inverse.p, pq_3.q)) ⋅ (pq_2.inverse.q // gcd (pq_3.p, pq_2.inverse.q)))
						)

					broken_morphism:
						pq_1.abs < zero.p -- An integer overflow might yield a negative rational "absolute" value.
							or
							-- An implementation might allow some overflows that
							-- do not yield a denominator equal to 0.
						product_overflows (
							pq_3.p // gcd (pq_3.p, pq_1.inverse.q),
							pq_1.inverse.p // gcd (pq_1.inverse.p, pq_3.q)
							) or
						product_overflows (
							pq_3.q // gcd (pq_1.inverse.p, pq_3.q),
							pq_1.inverse.q // gcd (pq_3.p, pq_1.inverse.q)
							) or
						product_overflows (
							pq_3.p // gcd (pq_3.p, pq_2.inverse.q),
							pq_2.inverse.p // gcd (pq_2.inverse.p, pq_3.q)
							) or
						product_overflows (
							pq_3.q // gcd (pq_2.inverse.p, pq_3.q),
							pq_2.inverse.q // gcd (pq_3.p, pq_2.inverse.q)
							)
				end
				broken_morphism := True
				retry
			end
		end

	reciprocal_ok (pq_1, pq_2: STS_RATIONAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for
			-- {STS_RATIONAL_NUMBER}.reciprocal?
		local
			broken_morphism: BOOLEAN
		do
			if pq_1.is_invertible then
				check
						good_divisor: one.divisible (pq_1) -- pq_1.is_invertible and good_factor and one.rational_multipliable (pq_1)							
					definition: pq_1.reciprocal ≍ (one / pq_1)
				end
				if pq_2.is_invertible then
					if not broken_morphism then
						check
							quasi_oppositely_homomorphic: (pq_1.abs ≤ pq_2.abs) =
								(pq_1.reciprocal.abs ≥ pq_2.reciprocal.abs)
							quasi_oppositely_isomorphic: (pq_1.abs < pq_2.abs) =
								(pq_1.reciprocal.abs > pq_2.reciprocal.abs)
						end
					end
				end
			end
			Result := True
		rescue
			print (pq_1) print (", ")
			print (pq_2) print ('%N')
			if not broken_morphism and (pq_1.abs ≤ pq_2.abs) /= (pq_1.reciprocal.abs ≥ pq_2.reciprocal.abs) then
				check
					i: attached pq_1.truncated_to_integer as i
					negative_abs: pq_1.is_integer and (i.min_value_exists and then i ≍ i.min_value) and pq_1.abs < zero
				end
				broken_morphism := True
				retry
			end
		end

	inverse_ok (pq_1, pq_2: STS_RATIONAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for
			-- {STS_RATIONAL_NUMBER}.inverse?
		local
			broken_morphism: BOOLEAN
		do
			if pq_1.is_invertible then
				check
						good_divisor: one.divisible (pq_1) -- pq_1.is_invertible and good_factor and one.rational_multipliable (pq_1)							
					definition: pq_1.inverse ≍ (one / pq_1)
				end
				if pq_2.is_invertible then
					if not broken_morphism then
						check
							quasi_oppositely_homomorphic: (pq_1.abs ≤ pq_2.abs) =
								(pq_1.inverse.abs ≥ pq_2.inverse.abs)
							quasi_oppositely_isomorphic: (pq_1.abs < pq_2.abs) =
								(pq_1.inverse.abs > pq_2.inverse.abs)
						end
					end
				end
			end
			Result := True
		rescue
			print (pq_1) print (", ")
			print (pq_2) print ('%N')
			if not broken_morphism and (pq_1.abs ≤ pq_2.abs) /= (pq_1.inverse.abs ≥ pq_2.inverse.abs) then
				check
					i: attached pq_1.truncated_to_integer as i
					negative_abs: pq_1.is_integer and (i.min_value_exists and then i ≍ i.min_value) and pq_1.abs < zero
				end
				broken_morphism := True
				retry
			end
		end

feature -- Properties (Math)

	gcd_ok (pq: STS_RATIONAL_NUMBER; i, j, k: STS_INTEGER_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for
			-- {STS_RATIONAL_NUMBER}.gcd?
		local
			broken_product_divisor,
			broken_factor,
			broken_sameness,
			broken_quotient,
			broken_function,
			broken_coprimes: BOOLEAN
		do
			if
				not (
					broken_product_divisor or
					broken_factor or
					broken_sameness or
					broken_quotient or
					broken_function or
					broken_coprimes
					)
			then
				check
					common_divisor: k | i and k | j implies k | pq.gcd (i, j)
						i_divisible_gcd_i_j:
								-- pq.gcd (i, j) /= 0 ⇐ i /= 0 ⇐ i | (j ⋅ k)
							i | (j ⋅ k) implies i.divisible (pq.gcd (i, j))
					product_divisor: i | (j ⋅ k) implies (i // pq.gcd (i, j)) | k
				end
			end
			if
				not (
					broken_factor or
					broken_sameness or
					broken_quotient or
					broken_function or
					broken_coprimes
					)
			then
				check
					common_factor: zero ≤ k implies
						pq.gcd (i ⋅ k, j ⋅ k) ≍ (pq.gcd (i, j) ⋅ k)
				end
			end
			if
				not (
					broken_sameness or
					broken_quotient or
					broken_function or
					broken_coprimes
					)
			then
				check
					same_gcd: pq.gcd (i + j ⋅ k, j) ≍ pq.gcd (i, j)
				end
			end
			if not (broken_quotient or broken_function or broken_coprimes) then
				check
						i_divisible_k: zero < k implies i.divisible (k)
						j_divisible_k: zero < k implies j.divisible (k)
						gcd_ij_divisible_k: zero < zero.p implies
							pq.gcd (i, j).divisible (k)
					quotient: zero < k and k | i and k | j implies
						pq.gcd (i // k, j // k) ≍ (pq.gcd (i, j) // k)
				end
			end
			if not (broken_function or broken_coprimes) then
				check
					multiplicative_function: pq.gcd (i, j) ≍ one.p implies
						pq.gcd (i ⋅ j, k) ≍ (pq.gcd (i, k) ⋅ pq.gcd (j, k))
				end
			end
			if not broken_coprimes then
				check
					coprimes: pq.gcd (i, j ⋅ k) ≍ one.p =
						(pq.gcd (i, j) ≍ one.p and pq.gcd (i, k) ≍ one.p)
				end
			end
			check
				commutative: pq.gcd (i, j) ≍ pq.gcd (j, i)
				associative: pq.gcd (i, pq.gcd (j, k)) ≍
					pq.gcd (pq.gcd (i, j), k)
			end
			Result := True
		rescue
			print (pq) print (", ")
			print (i) print (", ")
			print (j) print (", ")
			print (k) print ('%N')
				check
					i_divisible_gcd_i_j:
							-- pq.gcd (i, j) /= 0 ⇐ i /= 0 ⇐ i | (j ⋅ k)
						i | (j ⋅ k) implies i.divisible (pq.gcd (i, j))
					i_divisible_k: zero < k implies i.divisible (k)
					j_divisible_k: zero < k implies j.divisible (k)
					gcd_ij_divisible_k: zero < zero.p implies
						pq.gcd (i, j).divisible (k)
				end
			if
				not broken_product_divisor and
				i | (j ⋅ k) and then not ((i // pq.gcd (i, j)) | k)
			then
				check
					broken_product_divisor: product_overflows (j, k)
				end
				broken_product_divisor := True
				retry
			elseif
				not broken_factor and
				zero ≤ k and pq.gcd (i ⋅ k, j ⋅ k) ≭ (pq.gcd (i, j) ⋅ k)
			then
				check
					broken_factor: product_overflows (i, k) or
						product_overflows (j, k) or
						product_overflows (pq.gcd (i, j), k)
				end
				broken_factor := True
				retry
			elseif
				not broken_sameness and
				pq.gcd (i + j ⋅ k, j) ≭ pq.gcd (i, j)
			then
				check
					broken_sameness: product_overflows (j, k) or
						sum_overflows (i, j ⋅ k)
				end
				broken_sameness := True
				retry
			elseif
				not broken_quotient and
				(zero < k and k | i and k | j) and then (
					pq.gcd (i // k, j // k) ≭ (pq.gcd (i, j) // k)
					)
			then
				check
						-- Possible when integers are implemented as two's
						-- complement binary numbers.
					broken_quotient: pq.gcd (i, j) < zero.p
				end
				broken_quotient := True
				retry
			elseif
				not broken_function and
				pq.gcd (i, j) ≍ one.p and
				pq.gcd (i ⋅ j, k) ≭ (pq.gcd (i, k) ⋅ pq.gcd (j, k))
			then
				check
					broken_function: product_overflows (i, j) or
						product_overflows (pq.gcd (i, k), pq.gcd (j, k))
				end
				broken_function := True
				retry
			elseif
				not broken_coprimes and (
					pq.gcd (i, j ⋅ k) ≍ one.p /=
					(pq.gcd (i, j) ≍ one.p and pq.gcd (i, k) ≍ one.p)
					)
			then
				check
					broken_coprimes: product_overflows (j, k)
				end
				broken_coprimes := True
				retry
			end
		end

	div_ok (pq: STS_RATIONAL_NUMBER; i, j: STS_INTEGER_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for
			-- {STS_RATIONAL_NUMBER}.div?
		local
			broken_divisor, broken_dividend: BOOLEAN
			k: STS_INTEGER_NUMBER
		do
			if i.divisible (j) then
				k := if (i \\ j) ≍ zero.p then zero.p else One.p end
				if not (broken_divisor or broken_dividend) then
					check
						good_divisor:
								-- j.rational_abs.value /= 0 ⇐ i.divisible (j)
							i.real_value.divisible (j.rational_abs.value)
						quotient_floor: attached pq.real_from_value (
							Real_math_anchor.floor (i.real_value / j.real_value)
							).truncated_to_integer as quotient_floor
						definition: pq.div (i, j) ≍ quotient_floor
						consistent: (pq.div (i, j) ⋅ j + pq.rem (i, j)) ≍ i
						opposite_divisor:
							pq.div (i, -j) ≍ (- pq.div (i, j) - k)
					end
				end
				if not broken_dividend then
					check
						opposite_dividend:
							pq.div (-i, j) ≍ (- pq.div (i, j) - k)
					end
				end
			end
			Result := True
		rescue
			check attached k then -- If not, nothing else matters. :-)
				if
					not broken_divisor and
					pq.div (i, -j) ≭ (- pq.div (i, j) - k)
				then
					check
							-- Min value of a two's complement integer...
						broken_divisor: -j ≍ j
						broken_quotient: pq.div (i, -j) ≍ pq.div (i, j)
					end
					broken_divisor := True
					retry
				elseif
					not broken_dividend and
					pq.div (-i, j) ≭ (- pq.div (i, j) - k)
				then
					check
							-- Min value of a two's complement integer...
						broken_dividend: -i ≍ i
						broken_quotient: pq.div (-i, j) ≍ pq.div (i, j)
					end
					broken_dividend := True
					retry
				end
			end
		end

	rem_ok (pq: STS_RATIONAL_NUMBER; i, j, k: STS_INTEGER_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for
			-- {STS_RATIONAL_NUMBER}.rem?
		local
			broken_abs,
			broken_divisor,
			broken_dividend,
			broken_sum,
			broken_product: BOOLEAN
			l: STS_INTEGER_NUMBER
		do
			if i.divisible (j) then
				if
					not (
						broken_abs or
						broken_divisor or
						broken_dividend or
						broken_sum or
						broken_product
						)
				then
					check
						consistent: (pq.div (i, j) ⋅ j + pq.rem (i, j)) ≍ i
						signal_preserving: pq.rem (i, j) ≭ zero.p implies pq.rem (i, j).sign ≍ j.sign
						bounded: pq.rem (i, j).abs < j.abs
					end
				end
				l := if (i \\ j) ≍ zero.p then zero.p else One.p end
				if
					not (
						broken_divisor or
						broken_dividend or
						broken_sum or
						broken_product
						)
				then
					check
						identity: pq.rem (pq.rem (i, j), j) ≍ pq.rem (i, j)
						opposite_divisor:
							pq.rem (i, -j) ≍ (pq.rem (i, j) - j⋅l)
					end
				end
				if not (broken_dividend or broken_sum or broken_product) then
					check
						opposite_dividend:
							pq.rem (-i, j) ≍ (- pq.rem (i, j) + j⋅l)
					end
				end
			end
			if k ≭ zero.p then
				if not (broken_sum or broken_product) then
					check
						distribution_over_sum: pq.rem (i + j, k) ≍
							pq.rem (pq.rem (i, k) + pq.rem (j, k), k)
					end
				end
				if not broken_product then
					check
						distribution_over_product: pq.rem (i ⋅ j, k) ≍
							pq.rem (pq.rem (i, k) ⋅ pq.rem (j, k), k)
					end
				end
			end
			Result := True
		rescue
			print (i) print (", ")
			print (j) print (", ")
			print (k) print ('%N')
			if not broken_abs and pq.rem (i, j).abs ≥ j.abs then
				check
						-- Min value of a two's complement integer...
					broken_abs: j.abs < zero.p
					broken_bound: pq.rem (i, j).abs > j.abs
				end
				broken_abs := True
				retry
			elseif
				not broken_divisor and attached l and then
				pq.rem (i, -j) ≭ (pq.rem (i, j) - j⋅l)
			then
				check
						-- Min value of a two's complement integer...
					broken_divisor: -j ≍ j
					broken_rem: pq.rem (i, -j) ≍ pq.rem (i, j)
				end
				broken_divisor := True
				retry
			elseif
				not broken_dividend and attached l and then
				pq.rem (-i, j) ≭ (- pq.rem (i, j) + j⋅l)
			then
				check
						-- Min value of a two's complement integer...
					broken_dividend: -i ≍ i
					broken_rem: pq.rem (-i, j) ≍ pq.rem (i, j)
				end
				broken_dividend := True
				retry
			elseif
				not broken_sum and k ≭ zero.p and then
				pq.rem (i + j, k) ≭ pq.rem (pq.rem (i, k) + pq.rem (j, k), k)
			then
				check
					broken_sum: sum_overflows (i, j) or
						sum_overflows (pq.rem (i, k), pq.rem (j, k))
				end
				broken_sum := True
				retry
			elseif
				not broken_product and k ≭ zero.p and then
				pq.rem (i ⋅ j, k) ≭ pq.rem (pq.rem (i, k) ⋅ pq.rem (j, k), k)
			then
				check
					broken_product: product_overflows (i, j) or
						product_overflows (pq.rem (i, k), pq.rem (j, k))
				end
				broken_product := True
				retry
			end
		end

feature -- Properties (Predicate)

	integer_product_overflows_ok
		(pq: STS_RATIONAL_NUMBER; i, j: STS_INTEGER_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for
			-- {STS_RATIONAL_NUMBER}.integer_product_overflows?
		do
			check
				unexpected_zero_product: i ≭ zero.p and
					j ≭ zero.p and
					i ⋅ j ≍ zero.p implies
					pq.integer_product_overflows (i, j)
			end
			Result := True
		end

feature -- Predicate

	sum_overflows (i, j: STS_INTEGER_NUMBER): BOOLEAN
			-- Does the sum `i' + `j' overflow?
		do
			Result := i < zero.p and j < zero.p and i.min_value_exists and then
				i.min_value - j > i or
				i > zero.p and j > zero.p and i.max_value_exists and then
				i > i.max_value - j
		ensure
			definition: Result = (
				i < zero.p and j < zero.p and i.min_value_exists and then
				i.min_value - j > i or
				i > zero.p and j > zero.p and i.max_value_exists and then
				i > i.max_value - j
				)
		end

	difference_overflows (i, j: STS_INTEGER_NUMBER): BOOLEAN
			-- Does the difference `i' - `j' overflow?
		do
			Result := i < zero.p and j > zero.p and i.min_value_exists and then
				i.min_value + j > i or
				i > zero.p and j < zero.p and i.max_value_exists and then
				i > i.max_value + j or
				i ≍ zero.p and j.min_value_exists and then j ≍ j.min_value
		ensure
			definition: Result = (
				i < zero.p and j > zero.p and i.min_value_exists and then
				i.min_value + j > i or
				i > zero.p and j < zero.p and i.max_value_exists and then
				i > i.max_value + j or
				i ≍ zero.p and j.min_value_exists and then j ≍ j.min_value
				)
		end

	product_overflows (i, j: STS_INTEGER_NUMBER): BOOLEAN
			-- Does the product `i' ⋅ `j' overflow?
			-- TODO: DRY {STS_RATIONAL_NUMBER}!
		do
			Result := i < zero.p and j < zero.p and i.max_value_exists and then
				i < i.max_value // j or
				i < zero.p and j > zero.p and i.min_value_exists and then
				i < i.min_value // j or
				i > zero.p and j < zero.p and i.min_value_exists and then
				i > i.min_value // j or
				i > zero.p and j > zero.p and i.max_value_exists and then
				i > i.max_value // j
		ensure
			definition: Result = (
				i < zero.p and j < zero.p and i.max_value_exists and then
				i < i.max_value // j or
				i < zero.p and j > zero.p and i.min_value_exists and then
				i < i.min_value // j or
				i > zero.p and j < zero.p and i.min_value_exists and then
				i > i.min_value // j or
				i > zero.p and j > zero.p and i.max_value_exists and then
				i > i.max_value // j
				)
		end

feature -- Math

	gcd (i, j: STS_INTEGER_NUMBER): STS_INTEGER_NUMBER
			-- Greatest common divisor of `i' and `j'
			-- TODO: DRY!
		do
			if j ≍ zero.p then
				Result := i.abs
			else
					check
						good_divisor: i.divisible (j) -- j /= 0
					end
				Result := gcd (j, i \\ j)
			end
		ensure
			base: j ≍ zero.p implies Result ≍ i.abs
			good_divisor: j ≭ zero.p implies i.divisible (j)
			induction: j ≭ zero.p implies Result ≍ gcd (j, i \\ j)
		end

feature -- Anchor

	rational_anchor: STS_RATIONAL_NUMBER
			-- Anchor for rational naumbers
		deferred
		ensure
		end

--	rational_universe_anchor: STS_RATIONAL_SET
--			-- Anchor for `current_rational_universe'
--		deferred
--		end

note
	copyright: "Copyright (c) 2012-2022, Rosivaldo Fernandes Alves"
	license: "[
		Eiffel Forum License v2
		(see http://www.eiffel.com/licensing/forum.txt)
		]"
	source: "https://launchpad.net/set-theory"
end
