﻿note
	description: "[
		Object that checks whether the properties verified within set theory hold for an implementation of {STS_REAL_NUMBER}

		Notice that many straightforward algebraic properties of real numbers break due to intricacies of a floating point implementation of their values. Such
		breaks are documented as (many!) exceptions caught and handled when appropriate.
		]"
	author: "Rosivaldo Fernandes Alves"
	date: "$Date$"
	revision: "$Revision$"

deferred class
	REAL_NUMBER_PROPERTIES

inherit
	COMPLEX_NUMBER_PROPERTIES
		rename
			zero as complex_zero, -- TODO: Such a renaming should be necessary only in descendants.
			is_in_ok as complex_is_in_ok,
			is_not_in_ok as complex_is_not_in_ok,
			zero_ok as complex_zero_ok,
			one_ok as complex_one_ok,
			is_infinite_ok as complex_is_infinite_ok,
			is_rational_ok as complex_is_rational_ok,
			is_integer_ok as complex_is_integer_ok,
			equals_ok as complex_equals_ok,
			unequals_ok as complex_unequals_ok,
			modulus_ok as complex_modulus_ok,
			abs_ok as complex_abs_ok,
			argument_ok as complex_argument_ok,
			phase_ok as complex_phase_ok,
			plus_ok as complex_plus_ok,
			identity_ok as complex_identity_ok,
			minus_ok as complex_minus_ok,
			opposite_ok as complex_opposite_ok,
			product_ok as complex_product_ok,
			quotient_ok as complex_quotient_ok,
			reciprocal_ok as complex_reciprocal_ok,
			inverse_ok as complex_inverse_ok
		end

feature -- Access

	zero: like real_anchor
			-- The real number 0
		deferred
		ensure
			definition: Result.value = 0
		end

	one: like real_anchor
			-- The real number 1
		deferred
		end

feature -- Properties (Membership)

	is_in_ok (x: STS_REAL_NUMBER; s: STS_SET [STS_REAL_NUMBER, STS_EQUALITY [STS_REAL_NUMBER]]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_REAL_NUMBER}.is_in?
		do
			check
				require_non_emptiness: x ∈ s implies not s.is_empty
			end
			Result := True
		end

	is_not_in_ok (x: STS_REAL_NUMBER; s: STS_SET [STS_REAL_NUMBER, STS_EQUALITY [STS_REAL_NUMBER]]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_REAL_NUMBER}.is_not_in?
		do
			check
				definition: x ∉ s = s ∌ x
				sufficient_emptiness: s.is_empty implies x ∉ s
			end
			Result := True
		end

feature -- Properties (Access)

	zero_ok (x, y: STS_REAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_REAL_NUMBER}.zero?
		local
			inherited_ok, broken_left_absorption, broken_right_absorption, broken_dividend_absorption, broken_negative_exponent: BOOLEAN
		do
			if not (broken_left_absorption or broken_right_absorption or broken_dividend_absorption or broken_negative_exponent) then
				inherited_ok := b_ok (x) and imaginary_ok (x) and complex_zero_ok (x, y)
				if inherited_ok then
					check
						neutral_left_term: (x.zero + y) ≍ y
						neutral_right_term: (x + y.zero) ≍ x
						quasi_neutral_minuend: (x.zero - y) ≍ - y
						neutral_subtrahend: (x - y.zero) ≍ x
						absorbing_left_factor: (x.zero ⋅ y) ≍ x.zero
					end
				end
			end
			if inherited_ok and not (broken_right_absorption or broken_dividend_absorption or broken_negative_exponent) then
				check
					absorbing_right_factor: (x ⋅ y.zero) ≍ y.zero
				end
			end
			if inherited_ok and not (broken_dividend_absorption or broken_negative_exponent) then
				check
					absorbing_dividend: x.zero.divisible (y) implies (x.zero / y) ≍ x.zero
				end
			end
			if inherited_ok and not broken_negative_exponent then
				check
					negative_exponent: y < zero implies (x.zero ^ y).is_positive_infinity
				end
			end
			if inherited_ok then
				check
					to_power_zero: (x.zero ^ zero) ≍ one
					absorbing_base: zero < y implies (x.zero ^ y) ≍ x.zero
					zero_exponent: (x ^ y.zero) ≍ one
				end
				Result := True
			end
		rescue
			print (x) print ("; ")
			print (y) print ('%N')
			if not broken_left_absorption and (x.zero ⋅ y) ≭ x.zero then
				check
					broken_left_absorption_cause: not y.is_finite
					broken_left_absorption_nan_product: (x.zero ⋅ y).is_nan
				end
				broken_left_absorption := True
				retry
			elseif not broken_right_absorption and (x ⋅ y.zero) ≭ y.zero then
				check
					broken_right_absorption_cause: not x.is_finite
					broken_right_absorption_nan_product: (x ⋅ y.zero).is_nan
				end
				broken_right_absorption := True
				retry
			elseif not broken_dividend_absorption and x.zero.divisible (y) and (x.zero / y) ≭ x.zero then
				check
					broken_dividend_absorption_cause: y.is_nan or y ≍ zero -- For the case of an implementation that accepts division by zero.
					broken_dividend_absorption_nan_quotient: (x.zero / y).is_nan
				end
				broken_dividend_absorption := True
				retry
			elseif not broken_negative_exponent and y < zero and not (x.zero ^ y).is_positive_infinity then
				check
					broken_negative_exponent_cause: y.is_nan
					broken_negative_exponent_nan_power: (x.zero ^ y).is_nan
				end
				broken_negative_exponent := True
				retry
			end
		end

	one_ok (x, y: STS_REAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_REAL_NUMBER}.one?
		local
			inherited_ok, broken_absorbing_base: BOOLEAN
		do
			if not broken_absorbing_base then
				inherited_ok := complex_one_ok (x, y)
				if inherited_ok then
					check
						neutral_left_factor: (x.one ⋅ y) ≍ y
						neutral_right_factor: (x ⋅ y.one) ≍ x
							good_divisor: x.divisible (y.one) -- one.is_invertible
						neutral_divisor: (x / y.one) ≍ x
						neutral_exponent: (x ^ y.one) ≍ x
						absorbing_base: (x.one ^ y) ≍ x.one
					end
				end
			end
			Result := True
		rescue
			print (x) print ("; ")
			print (y) print ('%N')
			if not broken_absorbing_base and (x.one ^ y) ≭ x.one then
				check
					broken_absorbing_base_cause: not y.is_finite and (x.one ^ y).is_nan
					broken_absorbing_base_power_nan: (x.one ^ y).is_nan
				end
				broken_absorbing_base := True
				retry
			end
		end

	machine_epsilon_ok (x: STS_REAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for
			-- {STS_REAL_NUMBER}.machine_epsilon?
		do
			check
				real: x.machine_epsilon.is_real
				proper_real: x.machine_epsilon.is_proper_real
				not_imaginary: not x.machine_epsilon.is_imaginary
				not_proper_imaginary: not x.machine_epsilon.is_proper_imaginary
				not_proper_complex: not x.machine_epsilon.is_proper_complex
				positive: x.machine_epsilon > zero
			end
			Result := True
		end

	ulp_ok (x: STS_REAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_REAL_NUMBER}.ulp?
		local
			broken_ulp: BOOLEAN
		do
			if not broken_ulp then
				check
					real: x.ulp.is_real
					proper_real: x.ulp.is_proper_real
					not_imaginary: not x.ulp.is_imaginary
					not_proper_imaginary: not x.ulp.is_proper_imaginary
					not_proper_complex: not x.ulp.is_proper_complex
					positive: x.ulp > zero
				end
			end
			Result := True
		rescue
			if not broken_ulp and x.ulp ≤ zero then
				check
					x_is_nan: x.is_nan
					broken_ulp: x.ulp.is_nan
				end
				broken_ulp := True
				retry
			end
		end

	relative_epsilon_ok (x: STS_REAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_REAL_NUMBER}.relative_epsilon?
		local
			broken_epsilon: BOOLEAN
		do
			if not broken_epsilon then
				check
					real: x.relative_epsilon.is_real
					proper_real: x.relative_epsilon.is_proper_real
					not_imaginary: not x.relative_epsilon.is_imaginary
					not_proper_imaginary: not x.relative_epsilon.is_proper_imaginary
					not_proper_complex: not x.relative_epsilon.is_proper_complex
					positive: x.relative_epsilon > zero
				end
			end
			Result := True
		rescue
			if not broken_epsilon and x.relative_epsilon ≤ zero then
				check
					x_is_nan: x.is_nan
					broken_epsilon: x.relative_epsilon.is_nan
				end
				broken_epsilon := True
				retry
			end
		end

feature -- Properties (Quality)

	is_nan_ok (x, y: STS_REAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for
			-- {STS_REAL_NUMBER}.is_nan?
		do
			check
				absorbing_left_term: x.is_nan implies (x + y).is_nan
				absorbing_right_term: y.is_nan implies (x + y).is_nan
				absorbing_minuend: x.is_nan implies (x - y).is_nan
				absorbing_subtrahend: y.is_nan implies (x - y).is_nan
				absorbing_left_factor: x.is_nan implies (x ⋅ y).is_nan
				absorbing_right_factor: y.is_nan implies (x ⋅ y).is_nan
				absorbing_dividend: x.is_nan and x.divisible (y) implies
					(x / y).is_nan
					good_divisor: y.is_nan implies x.divisible (y) -- y /= 0
				propagating_nan_divisor: y.is_nan implies (x / y).is_nan
				propagating_nan_base: x.is_nan and y ≭ Zero implies
					(x ^ y).is_nan
				propagating_nan_exponent: y.is_nan implies (x ^ y).is_nan
			end
			Result := True
		end

	is_negative_infinity_ok (x, y: STS_REAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_REAL_NUMBER}.is_negative_infinity?
		do
			check
				absorbing_left_term: x.is_negative_infinity and
					not y.is_nan and
					not y.is_positive_infinity implies
					(x + y).is_negative_infinity
				absorbing_right_term: not x.is_nan and
					not x.is_positive_infinity and
					y.is_negative_infinity implies
					(x + y).is_negative_infinity
				absorbing_minuend: x.is_negative_infinity and
					not y.is_nan and
					not y.is_negative_infinity implies
					(x - y).is_negative_infinity
				quasi_absorbing_subtrahend: not x.is_nan and
					not x.is_negative_infinity and
					y.is_negative_infinity implies (x - y).is_positive_infinity
				quasi_absorbing_left_factor: x.is_negative_infinity and
					not y.is_nan and y ≭ zero implies
						if y < zero then
							(x ⋅ y).is_positive_infinity
						else
							(x ⋅ y).is_negative_infinity
						end
				quasi_absorbing_right_factor: not x.is_nan and
					x ≭ zero and
					y.is_negative_infinity implies
						if x < zero then
							(x ⋅ y).is_positive_infinity
						else
							(x ⋅ y).is_negative_infinity
						end
				quasi_absorbing_dividend: x.is_negative_infinity and
					x.divisible (y) and
					y.is_finite implies
						if y < zero then
							(x / y).is_positive_infinity
						else
							(x / y).is_negative_infinity
						end
					good_divisor: y.is_negative_infinity implies
						x.divisible (y) -- y /= 0
				nullifying_divisor: x.is_finite and
					y.is_negative_infinity implies
					(x / y) ≍ zero
			end
			Result := True
		end

	is_negative_zero_ok (x, y: STS_REAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for
			-- {STS_REAL_NUMBER}.is_negative_zero?
		local
			broken_left_absorption,
			broken_right_absorption,
			broken_dividend_absorption,
			broken_negative_exponent,
			broken_negative_zero_times_negative_number,
			broken_negative_zero_times_positive_number: BOOLEAN
		do
			if
				not (
					broken_left_absorption or
					broken_right_absorption or
					broken_dividend_absorption or
					broken_negative_exponent or
					broken_negative_zero_times_negative_number or
					broken_negative_zero_times_positive_number
					)
			then
				check
					neutral_left_term: x.is_negative_zero implies (x + y) ≍ y
					neutral_right_term: y.is_negative_zero implies (x + y) ≍ x
					quasi_neutral_minuend: x.is_negative_zero implies
						(x - y) ≍ - y
					neutral_subtrahend: y.is_negative_zero implies (x - y) ≍ x
					absorbing_left_factor: x.is_negative_zero implies x ⋅ y ≍ x
				end
			end
			if
				not (
					broken_right_absorption or
					broken_dividend_absorption or
					broken_negative_exponent or
					broken_negative_zero_times_negative_number or
					broken_negative_zero_times_positive_number
					)
			then
				check
					absorbing_right_factor: y.is_negative_zero implies
						(x ⋅ y) ≍ y
				end
			end
			if not (
				broken_dividend_absorption or
				broken_negative_exponent or
				broken_negative_zero_times_negative_number or
				broken_negative_zero_times_positive_number
				)
			then
				check
					absorbing_dividend: x.is_negative_zero and
						x.divisible (y) implies
						(x / y) ≍ x
				end
			end
			if
				not (
					broken_negative_exponent or
					broken_negative_zero_times_negative_number or
					broken_negative_zero_times_positive_number
					)
			then
				check
					negative_exponent: x.is_negative_zero and y < zero implies
						(x ^ y).is_infinite -- TODO: Split negative and positive infinities.
				end
			end
			if
				not (
					broken_negative_zero_times_negative_number or
					broken_negative_zero_times_positive_number
					)
			then
				check
					to_power_zero: x.is_negative_zero implies (x ^ zero) ≍ one
					absorbing_base: x.is_negative_zero and y > zero implies
						(x ^ y) ≍ x
					zero_exponent: y.is_negative_zero implies (x ^ y) ≍ one
					negative_zero_times_negative_number: x.is_negative_zero and
						y < zero implies
						(x ⋅ y).is_positive_zero
				end
			end
			if not broken_negative_zero_times_positive_number then
				check
					negative_zero_times_negative_zero: x.is_negative_zero and
						y.is_negative_zero implies
						(x ⋅ y).is_positive_zero
					negative_zero_times_positive_zero: x.is_negative_zero and
						y.is_positive_zero implies
						(x ⋅ y).is_negative_zero
					negative_zero_times_positive_number: x.is_negative_zero and
						y > zero implies
						(x ⋅ y).is_negative_zero
				end
			end
			check
				negative_zero_by_negative_number: x.is_negative_zero and
					y < zero implies
					(x / y).is_positive_zero
				negative_zero_by_positive_number: x.is_negative_zero and
					y > zero implies
					(x / y).is_negative_zero
			end
			Result := True
		rescue
			print (x) print (", ")
			print (y) print ('%N')
			if
				not broken_left_absorption and x.is_negative_zero and x ⋅ y ≭ x
			then
				check
					broken_left_absorption_cause: not y.is_finite
					broken_left_absorption_nan_product: (x ⋅ y).is_nan
				end
				broken_left_absorption := True
				retry
			elseif
				not broken_right_absorption and
				y.is_negative_zero and
				(x ⋅ y) ≭ y
			then
				check
					broken_right_absorption_cause: not x.is_finite
					broken_right_absorption_nan_product: (x ⋅ y).is_nan
				end
				broken_right_absorption := True
				retry
			elseif
				not broken_dividend_absorption and
				x.is_negative_zero and
				x.divisible (y) and
				(x / y) ≭ x
			then
				check
					broken_dividend_absorption_cause:
							-- For the case of an implementation that accepts
							-- division by zero.
						y.is_nan or y ≍ zero
					broken_dividend_absorption_nan_quotient: (x / y).is_nan
				end
				broken_dividend_absorption := True
				retry
			elseif
				not broken_negative_exponent and
				x.is_negative_zero and
				y < zero and
				not (x ^ y).is_positive_infinity
			then
				check
					broken_negative_exponent_cause: y.is_nan
					broken_negative_exponent_nan_power: (x ^ y).is_nan
				end
				broken_negative_exponent := True
				retry
			elseif
				not broken_negative_zero_times_negative_number
				and x.is_negative_zero
				and y < zero and
				not (x ⋅ y).is_positive_zero
			then
				check
					broken_negative_zero_times_negative_number_cause:
						not y.is_finite
					broken_negative_zero_times_negative_number: (x ⋅ y).is_nan
				end
				broken_negative_zero_times_negative_number := True
				retry
			elseif
				not broken_negative_zero_times_positive_number and
				x.is_negative_zero and
				y > zero and
				not (x ⋅ y).is_negative_zero
			then
				check
					broken_negative_zero_times_positive_number_cause:
						y.is_infinite
					broken_negative_zero_times_positive_number: (x ⋅ y).is_nan
				end
				broken_negative_zero_times_positive_number := True
				retry
			end
		end

	is_positive_zero_ok (x, y: STS_REAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for
			-- {STS_REAL_NUMBER}.is_positive_zero?
		local
			broken_left_absorption,
			broken_right_absorption,
			broken_dividend_absorption,
			broken_negative_exponent,
			broken_positive_zero_times_negative_number: BOOLEAN
		do
			if
				not (
					broken_left_absorption or
					broken_right_absorption or
					broken_dividend_absorption or
					broken_negative_exponent or
					broken_positive_zero_times_negative_number
					)
			then
				check
					neutral_left_term: x.is_positive_zero implies (x + y) ≍ y
					neutral_right_term: y.is_positive_zero implies (x + y) ≍ x
					quasi_neutral_minuend: x.is_positive_zero implies
						(x - y) ≍ - y
					neutral_subtrahend: y.is_positive_zero implies (x - y) ≍ x
					absorbing_left_factor: x.is_positive_zero implies x ⋅ y ≍ x
				end
			end
			if
				not (
					broken_right_absorption or
					broken_dividend_absorption or
					broken_negative_exponent or
					broken_positive_zero_times_negative_number
					)
			then
				check
					absorbing_right_factor: y.is_positive_zero implies
						(x ⋅ y) ≍ y
				end
			end
			if
				not (
					broken_dividend_absorption or
					broken_negative_exponent or
					broken_positive_zero_times_negative_number
					)
			then
				check
					absorbing_dividend: x.is_positive_zero and
						x.divisible (y) implies
						(x / y) ≍ x
				end
			end
			if
				not (
					broken_negative_exponent or
					broken_positive_zero_times_negative_number
					)
			then
				check
					negative_exponent: x.is_positive_zero and
						y < zero implies
						(x ^ y).is_positive_infinity
				end
			end
			if  not broken_positive_zero_times_negative_number then
				check
					to_power_zero: x.is_positive_zero implies (x ^ zero) ≍ one
					absorbing_base: x.is_positive_zero and y > zero implies
						(x ^ y) ≍ x
					zero_exponent: y.is_positive_zero implies (x ^ y) ≍ one
					positive_zero_times_negative_number: x.is_positive_zero and
						y < zero implies
						(x ⋅ y).is_negative_zero
				end
			end
			check
				positive_zero_times_negative_zero: x.is_positive_zero and
					y.is_negative_zero implies
					(x ⋅ y).is_negative_zero
				positive_zero_times_positive_zero: x.is_positive_zero and
					y.is_positive_zero implies
					(x ⋅ y).is_positive_zero
				positive_zero_times_positive_number: x.is_positive_zero and
					y > zero implies
					(x ⋅ y).is_positive_zero
				positive_zero_by_negative_number: x.is_positive_zero and
					y < zero implies
					(x / y).is_negative_zero
				positive_zero_times_positive_number: x.is_positive_zero and
					y > zero implies
					(x / y).is_positive_zero
			end
			Result := True
		rescue
			print (x) print (", ")
			print (y) print ('%N')
			if
				not broken_left_absorption and x.is_positive_zero and x ⋅ y ≭ x
			then
				check
					broken_left_absorption_cause: not y.is_finite
					broken_left_absorption_nan_product: (x ⋅ y).is_nan
				end
				broken_left_absorption := True
				retry
			elseif
				not broken_right_absorption and
				y.is_positive_zero and
				(x ⋅ y) ≭ y
			then
				check
					broken_right_absorption_cause: not x.is_finite
					broken_right_absorption_nan_product: (x ⋅ y).is_nan
				end
				broken_right_absorption := True
				retry
			elseif
				not broken_dividend_absorption and
				x.is_positive_zero and
				x.divisible (y) and
				(x / y) ≭ x
			then
				check
					broken_dividend_absorption_cause:
							-- For the case of an implementation that accepts
							-- division by zero.
						y.is_nan or y ≍ zero
					broken_dividend_absorption_nan_quotient: (x / y).is_nan
				end
				broken_dividend_absorption := True
				retry
			elseif
				not broken_negative_exponent and
				x.is_positive_zero and
				y < zero and
				not (x ^ y).is_positive_infinity
			then
				check
					broken_negative_exponent_cause: y.is_nan
					broken_negative_exponent_nan_power: (x ^ y).is_nan
				end
				broken_negative_exponent := True
				retry
			elseif
				not broken_positive_zero_times_negative_number and
				x.is_positive_zero and
				y < zero and
				not (x ⋅ y).is_negative_zero
			then
				check
					broken_positive_zero_times_negative_number_cause:
						not y.is_finite
					broken_positive_zero_times_negative_number: (x ⋅ y).is_nan
				end
				broken_positive_zero_times_negative_number := True
				retry
			end
		end

	is_positive_infinity_ok (x, y: STS_REAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for
			-- {STS_REAL_NUMBER}.is_positive_infinity?
		do
			check
				absorbing_left_term: x.is_positive_infinity and
					not y.is_nan and
					not y.is_negative_infinity implies
					(x + y).is_positive_infinity
				absorbing_right_term: not x.is_nan and
					not x.is_negative_infinity and
					y.is_positive_infinity implies
					(x + y).is_positive_infinity
				absorbing_minuend: x.is_positive_infinity and
					not y.is_nan and
					not y.is_positive_infinity implies
					(x - y).is_positive_infinity
				quasi_absorbing_subtrahend: not x.is_nan and
					not x.is_positive_infinity and
					y.is_positive_infinity implies
					(x - y).is_negative_infinity
				quasi_absorbing_left_factor: x.is_positive_infinity and
					not y.is_nan and y ≭ zero implies
						if y < zero then
							(x ⋅ y).is_negative_infinity
						else
							(x ⋅ y).is_positive_infinity
						end
				quasi_absorbing_right_factor: not x.is_nan and
					x ≭ zero and
					y.is_positive_infinity implies
						if x < zero then
							(x ⋅ y).is_negative_infinity
						else
							(x ⋅ y).is_positive_infinity
						end
				quasi_absorbing_dividend: x.is_positive_infinity and
					x.divisible (y) and
					y.is_finite implies
						if y < zero then
							(x / y).is_negative_infinity
						else
							(x / y).is_positive_infinity
						end
					good_divisor: y.is_positive_infinity implies
						x.divisible (y) -- y /= 0
				nullifying_divisor: x.is_finite and
					y.is_positive_infinity implies
					(x / y) ≍ zero
			end
			Result := True
		end

	is_infinite_ok (x, y: STS_REAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_REAL_NUMBER}.is_infinite?
		do
			if complex_is_infinite_ok (x, y) then
				check
					absorbing_left_term: x.is_infinite and y.is_finite implies (x + y).is_infinite
					absorbing_right_term: x.is_finite and y.is_infinite implies (x + y).is_infinite
					absorbing_minuend: x.is_infinite and y.is_finite implies (x - y).is_infinite
					absorbing_subtrahend: x.is_finite and y.is_infinite implies (x - y).is_infinite
					absorbing_left_factor: x.is_infinite and (y.is_finite and y ≭ zero or y.is_infinite) implies (x ⋅ y).is_infinite
					absorbing_right_factor: (x.is_finite and x ≭ zero or x.is_infinite) and y.is_infinite implies (x ⋅ y).is_infinite
					absorbing_dividend: x.is_infinite and y.is_finite and x.divisible (y) implies (x / y).is_infinite
						good_divisor: y.is_infinite implies x.divisible (y) -- y.is_invertible ⇐ y.is_infinite
					nullifying_divisor: x.is_finite and y.is_infinite implies (x / y) ≍ zero
				end
				Result := True
			end
		end

	is_rational_ok (x: STS_REAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_REAL_NUMBER}.is_rational?
		do
			if complex_is_rational_ok (x) then
				check
					is_not_nan: x.is_rational implies not x.is_nan
					is_not_negative_infinity: x.is_rational implies
						not x.is_negative_infinity
					is_not_negative_nero: x.is_rational implies
						not x.is_negative_zero
					is_not_positive_infinity: x.is_rational implies
						not x.is_positive_infinity
					is_not_infinite: x.is_rational implies not x.is_infinite
					is_finite: x.is_rational implies x.is_finite
				end
				Result := True
			end
		end

	is_integer_ok (x: STS_REAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_REAL_NUMBER}.is_integer?
		do
			if complex_is_integer_ok (x) then
				check
					is_not_nan: x.is_integer implies not x.is_nan
					is_not_negative_infinity: x.is_integer implies
						not x.is_negative_infinity
					is_not_negative_nero: x.is_integer implies
						not x.is_negative_zero
					is_not_positive_infinity: x.is_integer implies
						not x.is_positive_infinity
					is_not_infinite: x.is_integer implies not x.is_infinite
					is_finite: x.is_integer implies x.is_finite
				end
				Result := True
			end
		end

feature -- Properties (Comparison)

	equals_ok (x, y, z: STS_REAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_REAL_NUMBER}.equals?
		do
			if complex_equals_ok (x, y, z) then
				check
					r: attached current_real_universe as r
					definition: (2 ^ # r) ≤ max_subsets implies x ≍ y = r.powerset |∀
						agent (s: STS_SET [STS_REAL_NUMBER, STS_REAL_EQUALITY]; ia_x, ia_y: STS_REAL_NUMBER): BOOLEAN
							do
								Result := ia_x ∈ s = ia_y ∈ s
							end (?, x, y)
					by_quantification: x ≍ y = (r |∀ agent (w: STS_REAL_NUMBER; ia_x, ia_y: STS_REAL_NUMBER): BOOLEAN
						do
							Result := (w.value = ia_x.value) = (w.value = ia_y.value)
						end (?, x, y))
					complex_definition: x ≍ y = x.complex_equals (y)
					reflexive: x ≍ x
					symmetric: x ≍ y implies y ≍ x
					transitive: x ≍ y and y ≍ z implies x ≍ z
					euclidian: x ≍ z and y ≍ z implies x ≍ y
				end
				Result := True
			end
		end

	unequals_ok (x, y: STS_REAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_REAL_NUMBER}.unequals?
		do
			if complex_unequals_ok (x, y) then
				check
					r: attached current_real_universe as r
					definition: x ≭ y = (x.value /= y.value)
					by_existence: x ≭ y = (r |∃ agent (z: STS_REAL_NUMBER; ia_x, ia_y: STS_REAL_NUMBER): BOOLEAN
						do
							Result := (z.value = ia_x.value) /= (z.value = ia_y.value)
						end (?, x, y))
					by_membership: (2 ^ # r) ≤ max_subsets implies x ≭ y = r.powerset |∃
						agent (s: STS_SET [STS_REAL_NUMBER, STS_REAL_EQUALITY]; ia_x, ia_y: STS_REAL_NUMBER): BOOLEAN
							do
								Result := ia_x ∈ s /= ia_y ∈ s
							end (?, x, y)
					complex_definition: x ≭ y = x.complex_unequals (y)
					irreflexive: not (x ≭ x)
				end
				Result := True
			end
		end

	is_less_ok (x, y, z: STS_REAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_REAL_NUMBER}.is_less?
		do
			check
				irreflexive: not (x < x)
				transitive: x < y and y < z implies x < z
				relative_epsilon: x < y implies (y - x) ≥ x.relative_epsilon
				absolute_minimum: x < y implies not y.is_nan
			end
			Result := True
		end

	is_less_equal_ok (x, y, z: STS_REAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_REAL_NUMBER}.is_less_equal?
		do
			check
				reflexive: x ≤ x
				transitive: x ≤ y and y ≤ z implies x ≤ z
				antisymmetric: x ≤ y and y ≤ x implies x ≍ y
				absolute_minimum: x.is_nan implies x ≤ y
			end
			Result := True
		end

	is_greater_ok (x, y, z: STS_REAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_REAL_NUMBER}.is_greater?
		do
			check
				irreflexive: not (x > x)
				transitive: x > y and y > z implies x > z
				relative_epsilon: x > y implies (x - y) ≥ y.relative_epsilon
				absolute_maximum: x > y implies not y.is_positive_infinity
			end
			Result := True
		rescue
			print (x) print (", ")
			print (y) print (", ")
			print (z) print ('%N')
		end

	is_greater_equal_ok (x, y, z: STS_REAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_REAL_NUMBER}.is_greater_equal?
		do
			check
				reflexive: x ≥ x
				transitive: x ≥ y and y ≥ z implies x ≥ z
				antisymmetric: x ≥ y and y ≥ x implies x ≍ y
				absolute_maximum: x.is_positive_infinity implies x ≥ y
			end
			Result := True
		end

	min_ok (x, y, z: STS_REAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_REAL_NUMBER}.min?
		do
			check
				idempotent: (x ∧ x) ≍ x
				commutative: (x ∧ y) ≍ (y ∧ x)
				associative: ((x ∧ y) ∧ z) ≍ (x ∧ (y ∧ z))
				absolute_minimum: x.is_nan implies (x ∧ y).is_nan
			end
			Result := True
		end

	max_ok (x, y, z: STS_REAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_REAL_NUMBER}.max?
		do
			check
				idempotent: (x ∨ x) ≍ x
				commutative: (x ∨ y) ≍ (y ∨ x)
				associative: ((x ∨ y) ∨ z) ≍ (x ∨ (y ∨ z))
				absolute_maximum: x.is_positive_infinity implies (x ∨ y).is_positive_infinity
			end
			Result := True
		end

feature -- Properties (Operation)

	modulus_ok (x: STS_REAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_REAL_NUMBER}.modulus?
		do
			if complex_modulus_ok (x) then
				check
					when_negative: x < zero implies x.modulus ≍ - x
					when_non_negative: x ≥ zero implies x.modulus ≍ x
				end
				Result := True
			end
		rescue
			print (x) print ('%N')
		end

	abs_ok (x: STS_REAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_REAL_NUMBER}.abs?
		do
			if complex_abs_ok (x) then
				check
					when_negative: x < zero implies x.abs ≍ - x
					when_non_negative: x ≥ zero implies x.abs ≍ x
				end
				Result := True
			end
		rescue
			print (x) print ('%N')
		end

	argument_ok (x: STS_REAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_REAL_NUMBER}.argument?
		local
			broken_argument, inherited_ok: BOOLEAN
		do
			if not broken_argument then
				inherited_ok := complex_argument_ok (x)
				if inherited_ok then
					check
						when_negative: x.sign_bit ≍ one.truncated_to_integer implies x.argument.value = x.adjusted_value (Pi)
						when_positive: x.sign_bit ≍ zero.truncated_to_integer implies x.argument.is_positive_zero
					end
				end
			end
			Result := inherited_ok
		rescue
			if not broken_argument and x < zero and x.argument.value /= x.adjusted_value (Pi) then
				check
					x_is_nan: x.is_nan
					broken_argument: x.argument.is_nan
				end
				broken_argument := True
				retry
			end
		end

	phase_ok (x: STS_REAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_REAL_NUMBER}.phase?
		local
			l_broken_phase, inherited_ok: BOOLEAN
		do
			if not l_broken_phase then
				inherited_ok := complex_phase_ok (x)
				if inherited_ok then
					check
						when_negative: x.sign_bit ≍ one.truncated_to_integer implies x.phase.value = x.adjusted_value (Pi)
						when_positive: x.sign_bit ≍ zero.truncated_to_integer implies x.phase.is_positive_zero
					end
				end
			end
			Result := inherited_ok
		rescue
			if not l_broken_phase and x < zero and x.phase.value /= x.adjusted_value (Pi) then
				check
					x_is_nan: x.is_nan
					broken_phase: x.phase.is_nan
				end
				l_broken_phase := True
				retry
			end
		end

	plus_ok (x, y, z: STS_REAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_REAL_NUMBER}.plus?
		local
			inherited_ok, broken_homorphism: BOOLEAN
		do
			if not broken_homorphism then
				inherited_ok := complex_plus_ok (x, y)
				if inherited_ok then
					check
						definition: (x + y).complex_equals (x.complex_plus (y))
						neutral_left_term: (zero + x) ≍ x
						neutral_right_term: (x + zero) ≍ x
						commutative: (x + y) ≍ (y + x)
						homomorphic: x ≤ y implies (x + z) ≤ (y + z)
					end
				end
			end
			if inherited_ok then
				check
					propagating_left_nan_term: x.is_nan implies (x + y).is_nan
					propagating_right_nan_term: x.is_nan implies (x + y).is_nan
					impossible_sum: x.is_negative_infinity and y.is_positive_infinity implies (x + y).is_nan
					commuted_impossible_sum: x.is_positive_infinity and y.is_negative_infinity implies (x + y).is_nan
				end
				Result := True
			end
		rescue
			print (x) print (", ")
			print (y) print (", ")
			print (z) print ('%N')
			if not broken_homorphism and x ≤ y and (x + z) > (y + z) then
				check
					broken_homomorphism: not x.is_nan and y.is_positive_infinity and z.is_negative_infinity
					broken_sum: (y + z).is_nan
				end
				broken_homorphism := True
				retry
			end
		end

	identity_ok (x: STS_REAL_NUMBER; z: STS_COMPLEX_NUMBER): BOOLEAN -- TODO: Remove `z'.
			-- Do the properties verified within set theory hold for {STS_REAL_NUMBER}.identity?
		do
			if a_ok (x) and real_ok (x) and conjugate_ok (x, z) and complex_identity_ok (x) then -- TODO: No conjugate_ok!
				check
					as_left_term: + x ≍ (x + zero)
					as_right_term: + x ≍ (zero + x)
					as_left_factor: + x ≍ (x ⋅ one)
					as_right_factor: + x ≍ (one ⋅ x)
						good_divisor: x.divisible (one) -- one.is_invertible
					as_quotient: + x ≍ (x / one)
				end
				Result := True
			end
		end


	minus_ok (x, y, z: STS_REAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_REAL_NUMBER}.minus?
		local
			inherited_ok, broken_homorphism: BOOLEAN
		do
			if not broken_homorphism then
				inherited_ok := complex_minus_ok (x, y)
				if inherited_ok then
					check
						definition: (x - y).complex_equals (x.complex_minus (y))
						quasi_neutral_minuend: (zero - x) ≍ - x
						neutral_subtrahend: (x - zero) ≍ x
						quasi_commutative: (x - y).abs ≍ (y - x).abs
						oppositely_homomorphic: x ≤ y implies (z - x) ≥ (z - y)
					end
				end
			end
			if inherited_ok then
				check
					propagating_nan_minuend: x.is_nan implies (x - y).is_nan
					propagating_nan_subtrahend: x.is_nan implies (x - y).is_nan
					impossible_subtraction: x.is_negative_infinity and y.is_negative_infinity implies (x - y).is_nan
					commuted_impossible_subtraction: x.is_positive_infinity and y.is_positive_infinity implies (x - y).is_nan
				end
				Result := True
			end
		rescue
			print (x) print (", ")
			print (y) print (", ")
			print (z) print ('%N')
			if not broken_homorphism and x ≤ y and (z - x) < (z - y) then
				check
					broken_homomorphism: (x.is_nan and not y.is_nan and not z.is_nan or x.is_negative_infinity and not y.is_negative_infinity and z.is_negative_infinity)
					broken_subtraction: (z - x).is_nan
				end
				broken_homorphism := True
				retry
			end
		end

	opposite_ok (x: STS_REAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for
			-- {STS_REAL_NUMBER}.opposite?
		local
			inherited_ok,
			broken_left_opposition,
			broken_right_opposition,
			broken_sign_inversion: BOOLEAN
		do
			if
				not (
					broken_left_opposition or
					broken_right_opposition or
					broken_sign_inversion
					)
			then
				inherited_ok := complex_opposite_ok (x)
				if inherited_ok then
					check
						opposite_left_term: ((- x) + x) ≍ zero
					end
				end
			end
			if
				inherited_ok and
				not (broken_right_opposition or broken_sign_inversion)
			then
				check
					opposite_right_term: (x + (- x)) ≍ zero
				end
			end
			if inherited_ok and not broken_sign_inversion then
				check
					as_subtrahend: - x ≍ (zero - x)
					as_left_factor: - x ≍ (x ⋅ (- one))
					as_right_factor: - x ≍ ((- one) ⋅ x)
						divisible: x.divisible (- one) -- (- one).is_invertible
					as_quotient: - x ≍ (x / (- one))
					propagating_nan: x.is_nan implies (- x).is_nan
					inverted_sign: (- x).sign ≍ - x.sign
				end
			end
			Result := inherited_ok
		rescue
			print (x) print ('%N')
			if not broken_left_opposition and ((- x) + x) ≭ zero then
				check
					not_finite: not x.is_finite
					broken_left_opposition: ((- x) + x).is_nan
				end
				broken_left_opposition := True
				retry
			elseif not broken_right_opposition and (x + (- x)) ≭ zero then
				check
					not_finite: not x.is_finite
					broken_right_opposition: (x + (- x)).is_nan
				end
				broken_right_opposition := True
				retry
			elseif not broken_sign_inversion and (- x).sign ≭ - x.sign then
				check
					is_nan: x.is_nan
					negative_sign: x.sign < zero.truncated_to_integer
					negative_opposite_sign: (- x).sign < zero.truncated_to_integer
				end
				broken_sign_inversion := True
				retry
			end
		end

	product_ok (x, y, z: STS_REAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_REAL_NUMBER}.product?
		local
			inherited_ok, broken_left_absorption, broken_right_absorption, broken_morphism: BOOLEAN
		do
			if not broken_left_absorption and not broken_right_absorption and not broken_morphism then
				inherited_ok := complex_product_ok (x, y, z)
				if inherited_ok then
					check
						definition: (x ⋅ y).complex_equals (x.complex_product (y))
						neutral_left_factor: (one ⋅ x) ≍ x
						neutral_right_factor: x ⋅ one ≍ x
						absorbing_left_factor: (zero ⋅ x) ≍ zero
					end
				end
			end
			if inherited_ok and not broken_right_absorption and not broken_morphism then
				check
					absorbing_right_factor: x ⋅ zero ≍ zero
				end
			end
			if inherited_ok and not broken_morphism then
				check
					commutative: (x ⋅ y) ≍ (y ⋅ x)
					quasi_homomorphic: x.abs ≤ y.abs implies (x ⋅ z).abs ≤ (y ⋅ z).abs
				end
			end
			if inherited_ok then
				check
					impossible_product: x ≍ zero and y.is_infinite implies (x ⋅ y).is_nan
					commuted_impossible_product: x.is_infinite and y ≍ zero implies (x ⋅ y).is_nan
					propagating_left_nan: x.is_nan implies (x ⋅ y).is_nan
					propagating_right_nan: y.is_nan implies (x ⋅ y).is_nan
				end
				Result := True
			end
		rescue
			if not broken_left_absorption and (zero ⋅ x) ≭ zero then
				check
					broken_left_absorption: x.is_nan or x.is_infinite
					left_broken_product: (zero ⋅ x).is_nan
				end
				broken_left_absorption := True
				retry
			elseif not broken_right_absorption and x ⋅ zero ≭ zero then
				check
					broken_right_absorption: x.is_nan or x.is_infinite
					right_broken_product: (x ⋅ zero).is_nan
				end
				broken_right_absorption := True
				retry
			elseif not broken_morphism and x.abs ≤ y.abs and (x ⋅ z).abs > (y ⋅ z).abs then
				check
					broken_morphism: y ≍ zero and z.is_infinite or y.is_infinite and z ≍ zero
					broken_y_times_z: (y ⋅ z).is_nan
				end
				broken_morphism := True
				retry
			end
		end

	quotient_ok (x, y, z: STS_REAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_REAL_NUMBER}.quotient?
		local
			inherited_ok, broken_definition, broken_absorption, broken_morphism: BOOLEAN
		do
			if not (broken_definition or broken_absorption or broken_morphism) then
				inherited_ok := complex_quotient_ok (x, y, z)
				if inherited_ok then
					check
							induced_good_divisor: x.divisible (y) implies x.complex_divisible (y) -- Same definition.
						definition: x.divisible (y) implies (x / y).complex_equals (x.complex_quotient (y))
					end
				end
			end
			if inherited_ok and not (broken_absorption or broken_morphism) then
				check
					absorbing_dividend: x ≍ zero and x.divisible (y) implies (x / y) ≍ zero
				end
			end
			if inherited_ok and not broken_morphism then
				check
						good_divisor: x.divisible (one) -- one /= 0
					neutral_divisor: (x / one) ≍ x
					quasi_oppositely_homomorphic: x.abs ≤ y.abs and z.divisible (x) and z.divisible (y) implies (z / x).abs ≥ (z / y).abs
				end
			end
			if inherited_ok then
				check
						infinite_divisible_infinite: x.is_infinite and y.is_infinite implies x.divisible (y) -- y /= 0
					impossible_quotient: x.is_infinite and y.is_infinite implies (x / y).is_nan
				end
				Result := True
			end
		rescue
			print (x) print (", ")
			print (y) print (", ")
			print (z) print ('%N')
			if
				not broken_definition and x.divisible (y) and then
				(x.complex_divisible (y) and then (x / y).complex_unequals (x.complex_quotient (y)))
			then
				check
					zero_divisor: y ≍ zero
					nan_quotient: (x / y).is_nan
					same_real_part: (x / y) ≍ x.complex_quotient (y).a
					induced_imaginary_nan: x.complex_quotient (y).b.is_nan
				end
				broken_definition := True
				retry
			elseif not broken_absorption and x ≍ zero and x.divisible (y) and (x / y) ≭ zero then
				check
					y_is_rogue_divisor: y.is_nan or y ≍ zero
					broken_absorption: (x / y).is_nan
				end
				broken_absorption := True
				retry
			elseif not broken_morphism and x.abs ≤ y.abs and z.divisible (x) and z.divisible (y) and then (z / x).abs < (z / y).abs then
				check
					x_is_rogue_divisor: (x.is_nan or x ≍ zero and z ≍ zero or x.is_infinite and z.is_infinite) and
						not (y.is_nan or y ≍ zero and z ≍ zero or y.is_infinite and z.is_infinite)
						z_divisible_x: z.divisible (x) -- Otherwise quasi_oppositely_homomorphic would not fail.
					broken_morphism: (z / x).is_nan
				end
				broken_morphism := True
				retry
			end
		end

	reciprocal_ok (x, y: STS_REAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for
			-- {STS_REAL_NUMBER}.reciprocal?
		do
			if complex_reciprocal_ok (x, y) then
				check
					definition: x.is_invertible implies x.reciprocal ≍ (one / x)
				end
				Result := True
			end
		end

	inverse_ok (x, y: STS_REAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for
			-- {STS_REAL_NUMBER}.inverse?
		do
			if complex_inverse_ok (x, y) then
				check
					definition: x.is_invertible implies x.inverse ≍ (one / x)
				end
				Result := True
			end
		end

	power_ok (x: STS_REAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_REAL_NUMBER}.power?
		do
			check
				base_zero: zero < x implies (zero ^ x) ≍ zero
				base_one: x.is_finite implies (one ^ x) ≍ one
				exponent_zero: (x ^ zero) ≍ one
				exponent_one: (x ^ one) ≍ x
			end
			Result := True
		end

	square_root_ok (x: STS_REAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_REAL_NUMBER}.square_root?
		do
			if x ≥ zero then
				check
					non_negative: √¯ x ≥ zero
					zero: √¯ x.zero ≍ zero
					below_one: x ≤ one implies √¯ x ≥ x
					one: √¯ x.one ≍ one
					above_one: x ≥ one implies √¯ x ≤ x
					infinite: x.is_positive_infinity implies (√¯ x).is_positive_infinity
				end
			end
			Result := True
		end

feature -- Properties (Operation)

	to_rational_ok (x: STS_REAL_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_REAL_NUMBER}.to_rational?
		do
			check
				consistent: x.is_rational implies x.to_rational.value = x.value
			end
			Result := True
		rescue
			print (x) print ('%N')
		end

feature -- Properties (Math)

	splitted_ok (x: STS_REAL_NUMBER; q: STS_INTEGER_NUMBER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_REAL_NUMBER}.splitted?
		do
			check
				when_intinite: x.is_infinite implies x.splitted (q).a ≍ x and x.splitted (q).b ≍ q
				when_nan: x.is_nan implies x.splitted (q).a.is_nan and x.splitted (q).b ≍ q
				trivial_rational_splitting: x.is_rational implies
					x.splitted (one).a.abs ≍ x.to_rational.p.rational_abs and x.splitted (one).b.abs ≍ x.to_rational.q.rational_abs
				trivial_integer_splitting: x.is_integer implies x.splitted (one).a ≍ x and x.splitted (one).b ≍ one
			end
			Result := True
		rescue
			print (x) print (", ")
			print (q) print ('%N')
		end

note
	copyright: "Copyright (c) 2012-2022, Rosivaldo Fernandes Alves"
	license: "[
		Eiffel Forum License v2
		(see http://www.eiffel.com/licensing/forum.txt)
		]"
	source: "https://launchpad.net/set-theory"
end
