﻿note
	description: "Object that checks whether the properties verified within set theory hold for an implementation of {STS_NUMERIC_SET}"
	author: "Rosivaldo Fernandes Alves"
	date: "$Date$"
	revision: "$Revision$"

deferred class
	NUMERIC_SET_PROPERTIES [A -> STS_COMPLEX_NUMBER, EQ -> STS_NUMERIC_EQUALITY [A]]

inherit
	SET_PROPERTIES [A, EQ]
		rename
			equals_ok as sp_equals_ok,
			is_subset_ok as sp_is_subset_ok,
			is_superset_ok as sp_is_superset_ok,
			is_not_superset_ok as sp_is_not_superset_ok,
			is_comparable_ok as sp_is_comparable_ok,
			is_not_comparable_ok as sp_is_not_comparable_ok,
			is_strict_subset_ok as sp_is_strict_subset_ok,
			is_strict_superset_ok as sp_is_strict_superset_ok,
			is_proper_subset_ok as sp_is_proper_subset_ok,
			is_proper_superset_ok as sp_is_proper_superset_ok,
			is_disjoint_ok as sp_is_disjoint_ok,
			complemented_ok as sp_complemented_ok,
			united_ok as sp_united_ok,
			intersected_ok as sp_intersected_ok,
			subtracted_ok as sp_subtracted_ok,
			subtracted_symmetricaly_ok as sp_subtracted_symmetricaly_ok
		end

feature -- Access

	zero: A
			-- The number 0
		deferred
		ensure
			definition: Result ≍ Result.zero
		end

feature -- Properties (Access)

	zero_ok (s: STS_NUMERIC_SET [A, EQ]; z: A): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_NUMERIC_SET}.zero?
		local
			broken_left_absorption, broken_right_absorption, broken_dividend_absorption: BOOLEAN
		do
			if not (broken_left_absorption or broken_right_absorption or broken_dividend_absorption) then
				check
					real: s.zero.is_real
					imaginary: s.zero.is_imaginary
					not_proper_real: not s.zero.is_proper_real
					not_proper_imaginary: not s.zero.is_proper_imaginary
					not_proper_complex: not s.zero.is_proper_complex
					equal_to_opposite: s.zero ≍ - s.zero
					neutral_left_term: (s.zero + z) ≍ z
					neutral_right_term: (z + s.zero) ≍ z
					quasi_neutral_minuend: (s.zero - z) ≍ - z
					neutral_subtrahend: (z - s.zero) ≍ z
					absorbing_left_factor: s.zero ⋅ z ≍ s.zero
				end
			end
			if not (broken_right_absorption or broken_dividend_absorption) then
				check
					absorbing_right_factor: (z ⋅ s.zero) ≍ s.zero
				end
			end
			if not broken_dividend_absorption then
				check
					absorbing_dividend: s.zero.divisible (z) ⇒ (s.zero / z) ≍ s.zero
				end
			end
			Result := True
		rescue
			check
				print_s_z: (
					agent (ia_s: STS_NUMERIC_SET [A, EQ]; ia_z: A): BOOLEAN
						do
							print (ia_s) print (", ")
							print (ia_z) print ('%N')
							Result := True
						end (s, z)
					).item
			end
			if not broken_left_absorption and s.zero ⋅ z ≭ s.zero then
				check
					absorbing_left_factor_unfolded_definition_a: (s.zero ⋅ z).a ≍ if z ≍ zero then
							zero.a
						elseif z.is_proper_imaginary then
							- zero.a ⋅ z.b
						elseif z.is_proper_real then
							zero.a ⋅ z.a
						else
							zero.a * z.a - zero.a * z.b
						end
					absorbing_left_factor_unfolded_definition_b: (s.zero ⋅ z).b ≍ if z ≍zero then
							zero.a
						elseif z.is_proper_imaginary then
							zero.a
						elseif z.is_proper_real then
							zero.a
						else
							zero.a * z.b + zero.a * z.a
						end
					broken_left_absorption: z.is_proper_imaginary and not z.b.is_finite or
						z.is_proper_real and not z.a.is_finite or
						z.is_proper_complex and not (z.a.is_finite and z.b.is_finite)
					broken_left_absorption_a: z.is_proper_imaginary and not z.b.is_finite or
						z.is_proper_real and not z.a.is_finite or
						z.is_proper_complex and not (z.a.is_finite and z.b.is_finite) ⇒
						(s.zero ⋅ z).a.is_nan
					broken_left_absorption_b: z.is_proper_complex and not (z.a.is_finite and z.b.is_finite) ⇒ (s.zero ⋅ z).b.is_nan
				end
				broken_left_absorption := True
				retry
			elseif not broken_right_absorption and (z ⋅ s.zero) ≭ s.zero then
				check
					absorbing_right_factor_unfolded_definition_a: (z ⋅ s.zero).a ≍ if z ≍ zero then
							zero.a
						elseif z.is_proper_imaginary then
							- z.b ⋅ zero.a
						elseif z.is_proper_real then
							z.a ⋅ zero.a
						else
							z.a * zero.a - z.b * zero.a
						end
					absorbing_right_factor_unfolded_definition_b: (z ⋅ s.zero).b ≍ if z ≍ zero then
							zero.a
						elseif z.is_proper_imaginary then
							zero.a
						elseif z.is_proper_real then
							zero.a
						else
							z.a * zero.a + z.b * zero.a
						end
					broken_right_absorption: z.is_proper_imaginary and not z.b.is_finite or
						z.is_proper_real and not z.a.is_finite or
						z.is_proper_complex and not (z.a.is_finite and z.b.is_finite)
					broken_right_absorption_a: z.is_proper_imaginary and not z.b.is_finite or
						z.is_proper_real and not z.a.is_finite or
						z.is_proper_complex and not (z.a.is_finite and z.b.is_finite) ⇒
						(z ⋅ s.zero).a.is_nan
					broken_right_absorption_b: z.is_proper_complex and not (z.a.is_finite and z.b.is_finite) ⇒ (z ⋅ s.zero).b.is_nan
				end
				broken_right_absorption := True
				retry
			elseif not broken_dividend_absorption and s.zero.divisible (z) and (s.zero / z) ≭ s.zero then
				check
						zero_divisible_zb: z.is_proper_imaginary ⇒ zero.divisible (z.b) -- z.b /= 0 <== z.is_proper_imaginary
						zero_divisible_za: z.is_proper_real ⇒ zero.divisible (z.a) -- z.a /= 0 <== z.is_proper_real
					absorbing_dividend_unfolded_definition_a: (s.zero / z).a ≍ if z.is_proper_imaginary then
							zero.a / z.b
						elseif z.is_proper_real then
							zero.a / z.a
						else
							proper_complex_quotient_implementation (zero, z).a
						end
					absorbing_dividend_unfolded_definition_b: (s.zero / z).b ≍ if z.is_proper_imaginary then
							zero.a
						elseif z.is_proper_real then
							zero.a
						else
							proper_complex_quotient_implementation (zero, z).b
						end
					broken_dividend_absorption: z.is_proper_imaginary and z.b.is_nan or
						z.is_proper_real and z.a.is_nan or
						z.is_proper_complex and (z.a.is_nan or z.b.is_nan) or
						z ≍ zero
					broken_dividend_absorption_a: z.is_proper_imaginary and z.b.is_nan or
						z.is_proper_real and z.a.is_nan or
						z.is_proper_complex and (z.a.is_nan or z.b.is_nan) or
						z ≍ zero ⇒
						(s.zero / z).a.is_nan
					broken_dividend_absorption_b: z.b.is_proper_complex and (z.a.is_nan or z.b.is_nan) or z ≍ zero ⇒ (s.zero / z).b.is_nan
				end
				broken_dividend_absorption := True
				retry -- TODO: goto?
			end
		end

	u_ok (s: STS_NUMERIC_SET [A, EQ]; a: A): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_NUMERIC_SET}.u?
		do
			check
				is_not_empty: not s.u.is_empty
				strict_subset: s.u.others ⊂ s.u
				strict_superset: s.u ⊃ s.u.others
				has_everything: not (s.u ∌ a)
				itself: s.u & a ≍ s.u
				equality: s.u ≍ s = (attached {STS_UNIVERSE [A, EQ]} s or # s = # s.u)
				inclusion: s.u ⊆ s = s.u ≍ s
				always_a_superset: s.u ⊇ s
				never_not_a_superset: not (s.u ⊉ s)
				always_comparable: s.u.is_comparable (s)
				never_not_comparable: not s.u.is_not_comparable (s)
				never_a_strict_subset: not (s.u ⊂ s)
				superset: s.u ⊃ s = (# s.u > # s)
				is_trivial_subset: s.u.is_trivial_subset (s) = s.u ≍ s
				never_a_proper_subset: not s.u.is_proper_subset (s)
				is_proper_superset: s.u.is_proper_superset (s) = (s ≭ o and s.u ≭ s)
				is_disjoint: s.u.is_disjoint (s) = s.is_empty
				empty_complement: s.u ⊆ s ⇒ (s.u ∁ s) ≍ o
				union: (s.u ∪ s) ≍ s.u
				intersection: (s.u ∩ s) ≍ s
				symmetric_difference: (s.u ⊖ s) ≍ (s.u ∖ s)
			end
			Result := True
		end

	universe_ok (s: STS_NUMERIC_SET [A, EQ]; a: A): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_NUMERIC_SET}.universe?
		do
			check
				is_not_empty: not s.universe.is_empty
				strict_subset: s.universe.others ⊂ s.universe
				strict_superset: s.universe ⊃ s.universe.others
				has_everything: not (s.universe ∌ a)
				itself: s.universe & a ≍ s.universe
				equality: s.universe ≍ s = (attached {STS_UNIVERSE [A, EQ]} s or # s = # s.universe)
				inclusion: s.universe ⊆ s = s.universe ≍ s
				always_a_superset: s.universe ⊇ s
				never_not_a_superset: not (s.universe ⊉ s)
				always_comparable: s.universe.is_comparable (s)
				never_not_comparable: not s.universe.is_not_comparable (s)
				never_a_strict_subset: not (s.universe ⊂ s)
				superset: s.universe ⊃ s = (# s.universe > # s)
				is_trivial_subset: s.universe.is_trivial_subset (s) = s.universe ≍ s
				never_a_proper_subset: not s.universe.is_proper_subset (s)
				is_proper_superset: s.universe.is_proper_superset (s) = (s ≭ o and s.universe ≭ s)
				is_disjoint: s.universe.is_disjoint (s) = s.is_empty
				empty_complement: s.universe ⊆ s ⇒ (s.universe ∁ s) ≍ o
				union: (s.universe ∪ s) ≍ s.universe
				intersection: (s.universe ∩ s) ≍ s
				symmetric_difference: (s.universe ⊖ s) ≍ (s.universe ∖ s)
			end
			Result := True
		end

feature -- Properties (Comparison)

	equals_ok (s1: STS_NUMERIC_SET [A, EQ]; s2: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_NUMERIC_SET}.equals?
		do
			check
				when_a_universe: attached {STS_UNIVERSE [A, EQ]} s1 ⇒ s1 ≍ s2 = (# s1 = # s2)
				disguised_universe: # s1 = # s1.u ⇒ s1 ≍ s2 = (# s1 = # s2)
				compared_to_universe: attached {STS_UNIVERSE [A, EQ]} s2 ⇒ s1 ≍ s2 = (# s1 = # s2)
				compared_to_disguised_universe: # s2 = # s1.u ⇒ s1 ≍ s2 = (# s1 = # s2)
			end
			Result := True
		end

	is_subset_ok (s1: STS_NUMERIC_SET [A, EQ]; s2: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_NUMERIC_SET}.is_subset?
		do
			check
				when_a_universe: attached {STS_UNIVERSE [A, EQ]} s1 ⇒ s1 ⊆ s2 = s1 ≍ s2
				disguised_universe: # s1 = # s1.u ⇒ s1 ⊆ s2 = s1 ≍ s2
				compared_to_universe: attached {STS_UNIVERSE [A, EQ]} s2 ⇒ s1 ⊆ s2
				compared_to_disguised_universe: # s2 = # s1.u ⇒ s1 ⊆ s2
			end
			Result := True
		end

	is_superset_ok (s1: STS_NUMERIC_SET [A, EQ]; s2: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_NUMERIC_SET}.is_superset?
		do
			if sp_is_superset_ok (s1, s2) then
				check
					when_a_universe: attached {STS_UNIVERSE [A, EQ]} s1 ⇒ s1 ⊇ s2
					disguised_universe: # s1 = # s1.u ⇒ s1 ⊇ s2
					compared_to_universe: attached {STS_UNIVERSE [A, EQ]} s2 ⇒ s1 ⊇ s2 = s1 ≍ s2
					compared_to_disguised_universe: # s2 = # s1.u ⇒ s1 ⊇ s2 = s1 ≍ s2
				end
				Result := True
			end
		end

	is_not_superset_ok (s1: STS_NUMERIC_SET [A, EQ]; s2: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_NUMERIC_SET}.is_not_superset?
		do
			if sp_is_not_superset_ok (s1, s2) then
				check
					when_a_universe: attached {STS_UNIVERSE [A, EQ]} s1 ⇒ not (s1 ⊉ s2)
					disguised_universe: # s1 = # s1.u ⇒ not (s1 ⊉ s2)
					compared_to_universe: attached {STS_UNIVERSE [A, EQ]} s2 ⇒ s1 ⊉ s2 = s1 ≭ s2
					compared_to_disguised_universe: # s2 = # s1.u ⇒ s1 ⊉ s2 = s1 ≭ s2
				end
				Result := True
			end
		end

	is_comparable_ok (s1: STS_NUMERIC_SET [A, EQ]; s2: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_NUMERIC_SET}.is_comparable?
		do
			if sp_is_comparable_ok (s1, s2) then
				check
					when_a_universe: attached {STS_UNIVERSE [A, EQ]} s1 ⇒ s1.is_comparable (s2)
					disguised_universe: # s1 = # s1.u ⇒ s1.is_comparable (s2)
					compared_to_universe: attached {STS_UNIVERSE [A, EQ]} s2 ⇒ s1.is_comparable (s2)
					compared_to_disguised_universe: # s2 = # s1.u ⇒ s1.is_comparable (s2)
				end
				Result := True
			end
		end

	is_not_comparable_ok (s1: STS_NUMERIC_SET [A, EQ]; s2: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_NUMERIC_SET}.is_not_comparable?
		do
			if sp_is_not_comparable_ok (s1, s2) then
				check
					when_a_universe: attached {STS_UNIVERSE [A, EQ]} s1 ⇒ not s1.is_not_comparable (s2)
					disguised_universe: # s1 = # s1.u ⇒ not s1.is_not_comparable (s2)
					compared_to_universe: attached {STS_UNIVERSE [A, EQ]} s2 ⇒ not s1.is_not_comparable (s2)
					compared_to_disguised_universe: # s2 = # s1.u ⇒ not s1.is_not_comparable (s2)
				end
				Result := True
			end
		end

	is_strict_subset_ok (s1: STS_NUMERIC_SET [A, EQ]; s2: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_NUMERIC_SET}.is_strict_subset?
		do
			check
				when_a_universe: attached {STS_UNIVERSE [A, EQ]} s1 ⇒ not (s1 ⊂ s2)
				disguised_universe: # s1 = # s1.u ⇒ not (s1 ⊂ s2)
				compared_to_universe: attached {STS_UNIVERSE [A, EQ]} s2 ⇒ (s1 ⊂ s2) = (s1 ≭ s2)
				compared_to_disguised_universe: # s2 = # s1.u ⇒ (s1 ⊂ s2) = (s1 ≭ s2)
			end
			Result := True
		end

	is_strict_superset_ok (s1: STS_NUMERIC_SET [A, EQ]; s2: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_NUMERIC_SET}.is_strict_superset?
		do
			if sp_is_strict_superset_ok (s1, s2) then
				check
					when_a_universe: attached {STS_UNIVERSE [A, EQ]} s1 ⇒ (s1 ⊃ s2) = (s1 ≭ s2)
					disguised_universe: # s1 = # s1.u ⇒ (s1 ⊃ s2) = (s1 ≭ s2)
					compared_to_universe: attached {STS_UNIVERSE [A, EQ]} s2 ⇒ not (s1 ⊃ s2)
					compared_to_disguised_universe: # s2 = # s1.u ⇒ not (s1 ⊃ s2)
				end
				Result := True
			end
		end

	is_trivial_subset_ok (s1: STS_NUMERIC_SET [A, EQ]; s2: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_NUMERIC_SET}.is_trivial_subset?
		do
			check
				when_a_universe: attached {STS_UNIVERSE [A, EQ]} s1 ⇒ s1.is_trivial_subset (s2) = (s1 ≍ s2)
				disguised_universe: # s1 = # s1.u ⇒ s1.is_trivial_subset (s2) = (s1 ≍ s2)
			end
			Result := True
		end

	is_proper_subset_ok (s1: STS_NUMERIC_SET [A, EQ]; s2: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_NUMERIC_SET}.is_proper_subset?
		do
			if sp_is_proper_subset_ok (s1, s2) then
				check
					when_a_universe: attached {STS_UNIVERSE [A, EQ]} s1 ⇒ not s1.is_proper_subset (s2)
					disguised_universe: # s1 = # s1.u ⇒ not s1.is_proper_subset (s2)
					compared_to_universe: attached {STS_UNIVERSE [A, EQ]} s2 ⇒ s1.is_proper_subset (s2) = (s1 ≭ o and s1 ≭ s2)
					compared_to_disguised_universe: # s2 = # s1.u ⇒ s1.is_proper_subset (s2) = (s1 ≭ o and s1 ≭ s2)
				end
				Result := True
			end
		end

	is_proper_superset_ok (s1: STS_NUMERIC_SET [A, EQ]; s2: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_NUMERIC_SET}.is_proper_superset?
		do
			if sp_is_proper_superset_ok (s1, s2) then
				check
					when_a_universe: attached {STS_UNIVERSE [A, EQ]} s1 ⇒ s1.is_proper_superset (s2) = (s2 ≭ o and s1 ≭ s2)
					disguised_universe: # s1 = # s1.u ⇒ s1.is_proper_superset (s2) = (s2 ≭ o and s1 ≭ s2)
					compared_to_universe: attached {STS_UNIVERSE [A, EQ]} s2 ⇒ not s1.is_proper_superset (s2)
					compared_to_disguised_universe: # s2 = # s1.u ⇒ not s1.is_proper_superset (s2)
				end
				Result := True
			end
		end

	is_disjoint_ok (s1: STS_NUMERIC_SET [A, EQ]; s2: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_NUMERIC_SET}.is_disjoint?
		do
			if sp_is_disjoint_ok (s1, s2) then
				check
					when_a_universe: attached {STS_UNIVERSE [A, EQ]} s1 ⇒ s1.is_disjoint (s2) = (s2 ≍ o)
					disguised_universe: # s1 = # s1.u ⇒ s1.is_disjoint (s2) = (s2 ≍ o)
					compared_to_universe: attached {STS_UNIVERSE [A, EQ]} s2 ⇒ s1.is_disjoint (s2) = (s1 ≍ o)
					compared_to_disguised_universe: # s2 = # s1.u ⇒ s1.is_disjoint (s2) = (s1 ≍ o)
				end
				Result := True
			end
		end

feature -- Properties (Restriction)

	without_zero_ok (s: STS_NUMERIC_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_NUMERIC_SET}.without_zero?
		do
			check
				has_no_zero: zero ∉ s.without_zero
				subset: s.without_zero ⊆ s
				no_change: zero ∉ s ⇒ s.without_zero ≍ s
				everything_is_invertible: s.without_zero |∀ agent (z: A): BOOLEAN do Result := z.is_invertible end
			end
			Result := True
		end

feature -- Properties (Operation)

	complemented_ok (s01: STS_NUMERIC_SET [A, EQ]; s02, s: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_NUMERIC_SET}.complemented?
		do
			if sp_complemented_ok (s01, s02, s) then
				check
					when_a_universe: attached {STS_UNIVERSE [A, EQ]} s01 and s01 ⊆ s01 ⇒ (s01 ∁ s) ≍ o
					disguised_universe: # s01 = # s01.u and s01 ⊆ s01 ⇒ (s01 ∁ s) ≍ o
					complemented_from_universe: attached {STS_UNIVERSE [A, EQ]} s ⇒ s01 ⊆ s and then (s01 ∁ s) ≍ ∁ s01
					complemented_from_disguised_universe: # s = # s01.u ⇒ s01 ⊆ s and then (s01 ∁ s) ≍ ∁ s01
				end
				Result := True
			end
		end

	complement_ok (s1, s2: STS_NUMERIC_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_NUMERIC_SET}.complement?
		do
			check
				u: attached current_universe as u
				definition: ∁ s1 ≍ (u | agent s1.does_not_have)
				by_difference: ∁ s1 ≍ (u ∖ s1)
				by_reduction: ∁ s1 ≍ element_to_set.set_reduction (s1, u, agent {STS_SET [A, EQ]}.without)
				disjoint: (∁ s1).is_disjoint (s1)
				complementary: s1 ∪ (∁ s1) ≍ u
				involutive: ∁ ∁ s1 ≍ s1
				dual_inclusion: s1 ⊆ s2 = ∁ s1 ⊇ (∁ s2)
				full_complement: ∁ o ≍ u
				empty_complement: ∁ u ≍ o
			end
			Result := True
		end

	united_ok (a: A; s1: STS_NUMERIC_SET [A, EQ]; s2, s3, s4: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_NUMERIC_SET}.united?
		do
			if sp_united_ok (a, s1, s2, s3, s4) then
				check
					left_universe: attached {STS_UNIVERSE [A, EQ]} s1 ⇒ (s1 ∪ s2) ≍ s1
					disguised_left_universe: # s1 = # s1.u ⇒ (s1 ∪ s2) ≍ s1
					right_universe: attached {STS_UNIVERSE [A, EQ]} s2 ⇒ (s1 ∪ s2) ≍ s2
					disguised_right_universe: # s2 = # s1.u ⇒ (s1 ∪ s2) ≍ s2
				end
				Result := True
			end
		end

	intersected_ok (a: A; s1: STS_NUMERIC_SET [A, EQ]; s2, s3, s4: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_NUMERIC_SET}.intersected?
		do
			if sp_intersected_ok (a, s1, s2, s3, s4) then
				check
					left_universe: attached {STS_UNIVERSE [A, EQ]} s1 ⇒ (s1 ∩ s2) ≍ s2
					disguised_left_universe: # s1 = # s1.u ⇒ (s1 ∩ s2) ≍ s2
					right_universe: attached {STS_UNIVERSE [A, EQ]} s2 ⇒ (s1 ∩ s2) ≍ s1
					disguised_right_universe: # s2 = # s1.u ⇒ (s1 ∩ s2) ≍ s1
				end
				Result := True
			end
		end

	subtracted_ok (a: A; s1: STS_NUMERIC_SET [A, EQ]; s2, s3, s4, s5: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_NUMERIC_SET}.subtracted?
		do
			if sp_subtracted_ok (a, s1, s2, s3, s4, s5) then
				check
					right_universe: attached {STS_UNIVERSE [A, EQ]} s2 ⇒ (s1 ∖ s2) ≍ o
					disguised_right_universe: # s2 = # s1.u ⇒ (s1 ∖ s2) ≍ o
				end
				Result := True
			end
		end

	subtracted_symmetricaly_ok (a: A; s1: STS_NUMERIC_SET [A, EQ]; s2, s3: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_NUMERIC_SET}.subtracted_symmetricaly?
		do
			if sp_subtracted_symmetricaly_ok (a, s1, s2, s3) then
				check
					left_universe: attached {STS_UNIVERSE [A, EQ]} s1 ⇒ (s1 ⊖ s2) ≍ (s1 ∖ s2)
					disguised_left_universe: # s1 = # s1.u ⇒ (s1 ⊖ s2) ≍ (s1 ∖ s2)
					right_universe: attached {STS_UNIVERSE [A, EQ]} s2 ⇒ (s1 ⊖ s2) ≍ (s2 ∖ s1)
					disguised_right_universe: # s2 = # s1.u ⇒ (s1 ⊖ s2) ≍ (s2 ∖ s1)
				end
				Result := True
			end
		end

feature -- Anchor

	universe_anchor: STS_NUMERIC_SET [A, EQ]
			-- <Precursor>
		deferred
		end

feature {NONE} -- Implementation

	proper_complex_quotient_implementation (z1, z2: A): STS_COMPLEX_NUMBER
			-- Quotient of `z1' by `z2' as done by an implementation of {STS_COMPLEX_NUMBER}
			--| TODO: DRY.
		require
			good_divisor: z1.divisible (z2)
		do
			Result := z1 / z2
		ensure
			scale_log: attached Result.value_logb ((z2.a.abs ∨ z2.b.abs).value) as scale_log
			int_scale_log: attached if Result.value_is_finite (scale_log) then scale_log.truncated_to_integer else 0 end as int_scale_log
			scaled_c: attached (z2.a.value * 2 ^ (- int_scale_log)) as scaled_c
			scaled_d: attached (z2.b.value * 2 ^ (- int_scale_log)) as scaled_d
			denom: attached (scaled_c ^ 2 + scaled_d ^ 2) as denom
			invertible_denom: z2.is_proper_complex ⇒ denom /= 0 -- z2.a /= 0 and z2.b /= 0 <== z2.is_proper_complex

				-- (a+bi) / (c + di) = (ac+bd)/(c^2+d^2)+(bc-ad)i/(c^2+d^2)				
			complex_divisor_a_good_divisor:
				z2.is_proper_complex ⇒ (z1.a.value * scaled_c + z1.b.value * scaled_d).divisible (denom) -- denom /= 0 <== z2.is_proper_complex
			complex_divisor_b_good_divisor: z2.is_proper_complex ⇒ (z1.b.value * scaled_c - z1.a.value * scaled_d).divisible (denom)
			complex_divisor_quot_a: attached {STS_REAL_NUMBER} Result.real_from_double_value
				(((z1.a.value * scaled_c + z1.b.value * scaled_d) / denom) * 2 ^ (- int_scale_log)) as complex_by_complex_quot_a
			complex_divisor_quot_b: attached {STS_REAL_NUMBER} Result.real_from_double_value
				(((z1.b.value * scaled_c - z1.a.value * scaled_d) / denom) * 2 ^ (- int_scale_log)) as complex_by_complex_quot_b
			complex_divisor_a: z2.is_proper_complex ⇒
				(Result.a - complex_by_complex_quot_a).abs ≤ (Result.a.machine_epsilon ∨ complex_by_complex_quot_a.relative_epsilon)
			complex_divisor_b: z2.is_proper_complex ⇒
				(Result.b - complex_by_complex_quot_b).abs ≤ (Result.b.machine_epsilon ∨ complex_by_complex_quot_b.relative_epsilon)

			infinite_dividend: z1.is_infinite and z2.is_finite ⇒ Result.is_infinite
			infinite_divisor: z1.is_finite and z2.is_infinite ⇒ Result ≍ zero
			zero_divisor:
				(z1.is_finite and z1 ≭ zero or z1.is_infinite) and z2 ≍ zero ⇒ Result.is_infinite -- For an implementation that accepts a zero divisor.
		end

note
	copyright: "Copyright (c) 2012-2022, Rosivaldo Fernandes Alves"
	license: "[
		Eiffel Forum License v2
		(see http://www.eiffel.com/licensing/forum.txt)
		]"
	source: "https://launchpad.net/set-theory"
end
