﻿note
	description: "Object that checks whether the properties verified within set theory hold for an implementation of {STS_ORDERED_NUMERIC_SET}"
	author: "Rosivaldo Fernandes Alves"
	date: "$Date$"
	revision: "$Revision$"

deferred class
	ORDERED_NUMERIC_SET_PROPERTIES [A -> {STS_REAL_NUMBER}, EQ -> STS_NUMERIC_EQUALITY [A]]

inherit
	NUMERIC_SET_PROPERTIES [A, EQ]
		rename
			zero_ok as numeric_zero_ok
		end

feature -- Properties (Access)

	zero_ok (s: STS_ORDERED_NUMERIC_SET [A, EQ]; x: A): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_ORDERED_NUMERIC_SET}.zero?
		local
			inherited_ok, broken_left_absorption, broken_right_absorption, broken_dividend_absorption: BOOLEAN
		do
			if not (broken_left_absorption or broken_right_absorption or broken_dividend_absorption) then
				inherited_ok := numeric_zero_ok (s, x)
				if inherited_ok then
					check
						neutral_left_term: (s.zero + x) ≍ x
						neutral_right_term: (x + s.zero) ≍ x
						quasi_neutral_minuend: (s.zero - x) ≍ - x
						neutral_subtrahend: (x - s.zero) ≍ x
						absorbing_left_factor: s.zero ⋅ x ≍ s.zero
					end
				end
			end
			if inherited_ok and not (broken_right_absorption or broken_dividend_absorption) then
				check
					absorbing_right_factor: (x ⋅ s.zero) ≍ s.zero
				end
			end
			if inherited_ok and not broken_dividend_absorption then
				check
					absorbing_dividend: s.zero.divisible (x) implies (s.zero / x) ≍ s.zero
				end
			end
			Result := True
		rescue
			check
				print_s_x: (
					agent (ia_s: STS_ORDERED_NUMERIC_SET [A, EQ]; ia_x: A): BOOLEAN
						do
							print (ia_s) print (", ")
							print (ia_x) print ('%N')
							Result := True
						end (s, x)
					).item
			end
			if not broken_left_absorption and s.zero ⋅ x ≭ s.zero then
				check
					broken_left_absorption_cause: not x.is_finite
					broken_left_absorption: (s.zero ⋅ x).is_nan
				end
				broken_left_absorption := True
				retry
			elseif not broken_right_absorption and (x ⋅ s.zero) ≭ s.zero then
				check
					broken_right_absorption_cause: not x.is_finite
					broken_right_absorption: (x ⋅ s.zero).is_nan
				end
				broken_right_absorption := True
				retry
			elseif not broken_dividend_absorption and s.zero.divisible (x) and (s.zero / x) ≭ s.zero then
				check
					broken_dividend_absorption_cause: x.is_nan or x ≍ zero
					broken_dividend_absorption: (s.zero / x).is_nan
				end
				broken_dividend_absorption := True
				retry -- TODO: goto?
			end
		end

feature -- Properties (Restriction)

	non_positive_ok (s: STS_ORDERED_NUMERIC_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_ORDERED_NUMERIC_SET}.non_positive?
		do
			check
				definition: s.non_positive ≍ (s ∖ s.positive)
				by_intersection: s.non_positive ≍ (s ∩ (s.negative & zero))
				subset: s.non_positive ⊆ s
				partition: s.is_partitioned (s.non_positive.as_singleton & s.positive / o)
				quasi_disjoint_sets: s.non_positive ∩ s.non_negative ⊆ singleton (zero)
				without_zero: s.non_positive.without_zero ≍ s.negative
				non_positive_only: s.non_positive |∀ agent (x: A): BOOLEAN do Result := x ≤ zero end
				no_positive_number: s.non_positive |∄ agent (x: A): BOOLEAN do Result := zero < x end
				non_empty_interval: s |∃ agent (x: A): BOOLEAN do Result := x ≤ zero end implies not s.non_positive.is_empty
				possibly_singleton_interval: s |∀ agent (x: A): BOOLEAN do Result := zero ≤ x end implies s.non_positive ⊆ singleton (zero)
				empty_interval: s |∀ agent (x: A): BOOLEAN do Result := zero < x end implies s.non_positive.is_empty
			end
			Result := True
		end

	non_negative_ok (s: STS_ORDERED_NUMERIC_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_ORDERED_NUMERIC_SET}.non_negative?
		do
			check
				definition: s.non_negative ≍ (s ∖ s.negative)
				by_intersection: s.non_negative ≍ (s ∩ (s.positive & zero))
				subset: s.non_negative ⊆ s
				partition: s.is_partitioned (s.non_negative.as_singleton & s.negative / o)
				quasi_disjoint_sets: s.non_negative ∩ s.non_positive ⊆ singleton (zero)
				without_zero: s.non_negative.without_zero ≍ s.positive
				non_negative_only: s.non_negative |∀ agent (x: A): BOOLEAN do Result := zero ≤ x end
				no_negative_number: s.non_negative |∄ agent (x: A): BOOLEAN do Result := x < zero end
				non_empty_interval: s |∃ agent (x: A): BOOLEAN do Result := zero ≤ x end implies not s.non_negative.is_empty
				possibly_singleton_interval: s |∀ agent (x: A): BOOLEAN do Result := x ≤ zero end implies s.non_negative ⊆ singleton (zero)
				empty_interval: s |∀ agent (x: A): BOOLEAN do Result := x < zero end implies s.non_negative.is_empty
			end
			Result := True
		end

	negative_ok (s: STS_ORDERED_NUMERIC_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_ORDERED_NUMERIC_SET}.negative?
		do
			check
				definition: s.negative ≍ (s ∖ s.non_negative)
				by_intersection: s.negative ≍ (s ∩ (s.non_positive.without_zero))
				subset: s.negative ⊆ s
				partition: s.is_partitioned (s.negative.as_singleton & s.non_negative / o)
				without_zero: s.negative.without_zero ≍ s.negative
				negative_only: s.negative |∀ agent (x: A): BOOLEAN do Result := x < zero end
				no_non_negative_number: s.negative |∄ agent (x: A): BOOLEAN do Result := zero ≤ x end
				non_empty_interval: s |∃ agent (x: A): BOOLEAN do Result := x < zero end implies not s.negative.is_empty
				empty_interval: s |∀ agent (x: A): BOOLEAN do Result := zero ≤ x end implies s.negative.is_empty
			end
			Result := True
		end

	positive_ok (s: STS_ORDERED_NUMERIC_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_ORDERED_NUMERIC_SET}.positive?
		do
			check
				definition: s.positive ≍ (s ∖ s.non_positive)
				by_intersection: s.positive ≍ (s ∩ (s.non_negative.without_zero))
				subset: s.positive ⊆ s
				partition: s.is_partitioned (s.positive.as_singleton & s.non_positive / o)
				without_zero: s.positive.without_zero ≍ s.positive
				positive_only: s.positive |∀ agent (x: A): BOOLEAN do Result := zero < x end
				no_non_positive_number: s.positive |∄ agent (x: A): BOOLEAN do Result := x ≤ zero end
				non_empty_interval: s |∃ agent (x: A): BOOLEAN do Result := zero < x end implies not s.positive.is_empty
				empty_interval: s |∀ agent (x: A): BOOLEAN do Result := x ≤ zero end implies s.positive.is_empty
			end
			Result := True
		end

feature -- Properties (Interval)

	closed_interval_ok (s: STS_ORDERED_NUMERIC_SET [A, EQ]; a, b: A): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_ORDERED_NUMERIC_SET}.closed_interval?
			--| TODO: Feature's Basic text view is broken (due to unicode characteres)?
		do
			check
				definition: s [a, b] ≍ ((s ∖ s.left_unbounded_right_open_interval (a)) ∖ s.left_open_right_unbounded_interval (b))
				by_intersection: s [a, b] ≍ (s.left_closed_right_unbounded_interval (a) ∩ s.left_unbounded_right_closed_interval (b))
				subset: s [a, b] ⊆ s
				partition: a ≤ b implies
					s.is_partitioned (s [a, b].as_singleton & s.left_unbounded_right_open_interval (a) & s.left_open_right_unbounded_interval (b) / o)
				insider_only: s [a, b] |∀ agent (ia_a, x, ia_b: A): BOOLEAN do Result := ia_a ≤ x and x ≤ ia_b end (a, ?, b)
				no_outsider: s [a, b] |∄ agent (ia_a, x, ia_b: A): BOOLEAN do Result := x < ia_a or ia_b < x end (a, ?, b)
				non_empty_interval: (a ∈ s or b ∈ s) and a ≤ b implies not s [a, b].is_empty
				singleton_interval: a ∈ s and b ≍ a implies s [a, b].is_singleton
				empty_interval: a > b implies s [a, b].is_empty
			end
			Result := True
		end

	left_closed_right_open_interval_ok (s: STS_ORDERED_NUMERIC_SET [A, EQ]; a, b: A): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_ORDERED_NUMERIC_SET}.left_closed_right_open_interval?
		do
			check
				definition: s.left_closed_right_open_interval (a, b) ≍
					((s ∖ s.left_unbounded_right_open_interval (a)) ∖ s.left_closed_right_unbounded_interval (b))
				by_intersection: s.left_closed_right_open_interval (a, b) ≍
					(s.left_closed_right_unbounded_interval (a) ∩ s.left_unbounded_right_open_interval (b))
				subset: s.left_closed_right_open_interval (a, b) ⊆ s
				partition: a ≤ b implies s.is_partitioned (
					s.left_unbounded_right_open_interval (a).as_singleton &
					s.left_closed_right_open_interval (a, b) &
					s.left_closed_right_unbounded_interval (b) /
					o
					)
				discarded_b: b ∉ s.left_closed_right_open_interval (a, b)
				insider_only: s.left_closed_right_open_interval (a, b) |∀
					agent (ia_a, x, ia_b: A): BOOLEAN do Result := ia_a ≤ x and x < ia_b end (a, ?, b)
				no_outsider: s.left_closed_right_open_interval (a, b) |∄ agent (ia_a, x, ia_b: A): BOOLEAN do Result := x < ia_a or ia_b ≤ x end (a, ?, b)
				non_empty_interval: a ∈ s and a < b implies not s.left_closed_right_open_interval (a, b).is_empty
				empty_interval: a ≥ b implies s.left_closed_right_open_interval (a, b).is_empty
			end
			Result := True
		end

	left_open_right_closed_interval_ok (s: STS_ORDERED_NUMERIC_SET [A, EQ]; a, b: A): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_ORDERED_NUMERIC_SET}.left_open_right_closed_interval?
		do
			check
				definition: s.left_open_right_closed_interval (a, b) ≍
					((s ∖ s.left_unbounded_right_closed_interval (a)) ∖ s.left_open_right_unbounded_interval (b))
				by_intersection: s.left_open_right_closed_interval (a, b) ≍
					(s.left_open_right_unbounded_interval (a) ∩ s.left_unbounded_right_closed_interval (b))
				subset: s.left_open_right_closed_interval (a, b) ⊆ s
				partition: a ≤ b implies s.is_partitioned (
					s.left_unbounded_right_closed_interval (a).as_singleton &
					s.left_open_right_closed_interval (a, b) &
					s.left_open_right_unbounded_interval (b) /
					o
					)
				discarded_a: a ∉ s.left_open_right_closed_interval (a, b)
				insider_only: s.left_open_right_closed_interval (a, b) |∀
					agent (ia_a, x, ia_b: A): BOOLEAN do Result := ia_a < x and x ≤ ia_b end (a, ?, b)
				no_outsider: s.left_open_right_closed_interval (a, b) |∄ agent (ia_a, x, ia_b: A): BOOLEAN do Result := x ≤ ia_a or ia_b < x end (a, ?, b)
				non_empty_interval: a < b and b ∈ s implies not s.left_open_right_closed_interval (a, b).is_empty
				empty_interval: a ≥ b implies s.left_open_right_closed_interval (a, b).is_empty
			end
			Result := True
		end

	open_interval_ok (s: STS_ORDERED_NUMERIC_SET [A, EQ]; a, b: A): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_ORDERED_NUMERIC_SET}.open_interval?
		do
			check
				definition: s (a, b) ≍ ((s ∖ s.left_unbounded_right_closed_interval (a)) ∖ s.left_closed_right_unbounded_interval (b))
				by_intersection: s (a, b) ≍ (s.left_open_right_unbounded_interval (a) ∩ s.left_unbounded_right_open_interval (b))
				subset: s (a, b) ⊆ s
				partition: a < b implies
					s.is_partitioned (s.left_unbounded_right_closed_interval (a).as_singleton & s (a, b) & s.left_closed_right_unbounded_interval (b) / o)
				discarded_a: a ∉ s (a, b)
				discarded_b: b ∉ s (a, b)
				insider_only: s (a, b) |∀ agent (ia_a, x, ia_b: A): BOOLEAN do Result := ia_a < x and x < ia_b end (a, ?, b)
				no_outsider: s (a, b) |∄ agent (ia_a, x, ia_b: A): BOOLEAN do Result := x ≤ ia_a or ia_b ≤ x end (a, ?, b)
				empty_interval: a ≥ b implies s (a, b).is_empty
			end
			Result := True
		end

	left_closed_right_unbounded_interval_ok (s: STS_ORDERED_NUMERIC_SET [A, EQ]; a: A): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_ORDERED_NUMERIC_SET}.left_closed_right_unbounded_interval?
		do
			check
				definition: s.left_closed_right_unbounded_interval (a) ≍ ((s ∖ s.left_unbounded_right_open_interval (a)))
				subset: s.left_closed_right_unbounded_interval (a) ⊆ s
				partition: s.is_partitioned (s.left_unbounded_right_open_interval (a).as_singleton & s.left_closed_right_unbounded_interval (a) / o)
				insider_only: s.left_closed_right_unbounded_interval (a) |∀ agent (ia_a, x: A): BOOLEAN do Result := ia_a ≤ x end (a, ?)
				no_outsider: s.left_closed_right_unbounded_interval (a) |∄ agent (ia_a, x: A): BOOLEAN do Result := x < ia_a end (a, ?)
				non_empty_interval: a ∈ s implies not s.left_closed_right_unbounded_interval (a).is_empty
			end
			Result := True
		end

	left_open_right_unbounded_interval_ok (s: STS_ORDERED_NUMERIC_SET [A, EQ]; a: A): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_ORDERED_NUMERIC_SET}.left_open_right_unbounded_interval?
		do
			check
				definition: s.left_open_right_unbounded_interval (a) ≍ ((s ∖ s.left_unbounded_right_closed_interval (a)))
				subset: s.left_open_right_unbounded_interval (a) ⊆ s
				partition: s.is_partitioned (s.left_unbounded_right_closed_interval (a).as_singleton & s.left_open_right_unbounded_interval (a) / o)
				discarded_a: a ∉ s.left_open_right_unbounded_interval (a)
				insider_only: s.left_open_right_unbounded_interval (a) |∀ agent (ia_a, x: A): BOOLEAN do Result := ia_a < x end (a, ?)
				no_outsider: s.left_open_right_unbounded_interval (a) |∄ agent (ia_a, x: A): BOOLEAN do Result := x ≤ ia_a end (a, ?)
			end
			Result := True
		end

	left_unbounded_right_closed_interval_ok (s: STS_ORDERED_NUMERIC_SET [A, EQ]; b: A): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_ORDERED_NUMERIC_SET}.left_unbounded_right_closed_interval?
		do
			check
				definition: s.left_unbounded_right_closed_interval (b) ≍ ((s ∖ s.left_open_right_unbounded_interval (b)))
				subset: s.left_unbounded_right_closed_interval (b) ⊆ s
				partition: s.is_partitioned (s.left_unbounded_right_closed_interval (b).as_singleton & s.left_open_right_unbounded_interval (b) / o)
				insider_only: s.left_unbounded_right_closed_interval (b) |∀ agent (x, ia_b: A): BOOLEAN do Result := x ≤ ia_b end (?, b)
				no_outsider: s.left_unbounded_right_closed_interval (b) |∄ agent (x, ia_b: A): BOOLEAN do Result := ia_b < x end (?, b)
				non_empty_interval: b ∈ s implies not s.left_unbounded_right_closed_interval (b).is_empty
			end
			Result := True
		end

	left_unbounded_right_open_interval_ok (s: STS_ORDERED_NUMERIC_SET [A, EQ]; b: A): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_ORDERED_NUMERIC_SET}.left_unbounded_right_open_interval?
		do
			check
				definition: s.left_unbounded_right_open_interval (b) ≍ ((s ∖ s.left_closed_right_unbounded_interval (b)))
				subset: s.left_unbounded_right_open_interval (b) ⊆ s
				partition: s.is_partitioned (s.left_unbounded_right_open_interval (b).as_singleton & s.left_closed_right_unbounded_interval (b) / o)
				discarded_a: b ∉ s.left_unbounded_right_open_interval (b)
				insider_only: s.left_unbounded_right_open_interval (b) |∀ agent (x, ia_b: A): BOOLEAN do Result := x < ia_b end (?, b)
				no_outsider: s.left_unbounded_right_open_interval (b) |∄ agent (x, ia_b: A): BOOLEAN do Result := ia_b ≤ x end (?, b)
			end
			Result := True
		end

	unbounded_interval_ok (s: STS_ORDERED_NUMERIC_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_ORDERED_NUMERIC_SET}.unbounded_interval?
		do
			check
				same_emptiness: s.unbounded_interval ≍ o = s ≍ o
			end
			Result := True
		end

feature -- Properties (Element comparison)

	min_ok (s: STS_ORDERED_NUMERIC_SET [A, EQ]; x, y, z: A): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_ORDERED_NUMERIC_SET}.min?
		do
			check
				idempotent: s.min (x, x) ≍ x
				commutative: s.min (x, y) ≍ s.min (y, x)
				associative: s.min (s.min (x, y), z) ≍ s.min (x, s.min (y, z))
			end
			Result := True
		end

	max_ok (s: STS_ORDERED_NUMERIC_SET [A, EQ]; x, y, z: A): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_ORDERED_NUMERIC_SET}.max?
		do
			check
				idempotent: s.max (x, x) ≍ x
				commutative: s.max (x, y) ≍ s.max (y, x)
				associative: s.max (s.max (x, y), z) ≍ s.max (x, s.max (y, z))
			end
			Result := True
		end

feature -- Anchor

	universe_anchor: STS_ORDERED_NUMERIC_SET [A, EQ]
			-- <Precursor>
		deferred
		end

note
	copyright: "Copyright (c) 2012-2021, Rosivaldo Fernandes Alves"
	license: "[
		Eiffel Forum License v2
		(see http://www.eiffel.com/licensing/forum.txt)
		]"
	source: "https://launchpad.net/set-theory"
end
