﻿note
	description: "Object that checks whether the properties verified within set theory hold for an implementation of {STS_UNIVERSE}"
	author: "Rosivaldo Fernandes Alves"
	date: "$Date$"
	revision: "$Revision$"

deferred class
	UNIVERSE_PROPERTIES [A, EQ -> STS_EQUALITY [A]]

inherit
	SET_PROPERTIES [A, EQ]
		rename
			is_empty_ok as sp_is_empty_ok,
			others_ok as sp_others_ok,
			does_not_have_ok as sp_does_not_have_ok,
			with_ok as sp_with_ok,
			equals_ok as sp_equals_ok,
			is_subset_ok as sp_is_subset_ok,
			is_superset_ok as sp_is_superset_ok,
			is_not_superset_ok as sp_is_not_superset_ok,
			is_comparable_ok as sp_is_comparable_ok,
			is_not_comparable_ok as sp_is_not_comparable_ok,
			is_strict_subset_ok as sp_is_strict_subset_ok,
			is_strict_superset_ok as sp_is_strict_superset_ok,
			is_proper_subset_ok as sp_is_proper_subset_ok,
			is_proper_superset_ok as sp_is_proper_superset_ok,
			is_disjoint_ok as sp_is_disjoint_ok,
			complemented_ok as sp_complemented_ok,
			united_ok as sp_united_ok,
			intersected_ok as sp_intersected_ok,
			subtracted_symmetricaly_ok as sp_subtracted_symmetricaly_ok
		end

feature -- Properties (Primitive)

	is_empty_ok (u: STS_UNIVERSE [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_UNIVERSE}.is_empty?
		do
			if sp_is_empty_ok (u) then
				check
					is_not_empty: not u.is_empty
				end
				Result := True
			end
		end

	others_ok (u: STS_UNIVERSE [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_UNIVERSE}.others?
		do
			if sp_others_ok (u) then
				check
					strict_subset: u.others ⊂ u
					strict_superset: u ⊃ u.others
				end
				Result := True
			end
		end

feature -- Properties (Membership)

	does_not_have_ok (u: STS_UNIVERSE [A, EQ]; a: A): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_UNIVERSE}.does_not_have?
		do
			if sp_does_not_have_ok (u, a) then
				check
					has_everything: not (u ∌ a)
				end
				Result := True
			end
		end

feature -- Properties (Construction)

	with_ok (u: STS_UNIVERSE [A, EQ]; a: A): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_UNIVERSE}.with?
		do
			if sp_with_ok (u, a) then
				check
					itself: u & a ≍ u
				end
				Result := True
			end
		end

feature -- Properties (Comparison)

	equals_ok (u: STS_UNIVERSE [A, EQ]; s: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_UNIVERSE}.equals?
		do
			check
				definition: u ≍ s = (attached {STS_UNIVERSE [A, EQ]} s or # s = # u)
			end
			Result := True
		end

	is_subset_ok (u: STS_UNIVERSE [A, EQ]; s: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_UNIVERSE}.is_subset?
		do
			check
				definition: u ⊆ s = u ≍ s
			end
			Result := True
		end

	is_superset_ok (u: STS_UNIVERSE [A, EQ]; s: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_UNIVERSE}.is_superset?
		do
			if sp_is_superset_ok (u, s) then
				check
					always_a_superset: u ⊇ s
				end
				Result := True
			end
		end

	is_not_superset_ok (u: STS_UNIVERSE [A, EQ]; s: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_UNIVERSE}.is_not_superset?
		do
			if sp_is_not_superset_ok (u, s) then
				check
					always_a_superset: not (u ⊉ s)
				end
				Result := True
			end
		end

	is_comparable_ok (u: STS_UNIVERSE [A, EQ]; s: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_UNIVERSE}.is_comparable?
		do
			if sp_is_comparable_ok (u, s) then
				check
					always_comparable: u.is_comparable (s)
				end
				Result := True
			end
		end

	is_not_comparable_ok (u: STS_UNIVERSE [A, EQ]; s: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_UNIVERSE}.is_not_comparable?
		do
			if sp_is_not_comparable_ok (u, s) then
				check
					always_comparable: not u.is_not_comparable (s)
				end
				Result := True
			end
		end

	is_strict_subset_ok (u: STS_UNIVERSE [A, EQ]; s: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_UNIVERSE}.is_strict_subset?
		do
			check
				never_a_strict_subset: not (u ⊂ s)
			end
			Result := True
		end

	is_strict_superset_ok (u: STS_UNIVERSE [A, EQ]; s: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_UNIVERSE}.is_strict_superset?
		do
			if sp_is_strict_superset_ok (u, s) then
				check
					definition: u ⊃ s = (u ≭ s)
				end
				Result := True
			end
		end

	is_trivial_subset_ok (u: STS_UNIVERSE [A, EQ]; s: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_UNIVERSE}.is_trivial_subset?
		do
			check
				iff_equals: u.is_trivial_subset (s) = u ≍ s
			end
			Result := True
		end

	is_proper_subset_ok (u: STS_UNIVERSE [A, EQ]; s: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_UNIVERSE}.is_proper_subset?
		do
			if sp_is_proper_subset_ok (u, s) then
				check
					never_a_proper_subset: not u.is_proper_subset (s)
				end
				Result := True
			end
		end

	is_proper_superset_ok (u: STS_UNIVERSE [A, EQ]; s: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_UNIVERSE}.is_proper_superset?
		do
			if sp_is_proper_superset_ok (u, s) then
				check
					definition: u.is_proper_superset (s) = (s ≭ o and u ≭ s)
				end
				Result := True
			end
		end

	is_disjoint_ok (u: STS_UNIVERSE [A, EQ]; s: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_UNIVERSE}.is_disjoint?
		do
			if sp_is_disjoint_ok (u, s) then
				check
					iff_other_is_empty: u.is_disjoint (s) = s.is_empty
				end
				Result := True
			end
		end

feature -- Properties (Operation)

	complemented_ok (u: STS_UNIVERSE [A, EQ]; s: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_UNIVERSE}.complemented?
		do
			check
				empty_complement: u ⊆ s implies (u ∁ s) ≍ o
			end
			Result := True
		end

	united_ok (u: STS_UNIVERSE [A, EQ]; s: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_UNIVERSE}.united?
		do
			check
				itself: (u ∪ s) ≍ u
			end
			Result := True
		end

	intersected_ok (u: STS_UNIVERSE [A, EQ]; s: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_UNIVERSE}.intersected?
		do
			check
				other: (u ∩ s) ≍ s
			end
			Result := True
		end

	subtracted_symmetricaly_ok (u: STS_UNIVERSE [A, EQ]; s: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_UNIVERSE}.subtracted_symmetricaly?
		do
			check
				plain_difference: (u ⊖ s) ≍ (u ∖ s)
			end
			Result := True
		end

note
	copyright: "Copyright (c) 2012-2022, Rosivaldo Fernandes Alves"
	license: "[
		Eiffel Forum License v2
		(see http://www.eiffel.com/licensing/forum.txt)
		]"
	source: "https://launchpad.net/set-theory"
end
