﻿note
	description: "[
		Object that checks whether the properties verified within set theory hold for an implementation of {STS_TRANSFORMER}
		]"
	author: "Rosivaldo Fernandes Alves"
	date: "$Date: Sat 2018-12-01 19:25:58 -0300 $"
	revision: "$Revision$"

deferred class
	TRANSFORMER_PROPERTIES [A, B, EQ_A -> STS_EQUALITY [A], EQ_B -> STS_EQUALITY [B]]

inherit
	ELEMENT_PROPERTIES

feature -- Access

	o_b: like current_universe_b.subset_anchor
			-- The empty set of elements like {B}
		deferred
		ensure
			definition: Result.is_empty
		end

	unit_b: like current_universe_b.n_tuple_anchor
			-- The empty n-tuple of terms like {B}
		deferred
		ensure
			is_unit: Result.is_unit
		end

	eq_b: EQ_B
			-- Equality for objects like {B}
		deferred
		end

	current_universe_a: STS_TRANSFORMABLE_SET [A, B, EQ_A, EQ_B]
			-- The current "Universe", i.e. set with every object currently in system memory whose type conforms to {A}.
			-- Notice that this "universe" may change from a call to another.
		deferred
		end

	current_universe_b: STS_SET [B, EQ_B]
			-- The current "universe" of elements like {B}, i.e. set with every object currently in system memory whose type conforms to {B}.
			-- Notice that this "universe" may change from a call to another.
		deferred
		end

feature -- Properties (Operation)

	cartesian_product_ok (tr: STS_TRANSFORMER [A, B, EQ_A, EQ_B]; sa: STS_SET [A, EQ_A]; sb: STS_SET [B, EQ_B]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_TRANSFORMER}.cartesian_product?
		do
			check
				u_a: attached current_universe_a as u_a
				u_b: attached current_universe_b as u_b
				definition: tr.cartesian_product (sa, sb) ≍ (u_a × u_b).filtered_xy (agent (u_a.transformer).anded (agent sa.has, ?, agent sb.has, ?))
			end
			Result := True
		end

feature -- Properties (Transformation)

	set_map_ok (tr: STS_TRANSFORMER [A, B, EQ_A, EQ_B]; s: STS_SET [A, EQ_A]; f: FUNCTION [A, B]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_TRANSFORMER}.set_map?
		do
			check
				u_a: attached current_universe_a as u_a
				u_b: attached current_universe_b as u_b
				definition: tr.set_map (s, f) ≍ (
					u_b | agent  (ia_y: B ia_s: STS_SET [A, EQ_A] ia_f: FUNCTION [A, B]): BOOLEAN
						do
							Result := ia_s |∃ agent  (x: A y: B i2a_f: FUNCTION [A, B]): BOOLEAN
								do
									Result := eq_b (y, i2a_f (x))
								end (?, ia_y, ia_f)
						end (?, s, f)
					)
				by_reduction: tr.set_map (s, f) ≍ element_to_set.set_reduction (
					s, o_b, agent (acc: STS_SET [B, EQ_B] x: A; ia_f: FUNCTION [A, B]): STS_SET [B, EQ_B]
						do
							Result := acc & ia_f (x)
						end (?, ?, f)
					)
				cardinality: # tr.set_map (s, f) <= # s
			end
			Result := True
		end

	tuple_map_ok (tr: STS_TRANSFORMER [A, B, EQ_A, EQ_B]; t: STS_N_TUPLE [A, EQ_A]; f: FUNCTION [A, B]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_TRANSFORMER}.tuple_map?
		do
			check
				u_a: attached current_universe_a as u_a
				u_b: attached current_universe_b as u_b
				definition: tr.tuple_map (t, f).as_singleton ≍ (
					(u_b ^ t.n) | agent (ia_tb: STS_N_TUPLE [B, EQ_B]; ia_t: STS_N_TUPLE [A, EQ_A]; ia_f: FUNCTION [A, B]): BOOLEAN
						do
							Result := ia_t.for_all (
								agent (i: NATURAL; x_i: A; tb: STS_N_TUPLE [B, EQ_B]; i2a_f: FUNCTION [A, B]): BOOLEAN
									do
											check
													-- i ∈ {1..t.n} = {1..tb.n}
												positive: 0 < i
												small_enough: i <= tb.n
											end
										Result := eq_b (tb [i], i2a_f (x_i))
									end (?, ?, ia_tb, ia_f)
								)
						end (?, t, f)
					)
				n: tr.tuple_map (t, f).n = t.n
				lesser_definition: tr.tuple_map (t, f).for_all (
					agent (i: NATURAL y_i: B ia_t: STS_N_TUPLE [A, EQ_A] ia_f: FUNCTION [A, B]): BOOLEAN
						do
								check
										-- i ∈ {1..t.n}
									positive: 0 < i
									small_enough: i <= ia_t.n
								end
							Result := eq_b (y_i, ia_f (ia_t [i]))
						end (?, ?, t, f)
					)
				by_reduction: tr.tuple_map (t, f) ≍ term_to_tuple.tuple_right_reduction (
					t, agent (x: A; acc: STS_N_TUPLE [B, EQ_B] ia_f: FUNCTION [A, B]): STS_N_TUPLE [B, EQ_B]
						do
							Result := acc.after (ia_f (x))
						end (?, ?, f), unit_b
					)
			end
			Result := True
		rescue
			print (t) print ('%N')
		end

	tuple_left_reduction_ok (tr: STS_TRANSFORMER [A, B, EQ_A, EQ_B]; t: STS_N_TUPLE [A, EQ_A]; leftmost: B; f: FUNCTION [B, A, B]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_TRANSFORMER}.tuple_left_reduction?
		do
			check
				base: t.is_unit implies eq_b (tr.tuple_left_reduction (t, leftmost, f), leftmost)
				induction: not t.is_unit implies eq_b (tr.tuple_left_reduction (t, leftmost, f), f (tr.tuple_left_reduction (t.front, leftmost, f), t.last))
				reversed: eq_b (
					tr.tuple_left_reduction (t, leftmost, f), tr.tuple_right_reduction (
						t.reversed, agent  (x: A; acc: B; ia_f: FUNCTION [B, A, B]): B
							do
								Result := ia_f (acc, x)
							end (?, ?, f), leftmost
						)
					)
			end
			Result := True
		end

	tuple_indexed_left_reduction_ok
		(tr: STS_TRANSFORMER [A, B, EQ_A, EQ_B]; t: STS_N_TUPLE [A, EQ_A]; leftmost: B; i: NATURAL; f: FUNCTION [B, NATURAL, A, B]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_TRANSFORMER}.tuple_indexed_left_reduction?
		do
			check
				base: t.is_unit implies eq_b (tr.tuple_indexed_left_reduction (t, leftmost, i, f), leftmost)
				induction: not t.is_unit implies eq_b
					(tr.tuple_indexed_left_reduction (t, leftmost, i, f), f (tr.tuple_indexed_left_reduction (t.front, leftmost, i, f), t.n + i - 1, t.last))
				reversed: eq_b (
					tr.tuple_indexed_left_reduction (t, leftmost, i, f), tr.tuple_indexed_right_reduction (
						t.reversed, 0, agent (m, j: NATURAL x: A acc: B ia_f: FUNCTION [B, NATURAL, A, B]): B
							do
								Result := ia_f (acc, m - j - 1, x)
							end (t.n + i, ?, ?, ?, f), leftmost
						)
					)
			end
			Result := True
		end

	tuple_right_reduction_ok (tr: STS_TRANSFORMER [A, B, EQ_A, EQ_B]; t: STS_N_TUPLE [A, EQ_A]; f: FUNCTION [A, B, B]; rightmost: B): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_TRANSFORMER}.tuple_right_reduction?
		do
			check
				base: t.is_unit implies eq_b (tr.tuple_right_reduction (t, f, rightmost), rightmost)
				induction: not t.is_unit implies eq_b (tr.tuple_right_reduction (t, f, rightmost), tr.tuple_right_reduction (t.front, f, f (t.last, rightmost)))
				reversed: eq_b (
					tr.tuple_right_reduction (t, f, rightmost), tr.tuple_left_reduction (
						t.reversed, rightmost, agent  (acc: B; x: A; ia_f: FUNCTION [A, B, B]): B
							do
								Result := ia_f (x, acc)
							end (?, ?, f)
						)
					)
			end
			Result := True
		end

	tuple_indexed_right_reduction_ok
		(tr: STS_TRANSFORMER [A, B, EQ_A, EQ_B]; t: STS_N_TUPLE [A, EQ_A]; i: NATURAL; f: FUNCTION [NATURAL, A, B, B]; rightmost: B): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_TRANSFORMER}.tuple_indexed_right_reduction?
		do
			check
				base: t.is_unit implies eq_b (tr.tuple_indexed_right_reduction (t, i, f, rightmost), rightmost)
				induction: not t.is_unit implies eq_b (
					tr.tuple_indexed_right_reduction (t, i, f, rightmost), tr.tuple_indexed_right_reduction (t.front, i, f, f (i + t.n - 1, t.last, rightmost))
					)
				reversed: eq_b (
					tr.tuple_indexed_right_reduction (t, i, f, rightmost), tr.tuple_indexed_left_reduction (
						t.reversed, rightmost, 0, agent (m: NATURAL acc: B j: NATURAL x: A ia_f: FUNCTION [NATURAL, A, B, B]): B
							do
								Result := ia_f (m - j - 1, x, acc)
							end (t.n + i, ?, ?, ?, f)
						)
					)
			end
			Result := True
		end

feature -- Properties (Factory)

	pair_ok (tr: STS_TRANSFORMER [A, B, EQ_A, EQ_B]; a: A; b: B): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_TRANSFORMER}.pair?
		do
			check
				u_a: attached current_universe_a as u_a
				u_b: attached current_universe_b as u_b
				definition: tr.pair (a, b).as_singleton ≍
					((u_a × u_b).filtered_xy (agent (u_a.transformer).anded (agent u_a.equality_holds (?, a), ?, agent u_b.equality_holds (?, b), ?)))
			end
			Result := True
		end

feature -- Transformer

	element_to_set: STS_TRANSFORMER [A, STS_SET [B, EQ_B], EQ_A, STS_SET_EQUALITY [B, EQ_B]]
			-- Transformer between types derived from {A} and {STS_SET [B, EQ_B]}
		deferred
		end

	term_to_tuple: STS_TRANSFORMER [A, STS_N_TUPLE [B, EQ_B], EQ_A, STS_N_TUPLE_EQUALITY [B, EQ_B]]
			-- Transformer between types derived from {A} and {STS_N_TUPLE [B, EQ_B]}
		deferred
		end

note
	copyright: "Copyright (c) 2012-2021, Rosivaldo Fernandes Alves"
	license: "[
		Eiffel Forum License v2
		(see http://www.eiffel.com/licensing/forum.txt)
		]"
	source: "https://launchpad.net/set-theory"

end
