﻿note
	description: "Object that checks whether the properties verified within set theory hold for an implementation of {STS_N_TUPLE}"
	author: "Rosivaldo Fernandes Alves"
	date: "$Date: Sat 2018-12-01 19:25:58 -0300 $"
	revision: "$Revision$"

deferred class
	N_TUPLE_PROPERTIES [A, EQ -> STS_EQUALITY [A]]

inherit
	ELEMENT_PROPERTIES

feature -- Access

	o: like universe_anchor.subset_anchor
			-- The empty set
		deferred
		ensure
			is_empty: Result.is_empty
		end

	eq: EQ
			-- Equality for objects like {A}
		deferred
		end

	unit: like universe_anchor.n_tuple_anchor
			-- The unit n-tuple
		deferred
		ensure
			is_unit: Result.is_unit
		end

	current_universe: like universe_anchor
			-- The current "Universe", i.e. set with every object currently in system memory whose type conforms to {A}.
			-- Notice that this "universe" may change from a call to another.
		deferred
		end

	current_universe_n: like unit.indices
			-- The current "Universe" of natural numbers, i.e. set with every natural number currently in system memory.
			-- Notice that this "universe" may change from a call to another.
		deferred
		end

feature -- Properties (Primitive)

	is_unit_ok (t: STS_N_TUPLE [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_N_TUPLE}.is_unit?
		do
			check
				u: attached current_universe as u
				definition: t.is_unit = (t.n = 0)
				no_index: t.is_unit = t.indices.is_empty
				no_term: t.is_unit = t.terms.is_empty
				no_possible_term: t.is_unit = (u |∀ agent negated (agent t.has, ?))
			end
			Result := True
		end

	first_ok (t: STS_N_TUPLE [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_N_TUPLE}.first?
		do
			if not t.is_unit then
				check
					u: attached current_universe as u
						positive: 0 < 1 -- By definition
						small_enough: 1 ≤ t.n -- t.n > 0
					definition: eq (t.first, t [1])
					by_existence: singleton (t.first) ≍ (
						u | agent (x: A; ia_t: STS_N_TUPLE [A, EQ]): BOOLEAN
							do
								Result := ia_t.exists (
									agent (i: NATURAL; x_i, i2l_x: A): BOOLEAN
										do
											Result := i = 1 and eq (x_i, i2l_x)
										end (?, ?, x)
									)
							end (?, t)
						)
					building_up: t ≍ t.tail.after (t.first)
				end
			end
			Result := True
		end

	tail_ok (t: STS_N_TUPLE [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_N_TUPLE}.tail?
		do
			if t.is_unit then
				check
					unit: t.tail ≍ t
					no_term: t.tail.n = 0
				end
			else
				check
					building_up: t ≍ t.tail.after (t.first)
					one_less_term: t.tail.n = t.n - 1
				end
			end
			check
				definition: t.tail ≍ t.suffix (t.front.n)
				by_trimming: t.tail ≍ t.left_trimmed (1)
				front: t.n ≤ 1 implies t.tail ≍ t.front
			end
			Result := True
		end

feature -- Properties (Membership)

	has_ok (t: STS_N_TUPLE [A, EQ]; a: A): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_N_TUPLE}.has?
		do
			check
				definition: t.has (a) = t.exists (
					agent (i: NATURAL; x_i, ia_a: A): BOOLEAN
						do
							Result := eq (x_i, ia_a)
						end (?, ?, a)
					)
				possibly_repeated: t.has (a) = (
					0 < # (
						t.indices | agent (i: NATURAL; ia_a: A; ia_t: STS_N_TUPLE [A, EQ]): BOOLEAN
							do
								Result := eq (ia_t [i], ia_a)
							end (?, a, t)
						)
					)
				has_own_terms: t.for_all (
					agent (i: NATURAL; x_i: A; ia_t: STS_N_TUPLE [A, EQ]): BOOLEAN
						do
							Result := ia_t.has (x_i)
						end (?, ?, t)
					)
			end
			Result := True
		end

	does_not_have_ok (t: STS_N_TUPLE [A, EQ]; a: A): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_N_TUPLE}.does_not_have?
		do
			check
				definition: t.does_not_have (a) = t.does_not_exist (
					agent (i: NATURAL; x_i, ia_a: A): BOOLEAN
						do
							Result := eq (x_i, ia_a)
						end (?, ?, a)
					)
				zero_occurrences: t.does_not_have (a) = (
					0 = # (
						t.indices | agent (i: NATURAL; ia_a: A; ia_t: STS_N_TUPLE [A, EQ]): BOOLEAN
							do
								Result := eq (ia_t [i], ia_a)
							end (?, a, t)
						)
					)
				has_own_terms: t.does_not_exist (
					agent (i: NATURAL; x_i: A; ia_t: STS_N_TUPLE [A, EQ]): BOOLEAN
						do
							Result := ia_t.does_not_have (x_i)
						end (?, ?, t)
					)
			end
			Result := True
		end

feature -- Properties (Access)

	term_ok (t: STS_N_TUPLE [A, EQ]; i: NATURAL): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_N_TUPLE}.term?
		do
			if 0 < i and i ≤ t.n then
				check
					u: attached current_universe as u
					u_card: (
						agent (ia_u: STS_SET [A, EQ]): BOOLEAN
							do
								print (# ia_u) print ('%N')
								Result := True
							end (u)
						).item

					definition: singleton (t [i]) ≍ (
						u | agent (x: A; ia_i: NATURAL; ia_t: STS_N_TUPLE [A, EQ]): BOOLEAN
							do
								Result := ia_t.exists (
									agent (j: NATURAL; x_i: A; i2a_i: NATURAL; i2a_x: A): BOOLEAN
										do
											Result := j = i2a_i and eq (x_i, i2a_x)
										end (?, ?, ia_i, x)
									)
							end (?, i, t)
						)
					is_not_unit: not t.is_unit -- n ⊇ i > 0
					by_right_reduction: eq (t [i], t.front.indexed_right_reduced (1, agent indexed_term (i, ?, ?, ?), t.last))
				end
			end
			Result := True
		end

	terms_ok (t: STS_N_TUPLE [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_N_TUPLE}.terms?
		do
			check
				u: attached current_universe as u
				definition: t.terms ≍ (
					u | agent (x: A; ia_t: STS_N_TUPLE [A, EQ]): BOOLEAN
						do
							Result := ia_t.exists (
								agent (i: NATURAL; x_i: A; i2a_x: A): BOOLEAN
									do
										Result := eq (x_i, i2a_x)
									end (?, ?, x)
								)
						end (?, t)
					)
				by_left_reduction: t.terms ≍ term_to_set.tuple_left_reduction (t, o, agent {STS_SET [A, EQ]}.with)
				cardinality: # t.terms ≤ t.n
			end
			Result := True
		end

	front_ok (t: STS_N_TUPLE [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_N_TUPLE}.front?
		do
			if t.is_unit then
				check
					unit: t.front ≍ t
					no_term: t.front.n = 0
				then
				end
			else
				check
					building_up: t ≍ t.front.before (t.last)
					one_less_term: t.front.n = t.n - 1
				then
				end
			end
			check
				definition: t.front ≍ t.right_trimmed (1)
				tail: t.n ≤ 1 implies t.front ≍ t.tail
			end
			Result := True
		end

	last_ok (t: STS_N_TUPLE [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_N_TUPLE}.last?
		do
			if not t.is_unit then
				check
					u: attached current_universe as u
					u_card: (
						agent (ia_u: STS_SET [A, EQ]): BOOLEAN
							do
								print (# ia_u) print ('%N')
								Result := True
							end (u)
						).item

						positive: 0 < t.n -- not t.is_unit
						small_enough: t.n ≤ t.n -- By definition
					definition: singleton (t.last) ≍ (
						u | agent (x: A; ia_t: STS_N_TUPLE [A, EQ]): BOOLEAN
							do
								Result := ia_t.exists (
									agent (i: NATURAL; x_i, i2a_x: A; n: NATURAL): BOOLEAN
										do
											Result := i = n and eq (x_i, i2a_x)
										end (?, ?, x, ia_t.n)
									)
							end (?, t)
						)
					building_up: t ≍ t.front.before (t.last)
				end
			end
			Result := True
		end

feature -- Properties (Construction)

	before_ok (t: STS_N_TUPLE [A, EQ]; a: A): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_N_TUPLE}.before?
		do
			check
				one_more_term: t.before (a).n = t.n + 1
			end
			Result := True
		end

	after_ok (a: A; t: STS_N_TUPLE [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_N_TUPLE}.after?
		do
			check
				one_more_term: t.after (a).n = t.n + 1
			end
			Result := True
		end

feature -- Properties (Measurement)

	n_ok (t: STS_N_TUPLE [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_N_TUPLE}.n?
		do
			check
				n_indices: t.n = # t.indices
				at_most_n_terms: # t.terms ≤ t.n
			end
			Result := True
		end

feature -- Properties (Comparison)

	equals_ok (t1, t2: STS_N_TUPLE [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_N_TUPLE}.equals?
		do
			check
				u: attached current_universe as u
				i_s: attached if t1.n ≥ t2.n then t1.indices else t2.indices end as i_s
				definition: (t1 ≍ t2) = (i_s × u).for_all_xy (
					agent (i: NATURAL; x_i: A ia_t1, ia_t2: STS_N_TUPLE [A, EQ]): BOOLEAN
						do
							Result := ((0 < i and i ≤ ia_t1.n) and then eq (ia_t1 [i], x_i)) = ((0 < i and i ≤ ia_t2.n) and then eq (ia_t2 [i], x_i))
						end (?, ?, t1, t2)
					)
				lesser_definition: (t1 ≍ t2) = (
					(t1.indices ∪ t2.indices) |∀ agent (i: NATURAL; ia_t1, ia_t2: STS_N_TUPLE [A, EQ]): BOOLEAN
						do
							Result := (0 < i and i ≤ ia_t1.n and i ≤ ia_t2.n) and then eq (ia_t1 [i], ia_t2 [i])
						end (?, t1, t2)
					)
				by_membership: (2 ^ (# u ^ # i_s)) ≤ max_subsets implies t1 ≍ t2 = (
					(u ^ # i_s).powerset |∀ agent (s: STS_SET [STS_N_TUPLE [A, EQ], STS_N_TUPLE_EQUALITY [A, EQ]]; ia_t1, ia_t2: STS_N_TUPLE [A, EQ]): BOOLEAN
						do
							Result := ia_t1 ∈ s = ia_t2 ∈ s
						end (?, t1, t2)
					)
				same_n: t1 ≍ t2 implies t1.n = t2.n
			end
			Result := True
		end

	unequals_ok (t1, t2: STS_N_TUPLE [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_N_TUPLE}.unequals?
		do
			check
				u: attached current_universe as u
				i_s: attached if t1.n ≥ t2.n then t1.indices else t2.indices end as i_s
				definition: (t1 ≭ t2) = (
					(i_s × u).exist_xy (
						agent (i: NATURAL; x_i: A ia_t1, ia_t2: STS_N_TUPLE [A, EQ]): BOOLEAN
							do
								Result := ((0 < i and i ≤ ia_t1.n) and then eq (ia_t1 [i], x_i)) and ((0 >= i or i > ia_t2.n) or else not eq (ia_t2 [i], x_i))
							end (?, ?, t1, t2)
						) or
					(i_s × u).exist_xy (
						agent (i: NATURAL; x_i: A ia_t1, ia_t2: STS_N_TUPLE [A, EQ]): BOOLEAN
							do
								Result := ((0 >= i or i > ia_t1.n) or else not eq (ia_t1 [i], x_i)) and ((0 < i and i ≤ ia_t2.n) and then eq (ia_t2 [i], x_i))
							end (?, ?, t1, t2)
						)
					)
				lesser_definition: (t1 ≭ t2) = (
					(t1.indices ∪ t2.indices) |∃ agent (i: NATURAL; ia_t1, ia_t2: STS_N_TUPLE [A, EQ]): BOOLEAN
						do
							Result := (0 >= i or i > ia_t1.n or i > ia_t2.n) or else not eq (ia_t1 [i], ia_t2 [i])
						end (?, t1, t2)
					)
				by_membership: (2 ^ (# u ^ # i_s)) ≤ max_subsets implies t1 ≭ t2 = (
					(u ^ # i_s).powerset |∃ agent (s: STS_SET [STS_N_TUPLE [A, EQ], STS_N_TUPLE_EQUALITY [A, EQ]]; ia_t1, ia_t2: STS_N_TUPLE [A, EQ]): BOOLEAN
						do
							Result := ia_t1 ∈ s /= ia_t2 ∈ s
						end (?, t1, t2)
					)
				unequal_n: t1.n /= t2.n implies t1 ≭ t2
			end
			Result := True
		end

feature -- Properties (Quantifier)

	exists_ok (t: STS_N_TUPLE [A, EQ] p: PREDICATE [NATURAL, A]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_N_TUPLE}.exists?
		do
			check
				definition: t.exists (p) = not t.for_all (agent t.negated (p, ?, ?))
			end
			Result := True
		end

	does_not_exist_ok (t: STS_N_TUPLE [A, EQ] p: PREDICATE [NATURAL, A]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_N_TUPLE}.does_not_exist?
		do
			check
				definition: t.does_not_exist (p) = t.for_all (agent t.negated (p, ?, ?))
			end
			Result := True
		end

	for_all_ok (p: PREDICATE [NATURAL, A]; t: STS_N_TUPLE [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_N_TUPLE}.for_all?
		do
			check
				definition: t.for_all (p) = t.does_not_exist (agent t.negated (p, ?, ?))
			end
			Result := True
		end

feature -- Properties (Operation)

	left_trimmed_ok (m: NATURAL; t: STS_N_TUPLE [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_N_TUPLE}.left_trimmed?
		do
			check
				definition: t.left_trimmed (m) ≍ t.reversed.right_trimmed (m).reversed
				by_prefix: t.left_trimmed (m) ≍ t.reversed.prefix (t.n - m.min (t.n)).reversed
				nothing_left: m >= t.n implies t.left_trimmed (m).n = 0
				something_left: m < t.n implies t.left_trimmed (m).n = t.n - m
				n: t.left_trimmed (m).n = t.n - m.min (t.n)
				terms: t.left_trimmed (m).for_all (
					agent (i: NATURAL; x: A; ia_m: NATURAL; ia_t: STS_N_TUPLE [A, EQ]): BOOLEAN
						do
								check
										-- i ∈ {1..(ia_t.n - ia_m)}
									positive: 0 < i + ia_m
									small_enough: (i + ia_m) ≤ ia_t.n
								end
							Result := eq (x, ia_t [i + ia_m])
						end (?, ?, m, t)
					)
			end
			Result := True
		end

	right_trimmed_ok (m: NATURAL; t: STS_N_TUPLE [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_N_TUPLE}.right_trimmed?
		do
			check
				definition: t.right_trimmed (m) ≍ t.reversed.left_trimmed (m).reversed
				by_suffix: t.right_trimmed (m) ≍ t.reversed.suffix (t.n - m.min (t.n)).reversed
				nothing_left: m >= t.n implies t.right_trimmed (m).n = 0
				something_left: m < t.n implies t.right_trimmed (m).n = t.n - m
				n: t.right_trimmed (m).n = t.n - m.min (t.n)
				terms: t.right_trimmed (m).for_all (agent t.terms_are_equal)
			end
			Result := True
		end

	prefix_ok (m: NATURAL; t: STS_N_TUPLE [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_N_TUPLE}.prefix?
		do
			check
				definition: t.prefix (m) ≍ t.right_trimmed (t.n - m.min (t.n))
				by_suffix: t.prefix (m) ≍ t.reversed.suffix (m).reversed
				nothing_lost: m >= t.n implies t.prefix (m).n = t.n
				something_lost: m < t.n implies t.prefix (m).n = m
			end
			Result := True
		end

	suffix_ok (m: NATURAL; t: STS_N_TUPLE [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_N_TUPLE}.suffix?
		do
			check
				definition: t.suffix (m) ≍ t.left_trimmed (t.n - m.min (t.n))
				nothing_lost: m >= t.n implies t.suffix (m).n = t.n
				something_lost: m < t.n implies t.suffix (m).n = m
				n: t.suffix (m).n = m.min (t.n)
				terms: t.suffix (m).for_all (
					agent (i: NATURAL; x: A ia_m: NATURAL; ia_t: STS_N_TUPLE [A, EQ]): BOOLEAN
						do
								check
										-- i ∈ {1..ia_m.min (ia_t.n)}
									positive: 0 < ia_t.n - ia_m.min (ia_t.n) + i
									small_enough: (ia_t.n - ia_m.min (ia_t.n) + i) ≤ ia_t.n
								end
							Result := eq (x, ia_t [ia_t.n - ia_m.min (ia_t.n) + i])
						end (?, ?, m, t)
					)
			end
			Result := True
		end

	reversed_ok (t: STS_N_TUPLE [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_N_TUPLE}.reversed?
		do
			check
				definition: t.reversed ≍ term_to_tuple.tuple_left_reduction (t, unit, agent {STS_N_TUPLE [A, EQ]}.after)
				same_terms: t.reversed.terms ≍ t.terms
				involutive: t.reversed.reversed ≍ t
			end
			Result := True
		end

	replacing_ok (t: STS_N_TUPLE [A, EQ]; i: NATURAL; a: A): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_N_TUPLE}.replacing?
		do
			if 0 < i and i ≤ t.n then
				check
					definition: t.replacing (i, a) ≍ (t.prefix (i - 1) + t.suffix (t.n - i).after (a))
					n: t.replacing (i, a).n = t.n
					quantified: t.replacing (i, a).for_all (
						agent (ia_i, j: NATURAL; x_j, ia_a: A; ia_t: STS_N_TUPLE [A, EQ]): BOOLEAN
							do
								if j = ia_i then
									Result := eq (x_j, ia_a)
								else
										check
												-- j ∈ {1..t.n}
											positive: 0 < j
											small_enough: j ≤ ia_t.n
										end
									Result := eq (x_j, ia_t [j])
								end
							end (i, ?, ?, a, t)
						)
					before_i: t.replacing (i, a).prefix (i - 1) ≍ t.prefix (i - 1)
						small_enough: i ≤ t.replacing (i, a).n -- t.replacing (i, a).n = n
					at_i: eq (t.replacing (i, a) [i], a)
					after_i: t.replacing (i, a).suffix (t.n - i) ≍ t.suffix (t.n - i)
				end
			end
			Result := True
		end

	concatenated_ok (t1, t2: STS_N_TUPLE [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_N_TUPLE}.concatenated?
		do
			check
				definition: (t1 + t2).for_all (
					agent (i: NATURAL; x_i: A; ia_t1, ia_t2: STS_N_TUPLE [A, EQ]): BOOLEAN
						do
							if i ≤ ia_t1.n then
									check
										positive_1: 0 < i -- i ∈ {1..t1.n + t2.n}
									end
								Result := eq (x_i, ia_t1 [i])
							else
									check
											-- t1.n < i ⊆ t1.n + t2.n
										positive_2: 0 < i - ia_t1.n
										small_enough_2: (i - ia_t1.n) ≤ ia_t2.n
									end
								Result := eq (x_i, ia_t2 [i - ia_t1.n])
							end
						end (?, ?, t1, t2)
					)
			end
			Result := True
		end

feature -- Properties (Transformation)

	mapped_ok (t: STS_N_TUPLE [A, EQ]; f: FUNCTION [A, A]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_N_TUPLE}.mapped?
		do
			check
				u: attached current_universe as u
				definition: (# u ^ t.n) ≤ max_tuples implies (t ↦ f).as_singleton ≍ (
					(u ^ t.n) | agent (ia_ty, ia_t: STS_N_TUPLE [A, EQ]; ia_f: FUNCTION [A, A]): BOOLEAN
						do
							Result := ia_t.for_all (
								agent (i: NATURAL; x_i: A; ty: STS_N_TUPLE [A, EQ]; i2a_f: FUNCTION [A, A]): BOOLEAN
									do
											check
													-- i ∈ {1..t.n} = {1..ty.n}
												positive: 0 < i
												small_enough: i ≤ ty.n
											end
										Result := eq (ty [i], i2a_f (x_i))
									end (?, ?, ia_ty, ia_f)
								)
						end (?, t, f)
					)
				n: (t ↦ f).n = t.n
				lesser_definition: (t ↦ f).for_all (
					agent (i: NATURAL; y_i: A; ia_t: STS_N_TUPLE [A, EQ]; ia_f: FUNCTION [A, A]): BOOLEAN
						do
								check
										-- i ∈ {1..t.n}
									positive: 0 < i
									small_enough: i ≤ ia_t.n
								end
							Result := eq (y_i, ia_f (ia_t [i]))
						end (?, ?, t, f)
					)
				by_left_reduction: (t ↦ f) ≍ term_to_tuple.tuple_left_reduction (
					t, unit, agent (acc: STS_N_TUPLE [A, EQ] x: A; ia_f: FUNCTION [A, A]): STS_N_TUPLE [A, EQ]
						do
							Result := acc.before (ia_f (x))
						end (?, ?, f)
					)
				by_right_reduction: (t ↦ f) ≍ term_to_tuple.tuple_right_reduction (
					t, agent (x: A; acc: STS_N_TUPLE [A, EQ]; ia_f: FUNCTION [A, A]): STS_N_TUPLE [A, EQ]
						do
							Result := acc.after (ia_f (x))
						end (?, ?, f), unit
					)
			end
			Result := True
		end

	left_reduced_ok (t: STS_N_TUPLE [A, EQ]; leftmost: A; f: FUNCTION [A, A, A]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_N_TUPLE}.left_reduced?
		do
			check
				reversed: eq (
					t.left_reduced (leftmost, f),
					t.reversed.right_reduced (
						agent  (x, acc: A; ia_f: FUNCTION [A, A, A]): A
							do
								Result := ia_f (acc, x)
							end (?, ?, f), leftmost
						)
					)
			end
			Result := True
		end

	indexed_left_reduced_ok (t: STS_N_TUPLE [A, EQ] leftmost: A i: NATURAL f: FUNCTION [A, NATURAL, A, A]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_N_TUPLE}.indexed_left_reduced?
		do
			check
				reversed: eq (
					t.indexed_left_reduced (leftmost, i, f),
					t.reversed.indexed_right_reduced (
						0, agent (m, j: NATURAL; x, acc: A; ia_f: FUNCTION [A, NATURAL, A, A]): A
							do
								Result := ia_f (acc, m - j - 1, x)
							end (t.n + i, ?, ?, ?, f), leftmost
						)
					)
			end
			Result := True
		end

	proper_left_reduced_ok (t: STS_N_TUPLE [A, EQ]; f: FUNCTION [A, A, A]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_N_TUPLE}.proper_left_reduced?
		do
			if not t.is_unit then
				check
					reversed_is_not_unit: not t.reversed.is_unit
					reversed: eq (
						t.proper_left_reduced (f),
						t.reversed.proper_right_reduced (
							agent (x, acc: A; ia_f: FUNCTION [A, A, A]): A
								do
									Result := ia_f (acc, x)
								end (?, ?, f)
							)
						)
				end
			end
			Result := True
		end

	right_reduced_ok (t: STS_N_TUPLE [A, EQ]; f: FUNCTION [A, A, A]; rightmost: A): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_N_TUPLE}.right_reduced?
		do
			check
				reversed: eq (
					t.right_reduced (f, rightmost),
					t.reversed.left_reduced (
						rightmost, agent (acc, x: A; ia_f: FUNCTION [A, A, A]): A
						do
							Result := ia_f (x, acc)
						end (?, ?, f)
						)
					)
			end
			Result := True
		end

	indexed_right_reduced_ok (t: STS_N_TUPLE [A, EQ] i: NATURAL f: FUNCTION [NATURAL, A, A, A] rightmost: A): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_N_TUPLE}.indexed_right_reduced?
		do
			check
				reversed: eq (
					t.indexed_right_reduced (i, f, rightmost),
					t.reversed.indexed_left_reduced (
						rightmost, 0, agent (m: NATURAL; acc: A; j: NATURAL; x: A; ia_f: FUNCTION [NATURAL, A, A, A]): A
							do
								Result := ia_f (m - j - 1, x, acc)
							end (t.n + i, ?, ?, ?, f)
						)
					)
			end
			Result := True
		end

	proper_right_reduced_ok (t: STS_N_TUPLE [A, EQ]; f: FUNCTION [A, A, A]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_N_TUPLE}.proper_right_reduced?
		do
			if not t.is_unit then
				check
					reversed_is_not_unit: not t.reversed.is_unit
					reversed: eq (
						t.proper_right_reduced (f),
						t.reversed.proper_left_reduced (
							agent (acc, x: A; ia_f: FUNCTION [A, A, A]): A
								do
									Result := ia_f (x, acc)
								end (?, ?, f)
							)
						)
				end
			end
			Result := True
		end

feature -- Properties (Conversion)

	as_singleton_ok (t: STS_N_TUPLE [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_N_TUPLE}.as_singleton?
		do
			check
				u: attached current_universe as u
				definition: (# u ^ t.n) ≤ max_tuples implies t.as_singleton ≍ ((u ^ t.n) | agent t.equals)
				by_singleton: t.as_singleton ≍ (u ^ 0).singleton (t)
			end
			Result := True
		end

	first_tail_ok (t: STS_N_TUPLE [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_N_TUPLE}.first_tail?
		do
			if not t.is_unit then
				check
					u: attached current_universe as u
					first: singleton (t.first_tail.a) ≍ (u | agent u.equality_holds (?, t.first))
					second: (# u ^ (t.n - 1)) ≤ max_tuples implies t.first_tail.b.as_singleton ≍ ((u ^ (t.n - 1)) | agent {STS_N_TUPLE [A, EQ]}.equals (t.tail))
					reconstruction: t ≍ t.first_tail.b.after (t.first_tail.a)
				end
			end
			Result := True
		end

	front_last_ok (t: STS_N_TUPLE [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_N_TUPLE}.front_last?
		do
			if not t.is_unit then
				check
					u: attached current_universe as u
					first: (# u ^ (t.n - 1)) ≤ max_tuples implies t.front_last.a.as_singleton ≍ ((u ^ (t.n - 1)) | agent {STS_N_TUPLE [A, EQ]}.equals (t.front))
					second: singleton (t.front_last.b) ≍ (u | agent u.equality_holds (?, t.last))
					reconstruction: t ≍ t.front_last.a.before (t.front_last.b)
				end
			end
			Result := True
		end

feature -- Properties (Factory)

	unit_ok (t: STS_N_TUPLE [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_N_TUPLE}.unit?
		do
			check
				u: attached current_universe as u
				definition: t.unit.as_singleton ≍ (u ^ 0)
				has_nothing: u |∀ agent negated (agent (t.unit).has, ?)
				uniqueness: t.unit ≍ unit
				no_term: t.unit.n = 0
				empty_terms: t.unit.terms ≍ o
			end
			Result := True
		end

feature -- Factory

	singleton (a: A): STS_SET [A, EQ]
			-- Singleton in the form {`a'}
		do
			Result := o.singleton (a)
		ensure
			definition: Result ≍ o.singleton (a)
		end

feature -- Predicate

	negated (p: PREDICATE [A]; x: A): BOOLEAN
			-- Logical negation of `p' (`x'), i.e. is `p' (`x') false?
		do
			Result := not p (x)
		ensure
			class
			definition: Result = not p (x)
		end

feature -- Reduction

	indexed_term (i, j: NATURAL; x_j, acc: A): A
			-- The value of `x_j', if `i' = `j'; otherwise, the value of `acc'.
		do
			if j = i then
				Result := x_j
			else
				Result := acc
			end
		ensure
			indexed_term: j = i implies eq (Result, x_j)
			other_term: j /= i implies eq (Result, acc)
		end

feature -- Transformer

	term_to_tuple: STS_TRANSFORMER [A, STS_N_TUPLE [A, EQ], EQ, STS_N_TUPLE_EQUALITY [A, EQ]]
			-- Transformer between types derived from {A} and {STS_N_TUPLE [A, EQ]}
		deferred
		end

	term_to_set: STS_TRANSFORMER [A, STS_SET [A, EQ], EQ, STS_SET_EQUALITY [A, EQ]]
			-- Transformer between types derived from {A} and {STS_SET [A, EQ]}
		deferred
		end

feature {NONE} -- Anchor

	universe_anchor: STS_SET [A, EQ]
			-- Anchor for `current_universe'
		deferred
		end

feature {NONE} -- Implementation

	max_subsets: NATURAL
			-- Maximum number of subsets to be checked within property statements
			--| todo: DRY: {SET_PROPERTIES}
		deferred
		end

	max_tuples: NATURAL
			-- Maximum number of n-tuples to be checked within property statements
			--| todo: DRY: {SET_PROPERTIES}
		deferred
		end

note
	copyright: "Copyright (c) 2012-2021, Rosivaldo Fernandes Alves"
	license: "[
		Eiffel Forum License v2
		(see http://www.eiffel.com/licensing/forum.txt)
		]"
	source: "https://launchpad.net/set-theory"

end
