﻿note
	description: "Object that checks whether the properties verified within set theory hold for an implementation of {STS_SET_TUPLE}"
	author: "Rosivaldo Fernandes Alves"
	date: "$Date: Sat 2018-12-01 19:25:58 -0300 $"
	revision: "$Revision$"

deferred class
	SET_TUPLE_PROPERTIES [A, EQ -> STS_EQUALITY [A]]

inherit
	N_TUPLE_PROPERTIES [STS_SET [A, EQ], STS_SET_EQUALITY [A, EQ]]
		rename
			o as o_s,
			eq as eq_s,
			current_universe as current_universe_s,
			unit as unit_s,
			unit_ok as unit_s_ok,
			singleton as set_singleton,
			universe_anchor as universe_s_anchor
		end

feature -- Access

	o_s: like universe_s_anchor.subfamily_anchor
			-- The empty set of sets of elements like {A}
		deferred
		end

	u: like universe_s_anchor.family_union
			-- The "universe", i.e. set with every known element in a given theory.
			-- todo: redefine?
		do
			Result := ⋃ current_universe_s
		ensure
			definition: Result ≍ ⋃ old current_universe_s
		end

	unit: like universe_s_anchor.family_union.n_tuple_anchor
			-- N-tuple with no term, i.e. ().
		deferred
		ensure
			definition: Result.is_unit
		end

feature -- Properties (Quantifier)

	for_all_tuples_ok (st: STS_SET_TUPLE [A, EQ] p: PREDICATE [STS_N_TUPLE [A, EQ]]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET_TUPLE}.for_all_tuples?
		do
			check
				base: st.is_unit implies st.for_all_tuples (p) = p (unit)
				induction: not st.is_unit implies st.for_all_tuples (p) = (
					st.first |∀ agent (ia_p: PREDICATE [STS_N_TUPLE [A, EQ]]; ia_x: A; ia_st: STS_SET_TUPLE [A, EQ]): BOOLEAN
						do
							Result := ia_st.tail.for_all_tuples (
								agent  (i2a_p: PREDICATE [STS_N_TUPLE [A, EQ]]; x: A; t: STS_N_TUPLE [A, EQ]): BOOLEAN
									do
										Result := i2a_p (t.after (x))
									end (ia_p, ia_x, ?)
								)
						end (p, ?, st)
						)
				by_existence: st.for_all_tuples (p) = (∏ st) |∄ agent negated_pt (p, ?)
			end
			Result := True
		end

feature -- Properties (Operation)

	n_ary_union_ok (st: STS_SET_TUPLE [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET_TUPLE}.n_ary_union?
		do
			check
				definition: ⋃ st ≍ (
					u | agent (ia_x: A; ia_st: STS_SET_TUPLE [A, EQ]): BOOLEAN
						do
							Result := ia_st.exists (
								agent (i: NATURAL; s_i: STS_SET [A, EQ]; x: A): BOOLEAN
									do
										Result := s_i ∋ x
									end (?, ?, ia_x)
								)
						end (?, st)
						)
			end
			Result := True
		end

	n_ary_intersection_ok (st: STS_SET_TUPLE [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET_TUPLE}.n_ary_intersection?
		do
			check
				definition: ⋂ st ≍ (
					⋃ st | agent (ia_x: A; ia_st: STS_SET_TUPLE [A, EQ]): BOOLEAN
						do
							Result := ia_st.for_all (
								agent (i: NATURAL; s: STS_SET [A, EQ]; x: A): BOOLEAN
									do
										Result := s ∋ x
									end (?, ?, ia_x)
								)
						end (?, st)
					)
			end
			Result := True
		end

	n_ary_product_ok (st: STS_SET_TUPLE [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET_TUPLE}.n_ary_product?
		do
			check
				definition: # u ^ st.n <= max_tuples implies (∏ st) ≍ (
					(u ^ st.n) | agent (ia_t: STS_N_TUPLE [A, EQ]; ia_st: STS_SET_TUPLE [A, EQ]): BOOLEAN
						do
							Result := ia_t.n = ia_st.n and ia_st.for_all (
								agent (i: NATURAL; s_i: STS_SET [A, EQ]; t: STS_N_TUPLE [A, EQ]): BOOLEAN
									do
										if 0 < i and i <= t.n then
											Result := s_i ∋ t [i]
										end
									end (?, ?, ia_t)
								)
						end (?, st)
					)
				unit_yields_unit: st.is_unit implies unit ∈ ∏ st
				every_tuple: not st.is_unit implies (
					st.first |∀ agent (x: A; st_tail_product, st_product: STS_SET [STS_N_TUPLE [A, EQ], STS_N_TUPLE_EQUALITY [A, EQ]]): BOOLEAN
						do
							Result := (st_tail_product ↦ agent {STS_N_TUPLE [A, EQ]}.after (x)) ⊆
								st_product
						end (?, ∏ st.tail, ∏ st)
					)
				all_with_n_terms: (∏ st) |∀ agent (t: STS_N_TUPLE [A, EQ]; st_n: NATURAL): BOOLEAN
					do
						Result := t.n = st_n
					end (?, st.n)
				nothing_else: (∏ st) |∀ agent (t: STS_N_TUPLE [A, EQ]; ia_st: STS_SET_TUPLE [A, EQ]): BOOLEAN
					do
						Result := t.for_all (
							agent (i2a_st: STS_SET_TUPLE [A, EQ]; i: NATURAL; x_i: A): BOOLEAN
								do
										check
											positive: 0 < i -- i is in {1..t.n}
											small_enough: i <= i2a_st.n -- t.n = i2a_st.n <==`all_with_n_terms'
										end
									Result := i2a_st [i] ∋ x_i
								end (ia_st, ?, ?)
							)
					end (?, st)
				cardinality: # (∏ st) = set_to_natural.tuple_left_reduction (
					st, 1, agent (acc: NATURAL; s: STS_SET [A, EQ]): NATURAL
						do
							Result := acc * (# s)
						end
					)
			end
			Result := True
		end

feature -- Predicate

	negated_pt (p: PREDICATE [STS_N_TUPLE [A, EQ]]; t: STS_N_TUPLE [A, EQ]): BOOLEAN
			-- Logical negation of `p' (`t'), i.e. is `p' (`t') false?
		do
			Result := not p (t)
		ensure
			class
			definition: Result = not p (t)
		end

feature -- Transformer

	set_to_natural: STS_TRANSFORMER [STS_SET [A, EQ], NATURAL, STS_SET_EQUALITY [A, EQ], STS_OBJECT_EQUALITY [NATURAL]]
			-- Transformer between types derived from {STS_SET [A, EQ]} and {NATURAL}
		deferred
		end

feature {NONE} -- Anchor

	universe_s_anchor: STS_SET_FAMILY [A, EQ]
			-- Anchor for `current_universe_s'
		deferred
		end

note
	copyright: "Copyright (c) 2012-2021, Rosivaldo Fernandes Alves"
	license: "[
		Eiffel Forum License v2
		(see http://www.eiffel.com/licensing/forum.txt)
		]"
	source: "https://launchpad.net/set-theory"

end
