﻿note
	description: "Object that checks whether the properties verified within set theory hold for an implementation of {STS_TRANSFORMABLE_TUPLE}"
	author: "Rosivaldo Fernandes Alves"
	date: "$Date: Sat 2018-12-01 19:25:58 -0300 $"
	revision: "$Revision$"

deferred class
	TRANSFORMABLE_TUPLE_PROPERTIES [A, B, EQ_A -> STS_EQUALITY [A], EQ_B -> STS_EQUALITY [B]]

inherit
	N_TUPLE_PROPERTIES [A, EQ_A]
		rename
			o as o_a,
			unit as unit_a,
			eq as eq_a,
			current_universe as current_universe_a,
			mapped_ok as mapped_to_same_kind_ok,
			left_reduced_ok as left_reduced_to_term_like_ok,
			indexed_left_reduced_ok as indexed_left_reduced_to_term_like_ok,
			right_reduced_ok as right_reduced_to_term_like_ok,
			indexed_right_reduced_ok as indexed_right_reduced_to_term_like_ok,
			term_to_tuple as term_to_tuple_of_same_kind,
			universe_anchor as universe_a_anchor
		end

feature -- Access

	o_a: like current_universe_a.subset_anchor
			-- The empty set of elements like {A}
		deferred
		end

	unit_b: like current_universe_b.n_tuple_anchor
			-- The empty n-tuple of terms like {B}
		deferred
		ensure
			is_unit: Result.is_unit
		end

	eq_b: EQ_B
			-- Equality for objects like {B}
		deferred
		end

	current_universe_b: STS_SET [B, EQ_B]
			-- The current "universe" of elements like {B}, i.e. set with every object currently in system memory whose type conforms to {B}.
			-- Notice that this "universe" may change from a call to another.
		deferred
		end

feature -- Properties (Transformation)

	mapped_ok (t: STS_TRANSFORMABLE_TUPLE [A, B, EQ_A, EQ_B] f: FUNCTION [A, B]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET_TUPLE}.mapped?
		do
			check
				u_b: attached current_universe_b as u_b
				definition: (# u_b) ^ t.n <= max_tuples implies (t ↦ f).as_singleton ≍ (
					(u_b ^ t.n) | agent (ia_tb: STS_N_TUPLE [B, EQ_B]; ia_t: STS_TRANSFORMABLE_TUPLE [A, B, EQ_A, EQ_B]; ia_f: FUNCTION [A, B]): BOOLEAN
						do
							Result := ia_t.for_all (
								agent (i: NATURAL; x_i: A; tb: STS_N_TUPLE [B, EQ_B]; i2a_f: FUNCTION [A, B]): BOOLEAN
									do
											check
													-- i ∈ {1..t.n} = {1..tb.n}
												positive: 0 < i
												small_enough: i <= tb.n
											end
										Result := eq_b (tb [i], i2a_f (x_i))
									end (?, ?, ia_tb, ia_f)
								)
						end (?, t, f)
					)
				n: (t ↦ f).n = t.n
				lesser_definition: (t ↦ f).for_all (
					agent (i: NATURAL; y_i: B; ia_t: STS_TRANSFORMABLE_TUPLE [A, B, EQ_A, EQ_B]; ia_f: FUNCTION [A, B]): BOOLEAN
						do
								check
										-- i ∈ {1..t.n}
									positive: 0 < i
									small_enough: i <= ia_t.n
								end
							Result := eq_b (y_i, ia_f (ia_t [i]))
						end (?, ?, t, f)
					)
				by_left_reduction: (t ↦ f) ≍ term_to_tuple.tuple_left_reduction (
					t, unit_b, agent (acc: STS_N_TUPLE [B, EQ_B]; x: A; ia_f: FUNCTION [A, B]): STS_N_TUPLE [B, EQ_B]
						do
							Result := acc.before (ia_f (x))
						end (?, ?, f)
					)
				by_right_reduction: (t ↦ f) ≍ term_to_tuple.tuple_right_reduction (
					t, agent (x: A; acc: STS_N_TUPLE [B, EQ_B]; ia_f: FUNCTION [A, B]): STS_N_TUPLE [B, EQ_B]
						do
							Result := acc.after (ia_f (x))
						end (?, ?, f), unit_b
					)
			end
			Result := True
		end

	left_reduced_ok (t: STS_TRANSFORMABLE_TUPLE [A, B, EQ_A, EQ_B]; leftmost: B; f: FUNCTION [B, A, B]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET_TUPLE}.left_reduced?
		do
			check
				base: t.is_unit implies eq_b (t.left_reduced (leftmost, f), leftmost)
				induction: not t.is_unit implies eq_b (t.left_reduced (leftmost, f), f (t.front.left_reduced (leftmost, f), t.last))
				reversed: eq_b (
					t.left_reduced (leftmost, f),
					t.reversed.right_reduced (
						agent (x: A; acc: B; ia_f: FUNCTION [B, A, B]): B
							do
								Result := ia_f (acc, x)
							end (?, ?, f), leftmost
						)
					)
			end
			Result := True
		end

	indexed_left_reduced_ok (t: STS_TRANSFORMABLE_TUPLE [A, B, EQ_A, EQ_B]; leftmost: B; i: NATURAL; f: FUNCTION [B, NATURAL, A, B]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET_TUPLE}.indexed_left_reduced?
		do
			check
				base: t.is_unit implies eq_b (t.indexed_left_reduced (leftmost, i, f), leftmost)
				induction: not t.is_unit implies eq_b (t.indexed_left_reduced (leftmost, i, f), f (t.front.indexed_left_reduced (leftmost, i, f), t.n + i - 1, t.last))
				reversed: eq_b (
					t.indexed_left_reduced (leftmost, i, f),
					t.reversed.indexed_right_reduced (
						0, agent (m, j: NATURAL; x: A; acc: B; ia_f: FUNCTION [B, NATURAL, A, B]): B
							do
								Result := ia_f (acc, m - j - 1, x)
							end (t.n + i, ?, ?, ?, f), leftmost
						)
					)
			end
			Result := True
		end

	right_reduced_ok (t: STS_TRANSFORMABLE_TUPLE [A, B, EQ_A, EQ_B]; f: FUNCTION [A, B, B]; rightmost: B): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET_TUPLE}.right_reduced?
		do
			check
				base: t.is_unit implies eq_b (t.right_reduced (f, rightmost), rightmost)
				induction: not t.is_unit implies eq_b (t.right_reduced (f, rightmost), t.front.right_reduced (f, f (t.last, rightmost)))
				reversed: eq_b (
					t.right_reduced (f, rightmost),
					t.reversed.left_reduced (
						rightmost, agent (acc: B; x: A; ia_f: FUNCTION [A, B, B]): B
							do
								Result := ia_f (x, acc)
							end (?, ?, f)
						)
					)
			end
			Result := True
		end

	indexed_right_reduced_ok (t: STS_TRANSFORMABLE_TUPLE [A, B, EQ_A, EQ_B]; i: NATURAL; f: FUNCTION [NATURAL, A, B, B]; rightmost: B): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET_TUPLE}.indexed_right_reduced?
		do
			check
				base: t.is_unit implies eq_b (t.indexed_right_reduced (i, f, rightmost), rightmost)
				induction: not t.is_unit implies eq_b (t.indexed_right_reduced (i, f, rightmost), t.front.indexed_right_reduced (i, f, f (i + t.n - 1, t.last, rightmost)))
				reversed: eq_b (
					t.indexed_right_reduced (i, f, rightmost),
					t.reversed.indexed_left_reduced (
						rightmost, 0, agent (m: NATURAL; acc: B; j: NATURAL; x: A; ia_f: FUNCTION [NATURAL, A, B, B]): B
							do
								Result := ia_f (m - j - 1, x, acc)
							end (t.n + i, ?, ?, ?, f)
						)
					)
			end
			Result := True
		end

feature -- Transformer

	term_to_tuple: STS_TRANSFORMER [A, STS_N_TUPLE [B, EQ_B], EQ_A, STS_N_TUPLE_EQUALITY [B, EQ_B]]
			-- Transformer between types derived from {A} and {STS_N_TUPLE [B, EQ_B]}
		deferred
		end

note
	copyright: "Copyright (c) 2012-2021, Rosivaldo Fernandes Alves"
	license: "[
		Eiffel Forum License v2
		(see http://www.eiffel.com/licensing/forum.txt)
		]"
	source: "https://launchpad.net/set-theory"

end
