﻿note
	description: "[
		Object that checks whether the properties verified within set theory hold for an implementation of {STS_ORDERED_PAIR}
		]"
	author: "Rosivaldo Fernandes Alves"
	date: "$Date: Wed 2018-11-21 07:53:55 -0300 $"
	revision: "$Revision$"

deferred class
	ORDERED_PAIR_PROPERTIES [A, B, EQ_A -> STS_EQUALITY [A], EQ_B -> STS_EQUALITY [B]]

inherit
	ELEMENT_PROPERTIES
		rename
			is_in_ok as element_is_in_ok,
			is_not_in_ok as element_is_not_in_ok
		end

feature -- Access

	eq_a: EQ_A
			-- Equality for objects like {A}
		deferred
		end

	eq_b: EQ_B
			-- Equality for objects like {B}
		deferred
		end

	current_universe_a: STS_TRANSFORMABLE_SET [A, B, EQ_A, EQ_B]
			-- The current "Universe", i.e. set with every object currently in system memory whose type conforms to {A}.
			-- Notice that this "universe" may change from a call to another.
		deferred
		end

	current_universe_b: STS_TRANSFORMABLE_SET [B, A, EQ_B, EQ_A]
			-- The current "universe" of elements like {B}, i.e. set with every object currently in system memory whose type conforms to {B}.
			-- Notice that this "universe" may change from a call to another.
		deferred
		end

feature -- Properties (Membership)

	is_in_ok
		(ab: STS_ORDERED_PAIR [A, B, EQ_A, EQ_B]; s: STS_SET [STS_ORDERED_PAIR [A, B, EQ_A, EQ_B], STS_EQUALITY [STS_ORDERED_PAIR [A, B, EQ_A, EQ_B]]]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_ORDERED_PAIR}.is_in?
		do
			check
				necessary_non_emptiness: ab ∈ s implies not s.is_empty
			end
			Result := True
		end

	is_not_in_ok
		(ab: STS_ORDERED_PAIR [A, B, EQ_A, EQ_B]; s: STS_SET [STS_ORDERED_PAIR [A, B, EQ_A, EQ_B], STS_EQUALITY [STS_ORDERED_PAIR [A, B, EQ_A, EQ_B]]]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_ORDERED_PAIR}.is_not_in?
		do
			check
				definition: ab ∉ s = s ∌ ab
				sufficient_emptiness: s.is_empty implies ab ∉ s
			end
			Result := True
		end

feature -- Properties (Comparison)

	equals_ok (a1b1, a2b2: STS_ORDERED_PAIR [A, B, EQ_A, EQ_B]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_ORDERED_PAIR}.equals?
		do
			check
				u_a: attached current_universe_a as u_a
				u_b: attached current_universe_b as u_b
				definition: a1b1 ≍ a2b2 = (u_a × u_b).for_all_xy (
					agent (x: A y: B; a1: A b1: B; a2: A b2: B): BOOLEAN
						do
							Result := eq_a (x, a1) = eq_a (x, a2) and eq_b (y, b1) = eq_b (y, b2)
						end (?, ?, a1b1.a, a1b1.b, a2b2.a, a2b2.b)
					)
				by_membership: (2 ^ (# u_a * # u_b)) ≤ max_supersets implies a1b1 ≍ a2b2 = (
					(u_a × u_b).powerset |∀ agent (
						s: STS_SET [STS_ORDERED_PAIR [A, B, EQ_A, EQ_B], STS_ORDERED_PAIR_EQUALITY [A, B, EQ_A, EQ_B]]
						ia_a1b1, ia_a2b2: STS_ORDERED_PAIR [A, B, EQ_A, EQ_B]
						): BOOLEAN
						do
							Result := ia_a1b1 ∈ s = ia_a2b2 ∈ s
						end (?, a1b1, a2b2)
					)
				reflexive: a1b1 ≍ a1b1
			end
			Result := True
		end

	unequals_ok (a1b1, a2b2: STS_ORDERED_PAIR [A, B, EQ_A, EQ_B]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_ORDERED_PAIR}.unequals?
		do
			check
				u_a: attached current_universe_a as u_a
				u_b: attached current_universe_b as u_b
				definition: a1b1 ≭ a2b2 = (u_a × u_b).exist_xy (
					agent (x: A y: B; a1: A b1: B; a2: A b2: B): BOOLEAN
						do
							Result :=
								not (eq_a (x, a1) = eq_a (x, a2)) or not (eq_b (y, b1) = eq_b (y, b2))
						end (?, ?, a1b1.a, a1b1.b, a2b2.a, a2b2.b)
					)
				lesser_definition: a1b1 ≭ a2b2 = (not eq_a (a1b1.a, a2b2.a) or not eq_b (a1b1.b, a2b2.b))
				by_membership: 2 ^ (# u_a * # u_b) <= max_supersets implies a1b1 ≭ a2b2 = (
					(u_a × u_b).powerset |∃ agent (
						s: STS_SET [STS_ORDERED_PAIR [A, B, EQ_A, EQ_B], STS_ORDERED_PAIR_EQUALITY [A, B, EQ_A, EQ_B]]
						ia_a1b1, ia_a2b2: STS_ORDERED_PAIR [A, B, EQ_A, EQ_B]
						): BOOLEAN
						do
							Result := ia_a1b1 ∈ s /= ia_a2b2 ∈ s
						end (?, a1b1, a2b2))
				irreflexive: not (a1b1 ≭ a1b1)
			end
			Result := True
		end

	succeeds_ok
		(a2b2: STS_ORDERED_PAIR [A, B, EQ_A, EQ_B]; w1a1: STS_ORDERED_PAIR [detachable separate ANY, A, STS_EQUALITY [detachable separate ANY], EQ_A]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_ORDERED_PAIR}.succeeds?
		do
			check
				u_a: attached current_universe_a as u_a
				definition: a2b2 ≽ w1a1 = (
					u_a |∀ agent  (x, a2, a1: A): BOOLEAN
						do
							Result := eq_a (x, a2) = eq_a (x, a1)
						end (?, a2b2.a, w1a1.second)
					)
			end
			Result := True
		end

feature -- Properties (Operation)

	inverse_ok (ab: STS_ORDERED_PAIR [A, B, EQ_A, EQ_B]): BOOLEAN
			-- Do the properties verified within set theory hold for
			-- {STS_ORDERED_PAIR}.inverse?
		do
			check
				u_a: attached current_universe_a as u_a
				u_b: attached current_universe_b as u_b
				definition: ab.inverse.as_singleton ≍ (
					(u_b × u_a).filtered_xy (
						agent  (y: B x: A; a: A b: B): BOOLEAN
							do
								Result := eq_b (y, b) and eq_a (x, a)
							end (?, ?, ab.a, ab.b)
						)
					)
				involutive: ab.inverse.inverse ≍ ab
			end
			Result := True
		end

	composed_with_homogeneous_ok (ab: STS_ORDERED_PAIR [A, B, EQ_A, EQ_B]; aa_2, aa_1: STS_ORDERED_PAIR [A, A, EQ_A, EQ_A]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_ORDERED_PAIR}.composed_with_homogeneous?
		do
			if ab ≽ aa_2 then
				check
					u_a: attached current_universe_a as u_a
					u_b: attached current_universe_b as u_b
					definition: (ab.composed_with_homogeneous (aa_2)).as_singleton ≍ (
						(u_a × u_b).filtered_xy (
							agent (x: A y: B; a: A b: B): BOOLEAN
								do
									Result := eq_a (x, a) and eq_b (y, b)
								end (?, ?, aa_2.first, ab.b)
							)
						)
				end
				if aa_2 ≽ aa_1 then
					check
							ab_oh_aa_2_succeeds_aa_1: ab.composed_with_homogeneous (aa_2) ≽ aa_1
							ab_succeeds_aa_2_oh_aa_1: ab ≽ aa_2.composed_with_homogeneous (aa_1)
						associative: ab.composed_with_homogeneous (aa_2).composed_with_homogeneous (aa_1) ≍
							ab.composed_with_homogeneous (aa_2.composed_with_homogeneous (aa_1))
					end
				end
			end
			Result := True
		end

	composed_with_inverse_like_ok
		(ab: STS_ORDERED_PAIR [A, B, EQ_A, EQ_B]; ba: STS_ORDERED_PAIR [B, A, EQ_B, EQ_A]; bb: STS_ORDERED_PAIR [B, B, EQ_B, EQ_B]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_ORDERED_PAIR}.composed_with_inverse_like?
		do
			if ab ≽ ba then
				check
					u_b: attached current_universe_b as u_b
					definition: ab.composed_with_inverse_like (ba).as_singleton ≍ (
						(× u_b).filtered_xy (
							agent (x, y: B; a, b: B): BOOLEAN
								do
									Result := eq_b (x, a) and eq_b (y, b)
								end (?, ?, ba.first, ab.b)
							)
						)
				end
				if ba ≽ bb then
					check
							ab_oi_ba_succeeds_bb: ab.composed_with_inverse_like (ba) ≽ bb
							ab_succeeds_ba_oh_bb: ab ≽ ba.composed_with_homogeneous (bb)
						associative: ab.composed_with_inverse_like (ba).composed_with_homogeneous (bb) ≍
							ab.composed_with_inverse_like (ba.composed_with_homogeneous (bb))
					end
				end
			end
			Result := True
		end

feature -- Properties (Conversion)

	as_singleton_ok (ab: STS_ORDERED_PAIR [A, B, EQ_A, EQ_B]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_ORDERED_PAIR}.as_singleton?
		do
			check
				u_a: attached current_universe_a as u_a
				u_b: attached current_universe_b as u_b
				definition: ab.as_singleton ≍ (
					(u_a × u_b).filtered_xy (
						agent (x: A y: B; a: A b: B): BOOLEAN
							do
								Result := eq_a (x, a) and eq_b (y, b)
							end (?, ?, ab.a, ab.b)
						)
					)
			end
			Result := True
		end

feature -- Transformer

	ab_to_a: STS_TRANSFORMER [STS_ORDERED_PAIR [A, B, EQ_A, EQ_B], A, STS_EQUALITY [STS_ORDERED_PAIR [A, B, EQ_A, EQ_B]], EQ_A]
			-- Transformer between types derived from {STS_ORDERED_PAIR [A, B, EQ_A, EQ_B]} and {A}
		deferred
		end

	ab_to_b: STS_TRANSFORMER [STS_ORDERED_PAIR [A, B, EQ_A, EQ_B], b, STS_EQUALITY [STS_ORDERED_PAIR [A, B, EQ_A, EQ_B]], EQ_B]
			-- Transformer between types derived from {STS_ORDERED_PAIR [A, B, EQ_A, EQ_B]} and {B}
		deferred
		end

feature {NONE} -- Implementation

	max_supersets: NATURAL
			-- Maximum number of supersets to be checked within property statemens
		deferred
		end

note
	copyright: "Copyright (c) 2012-2021, Rosivaldo Fernandes Alves"
	license: "[
		Eiffel Forum License v2
		(see http://www.eiffel.com/licensing/forum.txt)
		]"
	source: "https://launchpad.net/set-theory"

end
