﻿
note
	description: "Object that checks whether the properties verified within set theory hold for an implementation of {STS_SET}"
	author: "Rosivaldo Fernandes Alves"
	date: "$Date: Sat 2018-12-01 19:25:58 -0300 $"
	revision: "$Revision$"

deferred class
	SET_PROPERTIES [A, EQ -> STS_EQUALITY [A]]

inherit
	ELEMENT_PROPERTIES
		rename
			is_in_ok as element_is_in_ok,
			is_not_in_ok as element_is_not_in_ok
		end

feature -- Access

	o: like universe_anchor.subset_anchor
			-- The empty set
		deferred
		ensure
			is_empty: Result.is_empty
		end

	unit: like universe_anchor.n_tuple_anchor
			-- The unit n-tuple, i.e. the n-tuple with no term.
		deferred
		ensure
			is_unit: Result.is_unit
		end

	eq: EQ
			-- Equality for objects like {A}
		deferred
		end

	current_universe: like universe_anchor
			-- The current "Universe", i.e. set with every object currently in system memory whose type conforms to {A}.
			-- Notice that this "universe" may change from a call to another.
		deferred
		end

feature -- Properties (Primitive)

	is_empty_ok (s: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.is_empty?
		do
			check
				u: attached current_universe as u
				definition: s.is_empty = (
					s ≍ (
						u | agent  (x: A): BOOLEAN
							do
								Result := False
							end
						)
					)
				has_nothing: s.is_empty = (u |∀ agent s.does_not_have)
				no_element: s.is_empty = (# s = 0)
				uniqueness: s.is_empty = (s ≍ o)
			then
				Result := True
			end
		end

	any_ok (s: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.any?
		do
			if not s.is_empty then
				check
					membership: s ∋ s.any
					building_up: s ≍ (s.others & s.any)
				then
				end
			end
			Result := True
		end

	others_ok (s: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.others?
		do
			if s.is_empty then
				check
					emptiness: s.others ≍ s
					no_element: # s.others = # s
				then
				end
			else
				check
					decomposing: s.others ≍ (s / s.any)
					strict_subset: s.others ⊂ s
					one_element_less: # s.others = # s - 1
				then
				end
			end
			check
				subset: s.others ⊆ s
			then
				Result := True
			end
		end

feature -- Properties (Membership)

	is_in_ok (s: STS_SET [A, EQ]; ss: STS_SET [STS_SET [A, EQ], STS_EQUALITY [STS_SET [A, EQ]]]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.is_in?
		do
			check
				require_non_emptiness: s ∈ ss ⇒ not ss.is_empty
			then
				Result := True
			end
		end

	is_not_in_ok (s: STS_SET [A, EQ]; ss: STS_SET [STS_SET [A, EQ], STS_EQUALITY [STS_SET [A, EQ]]]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.is_not_in?
		do
			check
				definition: s ∉ ss = ss ∌ s
				sufficient_emptiness: ss.is_empty ⇒ s ∉ ss
			then
				Result := True
			end
		end

	has_ok (s: STS_SET [A, EQ]; a: A): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.has?
		do
			check
				u: attached current_universe as u
				require_non_emptiness: s ∋ a ⇒ not s.is_empty
				universe_has_everything: u ∋ a
				uniqueness: s ∋ a = s |∃! agent s.equality_holds (?, a)
				has_own_elements: s |∀ agent s.has
			then
				Result := True
			end
		end

	does_not_have_ok (s: STS_SET [A, EQ]; a: A): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.does_not_have?
		do
			check
				definition: s ∌ a = s |∄ agent s.equality_holds (?, a)
				has_own_elements: s |∄ agent s.does_not_have
				empty_set_has_nothing: o ∌ a
			then
				Result := True
			end
		end

feature -- Properties (Construction)

	with_ok (s: STS_SET [A, EQ]; a: A): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.with?
		do
			check
				u: attached current_universe as u
				definition: (s & a) ≍ (u | agent u.ored (agent s.has, agent u.equality_holds (?, a), ?))
				inserted: (s & a) ∋ a
				nothing_lost: s ⊆ (s & a)
				nothing_else: (s & a) |∀ agent s.ored (agent s.has, agent s.equality_holds (?, a), ?)
				same_cardinality: s ∋ a ⇒ # (s & a) = # s
				incremented_cardinality: s ∌ a ⇒ # (s & a) = # s + 1
			then
				Result := True
			end
		end

	without_ok (s: STS_SET [A, EQ]; a: A): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.without?
		do
			check
				u: attached current_universe as u
				definition: (s / a) ≍ (u | agent anded (agent s.has, agent equality_does_not_hold (?, a), ?))
				excluded: (s / a) ∌ a
				every_other_element: s |∀ agent implied (agent negated(agent s.equality_holds (?, a), ?), agent (s / a).has, ?)
				nothing_else: (s / a) ⊆ s
				same_cardinality: s ∌ a ⇒ # (s / a) = # s
				decremented_cardinality: s ∋ a ⇒ # (s / a) = # s - 1
			then
				Result := True
			end
		end

feature -- Properties (Quality)

	is_singleton_ok (s: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.is_singleton?
		do
			check
				definition: s.is_singleton = (# s = 1)
				by_construction: s.is_singleton ⇒ not s.is_empty and then s ≍ singleton (s.any)
			then
				Result := True
			end
		end

feature -- Properties (Measurement)

	cardinality_ok (s: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.cardinality?
		do
			check
				definition: # s = s.as_tuple.n
			then
				Result := True
			end
		end

feature -- Properties (Comparison)

	equals_ok (s1, s2, s3: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.equals?
		do
			check
				u: attached current_universe as u
				definition: s1 ≍ s2 = (u |∀ agent u.iff (agent s1.has, agent s2.has, ?))
				u_ps_ps_card: attached (2 ^ (2 ^ # u)) as u_ps_ps_card
				set_comparisons_upper_bound: attached (2 * u_ps_ps_card) as set_comparisons_upper_bound
				element_comparisons_upper_bound: attached ((# u ^ 2) * set_comparisons_upper_bound) as
					element_comparisons_upper_bound
				checked_ps: attached {ISE_RUNTIME}.check_assert
					(element_comparisons_upper_bound ≤ max_asserted_elements) as checked_ps
					by_membership: element_comparisons_upper_bound ≤ max_elements ⇒ s1 ≍ s2 = (
						u.powerset.powerset |∀ agent
							(ss: STS_SET [STS_SET [A, EQ], STS_SET_EQUALITY [A, EQ]]; ia_s1, ia_s2: STS_SET [A, EQ]): BOOLEAN
							do
								Result := ia_s1 ∈ ss = ia_s2 ∈ ss
							end (?, s1, s2)
						)
				checked_ps_restored: attached {ISE_RUNTIME}.check_assert (checked_ps)
				by_inclusion: s1 ≍ s2 = (s1 ⊆ s2 and s2 ⊆ s1)
				reflexive: s1 ≍ s1
				symmetric: s1 ≍ s2 ⇒ s2 ≍ s1
				transitive: s1 ≍ s2 and s2 ≍ s3 ⇒ s1 ≍ s3
				same_cardinality: s1 ≍ s2 ⇒ (# s1 = # s2)
			then
				Result := True
			end
		end

	unequals_ok (s1, s2: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.unequals?
		do
			check
				u: attached current_universe as u
				definition_1: s1 ≭ s2 = (
					u |∃ agent u.ored (
						agent anded (agent s1.has, agent s2.does_not_have, ?),
						agent anded (agent s1.does_not_have, agent s2.has, ?), ?
						)
					)
				definition_2: s1 ≭ s2 = (
					(u |∃ agent anded (agent s1.has, agent s2.does_not_have, ?)) or
					(u |∃ agent anded (agent s2.has, agent s1.does_not_have, ?))
					)
				lesser_definition: s1 ≭ s2 = ((s1 |∃ agent s2.does_not_have) or (s2 |∃ agent s1.does_not_have))
				u_ps_ps_card: attached (2 ^ (2 ^ # u)) as u_ps_ps_card
				set_comparisons_upper_bound: attached (2 * u_ps_ps_card) as set_comparisons_upper_bound
				element_comparisons_upper_bound: attached ((# u ^ 2) * set_comparisons_upper_bound) as
					element_comparisons_upper_bound
				checked_ps: attached {ISE_RUNTIME}.check_assert
					(element_comparisons_upper_bound ≤ max_asserted_elements) as checked_ps
					by_membership: element_comparisons_upper_bound ≤ max_elements ⇒ s1 ≭ s2 = (
						u.powerset.powerset |∃ agent
							(s: STS_SET [STS_SET [A, EQ], STS_SET_EQUALITY [A, EQ]]; ia_s1, ia_s2: STS_SET [A, EQ]): BOOLEAN
							do
								Result := ia_s1 ∈ s /= ia_s2 ∈ s
							end (?, s1, s2)
						)
				checked_ps_restored: attached {ISE_RUNTIME}.check_assert (checked_ps)
				by_uninclusion: s1 ≭ s2 = (s1 ⊈ s2 or s2 ⊈ s1)
				irreflexive: not (s1 ≭ s1)
				symmetric: s1 ≭ s2 ⇒ s2 ≭ s1
				unequal_cardinalities: (# s1 /= # s2) ⇒ s1 ≭ s2
			then
				Result := True
			end
		end

	is_subset_ok (s1, s2, s3: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.is_subset?
		do
			check
				u: attached current_universe as u
				definition: s1 ⊆ s2 = (u |∀ agent implied (agent s1.has, agent s2.has, ?))
				by_difference: s1 ⊆ s2 = (s1 ∖ s2) ≍ o
				reflexive: s1 ⊆ s1
				transitive: s1 ⊆ s2 and s2 ⊆ s3 ⇒ s1 ⊆ s3
				antisymmetric: s1 ⊆ s2 and s2 ⊆ s1 ⇒ s1 ≍ s2
				everything_includes_o: o ⊆ s1
				u_includes_everything: s1 ⊆ u
				equality: s1 ≍ s2 = (s1 ⊆ s2 and s2 ⊆ s1)
				o_includes_only_o: s1 ⊆ o ⇒ s1 ≍ o
				only_u_includes_u: u ⊆ s1 ⇒ s1 ≍ u
				cardinality: s1 ⊆ s2 ⇒ # s1 ≤ # s2
			then
				Result := True
			end
		end

	is_not_subset_ok (s1, s2: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.is_not_subset?
		do
			check
				u: attached current_universe as u
				definition: s1 ⊈ s2 = (u |∃ agent anded (agent s1.has, agent s2.does_not_have, ?))
				lesser_definition: s1 ⊈ s2 = (s1 |∃ agent s2.does_not_have)
				by_difference: s1 ⊈ s2 = (s1 ∖ s2) ≭ o
				irreflexive: not (s1 ⊈ s1)
				unequality: s1 ≭ s2 = (s1 ⊈ s2 or s2 ⊈ s1)
				o_includes_only_o: s1 ⊈ o ⇒ s1 ≭ o
				only_u_includes_u: u ⊈ s1 ⇒ s1 ≭ u
				cardinality: # s1 > # s2 ⇒ s1 ⊈ s2
			then
				Result := True
			end
		end

	is_superset_ok (s1, s2: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.is_superset?
		do
			check
				definition: s1 ⊇ s2 = (s2 |∀ agent s1.has)
				cardinality: s1 ⊇ s2 ⇒ # s1 ≥ # s2
			then
				Result := True
			end
		end

	is_not_superset_ok (s1, s2: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.is_not_superset?
		do
			check
				definition: s1 ⊉ s2 = (s2 |∃ agent s1.does_not_have)
				cardinality: # s1 < # s2 ⇒ s1 ⊉ s2
				cardinality: s1 ⊇ s2 ⇒ # s1 ≥ # s2
			then
				Result := True
			end
		end

	is_comparable_ok (s1, s2: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.is_comparable?
		do
			check
				reflexive: s1.is_comparable (s1)
				symmetric: s1.is_comparable (s2) ⇒ s2.is_comparable (s1)
			then
				Result := True
			end
		end

	is_not_comparable_ok (s1, s2: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.is_not_comparable?
		do
			check
				u: attached current_universe as u
				definition: s1.is_not_comparable (s2) = (
						(u |∃ agent anded (agent s1.has, agent s2.does_not_have, ?))
						and
						(u |∃ agent anded (agent s2.has, agent s1.does_not_have, ?))
					)
				lesser_definition: s1.is_not_comparable (s2) = (
						(s1 |∃ agent s2.does_not_have) and (s2 |∃ agent s1.does_not_have)
					)
				by_inclusion: s1.is_not_comparable (s2) = (s1 ⊈ s2 and s2 ⊈ s1)
				proper_symmetric_difference: s1.is_not_comparable (s2) ⇒ (s1 ⊖ s2) ≭ o
				irreflexive: not s1.is_not_comparable (s1)
			then
				Result := True
			end
		end

	is_strict_subset_ok (s1, s2, s3: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.is_strict_subset?
		do
			check
				u: attached current_universe as u
				definition: s1 ⊂ s2 = (
						(u |∀ agent implied (agent s1.has, agent s2.has, ?)) and
						(u |∃ agent anded (agent s2.has, agent s1.does_not_have, ?))
					)
				lesser_definition: s1 ⊂ s2 = ((s1 |∀ agent s2.has) and (s2 |∃ agent s1.does_not_have))
				by_inclusion: s1 ⊂ s2 = (s1 ⊆ s2 and s2 ⊈ s1)
				irreflexive: not (s1 ⊂ s1)
				transitive: s1 ⊂ s2 and s2 ⊂ s3 ⇒ s1 ⊂ s3
				asymmetric: s1 ⊂ s2 ⇒ not (s2 ⊂ s1)
				proper_difference: s1 ⊂ s2 ⇒ (s2 ∖ s1) ≭ o
				cardinality: s1 ⊂ s2 ⇒ # s1 < # s2
			then
				Result := True
			end
		end

	is_not_strict_subset_ok (s1, s2: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.is_not_strict_subset?
		do
			check
				u: attached current_universe as u
				definition: s1 ⊄ s2 = (
						(u |∃ agent anded (agent s1.has, agent s2.does_not_have, ?)) or
						(u |∀ agent implied (agent s2.has, agent s1.has, ?))
					)
				lesser_definition: s1 ⊄ s2 = ((s1 |∃ agent s2.does_not_have) or (s2 |∀ agent s1.has))
				by_inclusion: s1 ⊄ s2 = (s1 ⊈ s2 or s2 ⊆ s1)
				reflexive: s1 ⊄ s1
				cardinality: # s1 ≥ # s2 ⇒ s1 ⊄ s2
			then
				Result := True
			end
		end

	is_strict_superset_ok (s1, s2: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.is_strict_superset?
		do
			check
				u: attached current_universe as u
				definition: s1 ⊃ s2 = (
					(u |∀ agent implied (agent s2.has, agent s1.has, ?)) and
					(u |∃ agent anded (agent s1.has, agent s2.does_not_have, ?))
					)
				lesser_definition: s1 ⊃ s2 = ((s1 |∃ agent s2.does_not_have) and (s2 |∀ agent s1.has))
				by_inclusion: s1 ⊃ s2 = (s1 ⊇ s2 and s1 ≭ s2)
				proper_difference: s1 ⊃ s2 ⇒ (s1 ∖ s2) ≭ o
				cardinality: s1 ⊃ s2 ⇒ # s1 > # s2
			then
				Result := True
			end
		end

	is_not_strict_superset_ok (s1, s2: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.is_not_strict_superset?
		do
			check
				u: attached current_universe as u
				definition: s1 ⊅ s2 = (
						(u |∃ agent anded (agent s2.has, agent s1.does_not_have, ?)) or
						(u |∀ agent implied (agent s1.has, agent s2.has, ?))
					)
				lesser_definition: s1 ⊅ s2 = ((s1 |∀ agent s2.has) or (s2 |∃ agent s1.does_not_have))
				by_inclusion: s1 ⊅ s2 = (s1 ⊉ s2 or s1 ≍ s2)
				cardinality: # s1 ≤ # s2 ⇒ s1 ⊅ s2
			then
				Result := True
			end
		end

	is_trivial_superset_ok (s1, s2: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.is_trivial_superset?
		do
			check
				definition: s1.is_trivial_superset (s2) = (s1 ≍ s2 or s2 ≍ o)
			then
				Result := True
			end
		end

	is_proper_subset_ok (s1, s2: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.is_proper_subset?
		do
			check
				definition: s1.is_proper_subset (s2) = (s1 ⊆ s2 and not s1.is_trivial_subset (s2))
				irreflexive: not s1.is_proper_subset (s1)
			then
				Result := True
			end
		end

	is_proper_superset_ok (s1, s2: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.is_proper_superset?
		do
			check
				definition: s1.is_proper_superset (s2) = (s1 ⊇ s2 and not s1.is_trivial_superset (s2))
			then
				Result := True
			end
		end

	is_disjoint_ok (s1, s2: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.is_disjoint?
		do
			check
				definition: s1.is_disjoint (s2) = s1 |∄ agent s2.has
				quasi_irreflexive: not s1.is_empty ⇒ not s1.is_disjoint (s1)
				symmetric: s1.is_disjoint (s2) = s2.is_disjoint (s1)
				left_disjoint: o.is_disjoint (s2)
				right_disjoint: s1.is_disjoint (o)
			then
				Result := True
			end
		end

	intersects_ok (s1, s2: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.intersects?
		do
			check
				definition: s1.intersects (s2) = (s1 ∩ s2) ≭ o
				by_existence: s1.intersects (s2) = (s1 |∃ agent s2.has)
				quasi_reflexive: not s1.is_empty ⇒ s1.intersects (s1)
				symmetric: s1.intersects (s2) = s2.intersects (s1)
			then
				Result := True
			end
		end

feature -- Properties (Quantifier)

	exists_ok (s: STS_SET [A, EQ]; p: PREDICATE [A]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.exists?
		do
			check
				definition: (s |∃ p) = not (s |∀ agent negated (p, ?))
				by_filtering: (s |∃ p) = (s | p) ≭ o
			end
			Result := True
		end

	does_not_exist_ok (s: STS_SET [A, EQ]; p: PREDICATE [A]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.does_not_exist?
		do
			check
				definition: s |∄ p = (s |∀ agent negated (p, ?))
				by_filtering: s |∄ p = (s | p) ≍ o
			end
			Result := True
		end

	exists_unique_ok (s: STS_SET [A, EQ]; p: PREDICATE [A]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.exists_unique?
		do
			check
				definition: s |∃! p = (s |∃ agent (ia_s: STS_SET [A, EQ] ia_p: PREDICATE [A] x: A): BOOLEAN
					do
						Result := ia_s |∀ agent ia_s.iff (ia_p, agent ia_s.equality_holds (?, x), ?)
					end (s, p, ?))
				by_inequality: s |∃! p = (s |∃ agent (ia_s: STS_SET [A, EQ] ia_p: PREDICATE [A] x: A): BOOLEAN
					do
						Result := ia_p (x) and ia_s |∄ agent anded (ia_p, agent equality_does_not_hold (?, x), ?)
					end (s, p, ?))
				by_equality: s |∃! p = (s |∃ agent (ia_s: STS_SET [A, EQ] ia_p: PREDICATE [A] x: A): BOOLEAN
					do
						Result := ia_p (x) and (ia_s |∀ agent implied (ia_p, agent ia_s.equality_holds (?, x), ?))
					end (s, p, ?))
				by_pairing: s |∃! p = (
					(s |∃ p) and (s.for_all_pairs (agent pair_implied (agent (s.element_to_element).anded (p, ?, p, ?), agent s.equality_holds, ?, ?)))
					)
				by_cardinality: s |∃! p = (# (s | p) = 1)
			end
			Result := True
		end

	exists_pair_ok (s: STS_SET [A, EQ]; p: PREDICATE [A, A]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.exists_pair?
		do
			check
				definition: s.exists_pair (p) = not s.for_all_pairs (agent pair_negated (p, ?, ?))
				by_filtering: (# s ^ 2) ≤ max_elements ⇒ s.exists_pair (p) = ((× s).filtered_xy (p)) ≭ (× o)
			end
			Result := True
		end

	does_not_exist_pair_ok (s: STS_SET [A, EQ]; p: PREDICATE [A, A]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.does_not_exist_pair?
		do
			check
				definition: s.does_not_exist_pair (p) = s.for_all_pairs (agent pair_negated (p, ?, ?))
				by_filtering: (# s ^ 2) ≤ max_elements ⇒ s.does_not_exist_pair (p) = ((× s).filtered_xy (p)) ≍ (× o)
			end
			Result := True
		end

	exists_distinct_pair_ok (s: STS_SET [A, EQ]; p: PREDICATE [A, A]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.exists_distinct_pair?
		do
			check
				definition: s.exists_distinct_pair (p) = not s.for_all_distinct_pairs (agent pair_negated(p, ?, ?))
				by_filtering: (# s ^ 2) ≤ max_elements ⇒ s.exists_distinct_pair (p) = (× s).filtered_xy (p) ⊈ ∆ s
				by_square: (# s ^ 2) ≤ max_elements ⇒ s.exists_distinct_pair (p) = (× s).exist_xy (agent pair_anded (p, agent equality_does_not_hold, ?, ?))
				by_inequality: s.exists_distinct_pair (p) = s.exists_pair (agent pair_anded (p, agent equality_does_not_hold, ?, ?))
			end
			Result := True
		end

	for_all_ok (s: STS_SET [A, EQ]; p: PREDICATE [A]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.for_all?
		do
			check
				definition: (s |∀ p) = (s |∄ agent negated (p, ?))
				by_filtering: (s |∀ p) = (s | p) ≍ s
			end
			Result := True
		end

	for_all_pairs_ok (s: STS_SET [A, EQ]; p: PREDICATE [A, A]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.for_all_pairs?
		do
			check
				definition: s.for_all_pairs (p) = s.does_not_exist_pair (agent pair_negated (p, ?, ?))
				by_filtering: (# s ^ 2) ≤ max_elements ⇒ s.for_all_pairs (p) = ((× s).filtered_xy (p) ≍ (× s))
			end
			Result := True
		end

	for_all_distinct_pairs_ok (s: STS_SET [A, EQ]; p: PREDICATE [A, A]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.for_all_distinct_pairs?
		do
			check
				definition: s.for_all_distinct_pairs (p) = not s.exists_distinct_pair (agent pair_negated (p, ?, ?))
				by_filtering: (# s ^ 2) ≤ max_elements ⇒ s.for_all_distinct_pairs (p) = ((× s).filtered_xy (agent pair_negated(p, ?, ?)) ⊆ ∆ s)
				by_square: (# s ^ 2) ≤ max_elements ⇒
					s.for_all_distinct_pairs (p) = (× s).for_all_xy (agent pair_implied (agent pair_negated (p, ?, ?), agent s.equality_holds, ?, ?))
				by_equality: s.for_all_distinct_pairs (p) = s.for_all_pairs (agent pair_implied (agent pair_negated (p, ?, ?), agent s.equality_holds, ?, ?))
			end
			Result := True
		end

feature -- Properties (Operation)

	subsets_ok (a: A; s0, s1, s2: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.subsets?
			--| Notice that `powerset_ok' checks the properties of {STS_SET}.powerset, that is synonym of {STS_SET}.subsets.
			--| Though it seems redundant, one must remember that synonym features are indeed not the same feature; they may
			--| eventually be redefined differently, hence the need for checking their properties separately.
		do
			check
				u: attached current_universe as u
				u_ss_card: attached (2 ^ # u) as u_ss_card
				checked_u_ss: attached {ISE_RUNTIME}.check_assert (u_ss_card ≤ max_asserted_subsets) as checked_u_ss
					definition: u_ss_card ≤ max_subsets ⇒ s1.subsets ≍ (u.subsets | agent s1.is_superset)
				checked_u_ss_restored: attached {ISE_RUNTIME}.check_assert (checked_u_ss)
				s1_ss_card: attached (2 ^ # s1) as s1_ss_card
				checked_s1_ss: attached {ISE_RUNTIME}.check_assert (s1_ss_card ≤ max_asserted_subsets) as checked_s1_ss
					by_reducing_without: s1_ss_card ≤ max_subsets ⇒ s1.subsets ≍ element_to_sets.set_reduction (
						s1, s1.as_singleton, agent (acc: STS_SET [STS_SET [A, EQ], STS_SET_EQUALITY [A, EQ]]; x: A):
							STS_SET [STS_SET [A, EQ], STS_SET_EQUALITY [A, EQ]]
							do
								Result := acc ∪ (acc ↦ agent {STS_SET [A, EQ]}.without (x))
							end
						)
					has_base_set: s1_ss_card ≤ max_subsets ⇒ s1 ∈ s1.subsets
					includes_lesser_subsets: s1_ss_card ≤ max_subsets ⇒ s1.others.subsets ⊆ s1.subsets
					includes_extended_subsets: s1_ss_card ≤ max_subsets and not s1.is_empty ⇒
						(s1.others.subsets ↦ agent {STS_SET [A, EQ]}.with (s1.any)) ⊆ s1.subsets
					cardinality: s1_ss_card ≤ max_subsets ⇒ # s1.subsets = 2 ^ (# s1)
					subset_membership: s1_ss_card ≤ max_subsets ⇒ (s0 ⊆ s1) = s0 ∈ s1.subsets
					induced_membership: s1_ss_card ≤ max_subsets ⇒ s1 ∋ a = singleton (a) ∈ s1.subsets
					has_empty_set: s1_ss_card ≤ max_subsets ⇒ o ∈ s1.subsets
					s2_ss_card: attached (2 ^ # s2) as s2_ss_card
					checked_s2_ss: attached {ISE_RUNTIME}.check_assert ((s1_ss_card ∨ s2_ss_card) ≤ max_asserted_subsets) as
						checked_s2_ss
						induced_inclusion: (s1_ss_card ∨ s2_ss_card) ≤ max_subsets ⇒ (s1 ⊆ s2) = (s1.subsets ⊆ s2.subsets)
						induced_equality: (s1_ss_card ∨ s2_ss_card) ≤ max_subsets ⇒ s1 ≍ s2 = s1.subsets ≍ s2.subsets
						disjointness_consequence: (s1_ss_card ∨ s2_ss_card) ≤ max_subsets ⇒
							(s1 ∩ s2) ≍ o ⇒ (s1.subsets ∩ s2.subsets) ≍ o.as_singleton
						induced_intersection: (s1_ss_card ∨ s2_ss_card) ≤ max_subsets ⇒
							(s1 ∩ s2).subsets ≍ (s1.subsets ∩ s2.subsets)
						quasi_included_difference: (s1_ss_card ∨ s2_ss_card) ≤ max_subsets ⇒
							(s1 ∖ s2).subsets ⊆ ((s1.subsets ∖ s2.subsets) ∪ o.as_singleton)
						s1_uni_s2_ss_card: attached (2 ^ # (s1 ∪ s2)) as s1_uni_s2_ss_card
						checked_s1_uni_s2_ss: attached {ISE_RUNTIME}.check_assert
							(s1_uni_s2_ss_card ≤ max_asserted_subsets) as checked_s1_uni_s2_ss
							included_union: s1_uni_s2_ss_card ≤ max_subsets ⇒ (s1.subsets ∪ s2.subsets) ⊆ (s1 ∪ s2).subsets
						checked_s1_uni_s2_ss_restored: attached {ISE_RUNTIME}.check_assert (checked_s1_uni_s2_ss)
					checked_s2_ss_restored: attached {ISE_RUNTIME}.check_assert (checked_s2_ss)
				checked_s1_ss_restored: attached {ISE_RUNTIME}.check_assert (checked_s1_ss)
			then
				Result := True
			end
		end

	powerset_ok (a: A; s0, s1, s2: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.powerset?
			--| Please see the comment at `subsets_ok' header.
		local
			n_s1_subsets, n_s2_subsets, n_s1_subtr_s2_subsets, n_s1_uni_s2_subsets: DOUBLE
			chk_ops: BOOLEAN
		do
			check
				u: attached current_universe as u
				definition: (2 ^ # u) ≤ max_subsets ⇒ s1.powerset ≍ (u.powerset | agent s1.is_superset)
			end
			chk_ops := {ISE_RUNTIME}.check_assert (False)
				n_s1_subsets := 2 ^ # s1
			chk_ops := {ISE_RUNTIME}.check_assert (chk_ops)
			if (n_s1_subsets ^ 2) ≤ max_subsets then
				check
					by_reducing_without: s1.powerset ≍ element_to_sets.set_reduction (
						s1, s1.as_singleton, agent
							(acc: STS_SET [STS_SET [A, EQ], STS_SET_EQUALITY [A, EQ]]; x: A): STS_SET [STS_SET [A, EQ], STS_SET_EQUALITY [A, EQ]]
							do
								Result := acc ∪ (acc ↦ agent {STS_SET [A, EQ]}.without (x))
							end
						)
				end
			end
			if n_s1_subsets ≤ max_subsets then
				check
					has_base_set: s1 ∈ s1.powerset
				end
			end
			if (n_s1_subsets ^ 2 / 4) ≤ max_subsets then
				check
					includes_lesser_subsets: s1.others.powerset ⊆ s1.powerset
					includes_extended_subsets: not s1.is_empty ⇒ (s1.others.powerset ↦ agent {STS_SET [A, EQ]}.with (s1.any)) ⊆ s1.powerset
				end
			end
			if n_s1_subsets ≤ max_subsets then
				check
					cardinality: # s1.powerset = 2 ^ (# s1)
					subset_membership: (s0 ⊆ s1) = s0 ∈ s1.powerset
					induced_membership: s1 ∋ a = singleton (a) ∈ s1.powerset
					has_empty_set: o ∈ s1.powerset
				end
			end
			chk_ops := {ISE_RUNTIME}.check_assert (False)
				n_s2_subsets := 2 ^ # s2
			chk_ops := {ISE_RUNTIME}.check_assert (chk_ops)
			if (n_s1_subsets ∨ n_s2_subsets) ≤ max_subsets then
				check
					induced_inclusion: (s1 ⊆ s2) = (s1.powerset ⊆ s2.powerset)
					induced_equality: s1 ≍ s2 = s1.powerset ≍ s2.powerset
					disjointness_consequence: (s1 ∩ s2) ≍ o ⇒ (s1.powerset ∩ s2.powerset) ≍ o.as_singleton
					induced_intersection: (s1 ∩ s2).powerset ≍ (s1.powerset ∩ s2.powerset)
				end
			end
			chk_ops := {ISE_RUNTIME}.check_assert (False)
				n_s1_subtr_s2_subsets := (2 ^ # (s1 ∖ s2)) ^ 2 / 2
			chk_ops := {ISE_RUNTIME}.check_assert (chk_ops)
			if (n_s1_subsets ∨ n_s2_subsets ∨ n_s1_subtr_s2_subsets) ≤ max_subsets then
				check
					quasi_included_difference: (s1 ∖ s2).powerset ⊆ ((s1.powerset ∖ s2.powerset) ∪ o.as_singleton)
				end
			end
			chk_ops := {ISE_RUNTIME}.check_assert (False)
				n_s1_uni_s2_subsets := (2 ^ # (s1 ∪ s2)) ^ 2 / 2
			chk_ops := {ISE_RUNTIME}.check_assert (chk_ops)
			if n_s1_uni_s2_subsets ≤ max_subsets then
				check
					included_union: (s1.powerset ∪ s2.powerset) ⊆ (s1 ∪ s2).powerset
				end
			end
			Result := True
		end

	trivial_subsets_ok (s: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.trivial_subsets?
		do
			if s.is_empty then
				check
					one_subset: # s.trivial_subsets = 1
				end
			else
				check
					two_subsets: # s.trivial_subsets = 2
				end
			end
			Result := True
		end

	proper_subsets_ok (s: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.proper_subsets?
		do
			check
				n: attached (# s) as n
				n_subsets: attached (2 ^ n) as n_subsets
				no_proper_subset: n_subsets ≤ max_subsets and n < 2 ⇒ # s.proper_subsets = 0
				some_proper_subset: n_subsets ≤ max_subsets and n ≥ 2 ⇒ # s.proper_subsets = 2 ^ n - 2
			end
			Result := True
		end

	filtered_ok (s: STS_SET [A, EQ]; p: PREDICATE [A]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.filtered?
		do
			check
				u: attached current_universe as u
				definition: s | p ≍ (u | agent anded (p, agent s.has, ?))
				by_intersection: s | p ≍ (s ∩ (u | p))
				subset: s | p ⊆ s
				only_compliant_elements: (s | p) |∀ p
				nothing_lost: (s ∖ (s | p)) |∄ p
			end
			Result := True
		end

	complemented_ok (s01, s02, s: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.complemented?
		local
			u: STS_SET [A, EQ]
			preconditions_tested, preconditions_ok: BOOLEAN
			chk: BOOLEAN
		do
			chk := {ISE_RUNTIME}.check_assert (False)
				u := current_universe
			chk := {ISE_RUNTIME}.check_assert (chk)
			if not preconditions_tested then
				check
					preconditions_ok: s01 ⊆ s and s02 ⊆ s
				end
				preconditions_ok := True
			end
			if preconditions_ok then
				check
					definition: (s01 ∁ s) ≍ (u | agent anded (agent s.has, agent s01.does_not_have, ?))
					by_difference: (s01 ∁ s) ≍ (s ∖ s01)
					by_reduction: (s01 ∁ s) ≍ element_to_set.set_reduction (s01, s, agent {STS_SET [A, EQ]}.without)
					disjoint: (s01 ∁ s).is_disjoint (s01)
					is_subset_too: (s01 ∁ s) ⊆ s
					complementary: (s01 ∪ (s01 ∁ s)) ≍ s
					involutive: ((s01 ∁ s) ∁ s) ≍ s01
					inverted_inclusion: s01 ⊆ s02 = ((s01 ∁ s) ⊇ (s02 ∁ s))
				end
			end
			check
					everything_includes_o: o ⊆ s -- {STS_SET}.is_subset definition
				full_complement: (o ∁ s) ≍ s
					reflexive_inclusion: s ⊆ s -- {STS_SET}.is_subset definition
				empty_complement: (s ∁ s) ≍ o
					u_includes_o: o ⊆ u -- {STS_SET}.is_subset definition
				dual_of_o: (o ∁ u) ≍ u
					u_includes_u: u ⊆ u -- {STS_SET}.is_subset definition
				dual_of_u: (u ∁ u) ≍ o
						-- Universe includes everything.
					u_includes_s01: s01 ⊆ u
					u_includes_c_s01: (s01 ∁ u) ⊆ u
				absolutely_involutive: ((s01 ∁ u) ∁ u) ≍ s01
					u_includes_s02: s02 ⊆ u -- Universe includes everything.
				dual_inclusion: s01 ⊆ s02 = ((s01 ∁ u) ⊇ (s02 ∁ u))
			end
			Result := True
		rescue
			if {EXCEPTIONS}.tag_name ~ "preconditions_ok" then
				preconditions_tested := True
				retry
			end
		end

	united_ok (a: A; s1, s2, s3, s4: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.united?
		do
			check
				u: attached current_universe as u
				definition: (s1 ∪ s2) ≍ (u | agent u.ored (agent s1.has, agent s2.has, ?))
				by_reduction: (s1 ∪ s2) ≍ element_to_set.set_reduction (s1, s2, agent {STS_SET [A, EQ]}.with)
				by_alternate_reduction: (s1 ∪ s2) ≍ element_to_set.set_reduction (s2, s1, agent {STS_SET [A, EQ]}.with)
				membership: (s1 ∪ s2) ∋ a = (s1 ∋ a or s2 ∋ a)
				includes_s1: s1 ⊆ (s1 ∪ s2)
				includes_s2: s2 ⊆ (s1 ∪ s2)
				keeps_superset: s1 ⊆ s2 = ((s1 ∪ s2) ≍ s2)
				superset: (s1 ⊆ s3 and s2 ⊆ s3) = ((s1 ∪ s2) ⊆ s3)
				subset: s3 ⊆ s1 or s3 ⊆ s2 ⇒ s3 ⊆ (s1 ∪ s2)
				induced_inclusion: s1 ⊆ s2 ⇒ (s1 ∪ s3) ⊆ (s2 ∪ s3)
				subunion: s1 ⊆ s2 and s3 ⊆ s4 ⇒ (s1 ∪ s3) ⊆ (s2 ∪ s4)
						-- Universe includes everything.
					u_includes_s1: s1 ⊆ u
					u_includes_s2: s2 ⊆ u
				inverted_inclusion: (s1 ∁ u) ⊆ (s2 ∁ u) = ((s1 ∪ s2) ≍ s1)
				identity: (o ∪ s1) ≍ s1
				absorption: (s1 ∪ u) ≍ u
				partitioned_universe: (s1 ∪ (s1 ∁ u)) ≍ u
				idempotent: (s1 ∪ s1) ≍ s1
				commutative: (s1 ∪ s2) ≍ (s2 ∪ s1)
				associative: ((s1 ∪ s2) ∪ s3) ≍ (s1 ∪ (s2 ∪ s3))
				universal_union: (s1 ∪ s2) ≍ u ⇒ (s1 ∁ u) ⊆ s2
				empty_union: (s1 ∪ s2) ≍ o = (s1 ≍ o and s2 ≍ o)
				pair_equality: paired_sets (s1, s2) ≍ paired_sets (s3, s4) ⇒ (s1 ∪ s2) ≍ (s3 ∪ s4)
				by_intersection_and_complement: (s1 ∪ s2) ≍ (s1 ∪ ((s1 ∁ u) ∩ s2))
			end
			Result := united_and_intersected_ok (s1, s2, s3, s4)
		end

	intersected_ok (a: A; s1, s2, s3, s4: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.intersected?
		do
			check
				u: attached current_universe as u
				definition: (s1 ∩ s2) ≍ (u | agent anded (agent s1.has, agent s2.has, ?))
				lesser_definition: (s1 ∩ s2) ≍ ((s1 ∪ s2) | agent anded (agent s1.has, agent s2.has, ?))
				by_reduction: (s1 ∩ s2) ≍ element_to_set.set_reduction (
					s1, o, agent (ia_s2, acc: STS_SET [A, EQ]; x: A): STS_SET [A, EQ]
						do
							if ia_s2 ∋ x then
								Result := acc & x
							else
								Result := acc
							end
						end (s2, ?, ?)
					)
				by_alternate_reduction: (s1 ∩ s2) ≍ element_to_set.set_reduction (
					s2, o, agent (ia_s1, acc: STS_SET [A, EQ]; x: A): STS_SET [A, EQ]
						do
							if ia_s1 ∋ x then
								Result := acc & x
							else
								Result := acc
							end
						end (s1, ?, ?)
					)
				right_filtered: (s1 ∩ s2) ≍ (s2 | agent s1.has)
				membership: (s1 ∩ s2) ∋ a = (s1 ∋ a and s2 ∋ a)
				is_subset_of_s1: (s1 ∩ s2) ⊆ s1
				is_subset_of_s2: (s1 ∩ s2) ⊆ s2
				keeps_subset: s1 ⊆ s2 = (s1 ∩ s2) ≍ s1
				subset: (s3 ⊆ s1 and s3 ⊆ s2) = (s3 ⊆ (s1 ∩ s2))
				superset: s1 ⊆ s3 and s2 ⊆ s3 ⇒ (s1 ∩ s2) ⊆ s3
				induced_inclusion: s1 ⊆ s2 ⇒ (s1 ∩ s3) ⊆ (s2 ∩ s3)
				subintersection: s1 ⊆ s2 and s3 ⊆ s4 ⇒ (s1 ∩ s3) ⊆ (s2 ∩ s4)
				absorption: (o ∩ s1) ≍ o
				identity: (s1 ∩ u) ≍ s1
					u_includes_s1: s1 ⊆ u -- Universe includes everything.
				disjoint_complement: (s1 ∩ (s1 ∁ u)) ≍ o
				idempotent: (s1 ∩ s1) ≍ s1
				commutative: (s1 ∩ s2) ≍ (s2 ∩ s1)
				associative: ((s1 ∩ s2) ∩ s3) ≍ (s1 ∩ (s2 ∩ s3))
					u_includes_s2: s2 ⊆ u -- Universe includes everything.
				with_complement_of_disjoint: (s1 ∩ s2) ≍ o ⇒ (s1 ∩ (s2 ∁ u)) ≍ s1
				inverted_inclusion: (s1 ∁ u) ⊆ (s2 ∁ u) = (s1 ∩ s2) ≍ s2
				induced_non_disjointness: (s1 ∩ s2) ≍ s1 and (s1 ∩ s3) ≭ o ⇒ (s2 ∩ s3) ≭ o
				pair_equality: paired_sets (s1, s2) ≍ paired_sets (s3, s4) ⇒ (s1 ∩ s2) ≍ (s3 ∩ s4)
				by_union_and_complement: (s1 ∩ s2) ≍ (s1 ∩ ((s1 ∁ u) ∪ s2))
			end
			Result := united_and_intersected_ok (s1, s2, s3, s4)
		end

	united_and_intersected_ok (s1, s2, s3, s4: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.united and {STS_SET}.intersected together?
		do
			check
				u: attached current_universe as u
				absorbed_union: (s1 ∩ (s1 ∪ s2)) ≍ s1
				absorbed_intersection: (s1 ∪ (s1 ∩ s2)) ≍ s1
				distributed_intersection: (s1 ∩ (s2 ∪ s3)) ≍ ((s1 ∩ s2) ∪ (s1 ∩ s3))
				distributed_union: (s1 ∪ (s2 ∩ s3)) ≍ ((s1 ∪ s2) ∩ (s1 ∪ s3))
						-- Universe includes everything.
					u_includes_s1: s1 ⊆ u
					u_includes_s2: s2 ⊆ u
					u_includes_intersection: (s1 ∩ s2) ⊆ u
					u_includes_union: (s1 ∪ s2) ⊆ u
				de_morgan_intersection: ((s1 ∩ s2) ∁ u) ≍ ((s1 ∁ u) ∪ (s2 ∁ u))
				de_morgan_union: ((s1 ∪ s2) ∁ u) ≍ ((s1 ∁ u) ∩ (s2 ∁ u))
				successive_inclusion: (s1 ⊆ s2 and s2 ⊆ s3) = ((s1 ∪ s2) ≍ (s2 ∩ s3))
				equality: (s1 ∪ s2) ≍ (s1 ∩ s2) = (s1 ≍ s2)
				empty_intersection: (s1 ∩ s2) ≍ o ⇒ (s1 ∪ (s2 ∁ u)) ≍ (s2 ∁ u)
				intersecting_two_unions: ((s1 ∪ s2) ∩ (s3 ∪ s4)) ≍ ((s1 ∩ s3) ∪ (s1 ∩ s4) ∪ (s2 ∩ s3) ∪ (s2 ∩ s4))
				uniting_two_intersections: ((s1 ∩ s2) ∪ (s3 ∩ s4)) ≍ ((s1 ∪ s3) ∩ (s1 ∪ s4) ∩ (s2 ∪ s3) ∩ (s2 ∪ s4))
				union_cardinality: # (s1 ∪ s2) = # s1 + # s2 - # (s1 ∩ s2)
			end
			Result := True
		end

	subtracted_ok (a: A; s1, s2, s3, s4, s5: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.subtracted?
		do
			check
				u: attached current_universe as u
				definition: (s1 ∖ s2) ≍ (u | agent anded (agent s1.has, agent s2.does_not_have, ?))
				by_reduction: (s1 ∖ s2) ≍ element_to_set.set_reduction (s2, s1, agent {STS_SET [A, EQ]}.without)
				membership: (s1 ∖ s2) ∋ a = (s1 ∋ a and s2 ∌ a)
				is_subset: (s1 ∖ s2) ⊆ s1
				by_complement: s2 ⊆ s1 ⇒ (s1 ∖ s2) ≍ (s2 ∁ s1)
				by_intersection: s1 ⊆ s5 and s2 ⊆ s5 ⇒ (s1 ∖ s2) ≍ (s1 ∩ (s2 ∁ s5))
					u_includes_s2: s2 ⊆ u -- Universe includes everything.
				by_absolute_complement: (s1 ∖ s2) ≍ (s1 ∩ (s2 ∁ u))
				subtract_o: (s1 ∖ o) ≍ s1
				subtract_from_o: (o ∖ s1) ≍ o
				subtract_u: (s1 ∖ u) ≍ o
					u_includes_s1: s1 ⊆ u -- Universe includes everything.
				subtract_from_u: (u ∖ s1) ≍ (s1 ∁ u)
				self_difference: (s1 ∖ s1) ≍ o
				subtract_complement: (s1 ∖ (s1 ∁ u)) ≍ s1
					u_includes_difference: (s1 ∖ s2) ⊆ u -- Universe includes everything.
				complemented_difference: ((s1 ∖ s2) ∁ u) ≍ ((s1 ∁ u) ∪ s2)
				reverted_difference: (s1 ∖ s2) ≍ ((s2 ∁ u) ∖ (s1 ∁ u))
				non_left_associative: ((s1 ∖ s2) ∖ s3) ≍ (s1 ∖ (s2 ∪ s3))
				non_right_associative: (s1 ∖ (s2 ∖ s3)) ≍ ((s1 ∖ s2) ∪ (s1 ∩ s3))
				no_union_distribution: (s1 ∪ (s2 ∖ s3)) ≍ ((s1 ∪ s2) ∖ (s3 ∖ s1))
				intersection_distributed: (s1 ∩ (s2 ∖ s3)) ≍ ((s1 ∩ s2) ∖ (s1 ∩ s3))
				no_left_distribution_over_union: (s1 ∖ (s2 ∪ s3)) ≍ ((s1 ∖ s2) ∩ (s1 ∖ s3))
				no_left_distribution_over_intersection: (s1 ∖ (s2 ∩ s3)) ≍ ((s1 ∖ s2) ∪ (s1 ∖ s3))
				right_distribution_over_union: ((s1 ∪ s2) ∖ s3) ≍ ((s1 ∖ s3) ∪ (s2 ∖ s3))
				right_distribution_over_intersection: ((s1 ∩ s2) ∖ s3) ≍ ((s1 ∖ s3) ∩ (s2 ∖ s3))
				subtracted_difference: (s1 ∖ (s1 ∖ s2)) ≍ (s1 ∩ s2)
				subtracted_from_difference: ((s1 ∖ s2) ∖ s2) ≍ (s1 ∖ s2)
				subset_of_union: (s1 ∖ s2) ⊆ (s1 ∪ s2)
				equality: s1 ≍ s2 = (s1 ∖ s2) ≍ (s2 ∖ s1)
				inclusion: s1 ⊆ s2 = ((s1 ∖ s2) ≍ o)
				disjointness: (s1 ∩ s2) ≍ o = (s1 ∖ s2) ≍ s1
				mutual_complement: s1 ⊆ s2 and s3 ≍ (s2 ∖ s1) ⇒ s1 ≍ (s2 ∖ s3)
				disjoint_complement: (s1 ∩ s2) ≍ o and (s1 ∪ s2) ≍ s3 ⇒ s1 ≍ (s3 ∖ s2)
				disjoint_to_right: ((s1 ∖ s2) ∩ s2) ≍ o
				united_to_right: ((s1 ∖ s2) ∪ s2) ≍ (s1 ∪ s2)
				right_subtracted_from_union: ((s1 ∪ s2) ∖ s2) ≍ (s1 ∖ s2)
				right_subtracted_from_intersection: ((s1 ∩ s2) ∖ s2) ≍ o
				left_minus_intersection: (s1 ∖ (s1 ∩ s2)) ≍ (s1 ∖ s2)
				left_intersected_with_difference: (s1 ∩ (s1 ∖ s2)) ≍ (s1 ∖ s2)
				disjoint_intersection_and_difference: ((s1 ∩ s2) ∩ (s1 ∖ s2)) ≍ o
				disjoint_mutual_differences: ((s1 ∖ s2) ∩ (s2 ∖ s1)) ≍ o
				symmetric_difference: ((s1 ∖ s2) ∪ (s2 ∖ s1)) ≍ ((s1 ∪ s2) ∖ (s1 ∩ s2))
				intersected_differences: ((s1 ∖ s2) ∩ (s3 ∖ s4)) ≍ ((s1 ∩ s3) ∖ (s2 ∪ s4))
			end
			Result := True
		end

	subtracted_symmetricaly_ok (a: A; s1, s2, s3: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.subtracted_symmetricaly?
		do
			check
				u: attached current_universe as u
				definition: (s1 ⊖ s2) ≍ (u | agent u.xored (agent s1.has, agent s2.has, ?))
				by_reduction: (s1 ⊖ s2) ≍ element_to_set.set_reduction (
					s1, s2, agent (acc: STS_SET [A, EQ]; x: A): STS_SET [A, EQ]
						do
							if acc ∋ x then
								Result := acc / x
							else
								Result := acc & x
							end
						end
					)
				by_alternate_reduction: (s1 ⊖ s2) ≍ element_to_set.set_reduction (
					s2, s1, agent  (acc: STS_SET [A, EQ]; x: A): STS_SET [A, EQ]
						do
							if acc ∋ x then
								Result := acc / x
							else
								Result := acc & x
							end
						end
					)
				symmetric_definition: (s1 ⊖ s2) ≍ ((s1 ∖ s2) ∪ (s2 ∖ s1))
				union_minus_intersection: (s1 ⊖ s2) ≍ ((s1 ∪ s2) ∖ (s1 ∩ s2))
				membership: (s1 ⊖ s2) ∋ a = (s1 ∋ a xor s2 ∋ a)
				subtract_o: (s1 ⊖ o) ≍ s1
					u_includes_s1: s1 ⊆ u -- Universe includes everything.
				subtract_u: (s1 ⊖ u) ≍ (s1 ∁ u)
				subtract_complement: (s1 ⊖ (s1 ∁ u)) ≍ u
				self_symmetric_difference: (s1 ⊖ s1) ≍ o
				commutative: (s1 ⊖ s2) ≍ (s2 ⊖ s1)
						 -- Universe includes everything.
					u_includes_s2: s2 ⊆ u
					u_includes_symmetric_difference: (s1 ⊖ s2) ⊆ u
				complemented: ((s1 ⊖ s2) ∁ u) ≍ ((s1 ∩ s2) ∪ ((s1 ∁ u) ∩ (s2 ∁ u)))
				associative: ((s1 ⊖ s2) ⊖ s3) ≍ (s1 ⊖ (s2 ⊖ s3))
				distributive_intersection: (s1 ∩ (s2 ⊖ s3)) ≍ ((s1 ∩ s2) ⊖ (s1 ∩ s3))
				non_distributive_union: (s1 ∪ (s2 ⊖ s3)) ≍ ((s1 ∪ s2 ∪ s3) ∖ ((s1 ∁ u) ∩ s2 ∩ s3))
				distributive_right_difference: ((s1 ⊖ s2) ∖ s3) ≍ ((s1 ∖ s3) ⊖ (s2 ∖ s3))
					u_includes_s3: s3 ⊆ u -- Universe includes everything.
				non_distributive_left_difference: (s1 ∖ (s2 ⊖ s3)) ≍ ((s1 ∩ s2 ∩ s3) ∪ (s1 ∩ (s2 ∁ u) ∩ (s3 ∁ u)))
				symdiff_of_complements: (s1 ⊖ s2) ≍ ((s1 ∁ u) ⊖ (s2 ∁ u))
				simply_complemented: ((s1 ⊖ s2) ∁ u) ≍ (s1 ⊖ (s2 ∁ u))
				difference: (s1 ∖ s2) ≍ (s1 ⊖ (s1 ∩ s2))
				disjoint_intersection: ((s1 ∩ s2) ∩ (s1 ⊖ s2)) ≍ o
				union: ((s1 ⊖ s2) ⊖ (s1 ∩ s2)) ≍ (s1 ∪ s2)
				minus_something: ((s1 ∪ s3) ⊖ (s2 ∪ s3)) ≍ ((s1 ⊖ s2) ∖ s3)
				subset_of_union: (s1 ⊖ s2) ⊆ (s1 ∪ s2)
				includes_difference: (s1 ∖ s2) ⊆ (s1 ⊖ s2)
				equality: (s1 ⊖ s2) ≍ o = s1 ≍ s2
				regular: (s1 ⊖ s3) ≍ (s2 ⊖ s3) ⇒ s1 ≍ s2
				equal_unions: (s1 ∪ s3) ≍ (s2 ∪ s3) = ((s1 ⊖ s2) ⊆ s3)
			end
			Result := True
		end

	crossed_ok (s1, s2, s3: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.crossed?
		do
			check
				u: attached current_universe as u
				n1: attached (# s1) as n1
				n2: attached (# s2) as n2
				n3: attached (# s3) as n3
				nu: attached (# u) as nu
				print_cardinalities: (
					agent (ia_n1, ia_n2, ia_n3, ia_nu: NATURAL): BOOLEAN
						do
							print (ia_n1) print (", ")
							print (ia_n2) print (", ")
							print (ia_n2) print (", ")
							print (ia_nu) print ('%N')
							Result := True
						end (n1, n2, n3, nu)
					).item
				definition: (nu ^ 2) ≤ max_elements ⇒ (s1 × s2) ≍ (× u).filtered_xy (
					agent predicate_holds_for_pair (agent (u.element_to_element).anded (agent s1.has, ?, agent s2.has, ?), ?, ?)
					)
				by_reduction: ((n1 × n2) ^ 2) ≤ max_elements ⇒ (s1 × s2) ≍ element_to_pairs.set_reduction (
					s1, (× o), agent (acc: STS_SET [STS_ORDERED_PAIR [A, A, EQ, EQ], STS_ORDERED_PAIR_EQUALITY [A, A, EQ, EQ]]; x: A; ia_s2: STS_SET [A, EQ]):
						STS_SET [STS_ORDERED_PAIR [A, A, EQ, EQ], STS_ORDERED_PAIR_EQUALITY [A, A, EQ, EQ]]
						do
							Result := acc ∪ element_to_pair.set_map (ia_s2, agent pair (x, ?))
						end (?, ?, s2)
					)
				by_alternate_reduction: ((n1 × n2) ^ 2) ≤ max_elements ⇒ (s1 × s2) ≍ element_to_pairs.set_reduction (
					s2, (× o), agent (ia_s1: STS_SET [A, EQ]; acc: STS_SET [STS_ORDERED_PAIR [A, A, EQ, EQ], STS_ORDERED_PAIR_EQUALITY [A, A, EQ, EQ]]; y: A):
						STS_SET [STS_ORDERED_PAIR [A, A, EQ, EQ], STS_ORDERED_PAIR_EQUALITY [A, A, EQ, EQ]]
						do
							Result := acc ∪ element_to_pair.set_map (ia_s1, agent pair (?, y))
						end (s1, ?, ?)
					)
				non_commutative: (n1 × n2) ≤ max_elements and s1 ≭ s2 and s1 ≭ o and s2 ≭ o ⇒ (s1 × s2) ≭ (s2 × s1)
				commutativity_requisites: (n1 × n2) ≤ max_elements ⇒ (s1 × s2 ≍ (s2 × s1) = (s1 ≍ o or s2 ≍ o or s1 ≍ s2))
				product_minus_square: (((n1 × n2) ∨ (n3 × n3)) ^ 2) ≤ max_elements ⇒ ((s1 × s2) ∖ (s3 × s3)) ≍ (((s1 ∖ s3) × s2) ∪ (s1 × (s2 ∖ s3)))
				square_minus_product: (((n1 × n1) ∨ (n2 × n3)) ^ 2) ≤ max_elements ⇒ ((s1 × s1) ∖ (s2 × s3)) ≍ (((s1 ∖ s2) × s1) ∪ (s1 × (s1 ∖ s3)))
				cardinality: # (s1 × s2) = # s1 * # s2
			end
			Result := True
		end

	squared_ok (s: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.squared?
		do
			check
				u: attached current_universe as u
				definition: (# u ^ 2) ≤ max_elements ⇒ (× s) ≍
					(× u).filtered_xy (agent predicate_holds_for_pair (agent (u.element_to_element).anded (agent s.has, ?, agent s.has, ?), ?, ?))
				by_reduction: (# s ^ 2) ≤ max_elements ⇒ (× s) ≍ element_to_pairs.set_reduction (
					s, (× o), agent (acc: STS_SET [STS_ORDERED_PAIR [A, A, EQ, EQ], STS_ORDERED_PAIR_EQUALITY [A, A, EQ, EQ]]; x: A; ia_s: STS_SET [A, EQ]):
						STS_SET [STS_ORDERED_PAIR [A, A, EQ, EQ], STS_ORDERED_PAIR_EQUALITY [A, A, EQ, EQ]]
						do
							Result := acc ∪ element_to_pair.set_map (ia_s, agent pair (x, ?))
						end (?, ?, s)
					)
				by_alternate_reduction: (# s ^ 2) ≤ max_elements ⇒ (× s) ≍ element_to_pairs.set_reduction (
					s, (× o), agent (ia_s: STS_SET [A, EQ]; acc: STS_SET [STS_ORDERED_PAIR [A, A, EQ, EQ], STS_ORDERED_PAIR_EQUALITY [A, A, EQ, EQ]]; y: A):
						STS_SET [STS_ORDERED_PAIR [A, A, EQ, EQ], STS_ORDERED_PAIR_EQUALITY [A, A, EQ, EQ]]
						do
							Result := acc ∪ element_to_pair.set_map (ia_s, agent pair (?, y))
						end (s, ?, ?)
					)
				cardinality: # (× s) = (# s) ^ 2
			end
			Result := True
		end

	diagonal_ok (s: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.diagonal?
		do
			check
				u: attached current_universe as u
				definition: (# u ^ 2) ≤ max_elements ⇒ ∆ s ≍ (× u).filtered_xy (
					agent predicate_holds_for_pair (
						agent  (ia_s: STS_SET [A, EQ]; x, y: A): BOOLEAN
							do
								Result := ia_s ∋ x and eq (x, y)
							end (s, ?, ?), ?, ?
						)
					)
				by_mapping: ∆ s ≍ element_to_pair.set_map (
					s, agent  (x: A): STS_ORDERED_PAIR [A, A, EQ, EQ]
						do
							Result := pair (x, x)
						end
					)
				cardinality: # (∆ s) = # s
			end
			Result := True
		end

	raised_ok (s: STS_SET [A, EQ]; n: NATURAL): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.raised?
		do
			check
				u: attached current_universe as u
				definition: ((# u) ^ n) ≤ max_tuples ⇒ (s ^ n) ≍ (
					(u ^ n) | agent (t: STS_N_TUPLE [A, EQ] ia_s: STS_SET [A, EQ]): BOOLEAN
						do
							Result := t.terms ⊆ ia_s
						end (?, s)
					)
				cardinality: ((# s) ^ n) ≤ max_tuples ⇒ # (s ^ n) = (# s) ^ n
			end
			Result := True
		end

	crossed_tuples_ok (
		s: STS_SET [A, EQ]
		ts: STS_SET [STS_N_TUPLE [A, EQ], STS_N_TUPLE_EQUALITY [A, EQ]]
		): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.crossed_tuples?
		do
			check
				u: attached current_universe as u
				ns: attached tuple_to_natural.set_map (
					ts, agent (t: STS_N_TUPLE [A, EQ]): NATURAL
						do
							Result := t.n + 1
						end
					) as ns
				m: attached natural_to_real_64.set_reduction (
					ns, 0, agent (ia_u: STS_SET [A, EQ]; acc: REAL_64; n: NATURAL): REAL_64
						do
							Result := acc + (# ia_u) ^ n
						end (u, ?, ?)
					) as m
				print_m: (
					agent (ia_m: DOUBLE): BOOLEAN
						do
							print (ia_m) print ('%N')
							Result := True
						end (m)
					).item
				definition: m ≤ max_tuples ⇒ s.crossed_tuples (ts) ≍ (
					natural_to_tuple_set.set_reduction (
						ns, ts.o, agent (
							ia_u: STS_SET [A, EQ]
							acc: STS_SET [STS_N_TUPLE [A, EQ], STS_N_TUPLE_EQUALITY [A, EQ]]
							n: NATURAL
							): STS_SET [STS_N_TUPLE [A, EQ], STS_N_TUPLE_EQUALITY [A, EQ]]
							do
								Result := acc ∪ (ia_u ^ n)
							end (u, ?, ?)
						) | agent
							(t: STS_N_TUPLE [A, EQ]; ia_s: STS_SET [A, EQ]; ia_ts: STS_SET [STS_N_TUPLE [A, EQ], STS_N_TUPLE_EQUALITY [A, EQ]]): BOOLEAN
							do
								Result := (not t.is_unit and then ia_s ∋ t.first) and ia_ts ∋ t.tail
							end (?, s, ts)
					)
				cardinality: # s.crossed_tuples (ts) = (# s) * (# ts)
			end
			Result := True
		end

feature -- Properties (Transformation)

	mapped_ok (s: STS_SET [A, EQ]; f: FUNCTION [A, A]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.mapped?
		do
			check
				u: attached current_universe as u
				definition: (s ↦ f) ≍ (
					u | agent (ia_y: A ia_s: STS_SET [A, EQ] ia_f: FUNCTION [A, A]): BOOLEAN
						do
							Result := ia_s |∃ agent (x, y: A i2a_f: FUNCTION [A, A]): BOOLEAN
								do
									Result := eq (y, i2a_f (x))
								end (?, ia_y, ia_f)
						end (?, s, f)
					)
				by_reduction: (s ↦ f) ≍ element_to_set.set_reduction (
					s, o, agent (acc: STS_SET [A, EQ] x: A; ia_f: FUNCTION [A, A]): STS_SET [A, EQ]
						do
							Result := acc & ia_f (x)
						end (?, ?, f)
					)
				cardinality: # (s ↦ f) ≤ # s
			end
			Result := True
		end

	proper_reduced_ok (s: STS_SET [A, EQ]; f: FUNCTION [A, A, A]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.proper_reduced?
		do
			if not s.is_empty then
				check
						is_not_unit: not s.as_tuple.is_unit -- not s.is_empty
					definition: eq (s.proper_reduced (f), s.as_tuple.proper_left_reduced (f))
				end
			end
			Result := True
		end

feature -- Properties (Conversion)

	as_singleton_ok (s: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.as_singleton?
		do
			check
				u: attached current_universe as u
				power_card: attached (2 ^ # u) as power_card
				definition: power_card ≤ max_subsets ⇒ s.as_singleton ≍ (u.powerset | agent s.equals)
				by_singleton: power_card ≤ max_subsets ⇒ s.as_singleton ≍ u.powerset.singleton (s)
			end
			Result := True
		end

	n_times_ok (s: STS_SET [A, EQ]; n: NATURAL): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.n_times?
		do
			check
				only_s: s.n_times (n).for_all (
					agent (i: NATURAL; xs, ia_s: STS_SET [A, EQ]): BOOLEAN
						do
--							Result := xs ≍ ia_s
							Result := xs ≍ ia_s
						end (?, ?, s)
					)
			end
			Result := True
		end

	as_tuple_ok (s: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.as_tuple?
		do
			check
				u: attached current_universe as u
				definition: (# u ^ # s) ≤ max_tuples ⇒ s.as_tuple.as_singleton ≍ (
					(u ^ (# s)) | agent (t: STS_N_TUPLE [A, EQ]; ia_s: STS_SET [A, EQ]): BOOLEAN
						do
							Result := t.for_all (
								agent (i: NATURAL x_i: A i2a_s: STS_SET [A, EQ]): BOOLEAN
									local
										i_th_element: FUNCTION [NATURAL, STS_SET [A, EQ], A]
									do
										i_th_element := agent
											(ia_i: NATURAL; i3a_s: STS_SET [A, EQ]; ia_i_th_element: FUNCTION [NATURAL, STS_SET [A, EQ], A]): A
											do
												if ia_i = 1 then
														check
															is_not_empty: not i3a_s.is_empty -- 0 < ia_i ≤ # i3a_s
														end
													Result := i3a_s.any
												else
													Result := ia_i_th_element (ia_i - 1, i3a_s.others, ia_i_th_element)
												end
											end
										Result := eq (x_i, i_th_element (i, i2a_s, i_th_element))
									end (?, ?, ia_s)
								)
						end (?, s)
					)
				n: s.as_tuple.n = # s
				terms: s.as_tuple.terms ≍ s
			end
			Result := True
		end

feature -- Properties (Factory)

	o_ok (s: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.o?
			--| To assess the relatioship between `o_ok' and `empty_set_ok', please see the note at the header of `subsets_ok'.
		do
			check
				u: attached current_universe as u
				definition: s.o ≍ (
					u | agent (x: A): BOOLEAN
						do
							Result := False
						end
					)
				has_nothing: u |∀ agent (s.o).does_not_have
				uniqueness: s.o ≍ o
				no_element: # s.o = 0
			end
			Result := True
		end

	empty_set_ok (s: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.empty_set?
			--| Please see the comment at `o_ok' header.
		do
			check
				u: attached current_universe as u
				definition: s.empty_set ≍ (
					u | agent (x: A): BOOLEAN
						do
							Result := False
						end
					)
				has_nothing: u |∀ agent (s.empty_set).does_not_have
				uniqueness: s.empty_set ≍ o
				no_element: # s.empty_set = 0
			end
			Result := True
		end

	singleton_ok (s: STS_SET [A, EQ]; a: A): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.singleton?
		do
			check
				u: attached current_universe as u
				definition: s.singleton (a) ≍ (u | agent u.equality_holds (?, a))
				exists_unique: s.singleton (a) |∃! agent s.equality_holds (?, a)
				nothing_else: s.singleton (a) |∀ agent s.equality_holds (?, a)
				is_not_empty: not s.singleton (a).is_empty
				any: eq (s.singleton (a).any, a)
				others: s.singleton (a).others ≍ o
				cardinality: # s.singleton (a) = 1
			end
			Result := True
		end

	unit_ok (s: STS_SET [A, EQ]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SET}.unit?
		do
			check
				u: attached current_universe as u
				definition: s.unit.as_singleton ≍ (u ^ 0)
				has_nothing: u |∀ agent negated (agent (s.unit).has, ?)
				uniqueness: s.unit ≍ unit
				no_term: s.unit.n = 0
				empty_terms: s.unit.terms ≍ o
			end
			Result := True
		end

feature -- Factory

	singleton (a: A): STS_SET [A, EQ]
			-- Singleton in the form {`a'}
		do
			Result := o.singleton (a)
		ensure
			definition: Result ≍ o.singleton (a)
		end

	paired_sets (s1, s2: STS_SET [A, EQ]): STS_SET [STS_SET [A, EQ], STS_SET_EQUALITY [A, EQ]]
			-- The sets `s1' and `s2' put into a set, i.e. {`s1', `s2'}
		do
			Result := s1.as_singleton & s2
		ensure
			definition: Result ≍ (s1.as_singleton & s2)
		end

	pair (x, y: A): STS_ORDERED_PAIR [A, A, EQ, EQ]
			-- The ordered pair (`x', `y')
		deferred
		ensure
			first: eq (Result.a, x)
			second: eq (Result.b, y)
		end

feature -- Predicate

	negated (p: PREDICATE [A]; x: A): BOOLEAN
			-- Logical negation of `p' (`x'), i.e. is `p' (`x') false?
		do
			Result := not p (x)
		ensure
			class
			definition: Result = not p (x)
		end

	anded (p1, p2: PREDICATE [A]; x: A): BOOLEAN
			-- Do `p1' (`x') and `p2' (`x') hold?
		do
			Result := p1 (x) and p2 (x)
		ensure
			class
			definition: Result = (p1 (x) and p2 (x))
		end

	implied (p1, p2: PREDICATE [A]; x: A): BOOLEAN
			-- If `p1' (`x') holds, does `p2' (`x') hold too, i.e. does `p1' (`x') imply `p2' (`x')?
		do
			Result := p1 (x) ⇒ p2 (x)
		ensure
			class
			definition: Result = (p1 (x) ⇒ p2 (x))
		end

	predicate_holds_for_pair (p: PREDICATE [A, A]; x1, x2: A): BOOLEAN
			-- Does `p' (`x1', `x2') hold?
		do
			Result := p (x1, x2)
		ensure
			class
			definition: Result = p (x1, x2)
		end

	pair_negated (p: PREDICATE [A, A]; x, y: A): BOOLEAN
			-- Logical negation of `p' (`x', `y'), i.e. is `p' (`x', `y') false?
		do
			Result := not p (x, y)
		ensure
			class
			definition: Result = not p (x, y)
		end

	pair_anded (p1, p2: PREDICATE [A, A]; x, y: A): BOOLEAN
			-- Do `p1' (`x', `y') and `p2' (`x', `y') hold?
		do
			Result := p1 (x, y) and p2 (x, y)
		ensure
			class
			definition: Result = (p1 (x, y) and p2 (x, y))
		end

	pair_implied (p1, p2: PREDICATE [A, A]; x, y: A): BOOLEAN
			-- If `p1' (`x', `y') holds, does `p2' (`x', `y') hold too, i.e. does `p1' (`x', `y') imply `p2' (`x', `y')?
		do
			Result := p1 (x, y) ⇒ p2 (x, y)
		ensure
			class
			definition: Result = (p1 (x, y) ⇒ p2 (x, y))
		end

	equality_does_not_hold (x1, x2: A): BOOLEAN
			-- Does not `eq' (`x1', `x2') hold?
		do
			Result := not eq (x1, x2)
		ensure
			definition: Result = not eq (x1, x2)
		end

feature -- Transformer

	set_to_set: STS_TRANSFORMER [STS_SET [A, EQ], STS_SET [A, EQ], STS_EQUALITY [STS_SET [A, EQ]], STS_SET_EQUALITY [A, EQ]]
			-- Transformer between types derived from {STS_SET [A, EQ]}
		deferred
		end

	element_to_pair: STS_TRANSFORMER [A, STS_ORDERED_PAIR [A, A, EQ, EQ], EQ, STS_ORDERED_PAIR_EQUALITY [A, A, EQ, EQ]]
			-- Transformer between types derived from {A} and {STS_ORDERED_PAIR [A, A, EQ, EQ]}
		deferred
		end

	element_to_pairs: STS_TRANSFORMER [
		A, STS_SET [STS_ORDERED_PAIR [A, A, EQ, EQ], STS_ORDERED_PAIR_EQUALITY [A, A, EQ, EQ]],
		EQ, STS_SET_EQUALITY [STS_ORDERED_PAIR [A, A, EQ, EQ], STS_ORDERED_PAIR_EQUALITY [A, A, EQ, EQ]]
		]
			-- Transformer between types derived from {A} and {STS_SET [STS_ORDERED_PAIR [A, A, EQ, EQ], STS_PAIR_EQUALITY [A, A, EQ, EQ]]}
		deferred
		end

	element_to_boolean: STS_TRANSFORMER [A, BOOLEAN, EQ, STS_OBJECT_EQUALITY [BOOLEAN]]
			-- Transformer between types derived from {A} and {BOOLEAN}
		deferred
		end

	tuple_to_natural: STS_TRANSFORMER [STS_N_TUPLE [A, EQ], NATURAL, STS_N_TUPLE_EQUALITY [A, EQ], STS_OBJECT_EQUALITY [NATURAL]]
			-- Transformer between types derived from {STS_N_TUPLE [A, EQ]} and {NATURAL}
		deferred
		end

	natural_to_real_64: STS_TRANSFORMER [NATURAL, REAL_64, STS_OBJECT_EQUALITY [NATURAL], STS_OBJECT_EQUALITY [REAL_64]]
			-- Transformer between types derived from {NATURAL} and {REAL_64}
		deferred
		end

	natural_to_tuple_set: STS_TRANSFORMER [
		NATURAL,
		STS_SET [STS_N_TUPLE [A, EQ], STS_N_TUPLE_EQUALITY [A, EQ]],
		STS_OBJECT_EQUALITY [NATURAL],
		STS_SET_EQUALITY [STS_N_TUPLE [A, EQ], STS_N_TUPLE_EQUALITY [A, EQ]]
		]
			-- Transformer between types derived from {NATURAL} and {STS_SET [STS_N_TUPLE [A, EQ], STS_N_TUPLE_EQUALITY [A, EQ]]}
		deferred
		end

	element_to_set: STS_TRANSFORMER [A, STS_SET [A, EQ], EQ, STS_SET_EQUALITY [A, EQ]]
			-- Transformer between types derived from {A} and {STS_SET [A, EQ]}
		deferred
		end

	element_to_sets: STS_TRANSFORMER [
		A, STS_SET [STS_SET [A, EQ], STS_SET_EQUALITY [A, EQ]],
		EQ, STS_SET_EQUALITY [STS_SET [A, EQ], STS_SET_EQUALITY [A, EQ]]
		]
			-- Transformer between types derived from {A} and {STS_SET [STS_SET [A, EQ], STS_SET_EQUALITY [A, EQ]]}
		deferred
		end

	tuple_to_set: STS_TRANSFORMER [
		STS_N_TUPLE [A, EQ], STS_SET [A, EQ],
		STS_N_TUPLE_EQUALITY [A, EQ], STS_SET_EQUALITY [A, EQ]
		]
			-- Transformer between types derived from {STS_N_TUPLE [A, EQ]} and {STS_SET [A, EQ]}
		deferred
		end

feature {NONE} -- Anchor

	universe_anchor: STS_SET [A, EQ]
			-- Anchor for `current_universe'
		deferred
		end

feature {NONE} -- Implementation

	natural_anchor: NATURAL
			-- Anchor for natural numbers
			--| TODO: Pull it up to a target-dependant class.
		do
		ensure
			class
		end

	max_asserted_elements: like natural_anchor
			-- Maximum number of elements be checked with assertions enabled
		deferred
		end

	max_elements: like natural_anchor
			-- Maximum number of elements to be checked within property statements
		deferred
		end

	max_asserted_subsets: like natural_anchor
			-- Maximum number of subsets of a set to be checked with assertions enabled
		deferred
		end

	max_subsets: like natural_anchor
			-- Maximum number of subsets to be checked within property statements
		deferred
		end

	max_tuples: like natural_anchor
			-- Maximum number of n-tuples to be checked within property statements
		deferred
		end

note
	copyright: "Copyright (c) 2012-2022, Rosivaldo Fernandes Alves"
	license: "[
		Eiffel Forum License v2
		(see http://www.eiffel.com/licensing/forum.txt)
		]"
	source: "https://launchpad.net/set-theory"

end
