﻿note
	description: "Object that checks whether the properties verified within set theory hold for an implementation of {STS_TRANSFORMABLE_SET}"
	author: "Rosivaldo Fernandes Alves"
	date: "$Date: Sat 2018-12-01 19:25:58 -0300 $"
	revision: "$Revision$"

deferred class
	TRANSFORMABLE_SET_PROPERTIES [A, B, EQ_A -> STS_EQUALITY [A], EQ_B -> STS_EQUALITY [B]]

inherit
	SET_PROPERTIES [A, EQ_A]
		rename
			o as o_a,
			eq as eq_a,
			current_universe as current_universe_a,
			equals_ok as sp_equals_ok,
			is_subset_ok as sp_is_subset_ok,
			is_disjoint_ok as sp_is_disjoint_ok,
			united_ok as sp_united_ok,
			intersected_ok as sp_intersected_ok,
			subtracted_ok as sp_subtracted_ok,
			subtracted_symmetricaly_ok as sp_subtracted_symmetricaly_ok,
			crossed_ok as crossed_with_same_kind_ok,
			mapped_ok as mapped_to_same_kind_ok,
			element_to_set as element_to_set_of_same_kind,
			universe_anchor as universe_a_anchor
		end

feature -- Access

	o_b: like current_universe_b.subset_anchor
			-- The empty set of elements like {B}
		deferred
		ensure
			is_empty: Result.is_empty
		end

	eq_b: EQ_B
			-- Equality for objects like {B}
		deferred
		end

	o_ab: like universe_a_anchor.crossed
			-- The empty set in the form `o_a' × `o_b'
		deferred
		ensure
			is_empty: Result.is_empty
		end

	current_universe_b: STS_SET [B, EQ_B]
			-- The current "universe" of elements like {B}, i.e. set with every object currently in system memory whose type conforms to {B}.
			-- Notice that this "universe" may change from a call to another.
		deferred
		end

feature -- Properties (Comparison)

	equals_ok (sa_1, sa_2: STS_TRANSFORMABLE_SET [A, B, EQ_A, EQ_B]; sb: STS_SET [B, EQ_B]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SUBSET_FAMILY_STRUCTURE}.equals?
		do
			check
				implied_equality: sb ≭ o_b and sa_1 × sb ≍ (sa_2 × sb) implies sa_1 ≍ sa_2
			end
			Result := True
		end

	is_subset_ok (sa_1, sa_2: STS_TRANSFORMABLE_SET [A, B, EQ_A, EQ_B]; sb_1: STS_TRANSFORMABLE_SET [B, A, EQ_B, EQ_A]; sb_2: STS_SET [B, EQ_B]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SUBSET_FAMILY_STRUCTURE}.is_subset?
		do
			check
				induced_product_inclusion: sa_1 ⊆ sa_2 implies sa_1 × sb_1 ⊆ (sa_2 × sb_1)
				induced_inverse_product_inclusion: sa_1 ⊆ sa_2 implies sb_1 × sa_1 ⊆ (sb_1 × sa_2)
				implied_inclusion: sb_1 ≭ o_b and sa_1 × sb_1 ⊆ (sa_2 × sb_1) implies sa_1 ⊆ sa_2
				inversely_implied_inclusion: sb_1 ≭ o_b and sb_1 × sa_1 ⊆ (sb_1 × sa_2) implies sa_1 ⊆ sa_2
				simultaneous_inclusion: sa_1 ≭ o_a and sb_1 ≭ o_b implies (sa_1 ⊆ sa_2 and sb_1 ⊆ sb_2) = (sa_1 × sb_1 ⊆ (sa_2 × sb_2))
			end
			Result := True
		end

	is_disjoint_ok (sa_1, sa_2: STS_TRANSFORMABLE_SET [A, B, EQ_A, EQ_B]; sb: STS_SET [B, EQ_B]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SUBSET_FAMILY_STRUCTURE}.is_disjoint?
		do
			check
				induced_disjointness: sb ≭ o_b implies sa_1.is_disjoint (sa_2) = (sa_1 × sb).is_disjoint (sa_2 × sb)
			end
			Result := True
		end

feature -- Properties (Operation)

	united_ok (sa_1, sa_2: STS_TRANSFORMABLE_SET [A, B, EQ_A, EQ_B]; sb: STS_TRANSFORMABLE_SET [B, A, EQ_B, EQ_A]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SUBSET_FAMILY_STRUCTURE}.united?
		do
			check
				distributes_right_product: (sa_1 ∪ sa_2) × sb ≍ ((sa_1 × sb) ∪ (sa_2 × sb))
				distributes_left_product: sb × (sa_1 ∪ sa_2) ≍ ((sb × sa_1) ∪ (sb × sa_2))
			end
			Result := True
		end

	intersected_ok (sa_1, sa_2: STS_TRANSFORMABLE_SET [A, B, EQ_A, EQ_B]; sb_1: STS_TRANSFORMABLE_SET [B, A, EQ_B, EQ_A]; sb_2: STS_SET [B, EQ_B]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SUBSET_FAMILY_STRUCTURE}.intersected?
		do
			check
				distributes_right_product: (sa_1 ∩ sa_2) × sb_1 ≍ ((sa_1 × sb_1) ∩ (sa_2 × sb_1))
				distributes_left_product: sb_1 × (sa_1 ∩ sa_2) ≍ ((sb_1 × sa_1) ∩ (sb_1 × sa_2))
				intersected_products: ((sa_1 × sb_1) ∩ (sa_2 × sb_2)) ≍ ((sa_1 ∩ sa_2) × (sb_1 ∩ sb_2))
				crossed_intersections: (sa_1 ∩ sa_2) × (sb_1 ∩ sb_2) ≍ ((sa_1 × sb_1) ∩ (sa_2 × sb_2))
			end
			Result := True
		end

	subtracted_ok (sa_1, sa_2: STS_TRANSFORMABLE_SET [A, B, EQ_A, EQ_B]; sb: STS_TRANSFORMABLE_SET [B, A, EQ_B, EQ_A]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SUBSET_FAMILY_STRUCTURE}.subtracted?
		do
			check
				distributes_right_product: (sa_1 ∖ sa_2) × sb ≍ ((sa_1 × sb) ∖ (sa_2 × sb))
				distributes_left_product: sb × (sa_1 ∖ sa_2) ≍ ((sb × sa_1) ∖ (sb × sa_2))
			end
			Result := True
		end

	subtracted_symmetricaly_ok (sa_1, sa_2: STS_TRANSFORMABLE_SET [A, B, EQ_A, EQ_B]; sb: STS_TRANSFORMABLE_SET [B, A, EQ_B, EQ_A]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SUBSET_FAMILY_STRUCTURE}.subtracted_symmetricaly?
		do
			check
				distributes_right_product: (sa_1 ⊖ sa_2) × sb ≍ ((sa_1 × sb) ⊖ (sa_2 × sb))
				distributes_left_product: sb × (sa_1 ⊖ sa_2) ≍ ((sb × sa_1) ⊖ (sb × sa_2))
			end
			Result := True
		end

	crossed_ok (sa: STS_TRANSFORMABLE_SET [A, B, EQ_A, EQ_B]; sb: STS_SET [B, EQ_B]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SUBSET_FAMILY_STRUCTURE}.crossed?
		do
			check
				u_a: attached current_universe_a as u_a
				u_b: attached current_universe_b as u_b
				definition: sa × sb ≍ (u_a × u_b).filtered_xy (agent (u_a.transformer).anded (agent sa.has, ?, agent sb.has, ?))
				cardinality: # (sa × sb) = # sa * # sb
				emptiness: (sa × sb) ≍ o_ab = (sa ≍ o_a or sb ≍ o_b)
			end
			Result := True
		end

feature -- Properties (Transformation)

	mapped_ok (s: STS_TRANSFORMABLE_SET [A, B, EQ_A, EQ_B] f: FUNCTION [A, B]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SUBSET_FAMILY_STRUCTURE}.mapped?
		do
			check
				u_b: attached current_universe_b as u_b
				print_u_b: (
					agent (ia_u_b: like current_universe_b): BOOLEAN
						do
							print (ia_u_b) print ('%N')
							Result := True
						end (u_b)
					).item

				definition: (s ↦ f) ≍ (
					u_b | agent (ia_y: B ia_s: STS_SET [A, EQ_A] ia_f: FUNCTION [A, B]): BOOLEAN
						do
							Result := ia_s |∃ agent (x: A y: B i2a_f: FUNCTION [A, B]): BOOLEAN
								do
									Result := eq_b (y, i2a_f (x))
								end (?, ia_y, ia_f)
						end (?, s, f)
					)
				by_reduction: (s ↦ f) ≍ element_to_set.set_reduction (
					s, o_b, agent (acc: STS_SET [B, EQ_B] x: A; ia_f: FUNCTION [A, B]): STS_SET [B, EQ_B]
						do
							Result := acc & ia_f (x)
						end (?, ?, f)
					)
				cardinality: # (s ↦ f) ≤ # s
			end
			Result := True
		rescue
			print (s) print (", ")
			print (s ↦ f) print ('%N')
		end

	reduced_ok (s: STS_TRANSFORMABLE_SET [A, B, EQ_A, EQ_B]; start: B; f: FUNCTION [B, A, B]): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_SUBSET_FAMILY_STRUCTURE}.reduced?
		do
			check
				definition: eq_b (s.reduced (start, f), s.as_tuple.left_reduced (start, f))
			end
			Result := True
		end

feature -- Transformer

	element_to_set: STS_TRANSFORMER [A, STS_SET [B, EQ_B], EQ_A, STS_SET_EQUALITY [B, EQ_B]]
			-- Transformer between types derived from {A} and {STS_SET [B, EQ_B]}
		deferred
		end

feature {NONE} -- Anchor

	universe_a_anchor: STS_TRANSFORMABLE_SET [A, B, EQ_A, EQ_B]
			-- Anchor for `current_universe_a'
		deferred
		end

note
	copyright: "Copyright (c) 2012-2021, Rosivaldo Fernandes Alves"
	license: "[
		Eiffel Forum License v2
		(see http://www.eiffel.com/licensing/forum.txt)
		]"
	source: "https://launchpad.net/set-theory"

end
