﻿note
	description: "[
		Model of a complex number, i.e. a number in the form a+bi, where a and
		b are real numbers and i is the imaginary unit, i.e. i^2 = -1.
		
		This specification tries to comply to the semantics of complex
		arithmetic exposed at 
			Annex G (IEC 60559-compatible complex arithmetic) 
		of 
			ISO/IEC9899:2017 (AKA ISO C17, please see 
			https://en.wikipedia.org/wiki/C17_(C_standard_revision).)
		available at https://tinyurl.com/244jc9uf.
		However, one must keep in mind that this specification does not try to
		offer a high-grade numeric class. Its intent is just to introduce
		examples of how several numeric types and sets may receive a
		set-theoretic treatment.
		
		It is noteworthy that this specification does not inherit from
		{NUMERIC}, since that class uses covariant arguments, which this
		library avoids as much as possible.
		]"
	author: "Rosivaldo Fernandes Alves"
	date: "$Date$"
	revision: "$Revision$"

deferred class
	COMPLEX_NUMBER

inherit
	ELEMENT
		rename
			is_in as element_is_in,
			is_not_in as element_is_not_in
		end

feature -- Primitive

	a,
	real: like real_anchor
			-- Real part of current complex number
		deferred
		end

	b,
	imaginary: like real_anchor
			-- Imaginary part of current complex number
		deferred
		end

feature -- Membership

	is_in alias "∈" (s: SET [COMPLEX_NUMBER, EQUALITY [COMPLEX_NUMBER]]): BOOLEAN
			-- Does `s` have current complex number?
		do
			Result := s ∋ Current
		ensure
			definition: Result = s ∋ Current
		end

	is_not_in alias "∉" (s: SET [COMPLEX_NUMBER, EQUALITY [COMPLEX_NUMBER]]): BOOLEAN
			-- Is not current complex number in `s'?
		do
			Result := not (Current ∈ s)
		ensure
			definition: Result = not (Current ∈ s)
		end

feature -- Access

	zero: like complex_anchor
			-- The complex number 0 + 0i
		deferred
		ensure
			real: Result.a.value = 0
			imaginary: Result.b.value = 0
		end

	one: like complex_anchor
			-- The complex number 1 + 0i
		deferred
		ensure
			real: Result.a.value = 1
			imaginary: Result.b.value = 0
		end

	two: like integer_anchor -- TODO: Natural instead
			-- The integer number 2
			--| Necessary for some specifications and properties
		deferred
		ensure
			definition: Result.value = 2
		end

	i,
	imaginary_unit: like complex_anchor
			-- The imaginary unit, i.e. the complex number 0 + 1i.
		deferred
		ensure
			real: Result.a ≍ zero.a
			imaginary: Result.b ≍ one.a
		end

feature -- Quality

	is_real: BOOLEAN
			-- Is current complex number a purely real number, i.e. is its imaginary part equal to zero?
		do
			Result := b ≍ zero.a
		ensure
			definition: Result = b ≍ zero.a
		end

	is_proper_real: BOOLEAN
			-- Is current complex number a real number unequal to zero?
		do
			Result := is_real and a ≭ zero.a
		ensure
			definition: Result = (is_real and a ≭ zero.a)
		end

	is_imaginary: BOOLEAN
			-- Is current complex number a purely imaginary number, i.e. is its real part equal to zero?
		do
			Result := a ≍ zero.argument
		ensure
			definition: Result = a ≍ zero.a
		end

	is_proper_imaginary: BOOLEAN
			-- Is current complex number an imaginary number unequal to zero?
		do
			Result := is_imaginary and b ≭ zero.a
		ensure
			definition: Result = (is_imaginary and b ≭ zero.a)
		end

	is_proper_complex: BOOLEAN
			-- Is current complex number neither a purely real nor a purely imaginary number?
		do
			Result := not is_real and not is_imaginary
		ensure
			definition: Result = (not is_real and not is_imaginary)
		end

	is_infinite: BOOLEAN
			-- Is current the representation of an infinity, i.e. is either `real' or `imaginary' the representation of an infinity?
			-- Notice that, if one part of current is an inﬁnity, current is an infinity too, even if its other part is a NaN.
		do
			Result := a.is_infinite or b.is_infinite
		ensure
			definition: Result = (a.is_infinite or b.is_infinite)
		end

	is_finite: BOOLEAN
			-- Is current the representation of a finite complex number, i.e. are both parts of it neither infinite nor NaN?
		do
			Result := a.is_finite and b.is_finite
		ensure
			definition: Result = (a.is_finite and b.is_finite)
		end

	is_rational: BOOLEAN
			-- Is there a rational representation of current complex number?
			-- TODO: Properties, tests
		do
			Result := is_real and a.is_rational
		ensure
			definition: Result = (is_real and a.is_rational)
		end

	is_integer: BOOLEAN
			-- Is there an integer representation of current complex number?
		do
			Result := is_real and a.is_integer
		ensure
			definition: Result = (is_real and a.is_integer)
		end

	is_invertible: BOOLEAN
			-- Does current complex number have a multiplicative `inverse' (AKA `reciprocal')?
			-- Notice that the quasi_definition post-condition allows an implementation to regard `zero' as invertible.
		deferred
		ensure
			scale_log: attached value_logb ((a.abs ∨ b.abs).value) as scale_log
			int_scale_log: attached if value_is_finite (scale_log) then scale_log.truncated_to_integer else 0 end as int_scale_log
			scaled_a: attached (a.value * 2 ^ (- int_scale_log)) as scaled_a
			scaled_b: attached (b.value * 2 ^ (- int_scale_log)) as scaled_b
			denom: attached (scaled_a ^ 2 + scaled_b ^ 2) as denom
			quasi_definition: denom /= 0 implies Result
		end

feature -- Comparison

	equals alias "≍" (z: COMPLEX_NUMBER): BOOLEAN
			-- Do current complex number and `z' have the same `real' and `imaginary' parts, respectively?
			--| Please have a look at the note in the header of {SET}.equals.
		do
			Result := a ≍ z.a and b ≍ z.b
		ensure
			definition: Result = (a ≍ z.a and b ≍ z.b)
		end

	unequals alias "≭" (z: COMPLEX_NUMBER): BOOLEAN
			-- Does not current complex number equal `z'?
		do
			Result := not (Current ≍ z)
		ensure
			definition: Result = not (Current ≍ z)
		end

feature -- Relationship

	divisible (z: COMPLEX_NUMBER): BOOLEAN
			-- May current complex number be divided by `z`?
		do
			Result := z.is_invertible
		ensure
			definition: Result = z.is_invertible
		end

feature -- Operation

	modulus,
	abs: like real_anchor
			-- Distance from current complex number to the origin of the complex plane
		do
			Result := hypotenuse (a, b)
		ensure
			definition: Result ≍ hypotenuse (a, b)
		end

	argument,
	phase: like real_anchor
			-- On the complex plane, the angle with the positive real semi-axis of the radius from the origin to current complex number.
		do
			Result := arc_tangent_yx (b, a)
		ensure
			Result ≍ arc_tangent_yx (b, a)
		end

	conjugate: like complex_anchor
			-- Complex number in the form `a' - `b'i
		deferred
		ensure
			real: Result.a ≍ a
			imaginary: Result.b ≍ - b
		end

	plus alias "+" (z: COMPLEX_NUMBER): like complex_anchor
			-- Sum of current complex number with `z`
		deferred
		ensure
			real: Result.a ≍ (a + z.a)
			imaginary: Result.b ≍ (b + z.b)
		end

	identity alias "+": like Current
			-- Unary plus; this number itself.
		do
			Result := Current
		ensure
			real: Result.a ≍ a
			imaginary: Result.b ≍ b
		end

	minus alias "-" alias "−" (z: COMPLEX_NUMBER): like complex_anchor
			-- Result of subtracting `z` from current complex number
		deferred
		ensure
			real: Result.a ≍ (a - z.a)
			imaginary: Result.b ≍ (b - z.b)
		end

	opposite alias "-" alias "−": like complex_anchor
			-- Opposite value of current number relative to addition, i.e.
			-- `Current' + (- `Current') = 0.
		deferred
		ensure
			real: Result.a ≍ - a
			imaginary: Result.b ≍ - b
		end

	product alias "*" alias "×" alias "⋅" (z: COMPLEX_NUMBER): like complex_anchor
			-- Current complex number multiplied by `z`
			--| NOTICE: This specification goes beyond ISO C17 (please see description of current class at its note clause above) in that it treats a purely real and purely imaginary
			--| number as if they had the type {REAL_NUMBER} or (a would-be) {IMAGINARY_NUMBER}, respectively. E.g. the product (a+0i)⋅(0+di) is simply adi (= 0+adi), i.e. a proper
			--| imaginary number, instead of (a⋅0-0d)+(ad+0⋅0)i (= (a⋅0-0d)+adi), which may be a proper complex, since a⋅0-0d may yield a NaN. Specifically, `zero' is treated as a
			--| proper imaginary, a proper real or a proper complex, following the respective qualities of the other factor.
		deferred
		ensure
			zero_times_zero_a: Current ≍ zero and z ≍ zero implies Result.a ≍ zero.a
			zero_times_zero_b: Current ≍ zero and z ≍ zero implies Result.b ≍ zero.a

				-- 0i * di = -0d
			zero_times_imaginary_a: Current ≍ zero and z.is_proper_imaginary implies Result.a ≍ (- b ⋅ z.b)
			zero_times_imaginary_b: Current ≍ zero and z.is_proper_imaginary implies Result.b ≍ zero.a

				-- 0 * c = 0c
			zero_times_real_a: Current ≍ zero and z.is_proper_real implies Result.a ≍ (a ⋅ z.a)
			zero_times_real_b: Current ≍ zero and z.is_proper_real implies Result.b ≍ zero.a

				-- (0+0i) * (c+di) = (0c-0d)+(0d+0c)i
			zero_times_complex_a: Current ≍ zero and z.is_proper_complex implies Result.a ≍ (a * z.a - b * z.b)
			zero_times_complex_b: Current ≍ zero and z.is_proper_complex implies Result.b ≍ (a * z.b + b * z.a)

				-- bi * 0i = -b*0
			imaginary_times_zero_a: is_proper_imaginary and z ≍ zero implies Result.a ≍ (- b ⋅ z.b)
			imaginary_times_zero_b: is_proper_imaginary and z ≍ zero implies Result.b ≍ zero.a

			imaginary_times_imaginary_a: is_proper_imaginary and z.is_proper_imaginary implies Result.a ≍ (- b ⋅ z.b)
			imaginary_times_imaginary_b: is_proper_imaginary and z.is_proper_imaginary implies Result.b ≍ zero.a

			imaginary_times_real_a: is_proper_imaginary and z.is_proper_real implies Result.a ≍ zero.a
			imaginary_times_real_b: is_proper_imaginary and z.is_proper_real implies Result.b ≍ (b ⋅ z.a)

			imaginary_times_complex_a: is_proper_imaginary and z.is_proper_complex implies Result.a ≍ (- b ⋅ z.b)
			imaginary_times_complex_b: is_proper_imaginary and z.is_proper_complex implies Result.b ≍ (b ⋅ z.a)

				-- a * 0 = a*0
			real_times_zero_a: is_proper_real and z ≍ zero implies Result.a ≍ (a ⋅ z.a)
			real_times_zero_b: is_proper_real and z ≍ zero implies Result.b ≍ zero.a

			real_times_imaginary_a: is_proper_real and z.is_proper_imaginary implies Result.a ≍ zero.a
			real_times_imaginary_b: is_proper_real and z.is_proper_imaginary implies Result.b ≍ (a ⋅ z.b)

			real_times_real_a: is_proper_real and z.is_proper_real implies Result.a ≍ (a ⋅ z.a)
			real_times_real_b: is_proper_real and z.is_proper_real implies Result.b ≍ zero.a

			real_times_complex_a: is_proper_real and z.is_proper_complex implies Result.a ≍ (a ⋅ z.a)
			real_times_complex_b: is_proper_real and z.is_proper_complex implies Result.b ≍ (a ⋅ z.b)

				-- (a+bi) * (0+0i) = (a*0-b*0)+(a*0+b*0)i
			complex_times_zero_a: is_proper_complex and z ≍ zero implies Result.a ≍ (a * z.a - b * z.b)
			complex_times_zero_b: is_proper_complex and z ≍ zero implies Result.b ≍ (a * z.b + b * z.a)

			complex_times_imaginary_a: is_proper_complex and z.is_proper_imaginary implies Result.a ≍ (- b ⋅ z.b)
			complex_times_imaginary_b: is_proper_complex and z.is_proper_imaginary implies Result.b ≍ (a ⋅ z.b)

			complex_times_real_a: is_proper_complex and z.is_proper_real implies Result.a ≍ (a ⋅ z.a)
			complex_times_real_b: is_proper_complex and z.is_proper_real implies Result.b ≍ (b ⋅ z.a)

				-- TODO: Bug on attached real_from_value... or attached {like real_anchor} real_from_value...
			prod_a: attached {REAL_NUMBER} real_from_value (a.value * z.a.value - b.value * z.b.value) as prod_a
			prod_b: attached {REAL_NUMBER} real_from_value (a.value * z.b.value + b.value * z.a.value) as prod_b
			complex_times_complex_a: is_proper_complex and z.is_proper_complex implies
				(Result.a - prod_a).abs ≤ (Result.a.machine_epsilon ∨ prod_a.relative_epsilon)
			complex_times_complex_b: is_proper_complex and z.is_proper_complex implies
				(Result.b - prod_b).abs ≤ (Result.b.machine_epsilon ∨ prod_b.relative_epsilon)

			infinite_left_factor: is_infinite and (z.is_finite and z ≭ zero or z.is_infinite) implies Result.is_infinite
			infinite_right_factor: (is_finite and Current ≭ zero or is_infinite) and z.is_infinite implies Result.is_infinite
		end

	quotient alias "/" alias "÷" (z: COMPLEX_NUMBER): like complex_anchor
			-- Division of current complex number by `z`
			--| NOTICE: The observations at the header of `product' also apply to this feature.
		require
			good_divisor: divisible (z)
		deferred
		ensure
				-- 0i / di = 0/d
			zero_by_imaginary_a_good_divisor: z.is_proper_imaginary implies b.divisible (z.b) -- z.b /= 0 ⇐ z.is_proper_imaginary
			zero_by_imaginary_a: Current ≍ zero and z.is_proper_imaginary implies Result.a ≍ (b / z.b)
			zero_by_imaginary_b: Current ≍ zero and z.is_proper_imaginary implies Result.b ≍ zero.a

				-- 0 / c = 0/c = 0/c
			zero_by_real_a_good_divisor: z.is_proper_real implies a.divisible (z.a) -- z.a /=  ⇐ z.is_proper_real
			zero_by_real_a: Current ≍ zero and z.is_proper_real implies Result.a ≍ (a / z.a)
			zero_by_real_b: Current ≍ zero and z.is_proper_real implies Result.b ≍ zero.a

				-- bi / di = b/d
			imaginary_by_imaginary_a_good_divisor: z.is_proper_imaginary implies b.divisible (z.b) -- z.b /= 0 ⇐ z.is_proper_imaginary
			imaginary_by_imaginary_a: is_proper_imaginary and z.is_proper_imaginary implies Result.a ≍ (b / z.b)
			imaginary_by_imaginary_b: is_proper_imaginary and z.is_proper_imaginary implies Result.b ≍ zero.a

				-- bi / c = bi/c
			imaginary_by_real_b_good_divisor: z.is_proper_real implies b.divisible (z.a) -- z.a /= 0 ⇐ z.is_proper_real
			imaginary_by_real_a: is_proper_imaginary and z.is_proper_real implies Result.a ≍ zero.a
			imaginary_by_real_b: is_proper_imaginary and z.is_proper_real implies Result.b ≍ (b / z.a)

				-- a / di = -ai/d
			real_by_imaginary_b_good_divisor: z.is_proper_imaginary implies a.divisible (z.b) -- z.b /= 0 ⇐ z.is_proper_imaginary
			real_by_imaginary_a: is_proper_real and z.is_proper_imaginary implies Result.a ≍ zero.a
			real_by_imaginary_b: is_proper_real and z.is_proper_imaginary implies Result.b ≍ - (a / z.b)

				-- a / c = a/c
			real_by_real_a_good_divisor: z.is_proper_real implies a.divisible (z.a) -- z.a /= 0 ⇐ z.is_proper_real
			real_by_real_a: is_proper_real and z.is_proper_real implies Result.a ≍ (a / z.a)
			real_by_real_b: is_proper_real and z.is_proper_real implies Result.b ≍ zero.a

				-- (a + bi) / di = b/d - ai/d
			complex_by_imaginary_a_good_divisor: z.is_proper_imaginary implies b.divisible (z.b) -- z.b /= 0 ⇐ z.is_proper_imaginary
			complex_by_imaginary_b_good_divisor: z.is_proper_imaginary implies a.divisible (z.b) -- z.b /= 0
			complex_by_imaginary_a: is_proper_complex and z.is_proper_imaginary implies Result.a ≍ (b / z.b)
			complex_by_imaginary_b: is_proper_complex and z.is_proper_imaginary implies Result.b ≍ - (a / z.b)

				-- (a + bi) / c = a/c + bi/c
			complex_by_real_a_good_divisor: z.is_proper_real implies a.divisible (z.a) -- z.a /= 0 ⇐ z.is_proper_real
			complex_by_real_b_good_divisor: z.is_proper_real implies b.divisible (z.a) -- z.a /= 0
			complex_by_real_a: is_proper_complex and z.is_proper_real implies Result.a ≍ (a / z.a)
			complex_by_real_b: is_proper_complex and z.is_proper_real implies Result.b ≍ (b / z.a)

			scale_log: attached value_logb ((z.a.abs ∨ z.b.abs).value) as scale_log
			int_scale_log: attached if value_is_finite (scale_log) then scale_log.truncated_to_integer else 0 end as int_scale_log
			scaled_c: attached (z.a.value * 2 ^ (- int_scale_log)) as scaled_c
			scaled_d: attached (z.b.value * 2 ^ (- int_scale_log)) as scaled_d
			denom: attached (scaled_c ^ 2 + scaled_d ^ 2) as denom
			invertible_denom: z.is_proper_complex implies denom /= 0 -- z.a /= 0 and z.b /= 0 ⇐ z.is_proper_complex

				-- (a+bi) / (c + di) = (ac+bd)/(c^2+d^2)+(bc-ad)i/(c^2+d^2)				
			complex_divisor_a_good_divisor: z.is_proper_complex implies (a.value * scaled_c + b.value * scaled_d).divisible (denom) -- denom /= 0 ⇐ z.is_proper_complex
			complex_divisor_b_good_divisor: z.is_proper_complex implies (b.value * scaled_c - a.value * scaled_d).divisible (denom)
			complex_divisor_quot_a: attached {REAL_NUMBER} real_from_double_value
				(((a.value * scaled_c + b.value * scaled_d) / denom) * 2 ^ (- int_scale_log)) as complex_by_complex_quot_a
			complex_divisor_quot_b: attached {REAL_NUMBER} real_from_double_value
				(((b.value * scaled_c - a.value * scaled_d) / denom) * 2 ^ (- int_scale_log)) as complex_by_complex_quot_b
			complex_divisor_a: z.is_proper_complex implies
				(Result.a - complex_by_complex_quot_a).abs ≤ (Result.a.machine_epsilon ∨ complex_by_complex_quot_a.relative_epsilon)
			complex_divisor_b: z.is_proper_complex implies
				(Result.b - complex_by_complex_quot_b).abs ≤ (Result.b.machine_epsilon ∨ complex_by_complex_quot_b.relative_epsilon)

			infinite_dividend: is_infinite and z.is_finite implies Result.is_infinite
			infinite_divisor: is_finite and z.is_infinite implies Result ≍ zero
			zero_divisor: (is_finite and Current ≭ zero or is_infinite) and z ≍ zero implies Result.is_infinite -- For the case of an implementation that regards zero invertible.
		end

	reciprocal,
	inverse: like complex_anchor
			-- Reciprocal (AKA inverse) value of current number relative to
			-- multiplication, i.e.
			-- `Current' ⋅ `reciprocal' = `Current' ⋅ `inverse' = 1.
		require
			is_invertible: is_invertible
		do
				check
					good_divisor: one.divisible (Current) -- `is_invertible'

				end
			Result := one / Current
		ensure
			good_divisor: one.divisible (Current) -- `is_invertible'
			real: Result.a ≍ (one / Current).a
			imaginary: Result.b ≍ (one / Current).b
		end

feature -- Math

	Pi: like native_real_anchor = 3.1415926535897932384626433832795029
			-- Value of the constant π
			--| TODO: Metrics tool doesn't see this comment.

	Pi_2: like native_real_anchor = 1.5707963267948966192313216916397514
			-- Value of the constant π/2

	Pi_4: like native_real_anchor = 0.7853981633974483096156608458198757
			-- Value of the constant π/4

	value_is_infinite (v: like native_real_anchor): BOOLEAN
			-- Is `v' the representation of an infinity, i.e. either is it a negative infinity or a positive_infinity?
		do
			Result := v.is_negative_infinity or v.is_positive_infinity
		ensure
			class
			definition: Result = (v.is_negative_infinity or v.is_positive_infinity)
		end

	value_is_finite (v: like native_real_anchor): BOOLEAN
			-- Is `v' the representation of a finite real number, i.e. is it neither a NaN, a negative infinity nor a positive infinity?
		do
			Result := not (v.is_nan or v.is_negative_infinity or v.is_positive_infinity)
		ensure
			class
			definition: Result = not (v.is_nan or v.is_negative_infinity or v.is_positive_infinity)
		end

	hypotenuse (x, y: REAL_NUMBER): like real_anchor
			-- Square root of the sum of the squares of x and y, without undue overﬂow or underﬂow.			
		deferred
		ensure
			ref_hypot: attached (√¯ (x ^ two + y ^ two)) as ref_hypot
			definition: (Result - ref_hypot).abs ≤ (Result.machine_epsilon ∨ ref_hypot.relative_epsilon)
			commutative: Result ≍ hypotenuse (y, x)
			when_unidimensional: y ≍ zero.a implies Result ≍ x.abs
			when_negatively_infinite: x.is_negative_infinity implies Result.is_positive_infinity
			when_positively_infinite: x.is_positive_infinity implies Result.is_positive_infinity
		end

	arc_tangent_yx (y, x: REAL_NUMBER): like real_anchor
			-- Arc tangent of `y'/`x', using the signs of both arguments to determine the quadrant of result.
		deferred
		ensure
				-- So many cases in order to conform to IEEE 754.
			either_is_nan: y.is_nan or x.is_nan implies Result.is_nan

			negative_zero_by_negative_zero: y.is_negative_zero and x.is_negative_zero implies Result.value = Result.adjusted_value (- Pi)
			positive_zero_by_negative_zero: y.is_positive_zero and x.is_negative_zero implies Result.value = Result.adjusted_value (Pi)

			negative_zero_by_positive_zero: y.is_negative_zero and x.is_positive_zero implies Result.is_negative_zero
			positive_zero_by_positive_zero: y.is_positive_zero and x.is_positive_zero implies Result.is_positive_zero

			negative_zero_by_negative_real: y.is_negative_zero and x.is_finite and x < zero.a implies Result.value = Result.adjusted_value (- Pi)
			positive_zero_by_negative_real: y.is_positive_zero and x.is_finite and x < zero.a implies Result.value = Result.adjusted_value (Pi)

			negative_zero_by_positive_real: y.is_negative_zero and x > zero.a and x.is_finite implies Result.is_negative_zero
			positive_zero_by_positive_real: y.is_positive_zero and x > zero.a and x.is_finite implies Result.is_positive_zero

			negative_real_by_negative_zero: y.is_finite and y < zero.a and x.is_negative_zero implies Result.value = Result.adjusted_value (- Pi_2)
			nevative_real_by_positive_zero: y.is_finite and y < zero.a and x.is_positive_zero implies Result.value = Result.adjusted_value (- Pi_2)

			positive_real_by_negative_zero: y > zero.a and y.is_finite and x.is_negative_zero implies Result.value = Result.adjusted_value (Pi_2)
			positive_real_by_positive_zero: y > zero.a and y.is_finite and x.is_positive_zero implies Result.value = Result.adjusted_value (Pi_2)

			negative_real_by_negative_infinity: y.is_finite and y < zero.a and x.is_negative_infinity implies Result.value = Result.adjusted_value (- Pi)
			positive_real_by_negative_infinity: y > zero.a and y.is_finite and x.is_negative_infinity implies Result.value = Result.adjusted_value (Pi)

			negative_real_by_positive_infinity: y.is_finite and y < zero.a and x.is_positive_infinity implies Result.is_negative_zero
			positive_real_by_positive_infinity: y > zero.a and y.is_finite and x.is_positive_infinity implies Result.is_positive_zero

			negative_infinity_by_finite_real: y.is_negative_infinity and x.is_finite implies Result.value = Result.adjusted_value (- Pi_2)
			positive_infinity_by_finite_real: y.is_positive_infinity and x.is_finite implies Result.value = Result.adjusted_value (Pi_2)

			negative_infinity_by_negative_infinity: y.is_negative_infinity and x.is_negative_infinity implies Result.value = Result.adjusted_value (- 3 × Pi_4)
			positive_infinity_by_negative_infinity: y.is_positive_infinity and x.is_negative_infinity implies Result.value = Result.adjusted_value (3 × Pi_4)

			negative_infinity_by_positive_infinity: y.is_negative_infinity and x.is_positive_infinity implies Result.value = Result.adjusted_value (- Pi_4)
			positive_infinity_by_positive_infinity: y.is_positive_infinity and x.is_positive_infinity implies Result.value = Result.adjusted_value (Pi_4)

			good_divisor: x ≭ zero.a implies y.value.divisible (x.value) -- x.value /= 0
			first_quadrant: y > zero.a and y.is_finite and x > zero.a and x.is_finite implies Result.value = Result.adjusted_value (arc_tangent (y.value / x.value))
			second_quadrant: y > zero.a and y.is_finite and x.is_finite and x < zero.a implies Result.value = Result.adjusted_value (arc_tangent (y.value / x.value) + Pi)
			third_quadrant: y.is_finite and y < zero.a and x.is_finite and x < zero.a implies Result.value = Result.adjusted_value (arc_tangent (y.value / x.value) - Pi)
			fourth_quadrant: y.is_finite and y < zero.a and x > zero.a and x.is_finite implies Result.value = Result.adjusted_value (arc_tangent (y.value / x.value))
		end

	arc_tangent (v: like native_real_anchor): like native_real_anchor
			-- Trigonometric arctangent of radian `v' in the range [-pi/2, +pi/2]
		do
			Result := {like real_math_anchor}.arc_tangent (v)
		ensure
			class
			definition: Result = {like real_math_anchor}.arc_tangent (v)
		end

	value_logb (v: like native_real_anchor): like native_real_anchor
			-- Exponent of `v', as a signed integer value in ﬂoating-point format. If `v' is subnormal it is treated as though it were normalized; thus, for
			-- positive ﬁnite `v', 1 ≤ `v' × (2 ^ − `value_logb' (`v')) < 2.
			--| Needed because {like real_math_anchor}.log_2 has rounding errors.
		external
			"C signature (float): float use <math.h>"
		alias
			"logbf"
		ensure
			class
			when_nan: v.is_nan implies Result.is_nan
			when_negative: not v.is_nan and v < 0 implies Result = value_logb (v.abs)
			when_zero: v = 0 implies Result.is_negative_infinity
			when_finite_positive: 0 < v and not v.is_positive_infinity implies 1 ≤ (v × (2 ^ − Result)) and v × (2 ^ − Result) < 2.
			when_positive_infinity: v.is_positive_infinity implies Result.is_positive_infinity
		end

	value_sign_bit (v: like native_real_anchor): like integer_anchor.value
			-- Status of the sign bit of `v', which is 1 for negative numbers but also, e.g. for a "negative" zero as specified by IEEE 754.
		external
			"C inline use <math.h>"
		alias
			"{
				return signbit ($v)? 1: 0;
				}"
		ensure
			class
			when_nan: v.is_nan implies Result = 0 or Result = 1
			when_negative: not v.is_nan and v < 0 implies Result = 1
			when_zero: v = 0 implies Result = 0 or Result = 1
			when_positive: 0 < v implies Result = 0
		end

feature -- Factory

	real_from_value (v: like native_real_anchor): like real_anchor
			-- Real number whose value is {REAL_NUMBER}.adjusted_value (`v')
		deferred
		ensure
			definition: Result.value = Result.adjusted_value (v)
		end

	real_from_double_value (dv: like native_real_anchor.to_double): like real_anchor
			-- Real number whose value is {REAL_NUMBER}.adjusted_value (`native_real_from_double_value' (`dv'))
		deferred
		ensure
			definition: Result.value = Result.adjusted_value (native_real_from_double_value (dv))
		end

	native_real_from_double_value (dv: like native_real_anchor.to_double): like Native_real_anchor
			-- Native real number equal to `dv'.`truncated_to_real'
			--| TODO: Move it to a platform-specific class.
		do
			Result := dv.truncated_to_real
		ensure
			class
			definition: Result = dv.truncated_to_real
		end

feature -- Anchor

	complex_anchor: COMPLEX_NUMBER
			-- Anchor for complex numbers
		deferred
		end

	native_real_anchor: REAL
			-- Anchor for native representation real numbers
		once
		ensure
			class
		end

	real_anchor: REAL_NUMBER
			-- Anchor for real numbers
		deferred
		end

	integer_anchor: INTEGER_NUMBER
			-- Anchor for integer numbers
		deferred
		end

	real_math_anchor: SINGLE_MATH
			-- Anchor for accessing basic mathematical operations on native real numbers
		once
			create Result
		ensure
			class
		end

invariant
	real: real ≍ a
	imaginary: imaginary ≍ b
	Pi: Pi = native_real_from_double_value ({like real_math_anchor}.Pi)
	Pi_2: Pi_2 = native_real_from_double_value ({like real_math_anchor}.Pi_2)
	Pi_4: Pi_4 = native_real_from_double_value ({like real_math_anchor}.Pi_4)

note
	copyright: "Copyright (c) 2012-2022, Rosivaldo Fernandes Alves"
	license: "[
		Eiffel Forum License v2
		(see http://www.eiffel.com/licensing/forum.txt)
		]"
	source: "https://launchpad.net/set-theory"

end
