﻿note
	description: "[
		Model of an integer number, i.e. a rational number whose denominator
		part equals one.
		]"
	author: "Rosivaldo Fernandes Alves"
	date: "$Date$"
	revision: "$Revision$"

deferred class
	INTEGER_NUMBER

inherit
	RATIONAL_NUMBER
		rename
			value as real_value,
			p as identity alias "+",
			numerator as identity alias "+",
			q as one,
			denominator as one,
			is_in as rational_is_in,
			is_not_in as rational_is_not_in,
			real_is_integer as is_integer,
			equals as rational_equals,
			unequals as rational_unequals,
			is_less as rational_is_less,
			is_less_equal as rational_is_less_equal,
			is_greater as rational_is_greater,
			is_greater_equal as rational_is_greater_equal,
			three_way_comparison as rational_three_way_comparison,
			min as rational_min,
			max as rational_max,
			divisible as rational_divisible,
			multipliable as rational_multipliable,
			modulus as rational_modulus,
			abs as rational_abs,
			plus as rational_plus,
			minus as rational_minus,
			opposite as rational_opposite,
			product as rational_product,
			quotient as rational_quotient,
			truncated_to_integer as identity alias "+",
			adjusted_value as real_adjusted_value
		undefine
			is_integer
		redefine
			rational_product
		end

feature -- Primitive

	value: like native_integer_anchor
			-- Native value of current integer number
		deferred
		ensure
			adjusted: Result = adjusted_value (Result)
			consistent: Result = real_value
		end

feature -- Membership

	is_in alias "∈" (s: SET [INTEGER_NUMBER, EQUALITY [INTEGER_NUMBER]]): BOOLEAN
			-- Does `s` have current integer number?
		do
			Result := s ∋ Current
		ensure
			definition: Result = s ∋ Current
		end

	is_not_in alias "∉" (s: SET [INTEGER_NUMBER, EQUALITY [INTEGER_NUMBER]]): BOOLEAN
			-- Is not current integer number in `s'?
		do
			Result := not (Current ∈ s)
		ensure
			definition: Result = not (Current ∈ s)
		end

feature -- Access

	zero: like integer_anchor
			-- The integer number 0
		deferred
		end

	one: like integer_anchor
			-- The integer number 1
		deferred
		end

	min_value: like integer_anchor
			-- Minimum value representable by an implementation of integer numbers
		require
			min_value_exists: min_value_exists
		deferred
		end

	max_value: like integer_anchor
			-- Maximum value representable by an implementation of integer numbers
		require
			max_value_exists: max_value_exists
		deferred
		end

feature -- Quality

--	is_invertible: BOOLEAN
--			-- Does current integer number have a multiplicative inverse (AKA
--			-- reciprocal)?
--			--| Notice that this feature is not inherited. An implementation of
--			--| {REAL_NUMBER}.is_invertible may regard a zero invertible, which
--			--| is normally not expected for an implementation of integer
--			--| numbers.
--		do
--			Result := Current ≭ zero
--		ensure then
--			definition: Result = Current ≭ zero
--		end

	min_value_exists: BOOLEAN
			-- Is there a minimum value for current representation of integer numbers?
		deferred
		end

	max_value_exists: BOOLEAN
			-- Is there a maximum value for current representation of integer numbers?
		deferred
		end

feature -- Comparison

	equals alias "≍" (i: INTEGER_NUMBER): BOOLEAN
			-- Do current integer number and `i' have the same numeric value?
			--| Please have a look at the note in the header of {SET}.equals.
		do
			Result := value = i.value
		ensure
			definition: Result = (value = i.value)
		end

	unequals alias "≭" (i: INTEGER_NUMBER): BOOLEAN
			-- Does not current integer number equal `i'?
		do
			Result := not (Current ≍ i)
		ensure
			definition: Result = not (Current ≍ i)
		end

	is_less alias "<" (i: INTEGER_NUMBER): BOOLEAN
			-- Is current integer number less than `i'?
		do
			Result := value < i.value
		ensure
			definition: Result = (value < i.value)
		end

	is_less_equal alias "<=" alias "≤" (i: INTEGER_NUMBER): BOOLEAN
			-- Is current integer number less than or equal to `i'?
		do
			Result := Current < i or Current ≍ i
		ensure
			definition: Result = (Current < i or Current ≍ i)
		end

	is_greater alias ">" (i: INTEGER_NUMBER): BOOLEAN
			-- Is current integer number greater than `i'?
		do
			Result := i < Current
		ensure
			definition: Result = (i < Current)
		end

	is_greater_equal alias ">=" alias "≥" (i: INTEGER_NUMBER): BOOLEAN
			-- Is current integer number greater than or equal to `i'?
		do
			Result := i ≤ Current
		ensure
			definition: Result = (i <= Current)
		end

	three_way_comparison alias "⋚" (i: INTEGER_NUMBER): like integer_anchor
			-- If current integer number equal to `i', 0; if smaller, -1; if greater, 1.
		do
			if Current < i then
				Result := - one
			elseif i < Current then
				Result := one
			else
				Result := zero
			end
		ensure
			equal_zero: (Result ≍ zero) = (Current ≍ i)
			smaller_negative: (Result ≍ - one) = (Current < i)
			greater_positive: (Result ≍ one) = (Current > i)
		end

	min alias "∧" (i: INTEGER_NUMBER): like integer_anchor
			-- The smaller of current integer number and `i'
		do
			if Current ≤ i then
				Result := Current
			else
				Result := i
			end
		ensure
			current_if_not_greater: Current <= i implies Result ≍ Current
			other_if_greater: Current > i implies Result ≍ i
		end

	max alias "∨" (i: INTEGER_NUMBER): like integer_anchor
			-- The greater of current integer number and `i'
		do
			if Current ≥ i then
				Result := Current
			else
				Result := i
			end
		ensure
			current_if_not_smaller: Current >= i implies Result ≍ Current
			other_if_smaller: Current < i implies Result ≍ i
		end

feature -- Relationship

	divisible (i: INTEGER_NUMBER): BOOLEAN
			-- May current integer number be divided by `i`?
		do
--			Result := value.divisible (i.value) -- TODO: Segmentation violation!
			Result := i.value /= 0
		ensure
--			definition: Result = value.divisible (i.value) -- TODO: Segmentation violation!
			definition: Result = (i.value /= 0)
		end

	divides alias "|" (i: INTEGER_NUMBER): BOOLEAN
			-- Does current integer number divide `i', i.e. is there an integer
			-- number k such that `i' = k ⋅ `Current'?
		do
			Result := i.divisible (Current) and then (i \\ Current) ≍ zero
		ensure
			definition: Result =
				(i.divisible (Current) and then (i \\ Current) ≍ zero)
		end

feature -- Operation

	rational_product (pq: RATIONAL_NUMBER): like rational_anchor
			-- Current integer number multiplied by `pq`
		require else
			always_good_factor: True
		local
			l_gcd: like gcd
		do
			l_gcd := gcd (identity, pq.q)
				check
					non_zero_gcd: l_gcd ≭ zero -- pq.q /= 0
					good_divisor: ((identity // l_gcd) * pq.p).divisible (pq.q // l_gcd) -- pq.q /= 0
				end
			Result := ((identity // l_gcd) * pq.p) / (pq.q // l_gcd)
		end

	modulus,
	abs: like integer_anchor
			-- Absolute value of current integer
			--| Notice that this feature is not inherited. {REAL_NUMBER}.abs always yields a non-negative number (or a NaN). If `Current' is represented
			--| by a two's-complement integer and and its value is the minimum possible value in such a representation, `abs' may yield an unexpected
			--| value, most possibly the same (still negative) value.
		deferred
		ensure
			definition: Result.value = adjusted_value (value.abs)
		end

	plus alias "+" (i: INTEGER_NUMBER): like integer_anchor
			-- Sum of current integer number with `i`
		deferred
		ensure
			definition: Result.value = adjusted_value (value + i.value)
		end

	minus alias "-" alias "−" (i: INTEGER_NUMBER): like integer_anchor
			-- Result of subtracting `i` from current integer number
		deferred
		ensure
			definition: Result.value = adjusted_value (value - i.value)
		end

	opposite alias "-" alias "−": like integer_anchor
			-- Opposite value of current number relative to `+' operation, i.e. `Current' + (- `Current') = 0.
			--| Notice that this feature is not inherited. If `Current' is represented by a two's-complement integer and and its value is the minimum possible
			--| value in such a representation, `opposite' may yield an unexpected value, most possibly the same (still negative) value, departing from
			--| `real_opposite' result.
		deferred
		ensure
			definition: Result.value = adjusted_value (- value)
		end

	product alias "*" alias "×" alias "⋅" (i: INTEGER_NUMBER): like integer_anchor
			-- Current integer number multiplied by `i`
		deferred
		ensure
			definition: Result.value = adjusted_value (value * i.value)
		end

	quotient alias "/" alias "÷" (i: INTEGER_NUMBER): like rational_anchor
			-- Rational number got by dividing current integer number by `i`
		require
			good_divisor: divisible (i)
		deferred
		ensure
			p_divisible_q: -- {RATIONAL_NUMBER} invariant: Result.q /= 0
				Result.p.divisible (Result.q)
			quotient: (Result.p // Result.q) ≍ (Current // i)
			p_rem_q_divisible_q: -- {RATIONAL_NUMBER} invariant: Result.q /= 0
				(Result.p \\ Result.q).divisible (Result.q)
			current_rem_i_divisible_i: -- i /= 0 ⇐ divisible (i)
				(Current \\ i).divisible (i)
			remainder: ((Result.p \\ Result.q) / Result.q) ≍
				((Current \\ i) / i)
		end

	integer_quotient alias "//" (i: INTEGER_NUMBER): like integer_anchor
			-- Division of current integer number by `i`
		require
			good_divisor: divisible (i)
		deferred
		ensure
--			good_divisor: value.divisible (i.value) -- i.value /= 0 ⇐ divisible (i) -- TODO: Segmentation violation.
			good_divisor: i.value /= 0
			definition: Result.value = adjusted_value (value // i.value)
		end

	integer_remainder alias "\\" (i: INTEGER_NUMBER): like integer_anchor
			-- Remainder of the integer division of `Current' by `i`
		require
			good_divisor: divisible (i)
		deferred
		ensure
--			good_divisor: value.divisible (i.value) -- i.value /= 0 ⇐ divisible (i) -- TODO: Segmentation violation.
			good_divisor: i.value /= 0
			definition: Result.value = adjusted_value (value \\ i.value)
		end

		--	reciprocal: like rational_anchor
		--			etc.

feature -- Implementation

	adjusted_value (v: like value): like value
			-- `v' adjusted to represent the `value' of current integer number
		deferred
		ensure
			stable: adjusted_value (Result) = Result
			consistent: Result = real_adjusted_value (Result)
		end

feature -- Anchor

	native_integer_anchor: INTEGER
			-- Anchor for the native representation of the value of current integer number
		do
		ensure
			class
		end

invariant
	min_value: min_value_exists implies min_value ≤ Current
--	max_value: max_value_exists implies Current ≤ max_value -- TODO: It fails when `max_value' is a once routine and is called by the first time.

note
	copyright: "Copyright (c) 2012-2022, Rosivaldo Fernandes Alves"
	license: "[
		Eiffel Forum License v2
		(see http://www.eiffel.com/licensing/forum.txt)
		]"
	source: "https://launchpad.net/set-theory"

end
