﻿note
	description: "[
		Real number in the form p/q, where p and q are integer numbers and
		q /= 0.

		This specification does not try to offer a high-grade numeric class.
		Its intent is just to introduce examples of how several numeric types
		and sets may receive a set-theoretic treatment.
		
		Notice that this specification also does not impose a canonical
		representation, e.g. gcd (p, q) = 1 or q > 0. Hence the implementation
		is free to impose, or not, such restrictions or whatever more is deemed
		necessary.
		
		I'm glad to thank the guys at Boost.org, whose Rational library 
		(https://www.boost.org/doc/libs/1_77_0/libs/rational/) provided me with
		great insights about rational number comparison.
		]"
	author: "Rosivaldo Fernandes Alves"
	date: "$Date$"
	revision: "$Revision$"

deferred class
	RATIONAL_NUMBER

inherit
	REAL_NUMBER
		rename
			is_in as real_is_in,
			is_not_in as real_is_not_in,
			i as imaginary_unit,
			is_integer as real_is_integer,
			is_invertible as real_is_invertible,
			equals as real_equals,
			unequals as real_unequals,
			is_less as real_is_less,
			is_less_equal as real_is_less_equal,
			is_greater as real_is_greater,
			is_greater_equal as real_is_greater_equal,
			three_way_comparison as real_three_way_comparison,
			min as real_min,
			max as real_max,
			divisible as real_divisible,
			plus as real_plus,
			minus as real_minus,
			opposite as real_opposite,
			product as real_product,
			quotient as real_quotient,
			inverse as real_inverse,
			reciprocal as real_reciprocal,
			to_rational as identity alias "+" -- TODO: Remove precondition? Aren't they ored?
		undefine
			modulus,
			abs
		end

feature -- Primitive

	p,
	numerator: like integer_anchor
			-- Numerator of current rational number
		deferred
		end

	q,
	denominator: like integer_anchor
			-- Denominator of current rational number
		deferred
		end

feature -- Membership

	is_in alias "∈" (s: SET [RATIONAL_NUMBER, EQUALITY [RATIONAL_NUMBER]]): BOOLEAN
			-- Does `s` have current rational number?
		do
			Result := s ∋ Current
		ensure
			definition: Result = s ∋ Current
		end

	is_not_in alias "∉" (s: SET [RATIONAL_NUMBER, EQUALITY [RATIONAL_NUMBER]]): BOOLEAN
			-- Is not current rational number in `s'?
		do
			Result := not (Current ∈ s)
		ensure
			definition: Result = not (Current ∈ s)
		end

feature -- Access

	zero: like rational_anchor
			-- The rational number 0/1
		deferred
		end

	one: like rational_anchor
			-- The rational number 1/1
		deferred
		end

feature -- Quality

	is_integer: BOOLEAN
			-- Does current rational number represent an integer number?
			-- Notice that, due to rounding and truncation, implementation of `real_is_integer' might give True even if `is_integer' gives False.
		do
				check
					good_divisor: p.divisible (q) -- Class invariant: q /= 0
				end
			Result := (p \\ q) ≍ zero.p
		ensure
			good_divisor: p.divisible (q) -- Class invariant: q /= 0
			definition: Result = (p \\ q) ≍ zero.p
		end

	is_invertible: BOOLEAN
			-- Does current rational number have a multiplicative inverse
			-- (AKA reciprocal)?
		do
--			Result := Current ≭ zero.p -- TODO: Fix it.
			Result := p ≭ zero.p
		ensure
--			definition: Result = Current ≭ zero
			definition: Result = p ≭ zero.p
		end

feature -- Comparison

	equals alias "≍" (other: RATIONAL_NUMBER): BOOLEAN
			-- Do current rational number and `other' have the same numerical
			-- value?
		do
				check
						-- Class invariant: q /= 0 and other.q /= 0
					good_divisor: p.divisible (q)
					other_good_divisor: other.p.divisible (other.q)
				end
			if is_integer then
				Result := other.is_integer and
				div (p, q) ≍ div (other.p, other.q)
			else
				if not other.is_integer then
						check
								-- not is_integer implies rem (p, q) /= 0 and
								-- not other.is_integer implies
								--	rem (other.p, other.q) /= 0
							q_divisible_p_rem_q: q.divisible (rem (p, q))
							other_q_divisible_other_p_rem_other_q:
								other.q.divisible (rem (other.p, other.q))
						end
					Result := div (p, q) ≍ div (other.p, other.q) and
						(q / rem (p, q)) ≍ (other.q / rem (other.p, other.q))
				end
			end
		ensure
				-- Class invariant: q /= 0 and other.q /= 0
			good_divisor: p.divisible (q)
			other_good_divisor: other.p.divisible (other.q)

				-- not is_integer implies rem (p, q) /= 0 and
				-- not other.is_integer implies rem (other.p, other.q) /= 0
			q_divisible_p_rem_q: not is_integer implies
				q.divisible (rem (p, q))
			other_q_divisible_other_p_rem_other_q: not other.is_integer implies
				other.q.divisible (rem (other.p, other.q))

			when_integer: is_integer implies Result =
				(other.is_integer and div (p, q) ≍ div (other.p, other.q))
			when_not_integer: not is_integer implies Result = (
				not other.is_integer and then (
					div (p, q) ≍ div (other.p, other.q) and then
					(q / rem (p, q)) ≍ (other.q / rem (other.p, other.q))
					)
				)
		end

	unequals alias "≭" (other: RATIONAL_NUMBER): BOOLEAN
			-- Does not current rational number equal `other'?
		do
			Result := not (Current ≍ other)
		ensure
			definition: Result = not (Current ≍ other)
		end

	is_less alias "<" (other: RATIONAL_NUMBER): BOOLEAN
			-- Is current rational number less than `other`?
		do
				check
						-- Class invariant: q /= 0 and other.q /= 0
					good_divisor: p.divisible (q)
					other_good_divisor: other.p.divisible (other.q)
				end
			if is_integer then
				if other.is_integer then
					Result := div (p, q) < div (other.p, other.q)
				else
					Result := div (p, q) ≤ div (other.p, other.q)
				end
			else
				if other.is_integer then
					Result := div (p, q) < div (other.p, other.q)
				else
						check
								-- not is_integer implies rem (p, q) /= 0 and
								-- not other.is_integer implies
								--	rem (other.p, other.q) /= 0
							q_divisible_p_rem_q: q.divisible (rem (p, q))
							other_p_divisible_other_p_rem_other_q:
								other.q.divisible (rem (other.p, other.q))
						end
					Result := div (p, q) < div (other.p, other.q) or
						div (p, q) ≍ div (other.p, other.q) and (
							(q / rem (p, q)) >
							(other.q / rem (other.p, other.q))
							)
				end
			end
		ensure
				-- Class invariant: q /= 0 and other.q /= 0
			good_divisor: p.divisible (q)
			other_good_divisor: other.p.divisible (other.q)

				-- not is_integer implies rem (p, q) /= 0 and
				-- not other.is_integer implies rem (other.p, other.q) /= 0
			q_divisible_p_rem_q: not is_integer implies
				q.divisible (rem (p, q))
			other_q_divisible_other_p_rem_other_q: not other.is_integer implies
				other.q.divisible (rem (other.p, other.q))

			when_both_integer: is_integer and other.is_integer implies
				Result = (div (p, q) < div (other.p, other.q))
			when_integer_fractional:
				is_integer and not other.is_integer implies
						-- ratio (q, rem (p, q)) = ∞ >
						-- ratio (other.q, rem (other.p, other.q))
					Result = div (p, q) ≤ div (other.p, other.q)
			when_fractional_integer:
				not is_integer and other.is_integer implies
						-- ratio (q, rem (p, q)) <
						-- ratio (other.q, rem (other.p, other.q)) = ∞
					Result = (div (p, q) < div (other.p, other.q))
			when_both_fractional:
				not is_integer and not other.is_integer implies Result = (
					div (p, q) < div (other.p, other.q) or
					div (p, q) ≍ div (other.p, other.q) and (
						(q / rem (p, q)) >
						(other.q / rem (other.p, other.q))
						)
					)
		end

	is_less_equal alias "<=" alias "≤" (pq: RATIONAL_NUMBER): BOOLEAN
			-- Is current rational number less than or equal to `pq`?
		do
				-- TODO: Optimize it?
			Result := Current < pq or Current ≍ pq
		ensure
			definition: Result = (Current < pq or Current ≍ pq)
		end

	is_greater alias ">" (pq: RATIONAL_NUMBER): BOOLEAN
			-- Is current rational number greater than `pq`?
		do
			Result := pq < Current
		ensure
			definition: Result = (pq < Current)
		end

	is_greater_equal alias ">=" alias "≥" (pq: RATIONAL_NUMBER): BOOLEAN
			-- Is current rational number greater than or equal to `pq'?
		do
			Result := pq ≤ Current
 		ensure
			definition: Result = (pq ≤ Current)
		end

	three_way_comparison alias "⋚" (pq: RATIONAL_NUMBER): like integer_anchor
			-- If current rational number equal to `pq`, 0; if smaller, -1; if
			-- greater, 1.
		do
			if Current < pq then
				Result := - one.p
			elseif Current ≍ pq then
				Result := zero.p
			else
				Result := one.p
			end
		ensure
			equal_zero: (Result ≍ zero.p) = (Current ≍ pq)
			smaller_negative: (Result ≍ - one.p) = (Current < pq)
			greater_positive: (Result ≍ one.p) = (Current > pq)
		end

	min alias "∧" (pq: RATIONAL_NUMBER): like rational_anchor
			-- The smaller of current rational number and `pq`
		do
			if Current ≤ pq then
				Result := Current
			else
				Result := pq
			end
		ensure
			current_if_not_greater: Current ≤ pq implies Result ≍ Current
			other_if_greater: pq < Current implies Result ≍ pq
		end

	max alias "∨" (pq: RATIONAL_NUMBER): like rational_anchor
			-- The greater of current rational number and `pq'
		do
			if Current ≥ pq then
				Result := Current
			else
				Result := pq
			end
		ensure
			current_if_not_smaller: Current ≥ pq implies Result ≍ Current
			other_if_smaller: Current < pq implies Result ≍ pq
		end

feature -- Relationship

	multipliable (other: RATIONAL_NUMBER): BOOLEAN
			-- May current rational number be multiplied by `other`?
			--| As explained at `product' header, an occasional overflow when
			--| multiplying denominators (both unequal to `zero' by definition)
			--| may produce an unexpected product equal to `zero'. Hence the need
			--| for this predicate.
		local
			l_gcd_1, l_gcd_2: like gcd
		do
			l_gcd_1 := gcd (other.p, q)
			l_gcd_2 := gcd (p, other.q)
				check
					good_divisor_1: -- l_gcd_1 /= 0 ⇐ q /= 0
						q.divisible (l_gcd_1)
					good_divisor_2: -- l_gcd_2 /= 0 ⇐ other.q /= 0
						other.q.divisible (l_gcd_2)
				end
			Result := not integer_product_overflows
				(q // l_gcd_1, other.q // l_gcd_2)
		ensure
				-- TODO: Debugger gets scrambled!
--			gcd_1: attached {like integer_anchor} gcd (other.p, q) as gcd_1
--			gcd_2: attached {like integer_anchor} gcd (p, other.q) as gcd_2
			gcd_1: attached {INTEGER_NUMBER} gcd (other.p, q) as gcd_1
			gcd_2: attached {INTEGER_NUMBER} gcd (p, other.q) as gcd_2
			good_divisor_1: q.divisible (gcd_1) -- gcd_1 /= 0 ⇐ q /= 0
			good_divisor_2: -- gcd_2 /= 0 ⇐ other.q /= 0
				other.q.divisible (gcd_2)
			good_factors:
					-- Otherwise we might get a denominator unexpectedly equal to
					-- zero.
				not integer_product_overflows
					(q // gcd_1, other.q // gcd_2) implies
				Result
		end

	divisible (other: RATIONAL_NUMBER): BOOLEAN
			-- May current rational number be divided by `other`?
		do
			Result := other.is_invertible and then
				multipliable (other.inverse)
		ensure
			definition: Result = (
				other.is_invertible and then multipliable (other.inverse)
				)
		end

feature -- Operation

	modulus,
	abs: like rational_anchor
			-- Distance from current rational number to the origin of the real
			-- number line
		do
			if Current < zero then
				Result := - Current
			else
				Result := Current
			end
		ensure then
			when_negative: Current < zero implies Result ≍ (- Current)
			when_non_negative: Current ≥ zero implies Result ≍ Current
		end

	plus alias "+" (other: RATIONAL_NUMBER): like rational_anchor
			-- Sum of current rational number with `other`
		local
			l_gcd, other_q_by_gcd: INTEGER_NUMBER
		do
			l_gcd := gcd (q, other.q)
				check
					non_zero_gcd: l_gcd ≭ zero.p -- q /= 0 and other.q /= 0
				end
			other_q_by_gcd := other.q // l_gcd
				check
					good_divisor: ( -- (other.q // l_gcd) /= 0
						p * other_q_by_gcd + other.p * (q // l_gcd)
						).divisible (q * other_q_by_gcd)
				end
			Result := (p * other_q_by_gcd + other.p * (q // l_gcd)) /
				(q * other_q_by_gcd)
		ensure
				-- a/b+c/d = (ad+cb)/bd =
				-- ((ad+cb)/g)/(bd/g) =
				-- (ad/g+cb/g)/(bd/g) =
				-- (a(d/g)+c(b/g))/(b(d/g)), where g = gcd (b, d)
			el_gcd: attached gcd (q, other.q) as el_gcd
			non_zero_gcd: el_gcd ≭ zero.p -- q /= 0 and other.q /= 0
			good_divisor: ( -- (other.q // el_gcd) /= 0
				p * (other.q // el_gcd) + other.p * (q // el_gcd)
				).divisible (q * (other.q // el_gcd))
			el_ratio: attached (
				(p * (other.q // el_gcd) + other.p * (q // el_gcd)) /
				(q * (other.q // el_gcd))
				) as el_ratio

			numerator: Result.p ≍ el_ratio.p
			denominator: Result.q ≍ el_ratio.q
		end

	minus alias "-" alias "−" (other: RATIONAL_NUMBER): like rational_anchor
			-- Result of subtracting `other` from current rational number
		local
			l_gcd, other_q_by_gcd: INTEGER_NUMBER
		do
			l_gcd := gcd (q, other.q)
				check
					non_zero_gcd: l_gcd ≭ zero.p -- q /= 0 and other.q /= 0
				end
			other_q_by_gcd := other.q // l_gcd
				check
					good_divisor: ( -- (other.q // l_gcd) /= 0
						p * other_q_by_gcd - other.p * (q // l_gcd)
						).divisible (q * other_q_by_gcd)
				end
			Result := (p * other_q_by_gcd - other.p * (q // l_gcd)) /
				(q * other_q_by_gcd)
		ensure
				-- Please have a look at `plus' definition.
			el_gcd: attached gcd (q, other.q) as el_gcd
			non_zero_gcd: el_gcd ≭ zero.p -- q /= 0 and other.q /= 0
			good_divisor: ( -- (other.q // el_gcd) /= 0
				p * (other.q // el_gcd) - other.p * (q // el_gcd)
				).divisible (q * (other.q // el_gcd))
			el_ratio: attached (
				(p * (other.q // el_gcd) - other.p * (q // el_gcd)) /
				(q * (other.q // el_gcd))
				) as el_ratio

			numerator: Result.p ≍ el_ratio.p
			denominator: Result.q ≍ el_ratio.q
		end

	opposite alias "-" alias "−": like rational_anchor
			-- Opposite value of current number relative to addition, i.e.
			-- `Current' + (- `Current') = 0.
			--| It departs from {REAL_NUMBER}.opposite in order to allow the
			--| latter to produce "negative" zeros without obliging current
			--| class to do the same.
			-- TODO: properties, tests
		deferred
		ensure
			sign: Result.sign ≍ - sign
			abs_value: Result.abs ≍ abs
		end

	product alias "*" alias "×" alias "⋅"
		(other: RATIONAL_NUMBER): like rational_anchor
			-- Current rational number multiplied by `other`
			--| NOTICE: There is a precondition! If the implementation of
			--| {INTEGER_NUMBER}.product produces overflows, we may get a
			--| denominator unexpectedly equal to `zero'. Since we can't assume
			--| that a particular overflow will consistently produce a value
			--| unequal to `zero', we better to avoid overflows altogether.
			--| Notice also that an occasional overflow on calculating the
			--| numerator, even if it yields a zero, does not produce an
			--| invalid rational number.
		require
			good_factor: multipliable (other)
		local
			l_gcd_1, l_gcd_2: like gcd
		do
			l_gcd_1 := gcd (p, other.q)
			l_gcd_2 := gcd (other.p, q)
				check
					non_zero_gcd_1: l_gcd_1 ≭ zero.p -- other.q /= 0
					non_zero_gcd_2: l_gcd_2 ≭ zero.p -- q /= 0
					good_divisor: -- q /= 0 and other.q /= 0
						((p // l_gcd_1) * (other.p // l_gcd_2)).divisible
							((q // l_gcd_2) * (other.q // l_gcd_1))
				end
			Result := ((p // l_gcd_1) * (other.p // l_gcd_2)) /
				((q // l_gcd_2) * (other.q // l_gcd_1))
		ensure
			gcd_1: attached gcd (p, other.q) as gcd_1
			gcd_2: attached gcd (other.p, q) as gcd_2
			non_zero_gcd_1: gcd_1 ≭ zero.p -- other.q /= 0
			non_zero_gcd_2: gcd_2 ≭ zero.p -- q /= 0
			good_divisor: ((p // gcd_1) * (other.p // gcd_2)).divisible
				((q // gcd_2) * (other.q // gcd_1)) -- q /= 0 and other.q /= 0
			el_ratio: attached (
				((p // gcd_1) * (other.p // gcd_2)) /
				((q // gcd_2) * (other.q // gcd_1))
				) as el_ratio

			numerator: Result.p ≍ el_ratio.p
			denominator: Result.q ≍ el_ratio.q
		end

	quotient alias "/" alias "÷" (other: RATIONAL_NUMBER): like rational_anchor
			-- Division of current rational number by `other`
		require
			good_divisor: divisible (other)
		do
				check
					good_factor: multipliable (other.inverse) -- divisible (other)
				end
			Result := Current ⋅ other.inverse
		ensure
				good_factor: multipliable (other.inverse) -- divisible (other)
			numerator: Result.p ≍ (Current ⋅ other.inverse).p
			denominator: Result.q ≍ (Current ⋅ other.inverse).q
		end

	reciprocal,
	inverse: like rational_anchor
			-- Reciprocal (AKA inverse) value of current rational number
			-- relative to multiplication, i.e.
			-- `Current' ⋅ `reciprocal' = `Current' ⋅ `inverse' = 1.
		require
			is_invertible: is_invertible
		do
				check
					good_divisor: q.divisible (p) -- p /= 0 ⇐ is_invertible
				end
			Result := q / p
		ensure
			good_divisor: q.divisible (p) -- p /= 0 ⇐ is_invertible
			numerator: Result.p ≍ (q / p).p
			denominator: Result.q ≍ (q / p).q
		end

feature -- Math

	gcd (i, j: INTEGER_NUMBER): like integer_anchor
			-- Greatest common divisor of `i' and `j'
		do
			if j ≍ zero.p then
				Result := converted_integer (i.abs)
			else
					check
						good_divisor: i.divisible (j) -- j /= 0
					end
				Result := gcd (j, i \\ j)
			end
		ensure
			base: j ≍ zero.p implies Result ≍ i.abs
			good_divisor: j ≭ zero.p implies i.divisible (j)
			induction: j ≭ zero.p implies Result ≍ gcd (j, i \\ j)
		end

	div (i, j: INTEGER_NUMBER): like integer_anchor
			-- The quotient `i'/`j' rounded towards negative infinity, i.e.
			-- ⌊`i'/`j'⌋.
		require
			good_divisor: i.divisible (j)
		do
			if (i \\ j) ≍ zero.p then
				Result := converted_integer (i // j)
			elseif i < zero.p then
				if j < zero.p then
					Result := converted_integer (i // j)
				else
						check
							j > zero.p -- i.divisible (j)
						end
					Result := converted_integer (i // j - one.p)
				end
			else
					check
						i > zero.p -- (i \\ j) /= 0
					end
				if j < zero.p then
					Result := converted_integer (i // j - one.p)
				else
						check
							j > zero.p -- i.divisible (j)
						end
					Result := converted_integer (i // j)
				end
			end
		ensure
			when_integer: (i \\ j) ≍ zero.p implies Result ≍ (i // j)
			when_both_negative:
				(i \\ j) ≭ zero.p and i < zero.p and j < zero.p implies
				Result ≍ (i // j)
			when_negative_by_positive:
				(i \\ j) ≭ zero.p and i < zero.p and j > zero.p implies
				Result ≍ (i // j - one.p)
			when_positive_by_negative:
				(i \\ j) ≭ zero.p and i > zero.p and j < zero.p implies
				Result ≍ (i // j - one.p)
			when_both_positive:
				(i \\ j) ≭ zero.p and i > zero.p and j > zero.p implies
				Result ≍ (i // j)
		end

	rem (i, j: INTEGER_NUMBER): like integer_anchor
			-- The remainder of the quotient `i'/`j' when the latter is rounded
			-- towards negative infinity, i.e. `i' - j⋅⌊`i'/`j'⌋.
		require
			good_divisor: i.divisible (j)
		do
			if (i \\ j) ≍ zero.p then
				Result := converted_integer (i \\ j)
			elseif i < zero.p then
				if j < zero.p then
					Result := converted_integer (i \\ j)
				else
						check
							j > zero.p -- i.divisible (j)
						end
					Result := converted_integer (i \\ j + j)
				end
			else
					check
						i > zero.p -- (i \\ j) /= 0
					end
				if j < zero.p then
					Result := converted_integer (i \\ j + j)
				else
						check
							j > zero.p -- i.divisible (j)
						end
					Result := converted_integer (i \\ j)
				end
			end
		ensure
			when_integer: (i \\ j) ≍ zero.p implies Result ≍ (i \\ j)
			when_both_negative:
				(i \\ j) ≭ zero.p and i < zero.p and j < zero.p implies
				Result ≍ (i \\ j)
			when_negative_by_positive:
				(i \\ j) ≭ zero.p and i < zero.p and j > zero.p implies
				Result ≍ (i \\ j + j)
			when_positive_by_negative:
				(i \\ j) ≭ zero.p and i > zero.p and j < zero.p implies
				Result ≍ (i \\ j + j)
			when_both_positive:
				(i \\ j) ≭ zero.p and i > zero.p and j > zero.p implies
				Result ≍ (i \\ j)
		end

feature -- Factory

	converted_integer (i: INTEGER_NUMBER): like integer_anchor
			-- `i' converted to a integer number like `integer_anchor'
		do
			Result := i
		ensure
			definition: Result.value = integer_anchor.adjusted_value (i.value)
		end

feature -- Predicate

	integer_product_overflows (i, j: INTEGER_NUMBER): BOOLEAN
			-- Does the product `i' ⋅ `j' overflow?
		do
			Result := i < zero.p and j < zero.p and i.max_value_exists and then
				i < i.max_value // j or
				i < zero.p and j > zero.p and i.min_value_exists and then
				i < i.min_value // j or
				i > zero.p and j < - one.p and i.min_value_exists and then
				i > i.min_value // j or
				i > zero.p and j > zero.p and i.max_value_exists and then
				i > i.max_value // j
		ensure
			definition: Result = (
				i < zero.p and j < zero.p and i.max_value_exists and then
				i < i.max_value // j or
				i < zero.p and j > zero.p and i.min_value_exists and then
				i < i.min_value // j or
				i > zero.p and j < - one.p and i.min_value_exists and then -- TODO: It assumes a two's-complement implementation, where i.min_value // j overflows.
				i > i.min_value // j or
				i > zero.p and j > zero.p and i.max_value_exists and then
				i > i.max_value // j
				)
		end

feature -- Anchor

	rational_anchor: RATIONAL_NUMBER
			-- Anchor for rational numbers
		deferred
		end

invariant
	numerator: numerator ≍ p
	denominator: denominator ≍ q
	non_zero_denominator: q ≭ zero.p
	identity: (+ Current) ≍ Current

note
	copyright: "Copyright (c) 2012-2022, Rosivaldo Fernandes Alves"
	license: "[
		Eiffel Forum License v2
		(see http://www.eiffel.com/licensing/forum.txt)
		]"
	source: "https://launchpad.net/set-theory"
end
