﻿note
	description: "[
		Model of a real number, i.e. a complex number whose imaginary part is
		zero.
		
		A finite representation, IEEE 754-like, is assumed, although with some
		minor changes it would allow even an implementation with real numbers
		of arbitrary precision. Anyway, this specification does not try to
		offer a high-grade numeric class. Its intent is just to introduce
		examples of how several numeric types and sets may receive a
		set-theoretic treatment.		
		
		It is Noteworthy that this specification does not inherit from
		{NUMERIC} or {COMPARABLE}, since those classes use covariant arguments,
		which this library avoids as much as possible.
		]"
	author: "Rosivaldo Fernandes Alves"
	date: "$Date$"
	revision: "$Revision$"

deferred class
	REAL_NUMBER

inherit
	COMPLEX_NUMBER
		rename
			a as identity alias "+",
			real as identity alias "+",
			b as zero,
			imaginary as zero,
			is_in as complex_is_in,
			is_not_in as complex_is_not_in,
			equals as complex_equals,
			unequals as complex_unequals,
			divisible as complex_divisible,
			plus as complex_plus,
			minus as complex_minus,
			product as complex_product,
			quotient as complex_quotient
		undefine
			is_rational
		redefine
			is_infinite,
			is_finite,
			is_integer,
			reciprocal, inverse
		end

feature -- Primitive

	value: like native_real_anchor
			-- Native value of current real number
		deferred
		ensure
			adjusted: Result = adjusted_value (Result)
		end

feature -- Membership

	is_in alias "∈" (s: SET [REAL_NUMBER, EQUALITY [REAL_NUMBER]]): BOOLEAN
			-- Does `s` have current real number?
		do
			Result := s ∋ Current
		ensure
			definition: Result = s ∋ Current
		end

	is_not_in alias "∉" (s: SET [REAL_NUMBER, EQUALITY [REAL_NUMBER]]): BOOLEAN
			-- Is not current real number in `s'?
		do
			Result := not (Current ∈ s)
		ensure
			definition: Result = not (Current ∈ s)
		end

feature -- Access

	sign: like integer_anchor
			-- Sign of current number as an integer value (0, -1 or 1)
		do
			Result := Current ⋚ zero
		ensure
			three_way: Result ≍ (Current ⋚ zero)
		end

	sign_bit: like integer_anchor
			-- Status of the sign bit of `value', which is 1 for negative numbers but also, e.g. for a "negative" zero as specified by IEEE 754.
			--| TODO: Natural instead?
		deferred
		ensure
			definition: Result.value = value_sign_bit (value)
		end

	zero: like real_anchor
			-- The real number 0
		deferred
		end

	one: like real_anchor
			-- The real number 1
		deferred
		end

	machine_epsilon: like real_anchor
			-- The difference between 1 and the least value greater than 1 that is representable by current kind of real number.
		do
			Result := one.relative_epsilon
		ensure
			definition: Result ≍ one.relative_epsilon
		end

	ulp,
	relative_epsilon: like real_anchor
			-- Unit in last place (ulp) of the floating representation of current real number
			--| This specifications follows the definition given at https://docs.julialang.org/en/v1/base/base/#Base.eps-Tuple{AbstractFloat}.
		do
			if previous_float.is_negative_infinity then
				Result := next_float - Current
			elseif next_float.is_positive_infinity then
				Result := Current - previous_float
			else
				Result := abs.next_float - abs
			end
		ensure
			when_huge_negative: previous_float.is_negative_infinity implies Result ≍ (next_float - Current)
			when_huge_positive: next_float.is_positive_infinity implies Result ≍ (Current - previous_float)
			ordinary_definition: not previous_float.is_negative_infinity and not next_float.is_positive_infinity implies Result ≍ (abs.next_float - abs)
		end

	previous_float: like real_anchor
			-- Greatest representable value that is less than current real number
		deferred
		ensure
			when_nan: is_nan implies Result.is_nan
			when_negative_infinity: is_negative_infinity implies Result.is_negative_infinity
			ordinary_case: not is_nan and not is_negative_infinity implies Result < Current
			greatest_below: -- Please see `previous_float' post-condition of `is_greater'
		end

	next_float: like real_anchor
			-- Least representable value that is greater than current real number
		deferred
		ensure
			when_nan: is_nan implies Result.is_nan
			when_positive_infinity: is_positive_infinity implies Result.is_positive_infinity
			ordinary_case: not is_nan and not is_positive_infinity implies Current < Result
			least_above: -- Please see `next_float' post-condition of `is_less'
		end

feature -- Quality

	is_nan: BOOLEAN
			-- Is current the representation of NaN, i.e. does it represent no valid real number?
		do
			Result := value.is_nan
		ensure
			definition: Result = value.is_nan
			total_order: Result implies Current ≍ Current
		end

	is_negative_infinity: BOOLEAN
			-- Is current the representation of negative_infinity, i.e. is it not greater than any valid real number?
		do
			Result := value.is_negative_infinity
		ensure
			definition: Result = value.is_negative_infinity
		end

	is_negative_zero: BOOLEAN
			-- Is current the representation of a "negative" zero, as specified by IEEE 754?
		do
			Result := Current ≍ zero and sign_bit ≍ one.truncated_to_integer
		ensure
			definition: Result = (Current ≍ zero and sign_bit ≍ one.truncated_to_integer)
		end

	is_positive_zero: BOOLEAN
			-- Is current the representation of a "positive" zero, as specified by IEEE 754?
		do
			Result := Current ≍ zero and sign_bit ≍ zero.truncated_to_integer
		ensure
			definition: Result = (Current ≍ zero and sign_bit ≍ zero.truncated_to_integer)
		end

	is_positive_infinity: BOOLEAN
			-- Is current the representation of positive_infinity, i.e. is it not less than any valid real number?
		do
			Result := value.is_positive_infinity
		ensure
			definition: Result = value.is_positive_infinity
		end

	is_infinite: BOOLEAN
			-- Is current the representation of an infinity, i.e. either it `is_negative_infinity' or `is_positive_infinity'?
		do
			Result := is_negative_infinity or is_positive_infinity
		ensure then
			real_definition: Result = (is_negative_infinity or is_positive_infinity)
		end

	is_finite: BOOLEAN
			-- Is current the representation of a finite real number, i.e. neither it `is_nan' or `is_infinite'?
		do
			Result := not (is_nan or is_infinite)
		ensure then
			real_definition: Result = not (is_nan or is_infinite)
		end

	is_rational: BOOLEAN
			-- Does current real number represent a rational number?
		deferred
		ensure then
			splitted_one: attached splitted (one) as splitted_one
			a: attached splitted_one.a as a
			b: attached splitted_one.b as b
			real_definition:; Result = (
				(a.is_integer or zero.truncated_to_integer.max_value_exists and then (a - one) ≍ zero.truncated_to_integer.max_value) and
				(b.is_integer or zero.truncated_to_integer.max_value_exists and then (b - one) ≍ zero.truncated_to_integer.max_value)
				)
		end

	is_integer: BOOLEAN
			-- Does current real number represent an integer number?
		do
			Result := Current ≍ truncated_to_integer
		ensure then
			real_definition: Result = Current ≍ truncated_to_integer
		end

	is_invertible: BOOLEAN
			-- Does current real number have a multiplicative inverse
			-- (AKA reciprocal)?
			-- Notice that the real_quasi_definition post-condition allows an
			-- implementation to regard zero as invertible.
		do
			Result := Current ≭ zero
		ensure then
			real_quasi_definition: Current ≭ zero implies Result
		end

feature -- Comparison

	equals alias "≍" (x: REAL_NUMBER): BOOLEAN
			-- Do current real number and `x' have the same numeric value?
			--| Please have a look at the note in the header of {SET}.equals.
		do
			Result := value = x.value
		ensure
			definition: Result = (value = x.value)
		end

	unequals alias "≭" (x: REAL_NUMBER): BOOLEAN
			-- Does not current real number equal `x'?
		do
			Result := not (Current ≍ x)
		ensure
			definition: Result = not (Current ≍ x)
		end

	is_less alias "<" (x: REAL_NUMBER): BOOLEAN
			-- Is current real number less than `x'?
		do
			Result := value < x.value
		ensure
			definition: Result = (value < x.value)
			previous_float: Result implies next_float ≤ x
		end

	is_less_equal alias "<=" alias "≤" (x: REAL_NUMBER): BOOLEAN
			-- Is current real number less than or equal to `x'?
		do
			Result := Current < x or Current ≍ x
		ensure
			definition: Result = (Current < x or Current ≍ x)
		end

	is_greater alias ">" (x: REAL_NUMBER): BOOLEAN
			-- Is current real number greater than `x'?
		do
			Result := x < Current
		ensure
			definition: Result = (x < Current)
			previous_float: Result implies x ≤ previous_float
		end

	is_greater_equal alias ">=" alias "≥" (x: REAL_NUMBER): BOOLEAN
			-- Is current real number greater than or equal to `x'?
		do
			Result := x ≤ Current
 		ensure
			definition: Result = (x ≤ Current)
		end

	three_way_comparison alias "⋚" (x: REAL_NUMBER): like integer_anchor
			-- If current real number equal to `x', 0; if smaller, -1; if greater, 1.
		do
			if Current < x then
				Result := - one.truncated_to_integer
			elseif Current ≍ x then
				Result := zero.truncated_to_integer
			else
				Result := one.truncated_to_integer
			end
		ensure
			equal_zero: (Result ≍ zero.truncated_to_integer) = (Current ≍ x)
			smaller_negative: (Result ≍ - one.truncated_to_integer) = (Current < x)
			greater_positive: (Result ≍ one.truncated_to_integer) = (Current > x)
		end

	min alias "∧" (x: REAL_NUMBER): like real_anchor
			-- The smaller of current real number and `x'
		do
			if Current ≤ x then
				Result := Current
			else
				Result := x
			end
		ensure
			current_if_not_greater: Current ≤ x implies Result ≍ Current
			other_if_greater: x < Current implies Result ≍ x
		end

	max alias "∨" (x: REAL_NUMBER): like real_anchor
			-- The greater of current real number and `x'
		do
			if Current ≥ x then
				Result := Current
			else
				Result := x
			end
		ensure
			current_if_not_smaller: Current ≥ x implies Result ≍ Current
			other_if_smaller: Current < x implies Result ≍ x
		end

feature -- Relationship

	divisible (x: REAL_NUMBER): BOOLEAN
			-- May current real number be divided by `x`?
		do
			Result := x.is_invertible
		ensure
			definition: Result = x.is_invertible
		end

feature -- Operation

	plus alias "+" (x: REAL_NUMBER): like real_anchor
			-- Sum of current real number with `x`
		do
			Result := real_from_value (value + x.value)
		ensure
			definition: Result ≍ real_from_value (value + x.value)
		end

	minus alias "-" alias "−" (x: REAL_NUMBER): like real_anchor
			-- Result of subtracting `x` from current real number
		do
			Result := real_from_value (value - x.value)
		ensure
			definition: Result ≍ real_from_value (value - x.value)
		end

	opposite alias "-" alias "−": like real_anchor
			-- <Precursor>
		do
			Result := real_from_value (- value)
		ensure then
			definition: Result ≍ real_from_value (- value)
		end

	product alias "*" alias "×" alias "⋅" (x: REAL_NUMBER): like real_anchor
			-- Current real number multiplied by `x`
		do
			Result := real_from_value (value × x.value)
		ensure
			definition: Result ≍ real_from_value (value × x.value)
		end

	quotient alias "/" alias "÷" (x: REAL_NUMBER): like real_anchor
			-- Division of current real number by `x`
			-- Notice that quasi_definition post-condition allows an implementation to regard `zero' a good divisor.
		require
			good_divisor: divisible (x)
		do
				check
					good_divisor: value.divisible (x.value) -- x.is_invertible <== divisible (x)
				end
			Result := real_from_value (value / x.value)
		ensure
			quasi_definition: value.divisible (x.value) implies Result ≍ real_from_value (value / x.value)
		end

	reciprocal,
	inverse: like real_anchor
			-- <Precursor>
		do
				check
					good_divisor: one.divisible (Current) -- `is_invertible'						
				end
			Result := one / Current
		end

	power alias "^" (x: REAL_NUMBER): like real_anchor
			-- Current real number to the power `x'
		do
			Result := real_from_double_value (value ^ x.value)
		ensure
			definition: Result ≍ real_from_double_value (value ^ x.value)
		end

	square_root alias "√¯": like real_anchor
			-- Square root of current real number
		require
			non_negative: Current >= zero
		do
			Result := real_from_value (sqrt (value))
		ensure
			definition: Result ≍ real_from_value (sqrt (value))
		end

feature -- Conversion

	to_rational: like rational_anchor
			-- Rational representation of current real number
			-- TODO: Properties, tests
		require
			is_rational: is_rational
		deferred
		ensure
			splitted_one: attached splitted (one) as splitted_one
			a: attached splitted_one.a as a
			b: attached splitted_one.b as b
			numerator: a.abs ≍ Result.p.rational_abs
			denominator: b.abs ≍ Result.q.rational_abs
			same_sign: Result.sign ≍ sign
		end

	truncated_to_integer: like integer_anchor
			-- Integer part (same sign, largest absolute value no greater
			-- than current real number's)
		deferred
		ensure
			definition: Result.value =
				Result.adjusted_value (value.truncated_to_integer)
		end

feature -- Math

	sqrt (v: like real_anchor.value): like real_anchor.value
			-- Square root of `v`
		do
			Result := {like real_math_anchor}.sqrt (v)
		ensure
			class
			definition: Result = {like real_math_anchor}.sqrt (v)
		end

	splitted (q: REAL_NUMBER): HOMOGENEOUS_ORDERED_PAIR [REAL_NUMBER, REAL_EQUALITY]
			-- Current real number represented as an ordered pair (a, b) = (`Current' ⋅ (2 ^ k), `q' ⋅ (2 ^ k)), k ∈ ℕ.
		deferred
		ensure
				-- TODO: Get rid of `floor_real_32'.
			when_integral: value = value.floor_real_32 implies Result.a ≍ Current and Result.b ≍ q
			when_fractional: value /= value.floor_real_32 implies Result ≍ (Current ⋅ two).splitted (q ⋅ two)
		end

feature -- Implementation

	adjusted_value (v: like value): like value
			-- `v' adjusted to represent the `value' of a real number like
			-- current
		deferred
		ensure
			stable: adjusted_value (Result) = Result
		end

feature -- Anchor

	rational_anchor: RATIONAL_NUMBER
			-- Anchor for rational numbers
		deferred
		end

invariant
	machine_epsilon: Current > one implies (Current - one) ≥ machine_epsilon
	modulus: modulus.value = value.abs
	abs: abs.value = value.abs
	identity: (+ Current).value = value

note
	copyright: "Copyright (c) 2012-2022, Rosivaldo Fernandes Alves"
	license: "[
		Eiffel Forum License v2
		(see http://www.eiffel.com/licensing/forum.txt)
		]"
	source: "https://launchpad.net/set-theory"
end
