﻿note
	description: "Set of ordered numbers"
	author: "Rosivaldo Fernandes Alves"
	date: "$Date$"
	revision: "$Revision$"

deferred class
	ORDERED_NUMERIC_SET [A -> {REAL_NUMBER}, EQ -> NUMERIC_EQUALITY [A]] -- TODO: Replace {REAL_NUMBER} by {ORDERED_NUMBER}?

inherit
	NUMERIC_SET [A, EQ]
		redefine
			set_anchor,
			subset_anchor,
			superset_anchor
		end

feature -- Restriction

	non_positive: like subset_anchor
			-- This set deprived of positive numbers, i.e. (←, 0].
		do
			Result := left_unbounded_right_closed_interval (zero)
		ensure
			definition: Result ≍ left_unbounded_right_closed_interval (zero)
		end

	non_negative: like subset_anchor
			-- This set deprived of negative numbers, i.e. [0, →).
		do
			Result := left_closed_right_unbounded_interval (zero)
		ensure
			definition: Result ≍ left_closed_right_unbounded_interval (zero)
		end

	negative: like subset_anchor
			-- The negative numbers in this set, i.e. (←, 0).
		do
			Result := left_unbounded_right_open_interval (zero)
		ensure
			definition: Result ≍ left_unbounded_right_open_interval (zero)
		end

	positive: like subset_anchor
			-- The positive numbers in this set, i.e. (0, →).
		do
			Result := left_open_right_unbounded_interval (Zero)
		ensure
			definition: Result ≍ left_open_right_unbounded_interval (Zero)
		end

feature -- Interval

	closed_interval alias "[]" (a, b: A): like subset_anchor
			-- Set of every number x in current set such that `a' ≤ x ≤ `b', i.e. [`a', `b'].
		do
			Result := Current | agent anded (agent number_is_less_equal (a, ?), agent number_is_less_equal (?, b), ?)
		ensure
			definition: Result ≍ (Current | agent anded (agent number_is_less_equal (a, ?), agent number_is_less_equal (?, b), ?))
		end

	left_closed_right_open_interval (a, b: A): like subset_anchor
			-- Set of every number x in current set such that `a' ≤ x < `b', i.e. [`a', `b').
		do
			Result := Current | agent anded (agent number_is_less_equal (a, ?), agent number_is_less (?, b), ?)
		ensure
			definition: Result ≍ (Current | agent anded (agent number_is_less_equal (a, ?), agent number_is_less (?, b), ?))
		end

	left_open_right_closed_interval (a, b: A): like subset_anchor
			-- Set of every number x in current set such that `a' < x ≤ `b', i.e. (`a', `b'].
		do
			Result := Current | agent anded (agent number_is_less (a, ?), agent number_is_less_equal (?, b), ?)
		ensure
			definition: Result ≍ (Current | agent anded (agent number_is_less (a, ?), agent number_is_less_equal (?, b), ?))
		end

	open_interval alias "()" (a, b: A): like subset_anchor
			-- Set of every number x in current set such that `a' < x < `b', i.e. (`a', `b').
		do
			Result := Current | agent anded (agent number_is_less (a, ?), agent number_is_less (?, b), ?)
		ensure
			definition: Result ≍ (Current | agent anded (agent number_is_less (a, ?), agent number_is_less (?, b), ?))
		end

	left_closed_right_unbounded_interval (a: A): like subset_anchor
			-- Set of every number x in current set such that `a' ≤ x, i.e. [`a', →).
		do
			Result := Current | agent number_is_less_equal(a, ?)
		ensure
			definition: Result ≍ (Current | agent number_is_less_equal(a, ?))
		end

	left_open_right_unbounded_interval (a: A): like subset_anchor
			-- Set of every number x in current set such that `a' < x, i.e. (`a', →).
		do
			Result := Current | agent number_is_less(a, ?)
		ensure
			definition: Result ≍ (Current | agent number_is_less(a, ?))
		end

	left_unbounded_right_closed_interval (b: A): like subset_anchor
			-- Set of every number x in current set such that x ≤ `b', i.e. (←, `b'].
		do
			Result := Current | agent number_is_less_equal(?, b)
		ensure
			definition: Result ≍ (Current | agent number_is_less_equal(?, b))
		end

	left_unbounded_right_open_interval (b: A): like subset_anchor
			-- Set of every number x in current set such that x < `b', i.e. (←, `b')
		do
			Result := Current | agent number_is_less(?, b)
		ensure
			definition: Result ≍ (Current | agent number_is_less(?, b))
		end

	unbounded_interval: like Current
			-- Current set itself, i.e. (←, →).
		do
			Result := Current
		ensure
			definition: Result ≍ Current
		end

feature -- Element comparison

	min (a, b: A): A
			-- The smaller of `a' and `b'
			--| `a' ∧ `b' doesn't work well with generic entities.
		do
			if a ≤ b then
				Result := a
			else
				Result := b
			end
		ensure
			class
			definition: Result ≍ (a ∧ b)
		end

	max (a, b: A): A
			-- The greater of `a' and `b'
			--| `a' ∨ `b' doesn't work well with generic entities.
		do
			if a ≥ b then
				Result := a
			else
				Result := b
			end
		ensure
			class
			definition: Result ≍ (a ∨ b)
		end

feature -- Predicate

	anded (p1, p2: PREDICATE [A]; x: A): BOOLEAN
			-- Does `p1' (`x') or `p2' (`x') hold?
			--| TODO: DRY.
		do
			Result := p1 (x) and p2 (x)
		ensure
			class
			definition: Result = (p1 (x) and p2 (x))
		end

	number_is_less (x, y: A): BOOLEAN
			-- Is `x' less than `y'?
		do
			Result := x < y
		ensure then
			class
			definition: Result = (x < y)
		end

	number_is_less_equal (x, y: A): BOOLEAN
			-- Is `x' less than or equal to `y'?
		do
			Result := x ≤ y
		ensure then
			class
			definition: Result = (x ≤ y)
		end

feature -- Anchor

	set_anchor,
	subset_anchor,
	superset_anchor: ORDERED_NUMERIC_SET [A, EQ]
			-- <Precursor>
		deferred
		end

note
	copyright: "Copyright (c) 2012-2021, Rosivaldo Fernandes Alves"
	license: "[
		Eiffel Forum License v2
		(see http://www.eiffel.com/licensing/forum.txt)
		]"
	source: "https://launchpad.net/set-theory"

end
