﻿note
	description: "[
		Object able to transform objects derived from a generic parameter into objects derived from another generic parameter.
		]"
	author: "Rosivaldo Fernandes Alves"
	date: "$Date: Sat 2018-12-01 19:25:58 -0300 $"
	revision: "$Revision$"

deferred class
	TRANSFORMER [A, B, EQ_A -> EQUALITY [A], EQ_B -> EQUALITY [B]]

inherit
	ELEMENT

feature -- Access

	eq_b: EQ_B
			-- Equality for elements like {B}
			--| Please have a look at the note in the header of {SET}.equals.
		deferred
		ensure
			class
		end

feature -- Operation

	cartesian_product (xs: SET [A, EQUALITY [A]] ys: SET [B, EQUALITY [B]]): like product_anchor
			-- Cartesian product of `xs' and `ys', i.e. every ordered pair whose first and second elements are in `xs' and in `ys', respectively.
		deferred
		ensure
			every_pair: xs |∀ agent transformer.every_pair_is_in_set (?, ys, Result)
			nothing_else: Result.for_all_xy (agent transformer.anded (agent xs.has, ?, agent ys.has, ?))
		end

feature -- Transformation
	-- todo: What if some `f' has a precondition?

	set_map (xs: SET [A, EQUALITY [A]]; f: FUNCTION [A, B]): like set_map_anchor
			-- Set whose elements are the values of `f' (x), for all x in `xs', i.e. {`f' (x) | x ∈ `xs'}.
		deferred
		ensure
			maps_everything: xs |∀ agent transformer.target_set_has_value (Result, f, ?)
			nothing_else: Result |∀ agent transformer.source_set_has_argument (f, xs, ?)
		end

	set_reduction (xs: SET [A, EQUALITY [A]]; start: B; f: FUNCTION [B, A, B]): B
			-- `xs' reduced by `f' to a value like `start'
			-- NOTICE: Since the order of the elements in a set is irrelevant, the value of `set_reduction' may differ for two equal sets, unless `f' is "commutative", as
			-- explained at "Painfully detailed information about ambiguous folds" (https://preview.tinyurl.com/y4qp7zre). Additionally, please see the comment at {STS_SET}.reduced
			-- header.
		do
			if xs.is_empty then
				Result := start
			else
				Result := set_reduction (xs.others, f (start, xs.any), f)
			end
		ensure
			class
			base: xs.is_empty implies eq_b (Result, start)
			induction: not xs.is_empty implies eq_b (Result, set_reduction (xs.others, f (start, xs.any), f))
		end

	tuple_map (xt: N_TUPLE [A, EQUALITY [A]]; f: FUNCTION [A, B]): like tuple_map_anchor
			-- N-tuple whose i-th term is the value of `f` (`xt' [i]), for all i in `xt`.`indices'.
		deferred
		ensure
			maps_everything: xt.for_all (agent transformer.target_tuple_has_value (Result, ?, f, ?))
			nothing_else: Result.for_all (agent transformer.source_tuple_has_argument (f, xt, ?, ?))
		end

	tuple_left_reduction (xt: N_TUPLE [A, EQUALITY [A]]; leftmost: B; f: FUNCTION [B, A, B]): B
			-- Value got by applying `f', from left to right, to the terms in `xt'; `leftmost' is the default result.
		do
			if xt.is_unit then
				Result := leftmost
			else
				Result := tuple_left_reduction (xt.tail, f (leftmost, xt.first), f)
			end
		ensure
			class
			base: xt.is_unit implies eq_b (Result, leftmost)
			induction: not xt.is_unit implies eq_b (Result, tuple_left_reduction (xt.tail, f (leftmost, xt.first), f))
		end

	tuple_indexed_left_reduction (xt: N_TUPLE [A, EQUALITY [A]]; leftmost: B; i: NATURAL; f: FUNCTION [B, NATURAL, A, B]): B
			-- Value got by applying `f', from left to right, to the terms in `xt', indexed sequentially from `i' upwards; `leftmost' is the default result.
		do
			if xt.is_unit then
				Result := leftmost
			else
				Result := tuple_indexed_left_reduction (xt.tail, f (leftmost, i, xt.first), i + 1, f)
			end
		ensure
			class
			base: xt.is_unit implies eq_b (Result, leftmost)
			induction: not xt.is_unit implies eq_b (Result, tuple_indexed_left_reduction (xt.tail, f (leftmost, i, xt.first), i + 1, f))
		end

	tuple_right_reduction (xt: N_TUPLE [A, EQUALITY [A]]; f: FUNCTION [A, B, B]; rightmost: B): B
			-- Value got by applying `f', from right to left, to the terms in `xt'; `rightmost' is the default value.
		do
			if xt.is_unit then
				Result := rightmost
			else
				Result := f (xt.first, tuple_right_reduction (xt.tail, f, rightmost))
			end
		ensure
			class
			base: xt.is_unit implies eq_b (Result, rightmost)
			induction: not xt.is_unit implies eq_b (Result, f (xt.first, tuple_right_reduction (xt.tail, f, rightmost)))
		end

	tuple_indexed_right_reduction (xt: N_TUPLE [A, EQUALITY [A]]; i: NATURAL; f: FUNCTION [NATURAL, A, B, B]; rightmost: B): B
			-- Value got by applying `f', from right to left, to the terms in `xt', indexed sequentially from `i' upwards; `rightmost' is the default value.
		do
			if xt.is_unit then
				Result := rightmost
			else
				Result := f (i, xt.first, tuple_indexed_right_reduction (xt.tail, i + 1, f, rightmost))
			end
		ensure
			class
			base: xt.is_unit implies eq_b (Result, rightmost)
			induction: not xt.is_unit implies eq_b (Result, f (i, xt.first, tuple_indexed_right_reduction (xt.tail, i + 1, f, rightmost)))
		end

feature -- Factory

	pair (a: A; b: B): ORDERED_PAIR [A, B, EQ_A, EQ_B]
			-- Ordered pair in the form (`a', `b')
		deferred
		ensure
			first: Result.eq_a (Result.a, a)
			second: Result.eq_b (Result.b, b)
		end

feature -- Predicate
	-- todo: Explain.

	anded (p: PREDICATE [A]; x: A; q: PREDICATE [B]; y: B): BOOLEAN
			-- Does `p' (`x') and `q' (`y') hold?
		do
			Result := p (x) and q (y)
		ensure
			class
			definition: Result = (p (x) and q (y))
		end

	every_pair_is_in_set (x: A; ys: SET [B, EQUALITY [B]]; ps: SET_OF_ORDERED_PAIRS [A, B, EQ_A, EQ_B]): BOOLEAN
			-- Does `ps' have every pair in the form (`x', y), where y is in `ys'?
		do
			Result := ys |∀ agent transformer.pair_is_in_set (x, ?, ps)
		ensure
			class
			definition: Result = (ys |∀ agent transformer.pair_is_in_set (x, ?, ps))
		end

	pair_is_in_set (x: A; y: B ps: SET_OF_ORDERED_PAIRS [A, B, EQ_A, EQ_B]): BOOLEAN
			-- Does `ps' have the pair (`x', `y')?
		do
			Result := transformer.pair (x, y) ∈ ps
		ensure
			class
			definition: Result = transformer.pair (x, y) ∈ ps
		end

	target_set_has_value (ys: SET [B, EQ_B]; f: FUNCTION [A, B] x: A): BOOLEAN
			-- Does `ys' have the value `f' (`x')?
		do
			Result := ys ∋ f (x)
		ensure
			class
			definition: Result = ys ∋ f (x)
		end

	source_set_has_argument (f: FUNCTION [A, B]; xs: SET [A, EQUALITY [A]]; y: B): BOOLEAN
			-- Does `xs' have some element x such that `f' (`x') = `y'?
		do
			Result := xs |∃ agent transformer.argument_is_mapped_to_value (f, ?, y)
		ensure
			class
			definition: Result = (xs |∃ agent transformer.argument_is_mapped_to_value (f, ?, y))
		end

	argument_is_mapped_to_value (f: FUNCTION [A, B]; x: A; y: b): BOOLEAN
			-- Does `f' map `x' to `y', i.e. `f' (`x') = `y'?
		do
			Result := eq_b (f (x), y)
		ensure
			class
			definition: Result = eq_b (f (x), y)
		end

	target_tuple_has_value (yt: N_TUPLE [B, EQ_B] i: NATURAL f: FUNCTION [A, B] x: A): BOOLEAN
			-- Does `yt' [`i'] equal `f' (`x')?
		do
			if 0 < i and i <= yt.n then
				Result := eq_b (yt [i], f (x))
			end
		ensure
			class
			invalid_index: i = 0 or i > yt.n implies not Result
			definition: 0 < i and i <= yt.n implies Result = eq_b (yt [i], f (x))
		end

	source_tuple_has_argument (f: FUNCTION [A, B] xt: N_TUPLE [A, EQUALITY [A]] i: NATURAL y: B): BOOLEAN
			-- Does `f' (`xt' [`i']) equal `y'?
		do
			if 0 < i and i <= xt.n then
				Result := eq_b (f (xt [i]), y)
			end
		ensure
			class
			invalid_index: i = 0 or i > xt.n implies not Result
			definition: 0 < i and i <= xt.n implies Result = eq_b (f (xt [i]), y)
		end

feature -- Transformer

	transformer: like Current
			-- Instance of this class to be used as target of agents
		deferred
		ensure
			class
		end

feature -- Anchor

	product_anchor: SET_OF_ORDERED_PAIRS [A, B, EQ_A, EQ_B]
			-- Anchor for Cartesian products
		deferred
		end

	set_map_anchor: SET [B, EQ_B]
			-- Anchor for set maps
		deferred
		end

	tuple_map_anchor: N_TUPLE [B, EQ_B]
			-- Anchor for n-tuple maps
		deferred
		end

note
	copyright: "Copyright (c) 2012-2021, Rosivaldo Fernandes Alves"
	license: "[
		Eiffel Forum License v2
		(see http://www.eiffel.com/licensing/forum.txt)
		]"
	source: "https://launchpad.net/set-theory"

end
