﻿note
	description: "[
		Element that models a mathematical n-tuple.
		From a purist outlook, it should be a kind of set, maybe following Kuratowski's definition (http://preview.tinyurl.com/y7z2j62p). But I found that it only complicates the
		design for no true gain.
		]"
	author: "Rosivaldo Fernandes Alves"
	date: "$Date: Sat 2018-12-01 19:25:58 -0300 $"
	revision: "$Revision$"

deferred class
	N_TUPLE [A, EQ -> EQUALITY [A]]

inherit
	ELEMENT
		rename
			is_in as element_is_in,
			is_not_in as element_is_not_in
		end

feature -- Primitive

	is_unit: BOOLEAN
			-- Is current n-tuple a unit, i.e. does it have no term?
		deferred
		end

	first: A
			-- First term in current n-tuple
		require
			is_not_unit: not is_unit
		deferred
		end

	tail: like trimmed_tuple_anchor
			-- N-tuple of terms that stand after the `first' term in current n-tuple
		deferred
		end

	eq: EQ
			-- Rule for testing equality between terms in current n-tuple
			--| Please have a look at the note in the header of {SET}.eq.
		deferred
		end

feature -- Membership

	is_in alias "∈" (s: SET [N_TUPLE [A, EQ], EQUALITY [N_TUPLE [A, EQ]]]): BOOLEAN
			-- Does `s` have current n-tuple?
		do
			Result := s ∋ Current
		ensure
			definition: Result = s ∋ Current
		end

	is_not_in alias "∉" (s: SET [N_TUPLE [A, EQ], EQUALITY [N_TUPLE [A, EQ]]]): BOOLEAN
			-- Is not current n-tuple in `s'?
		do
			Result := not (Current ∈ s)
		ensure
			definition: Result = not (Current ∈ s)
		end

	has (a: A): BOOLEAN
			-- Is `a` a term in current n-tuple?
		do
			Result := terms ∋ a
		ensure
			definition: Result = terms ∋ a
		end

	does_not_have (a: A): BOOLEAN
			-- Does not current n-tuple have `a'?
		do
			Result := not has (a)
		ensure
			definition: Result = not has (a)
		end

feature -- Access

	indices: TRANSFORMABLE_SET [NATURAL, A, OBJECT_EQUALITY [NATURAL], EQ]
			-- Set with the indices of current n-tuple
		deferred
		ensure
			n_indices: # Result = n
			positive: Result |∀ agent index_is_positive
			small_enough: Result |∀ agent index_is_less_equal (?, n)
		end

	term alias "[]" (i: NATURAL): like first
			-- The `i'-th term in current n-tuple
		require
			positive: 0 < i
			small_enough: i ≤ n
		do
				check
					is_not_unit: not is_unit -- n > 0
				end
			Result := tail.indexed_left_reduced (first, 2, agent indexed_term (i, ?, ?, ?))
		ensure
			is_not_unit: not is_unit -- n > 0
			definition: eq (Result, tail.indexed_left_reduced (first, 2, agent indexed_term (i, ?, ?, ?)))
		end

	terms: like terms_anchor
			-- Set with every term in current n-tuple
		do
				check
					positive: indices |∀ agent index_is_positive -- by definition
					small_enough: indices |∀ agent index_is_less_equal (?, n) -- by definition
				end
			Result := converted_terms (indices ↦ agent term)
		ensure
			positive: indices |∀ agent index_is_positive -- by definition
			small_enough: indices |∀ agent index_is_less_equal (?, n) -- by definition
			definition: Result ≍ (indices ↦ agent term)
		end

	front: like trimmed_tuple_anchor
			-- N-tuple of terms that stand before the `last' term in current n-tuple
		do
			Result := prefix (tail.n)
		ensure
			definition: Result ≍ prefix (tail.n)
		end

	last: like first
			-- Last term in current n-tuple
		require
			is_not_unit: not is_unit
		do
				check
					positive: 0 < n -- not `is_unit'
					small_enough: n ≤ n -- by definition
				end
			Result := Current [n]
		ensure
			positive: 0 < n -- not `is_unit'
			small_enough: n ≤ n -- by definition
			definition: eq (Result, Current [n])
		end

feature -- Construction

	before (a: A): like extended_tuple_anchor
			-- Current n-tuple followed by `a'
		deferred
		ensure
			is_not_unit: not Result.is_unit
			front: Result.front ≍ Current
			last: eq (Result.last, a)
		end

	after (a: A): like extended_tuple_anchor
			-- Current n-tuple preceded by `a'
		deferred
		ensure
			is_not_unit: not Result.is_unit
			first: eq (Result.first, a)
			tail: Result.tail ≍ Current
		end

feature -- Measurement

	n: NATURAL
			-- Number of terms in current n-tuple
		do
			Result := term_to_index.tuple_indexed_left_reduction (Current, 0, 1, agent just_index)
		ensure
			definition: Result = term_to_index.tuple_indexed_left_reduction (Current, 0, 1, agent just_index)
		end

feature -- Comparison

	equals alias "≍" (t: N_TUPLE [A, EQ]): BOOLEAN
			-- Does current n-tuple have the same terms in `t', at the same indices?
			--| Please have a look at the note in the header of {SET}.equals.
		do
			Result := for_all (agent t.terms_are_equal) and t.for_all (agent terms_are_equal)
		ensure
			definition: Result = (for_all (agent t.terms_are_equal) and t.for_all (agent terms_are_equal))
		end

	unequals alias "≭" (t: N_TUPLE [A, EQ]): BOOLEAN
			-- Does not current n-tuple equal `t'?
		do
			Result := not (Current ≍ t)
		ensure
			definition: Result = not (Current ≍ t)
		end

feature -- Quantifier

	exists (p: PREDICATE [NATURAL, A]): BOOLEAN
			-- Existential quantifier: does `p' hold for some index and its respective term in current n-tuple?
		do
			Result := indices |∃ agent predicate_holds_to_term (p, ?)
		ensure
			definition: Result = (indices |∃ agent predicate_holds_to_term (p, ?))
		end

	does_not_exist (p: PREDICATE [NATURAL, A]): BOOLEAN
			-- Existence negation: does `p' hold for no index and its respective term in current n-tuple?
		do
			Result := not exists (p)
		ensure
			definition: Result = not exists (p)
		end

	for_all (p: PREDICATE [NATURAL, A]): BOOLEAN
			-- Universal quantifier: does `p' hold for every index and its respective term in current n-tuple?
		do
			Result := indices |∀ agent predicate_holds_to_term (p, ?)
		ensure
			definition: Result = (indices |∀ agent predicate_holds_to_term (p, ?))
		end

feature -- Operation

	left_trimmed (m: NATURAL): like trimmed_tuple_anchor
			-- Current n-tuple trimmed from its first `m' terms
		do
			Result := suffix (n - m.min (n))
		ensure
			definition: Result ≍ suffix (n - m.min (n))
		end

	right_trimmed (m: NATURAL): like trimmed_tuple_anchor
			-- Current n-tuple trimmed from its last `m' terms
		do
			Result := prefix (n - m.min (n))
		ensure
			definition: Result ≍ prefix (n - m.min (n))
		end

	prefix (m: NATURAL): like trimmed_tuple_anchor
			-- The first `m' terms of current n-tuple
		deferred
		ensure
			n: Result.n = (m ∧ n)
			terms: Result.for_all (agent terms_are_equal)
		end

	suffix (m: NATURAL): like trimmed_tuple_anchor
			-- The last `m' terms of current n-tuple
		do
			Result := reversed.prefix (m).reversed
		ensure
			definition: Result ≍ reversed.prefix (m).reversed
		end

	reversed: like tuple_anchor
			-- The terms of current n-tuple in reversed order
		deferred
		ensure
			n: Result.n = n
			terms: Result.for_all (agent reversed_terms_are_equal)
		end

	replacing (i: NATURAL; a: A): like tuple_anchor
			-- N-tuple equal to current one in every term but the `i'-th one, which is replaced by `a'.
		require
			positive: 0 < i
			small_enough: i ≤ n
		do
			Result := prefix (i - 1).before (a) + suffix (n - i)
		ensure
			definition: Result ≍ (prefix (i - 1).before (a) + suffix (n - i))
		end

	concatenated alias "+" (t: N_TUPLE [A, EQ]): like extended_tuple_anchor
			-- N-tuple with `n' + `t'.`n' terms, whose first `n' terms equal those in current one and whose last `t'.`n' terms equal those in `t'.
		deferred
		ensure
			n: Result.n = n + t.n
			prefix: Result.prefix (n) ≍ Current
			suffix: Result.suffix (t.n) ≍ t
		end

feature -- Transformation

	mapped alias "↦" (f: FUNCTION [A, A]): like term_to_term.tuple_map
			-- Current n-tuple mapped by `f' to an n-tuple of same kind
		do
			Result := term_to_term.tuple_map (Current, f)
		ensure
			definition: Result ≍ term_to_term.tuple_map (Current, f)
		end

	left_reduced (leftmost: A; f: FUNCTION [A, A, A]): A
			-- Current n-tuple left-reduced by `f' to a value like `leftmost'
		do
			Result := term_to_term.tuple_left_reduction (Current, leftmost, f)
		ensure
			definition: eq (Result, term_to_term.tuple_left_reduction (Current, leftmost, f))
		end

	indexed_left_reduced (leftmost: A; i: NATURAL; f: FUNCTION [A, NATURAL, A, A]): A
			-- Current n-tuple, indexed from `i' upwards, left-reduced by `f' to a value like `leftmost'.
		do
			Result := term_to_term.tuple_indexed_left_reduction (Current, leftmost, i, f)
		ensure
			definition: eq (Result, term_to_term.tuple_indexed_left_reduction (Current, leftmost, i, f))
		end

	proper_left_reduced (f: FUNCTION [A, A, A]): A
			-- Current non-empty n-tuple left-reduced by `f' to a value like {A}
		require
			is_not_unit: not is_unit
		do
			Result := tail.left_reduced (first, f)
		ensure
			definition: eq (Result, tail.left_reduced (first, f))
		end

	right_reduced (f: FUNCTION [A, A, A]; rightmost: A): A
			-- Current n-tuple right-reduced by `f' to a value like `rightmost'
		do
			Result := term_to_term.tuple_right_reduction (Current, f, rightmost)
		ensure
			definition: eq (Result, term_to_term.tuple_right_reduction (Current, f, rightmost))
		end

	indexed_right_reduced (i: NATURAL; f: FUNCTION [NATURAL, A, A, A]; rightmost: A): A
			-- Current n-tuple, indexed from `i' upwards, right-reduced by `f' to a value like `rightmost'
		do
			Result := term_to_term.tuple_indexed_right_reduction (Current, i, f, rightmost)
		ensure
			definition: eq (Result, term_to_term.tuple_indexed_right_reduction (Current, i, f, rightmost))
		end

	proper_right_reduced (f: FUNCTION [A, A, A]): A
			-- Current non-empty n-tuple right-reduced to a value like {A}
		require
			is_not_unit: not is_unit
		do
			Result := front.right_reduced (f, last)
		ensure
			definition: eq (Result, front.right_reduced (f, last))
		end

feature -- Conversion

	as_singleton: SET [N_TUPLE [A, EQ], N_TUPLE_EQUALITY [A, EQ]]
			-- Current n-tuple alone into a singleton, i.e. {`Current'}.
		deferred
		ensure
			has_this_tuple: Current ∈ Result
			nothing_else: Result.is_singleton
		end

	first_tail: ORDERED_PAIR [A, N_TUPLE [A, EQ], EQ, N_TUPLE_EQUALITY [A, EQ]]
			-- Current n-tuple converted into an ordered pair in the form (`first', `tail')
		require
			is_not_unit: not is_unit
		deferred
		ensure
			first: eq (Result.a, first)
			second: Result.b ≍ tail
		end

	front_last: ORDERED_PAIR [N_TUPLE [A, EQ], A, N_TUPLE_EQUALITY [A, EQ], EQ]
			-- Current n-tuple converted into an ordered pair in the form (`front', `last')
		require
			is_not_unit: not is_unit
		deferred
		ensure
			first: Result.a ≍ front
			second: eq (Result.b, last)
		end

feature -- Factory

	unit: like trimmed_tuple_anchor
			-- N-tuple with no term, i.e. ().
		deferred
		ensure
			definition: Result.is_unit
		end

	converted_terms (s: SET [A, EQ]): like terms_anchor
			-- `s' converted to a set like `terms_anchor'
		deferred
		ensure
			definition: Result ≍ s
		end

feature -- Predicate

	predicate_holds_to_term (p: PREDICATE [NATURAL, A]; i: NATURAL): BOOLEAN
			-- Does `p' hold to `Current' [`i']?
		do
			if 0 < i and i ≤ n then
				Result := p (i, Current [i])
			end
		ensure
			invalid_index: i = 0 or i > n implies not Result
			definition: 0 < i and i ≤ n implies Result = p (i, Current [i])
		end

	negated (p: PREDICATE [NATURAL, A]; i: NATURAL; x: A): BOOLEAN
			-- Logical negation of `p' (`i', `x'), i.e. is `p' (`i', `x') false?
		do
			Result := not p (i, x)
		ensure
			class
			definition: Result = not p (i, x)
		end

	terms_are_equal (i: NATURAL; x_i: A): BOOLEAN
			-- Does `x_i' equal `Current' [`i']?
		do
			if 0 < i and i ≤ n then
				Result := eq (x_i, Current [i])
			end
		ensure
			invalid_index: i = 0 or i > n implies not Result
			definition: 0 < i and i ≤ n implies Result = eq (x_i, Current [i])
		end

	reversed_terms_are_equal (i: NATURAL; x_i: A): BOOLEAN
			-- Does `x_i' equal `Current' [`n'-`i'+1]?
		do
			if 0 < i and i ≤ n then
					check
						positive: 0 < n - i + 1 -- Plain algebra
						small_enough: (n - i + 1) ≤ n -- Plain algebra
					end
				Result := eq (x_i, Current [n - i + 1])
			end
		ensure
			invalid_index: i = 0 or i > n implies not Result
			positive: 0 < i and i ≤ n implies 0 < n - i + 1 -- Plain algebra
			small_enough: 0 < i and i ≤ n implies (n - i + 1) ≤ n -- Plain algebra
			definition: 0 < i and i ≤ n implies Result = eq (x_i, Current [n - i + 1])
		end

	index_is_positive (i: NATURAL): BOOLEAN
			-- Is `i' greater than zero?
		do
			Result := 0 < i
		ensure
			class
			definition: Result = (0 < i)
		end

	index_is_less_equal (i, j: NATURAL): BOOLEAN
			-- Is `i' less than or equal to `j'?
		do
			Result := i ≤ j
		ensure
			class
			definition: Result = (i ≤ j)
		end

feature -- Reduction

	just_index (acc, i: NATURAL; x_i: A): NATURAL
			-- Just the value of `i', ignoring `acc' and `x_i' altogether.
		do
			Result := i
		ensure
			class
			definition: Result = i
		end

	indexed_term (i: NATURAL; acc: A; j: NATURAL; x_j: A): A
			-- The value of `x_j', if `i' = `j'; otherwise, the value of `acc'.
		do
			if j = i then
				Result := x_j
			else
				Result := acc
			end
		ensure
			indexed_term: j = i implies eq (Result, x_j)
			other_term: j /= i implies eq (Result, acc)
		end

feature -- Transformer

	term_to_term: TRANSFORMER [A, A, EQ, EQ]
			-- Transformer between types derived from {A}
		deferred
		end

	term_to_index: TRANSFORMER [A, NATURAL, EQ, OBJECT_EQUALITY [NATURAL]]
			-- Transformer between types derived from {A} and {NATURAL}
		deferred
		end

feature -- Anchor

	tuple_anchor: N_TUPLE [A, EQ]
			-- Anchor for n-tuples like current one
		deferred
		end

	trimmed_tuple_anchor: N_TUPLE [A, EQ]
			-- Anchor for n-tuples like those got by trimming current one
		deferred
		end

	extended_tuple_anchor: N_TUPLE [A, EQ]
			-- Anchor for n-tuples like those got by extending current one
		deferred
		end

	terms_anchor: SET [A, EQ]
			-- Anchor for sets whose elements are like terms of current n-tuple
		deferred
		end

invariant
	consistent_unitness: is_unit implies tail.is_unit

note
	copyright: "Copyright (c) 2012-2021, Rosivaldo Fernandes Alves"
	license: "[
		Eiffel Forum License v2
		(see http://www.eiffel.com/licensing/forum.txt)
		]"
	source: "https://launchpad.net/set-theory"

end
