﻿note
	description: "[
		Element that models a mathematical ordered pair.
		From a purist outlook, it should be a kind of set, maybe following Kuratowski's definition (http://preview.tinyurl.com/yb3w85vl). But I found that it only complicates the design
		for no true gain.
		]"
	author: "Rosivaldo Fernandes Alves"
	date: "$Date: Wed 2018-11-21 07:53:55 -0300 $"
	revision: "$Revision$"

deferred class
	ORDERED_PAIR [A, B, EQ_A -> EQUALITY [A], EQ_B -> EQUALITY [B]]

inherit
	ELEMENT
		rename
			is_in as element_is_in,
			is_not_in as element_is_not_in
		end

feature -- Primitive

	a,
	first: A
			-- First element in current ordered pair
		deferred
		end

	b,
	second: B
			-- Second element in current ordered pair
		deferred
		end

	eq_a: EQ_A
			-- Equality for objects like `first'
			--| Please have a look at the note in the header of {SET}.eq.
		deferred
		end

	eq_b: EQ_B
			-- Equality for objects like `second'
			--| Please have a look at the note in the header of {SET}.eq.
		deferred
		end

feature -- Membership

	is_in alias "∈" (s: SET [ORDERED_PAIR [A, B, EQ_A, EQ_B], EQUALITY [ORDERED_PAIR [A, B, EQ_A, EQ_B]]]): BOOLEAN
			-- Does `s` have current ordered pair?
		do
			Result := s ∋ Current
		ensure
			definition: Result = s ∋ Current
		end

	is_not_in alias "∉" (s: SET [ORDERED_PAIR [A, B, EQ_A, EQ_B], EQUALITY [ORDERED_PAIR [A, B, EQ_A, EQ_B]]]): BOOLEAN
			-- Is not current ordered pair in `s'?
		do
			Result := not (Current ∈ s)
		ensure
			definition: Result = not (Current ∈ s)
		end

feature -- Comparison

	equals alias "≍" (xy: ORDERED_PAIR [A, B, EQ_A, EQ_B]): BOOLEAN
			-- Do current pair and `xy' have the same elements in the same order?
			--| Please have a look at the note in the header of {SET}.equals.
		do
			Result := eq_a (a, xy.a) and eq_b (b, xy.b)
		ensure
			definition: Result = (eq_a (a, xy.a) and eq_b (b, xy.b))
		end

	unequals alias "≭" (xy: ORDERED_PAIR [A, B, EQ_A, EQ_B]): BOOLEAN
			-- Does not current pair equal `xy'?
		do
			Result := not (Current ≍ xy)
		ensure
			definition: Result = not (Current ≍ xy)
		end

	succeeds alias "≽" (wx: ORDERED_PAIR [detachable separate ANY, A, EQUALITY [detachable separate ANY], EQ_A]): BOOLEAN
			-- Does the first element in current pair equal the second element in `wx'?
		do
			Result := eq_a (a, wx.b)
		ensure
			definition: Result = eq_a (a, wx.b)
		end

feature -- Operation

	inverse: ORDERED_PAIR [B, A, EQ_B, EQ_A]
			-- The elements of current pair in inverted order, i.e. (`b', `a').
		deferred
		ensure
			first: eq_b (Result.a, b)
			second: eq_a (Result.b, a)
		end

	composed_with_homogeneous (x1x2: ORDERED_PAIR [A, A, EQ_A, EQ_A]): ORDERED_PAIR [A, B, EQ_A, EQ_B]
			-- Composition of current pair and `x1x2', i.e. the pair (`x1x2'.`first', `second').
		require
			succeeds: Current ≽ x1x2
		deferred
		ensure
			first: eq_a (Result.a, x1x2.first)
			second: eq_b (Result.b, second)
		end

	composed_with_inverse_like (yx: ORDERED_PAIR [B, A, EQ_B, EQ_A]): COMPOSABLE_ORDERED_PAIR [A, B, B, EQ_A, EQ_B, EQ_B]
			-- Composition of current pair and `yx', i.e. the pair (`yx'.`first', `second').
		require
			succeeds: Current ≽ yx
		deferred
		ensure
			first: eq_b (Result.a, yx.first)
			second: eq_b (Result.b, second)
		end

feature -- Conversion

	as_singleton: SET_OF_ORDERED_PAIRS [A, B, EQ_A, EQ_B]
			-- Current pair alone into a singleton, i.e. {(`a', `b')}.
		deferred
		ensure
			has_current_pair: Current ∈ Result
			nothing_else: Result.is_singleton
		end

invariant
	first: eq_a (first, a)
	second: eq_b (second, b)

note
	copyright: "Copyright (c) 2012-2021, Rosivaldo Fernandes Alves"
	license: "[
		Eiffel Forum License v2
		(see http://www.eiffel.com/licensing/forum.txt)
		]"
	source: "https://launchpad.net/set-theory"

end
