﻿note
	description: "Set of subsets of a given superset"
	author: "Rosivaldo Fernandes Alves"
	date: "$Date: Sat 2018-11-24 09:13:02 -0300 $"
	revision: "$Revision$"

deferred class
	SUBSET_FAMILY [A, EQ -> EQUALITY [A]]

inherit
	SET_OF_SETS [A, EQ]
		rename
			universe as superset,
			member_sets_are_disjoint as subsets_are_disjoint,
			sets_union as family_union alias "⋃",
			sets_intersection as family_intersection alias "⋂",
			o_s as o_f,
			empty_set_of_sets as empty_family,
			unit_s as unit_f,
			empty_member_set as empty_subset,
			some_set_has as some_subset_has,
			every_set_has as every_subset_has,
			set_anchor as family_anchor,
			subset_anchor as subfamily_anchor,
			superset_anchor as superfamily_anchor
		redefine
			with,
			without,
			complemented,
			intersected,
			subtracted,
			subtracted_symmetricaly,
			family_anchor,
			subfamily_anchor,
			superfamily_anchor
		end

feature -- Construction

	with alias "&" (s: SET [A, EQ]): like superfamily_anchor
			-- Every subset in this family, together with `s`.
		do
			Result := Precursor {SET_OF_SETS} (s)
		ensure then
			extended_superset: Result.superset ≍ (superset ∪ s)
		end

	without alias "/" (s: SET [A, EQ]): like subfamily_anchor
			-- Every subset in this family but `s`
		do
			Result := Precursor {SET_OF_SETS} (s)
		ensure then
			same_superset: Result.superset ≍ superset
		end

feature -- Comparison

	same_family (sf: SUBSET_FAMILY [A, EQ]): BOOLEAN
			-- Do this family and `sf' have the same subsets and the same superset?
			-- Notice that two subset families regarded equal by `equals' will not be the `same_family' if their `superset's are unequal.
		do
			Result := Current ≍ sf and superset ≍ sf.superset
		ensure
			definition: Result = (Current ≍ sf and superset ≍ sf.superset)
		end

feature -- Operation

	filtered alias "|" (p: PREDICATE [SET [A, EQ]]): like subfamily_anchor
			-- Every subset in this family for which `p` holds
		deferred
		ensure then
			same_superset: Result.superset ≍ superset
		end

	complemented alias "∁" (ss: SET [SET [A, EQ], SET_EQUALITY [A, EQ]]): like family_anchor
			-- Complement of this subset family relative to `ss', i.e. every subset in `ss' that is not in this family.
		do
			Result := Precursor {SET_OF_SETS} (ss)
		ensure then
			other_superset: Result.superset ≍ converted (ss).superset
		end

	complemented_family alias "∁": like family_anchor
			-- Subset family with every subset of `superset' that is not in this family
			-- Notice that two subset families regarded equal by `equals' will have different `complemented_family' if their `superset's are unequal.
		do
				-- todo: how to avoid creating an entire powerset?
				check
					is_subset: Current ⊆ superset.powerset -- superset.powerset.superset ≍ superset
				end
			Result := Current ∁ superset.powerset
		ensure
			is_subset: Current ⊆ superset.powerset -- superset.powerset.superset ≍ superset
			definition: Result ≍ (Current ∁ superset.powerset)
			same_superset: Result.superset ≍ superset
		end

	united alias "∪" (ss: SET [SET [A, EQ], SET_EQUALITY [A, EQ]]): like superfamily_anchor
			-- Union of this subset family and `ss`, i.e. every subset that is in this family or in `ss`.
		deferred
		ensure then
			extended_superset: Result.superset ≍ (superset ∪ converted (ss).superset)
		end

	intersected alias "∩" (ss: SET [SET [A, EQ], SET_EQUALITY [A, EQ]]): like subfamily_anchor
			-- Intersection of this family and `ss`, i.e. every subset in this family that is in `ss`.
		do
			Result := Precursor {SET_OF_SETS} (ss)
		ensure then
			same_superset: Result.superset ≍ superset
		end

	subtracted alias "∖" (ss: SET [SET [A, EQ], SET_EQUALITY [A, EQ]]): like subfamily_anchor
			-- Difference of this subset family and `ss', i.e. every subset that is in this family but not in `ss'.
		do
			Result := Precursor {SET_OF_SETS} (ss)
		ensure then
			same_superset: Result.superset ≍ superset
		end

	subtracted_symmetricaly alias "⊖" (ss: SET [SET [A, EQ], SET_EQUALITY [A, EQ]]): like family_anchor
			-- Symmetric difference of this family and `ss`, i.e. every subset that is in this family or in `ss` but not in both.
		do
			Result := Precursor {SET_OF_SETS} (ss)
		ensure then
			extended_superset: Result.superset ≍ (superset ∪ converted (ss).superset)
		end

feature -- Factory

	singleton (s: SET [A, EQ]): like family_anchor
			-- Family that has `s` as its sole subset, i.e. {`s`}.
		deferred
		ensure then
			superset: Result.superset ≍ s -- todo: This assertion may run unchecked due to a compiler error.
		end

	converted (ss: SET [SET [A, EQ], SET_EQUALITY [A, EQ]]): like family_anchor
			-- `ss' converted to a subset family
		deferred
		ensure then
			when_a_family: attached {SUBSET_FAMILY [A, EQ]} ss as sf implies Result.superset ≍ sf.superset
			when_not_a_family: not attached {SUBSET_FAMILY [A, EQ]} ss implies Result.superset ≍ (⋃ Result)
		end

feature -- Anchor

	family_anchor,
	subfamily_anchor,
	superfamily_anchor: SUBSET_FAMILY [A, EQ]
			-- <Precursor>
		deferred
		end

invariant
	others_keeps_superset: others.superset ≍ superset
	o_f_keeps_superset: o_f.superset ≍ superset
	empty_family_keeps_superset: empty_family.superset ≍ superset

note
	copyright: "Copyright (c) 2012-2021, Rosivaldo Fernandes Alves"
	license: "[
		Eiffel Forum License v2
		(see http://www.eiffel.com/licensing/forum.txt)
		]"
	source: "https://launchpad.net/set-theory"

end
