﻿note
	description: "[
		Pool of features available for test classes
		Since AutoTest crashes when generating or extracting tests for a void-safe system, this class is the root of a (highly improvised) framework for manual
		testing of sti-library against stp-library.
		]"
	author: "Rosivaldo Fernandes Alves"
	date: "$Date: Wed 2018-11-21 07:53:55 -0300 $"
	revision: "$Revision$"
	testing: "type/manual"

deferred class
	SETS_TESTS_RESOURCES

inherit
	STS_ELEMENT
		undefine
			default_create
		end

	EQA_TEST_SET

feature -- Access

	Max_count: NATURAL = 5
			-- Maximum value for `some_count'
			-- Why so few? Believe me: you don't want to deal with all subsets of all subsets of a set with 6 elements!

	Min_character: CHARACTER = 'a'
			-- Minimum value for `some_character'

	some_count: NATURAL
			-- Randomly created natural number to be used for counting
		do
			Result := next_random_item.as_natural_32 \\ (Max_count + 1)
		ensure
			class
			upper_bound: Result ≤ Max_count
		end

	some_index: NATURAL
			-- Randomly created index
		do
				check
					positive: 0 < Max_count -- `Max_count' definition
				end
			Result := some_index_up_to (Max_count)
		ensure
			class
			positive: 0 < Result
			upper_bound: Result <= Max_count
		end

	some_index_up_to (n: NATURAL): NATURAL
			-- Randomly created index in the range {1..`n'}
		require
			positive: 0 < n
		do
			Result := next_random_item.as_natural_32 \\ n + 1
		ensure
			class
			positive: 0 < Result
			small_enough: Result <= n
		end

feature -- Factory (object)

	some_character_ref: detachable separate CHARACTER_REF
			-- Randomly fetched polymorphic character reference
		do
			inspect
				next_random_item \\ 2
			when 0 then
				Result := some_direct_character_ref
			when 1 then
				Result := some_character
			end
		end

	some_direct_character_ref: like some_character_ref
			-- Randomly fetched monomorphic character reference
		local
			new_c: FUNCTION [like new_character_ref]
			obj: like some_direct_instance
		do
			new_c := agent new_character_ref
			from
				obj := some_direct_instance (new_c)
			until
				attached obj -- TODO: System crashes when an agent gets a "separate" Void actual argument.
			loop
				obj := some_direct_instance (new_c)
			end
			if attached obj then
				Result := {like new_character_ref} / obj
					check
							-- `some_direct_instance' and `new_character_ref' definitions
						right_type: attached Result
						monomorphic: object_is_direct_instance (Result, {like new_character_ref})
					end
			end
		ensure
--			monomorphic: attached Result implies object_is_direct_instance (Result, {like some_character_ref}) -- TODO: why not?
			monomorphic: attached Result implies object_is_direct_instance (Result, {like new_character_ref})
		end

	new_character_ref: like some_direct_character_ref
			-- Randomly created monomorphic character reference
		local
			c_ref: like new_character_ref
		do
			create c_ref
			separate c_ref as sep_c_ref do
				sep_c_ref.set_item (some_character)
			end
			Result := c_ref
		ensure
			class
			monomorphic: attached Result implies object_is_direct_instance (Result, {like some_direct_character_ref})
		end

	some_character: CHARACTER
			-- Randomly created character
		local
			n: NATURAL
		do
			n := some_index
				check
					valid_character:
							-- 0 < n <= Max_count <== `some_index' definition
							-- class invariant: (Min_character.natural_32_code + Max_count - 1).is_valid_character_8_code
						(Min_character.natural_32_code + n - 1).is_valid_character_8_code
				end
			Result := (Min_character.natural_32_code + n - 1).to_character_8
--			set_object_pool (object_pool & Result)
		ensure
			class
		end

	some_integer_ref: detachable INTEGER_REF
			-- Randomly fetched polymorphic integer-number reference
		do
			inspect next_random_item \\ 2
			when 0 then
				Result := some_direct_integer_ref
			when 1 then
				Result := some_integer
			end
		end

	some_direct_integer_ref: like some_integer_ref
			-- Randomly fetched monomorphic integer-number reference
		local
			obj: like some_direct_instance
			new_i: FUNCTION [like new_integer_ref]
		do
			new_i := agent new_integer_ref
			from
				obj := some_direct_instance (new_i)
			until
				attached obj -- TODO: Some mapping e.g. f (Void) may require a "separate" Void. Please see TODO at `some_direct_character_ref'.
			loop
				obj := some_direct_instance (new_i)
			end
			if attached obj then
				Result := {like new_integer_ref} / obj
					check
							-- `some_direct_instance' and `new_integer_ref' definitions
						right_type: attached Result
						monomorphic: Result.generating_type ~ {like new_integer_ref}
					end
			end
		ensure
			monomorphic: attached Result implies Result.generating_type ~ {like some_integer_ref}
		end

	new_integer_ref: like some_direct_integer_ref
			-- Randomly created monomorphic integer-number reference
		do
			Result := some_integer.to_reference
		ensure
			class
			monomorphic: attached Result implies Result.generating_type ~ {like some_direct_integer_ref}
		end

	some_integer: INTEGER
			-- Randomly created integer number
		do
			Result := next_random_item \\ Max_count.as_integer_32 - Max_count.as_integer_32 // 2
--			set_object_pool (object_pool & Result)
		ensure
			class
		end

	some_natural: NATURAL
			-- Randomly created natural number
		do
			Result := (next_random_item \\ Max_count.as_integer_32).as_natural_32
--			set_object_pool (object_pool & Result)
		ensure
			class
		end

feature -- Factory (element)

	some_element: STS_ELEMENT
			-- Randomly fetched polymorphic element
		do
			inspect
				next_random_item \\ 7
			when 0 then
				Result := some_direct_element
			when 1 then
				Result := some_equality
			when 2 then
				Result := some_pair
			when 3 then
				Result := some_set
			when 4 then
				Result := some_pair_composer
			when 5 then
				Result := some_transformer
			when 6 then
				Result := some_n_tuple
			end
		end

	some_direct_element: like some_element
			-- Randomly fetched monomorphic element
		do
			check
					-- `some_direct_instance' and `new_element' definitions
				a: attached {like new_element} some_direct_instance (agent new_element) as a
				monomorphic: a.generating_type ~ {detachable like new_element}
			then
				Result := a
			end
		ensure
			monomorphic: Result.generating_type ~ {detachable like some_element}
		end

	new_element: like some_direct_element
			-- Randomly created monomorphic element
		do
			create Result
		ensure
			class
			monomorphic: Result.generating_type ~ {detachable like some_direct_element}
		end

	some_elements: STI_SET [STS_ELEMENT, STS_EQUALITY [STS_ELEMENT]]
			-- Randomly fetched polymorphic set of elements
		do
			inspect
				next_random_item \\ 2
			when 0 then
				Result := some_element_references
			when 1 then
				Result := some_element_objects
			end
		end

	some_element_references: STI_SET [STS_ELEMENT, STS_REFERENCE_EQUALITY [STS_ELEMENT]]
			-- Randomly created set of elements compared by references
		do
			across
				1 |..| some_count.as_integer_32 as i
			from
				create Result.make_empty
			loop
				Result := Result & some_element
			end
		end

	some_element_objects: STI_SET [STS_ELEMENT, STS_OBJECT_EQUALITY [STS_ELEMENT]]
			-- Randomly created set of elements compared by object values
		do
			across
				1 |..| some_count.as_integer_32 as i
			from
				create Result.make_empty
			loop
				Result := Result & some_element
			end
		end

feature -- Factory (equality)

	some_equality: STS_EQUALITY [detachable separate ANY]
			-- Randomly fetched polymorphic equality
		deferred
		end

feature -- Factory (ordered pair)

	some_pair: STS_ORDERED_PAIR [detachable separate ANY, detachable separate ANY, STS_EQUALITY [detachable separate ANY], STS_EQUALITY [detachable separate ANY]]
			-- Randomly fetched polymorphic ordered pair
		deferred
		end

feature -- Factory (set)

	some_set: STI_SET [detachable separate ANY, STS_EQUALITY [detachable separate ANY]]
			-- Randomly fetched polymorphic set
		deferred
		end

feature -- Factory (pair composer)

	some_pair_composer: STI_ORDERED_PAIR_COMPOSER [
		detachable separate ANY, detachable separate ANY, detachable separate ANY,
		STS_EQUALITY [detachable separate ANY],
		STS_EQUALITY [detachable separate ANY],
		STS_EQUALITY [detachable separate ANY]
		]
			-- Randomly fetched polymorphic pair composer
		deferred
		end

feature -- Factory (transformer)

	some_transformer: STI_TRANSFORMER [
		detachable separate ANY, detachable separate ANY,
		STS_EQUALITY [detachable separate ANY], STS_EQUALITY [detachable separate ANY]
		]
			-- Randomly fetched polymorphic transformer
		deferred
		end

feature -- Factory (n-tuple)

	some_n_tuple: STI_N_TUPLE [detachable separate ANY, STS_EQUALITY [detachable separate ANY]]
			-- Randomly fetched polymorphic n-tuple
		deferred
		end

feature {NONE} -- Implementation

	next_random_item: like random_sequence.item
			-- Item at next position of `random_sequence'
		do
				check
					not_after: not random_sequence.After -- {RANDOM}.After = `False'
				end
			random_sequence.forth
				check
					readable: random_sequence.Readable -- {RANDOM}.Readable = `True'
					not_off: not random_sequence.off -- {RANDOM}.is_empty = `False' and {RANDOM}.After = `False'
				end
			Result := random_sequence.item
		ensure
			class
		end

	random_sequence: RANDOM
			-- Sequence of random numbers
		local
			time: TIME
		once
			from
				create time.make_now
			until
				time.compact_time >= 0
			loop
				create time.make_now
			end
			create Result.set_seed (time.compact_time)
		ensure
			class
		end

	some_direct_instance (a_new_instance: FUNCTION [detachable separate ANY]): detachable separate ANY
			-- Randomly fetched object whose type equals the result type of `a_new_instance'
		local
			rt: like result_type
			direct_instances: SPECIAL [ANY]
			i: INTEGER
			tries: NATURAL
			chk_ops: BOOLEAN
		do
			if next_random_item \\ 2 = 0 then
				Result := new_direct_instance (a_new_instance)
			else
				rt := result_type (a_new_instance)

					-- TODO: Does {MEMORY}.objects_instance_of_type (rt.type_id) return objects partially garbage-collected?
				direct_instances := {MEMORY}.objects_instance_of_type (rt.type_id) -- TODO: It crahses frequently when gathering, e.g. sets of sets...
				if direct_instances.count = 0 then
					if rt.is_attached or next_random_item \\ 2 = 0 then
						Result := new_direct_instance (a_new_instance)
					else
						Result := Void
					end
				else
						check
							positive: 0 < direct_instances.count -- direct_instances.count /= 0
						end
					if rt.is_attached then
						i := some_index_up_to (direct_instances.count.as_natural_32).as_integer_32 - 1 -- TODO: Create an integer some_index_up_to.
					else
						i := some_index_up_to (direct_instances.count.as_natural_32 + 1).as_integer_32 - 1
					end
					if i < direct_instances.count then
							check
								valid_index: direct_instances.valid_index (i) -- 0 ≤ i < direct_instances.count
							end
						Result := direct_instances [i]
						separate Result as sep_res do
							chk_ops := {ISE_RUNTIME}.check_assert (True)
								sep_res.do_nothing -- Cause an invariant violation if Result is ill formed.
							chk_ops := {ISE_RUNTIME}.check_assert (chk_ops)
						end
					else
						Result := Void
					end
				end
			end
		ensure
			direct_instance_fetched: object_is_direct_instance (Result, result_type (a_new_instance))
		rescue
			if attached {EXCEPTIONS}.exception_trace as et and then et.has_substring (" _invariant ") then
					-- Assume that direct_instances has fetched an ill-formed object. Just retry.
				tries := tries + 1
				if tries < 20 then
					retry
				end
			end
		end

	new_direct_instance (a_new_instance: FUNCTION [detachable separate ANY]): detachable separate ANY
			-- Randomly created object whose type equals the result type of `a_new_instance'
		do
			Result := a_new_instance.item
		ensure
			direct_instance_created: object_is_direct_instance (Result, result_type (a_new_instance))
		end

	result_type (a_new_instance: FUNCTION [detachable separate ANY]): TYPE [detachable ANY]
			-- Result type of the function accessed via `a_new_instance' agent
		local
			l_new_instance_type: TYPE [detachable ANY]
		do
			l_new_instance_type := a_new_instance.generating_type
				check
					large_enough: l_new_instance_type.generic_parameter_count ≥ 1 -- FUNCTION [detachable separate ANY] has generic parameters.
					small_enough: l_new_instance_type.generic_parameter_count ≤ l_new_instance_type.generic_parameter_count -- By definition
				end
			Result := l_new_instance_type.generic_parameter_type (l_new_instance_type.generic_parameter_count)
		ensure
			class
			el_new_instance_type: attached a_new_instance.generating_type as el_new_instance_type
			large_enough: el_new_instance_type.generic_parameter_count ≥ 1 -- FUNCTION [detachable separate ANY] has generic parameters.
			small_enough: el_new_instance_type.generic_parameter_count ≤ el_new_instance_type.generic_parameter_count -- By definition
			definition: Result ~ el_new_instance_type.generic_parameter_type (el_new_instance_type.generic_parameter_count)
		end

--	object_is_direct_instance (x: detachable separate ANY; t: TYPE [detachable separate ANY]): BOOLEAN
--			-- Is `x' a direct instance of `t', i.e. does `x' conform to `t' and is it not a proper descendant of `t'?
--		do
--			if x = Void then
--				Result := not t.is_attached
--			elseif t.is_attached then
--				if attached {ANY} x then
--					Result := t ~ {attached like x}
--				else
--						-- `x' is separated.
--					Result := t ~ {attached separate like x}
--				end
--			else
--				if attached {ANY} x then
--					Result := t ~ {like x}
--				else
--						-- `x' is separated.
--					Result := t ~ {separate like x}
--				end
--			end
--		ensure
--			class
--			when_void: x = Void implies Result = not t.is_attached
--			when_attached: x /= Void and t.is_attached and attached {ANY} x implies Result = (t ~ {attached like x})
--			when_separate_attached: x /= Void and t.is_attached and not attached {ANY} x implies Result = (t ~ {attached separate like x})
--			when_detachable: x /= Void and not t.is_attached and attached {ANY} x implies Result = (t ~ {like x})
--			when_detachable: x /= Void and not t.is_attached and not attached {ANY} x implies Result = (t ~ {separate like x})
--		end

	object_is_direct_instance (x: detachable separate ANY; t: TYPE [detachable separate ANY]): BOOLEAN
			-- Is `x' a direct instance of `t', i.e. does `x' conform to `t' and is it not a proper descendant of `t'?
		do
			if x = Void then
				Result := not t.is_attached
			elseif t.is_attached then
				Result := t ~ {attached like x} or t ~ {attached separate like x}
			else
				Result := t ~ {like x} or t ~ {separate like x}
			end
		ensure
			class
			when_void: x = Void implies Result = not t.is_attached
			when_attached: x /= Void and t.is_attached implies Result = (t ~ {attached like x} or t ~ {attached separate like x})
			when_detachable: x /= Void and not t.is_attached implies Result = (t ~ {like x} or t ~ {separate like x})
		end

invariant
	valid_characters: (Min_character.natural_32_code + Max_count - 1).is_valid_character_8_code

note
	copyright: "Copyright (c) 2012-2021, Rosivaldo Fernandes Alves"
	license: "[
		Eiffel Forum License v2
		(see http://www.eiffel.com/licensing/forum.txt)
		]"
	source: "https://launchpad.net/set-theory"

end
