﻿note
	description: "[
		Object that checks whether the properties verified within set theory hold for an
		implementation of {STS_INTEGER}
	]"
	author: "Rosivaldo Fernandes Alves"
	date: "$Date: Sat 2018-11-10 18:52:01 -0300 $"
	revision: "$Revision$"

deferred class
	STP_INTEGER_PROPERTIES

inherit

	STP_ELEMENT_PROPERTIES

feature -- Access

	zero: STS_INTEGER
			-- Neutral element for "+" and "-"
		deferred
		ensure then
			definition: Result.value = 0
		end

	one: STS_INTEGER
			-- Neutral element for "*" and "/"
		deferred
		ensure then
			definition: Result.value = 1
		end

feature -- Properties (Access)

	zero_ok (i, j: STS_INTEGER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_INTEGER}.zero?
		do
			check
				left_neutral_for_plus: i.zero.integer_plus (j) ≍ j
				right_neutral_for_plus: (j + i.zero) ≍ j
				quasi_left_neutral_for_minus: i.zero.integer_minus (j) ≍ - j
				right_neutral_for_minus: (j - i.zero) ≍ j
			then
				Result := True
			end
		end

	one_ok (i, j: STS_INTEGER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_INTEGER}.one?
		do
			check
				left_neutral_for_times: i.one.integer_product (j) ≍ j
				right_neutral_for_times: (j * i.one) ≍ j
				quasi_left_neutral_for_divided: i.one.integer_divisible (j) implies - one <= i.one.integer_quotient (j) and i.one.integer_quotient (j) <= one
				good_divisor: j.divisible (i.one) -- i.one /= 0
				right_neutral_for_divided: (j // i.one) ≍ j
			then
				Result := True
			end
		end

feature -- Properties (Comparison)

	equals_ok (i, j, k: STS_INTEGER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_INTEGER}.equals?
		do
			check
				reflexive: i ≍ i
				symmetric: i ≍ j implies j ≍ i
				transitive: i ≍ j and j ≍ k implies i ≍ k
				euclidian: i ≍ k and j ≍ k implies i ≍ j
			then
				Result := True
			end
		end

	is_less_ok (i, j, k: STS_INTEGER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_INTEGER}.is_less?
		do
			check
				irreflexive: not (i < i)
				transitive: i < j and j < k implies i < k
			then
				Result := True
			end
		end

	is_less_equal_ok (i, j, k: STS_INTEGER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_INTEGER}.is_less_equal?
		do
			check
				reflexive: i <= i
				transitive: i <= j and j <= k implies i <= k
				antisymmetric: i <= j and j <= i implies i ≍ j
			then
				Result := True
			end
		end

feature -- Properties (Basic operations)

	plus_ok (i, j, k: STS_INTEGER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_INTEGER}.plus?
		local
			must_retry: BOOLEAN
		do
			if not must_retry then
				check
					left_neutral: (zero + i) ≍ i
					right_neutral: (i + zero) ≍ i
					associative: ((i + j) + k) ≍ (i + (j + k))
					homomorphic: i <= j implies i + k <= j + k
					isomorphic: i < j implies i + k < j + k
				then
					Result := True
				end
			else
				check
					overflow_hid_homorphism: sum_overflows (i, k) xor sum_overflows (j, k)
				then
					Result := True
				end
			end
		rescue
			if not must_retry then
				must_retry := True
				retry
			end
		end

	identity_ok (i: STS_INTEGER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_INTEGER}.identity?
		do
			check
				same_integer: + i ≍ i
				neutral_sum: + i ≍ (zero + i)
			then
				Result := True
			end
		end

	minus_ok (i, j, k: STS_INTEGER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_INTEGER}.minus?
		local
			must_retry: BOOLEAN
		do
			if not must_retry then
				check
					quasi_left_neutral: (zero - i) ≍ - i
					right_neutral: (i - zero) ≍ i
					oppositely_homomorphic: i <= j implies k - i >= k - j
					oppositely_isomorphic: i < j implies k - i > k - j
				then
					Result := True
				end
			else
				check
					overflow_hid_homorphism: subtraction_overflows (k, i) xor subtraction_overflows (k, j)
				then
					Result := True
				end
			end
		rescue
			if not must_retry then
				must_retry := True
				retry
			end
		end

	opposite_ok (i: STS_INTEGER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_INTEGER}.opposite?
		do
			check
				involutive: - - i ≍ i
				quasi_neutral_subtraction: - i ≍ (zero - i)
				as_product: - i ≍ (- one * i)
				divisible: i.divisible (- one) -- - one /= 0
				as_quotient: - i ≍ (i // ( - one))
			then
				Result := True
			end
		end

	product_ok (i, j, k: STS_INTEGER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_INTEGER}.product?
		local
			must_retry: BOOLEAN
		do
			if not must_retry then
				check
					left_neutral: (one * i) ≍ i
					right_neutral: (i * one) ≍ i
					associative: ((i * j) * k) ≍ (i * (j * k))
					quasi_homomorphic: i <= j implies (i * k).abs <= (j * k).abs
					quasi_isomorphic: i < j implies (i * k).abs < (j * k).abs
				then
					Result := True
				end
			else
				check
					overflow_hid_homorphism: product_overflows (i, k) xor product_overflows (j, k)
				then
					Result := True
				end
			end
		rescue
			if not must_retry then
				must_retry := True
				retry
			end
		end

	integer_quotient_ok (i, j, k: STS_INTEGER): BOOLEAN
			-- Do the properties verified within set theory hold for {STS_INTEGER}.integer_quotient?
		local
			must_retry: BOOLEAN
		do
			if not must_retry then
				check
					quasi_left_neutral: one.divisible (i) implies - one <= (one // i) and (one // i) <= one
					good_divisor: i.divisible (one) -- one /= 0
					right_neutral: (i // one) ≍ i
					oppositely_homomorphic: i <= j implies (k // i).abs >= (k // j).abs
					oppositely_isomorphic: i < j implies (k // i).abs > (k // j).abs
				then
					Result := True
				end
			else
				check
					overflow_hid_homorphism: quotient_overflows (k, i) xor quotient_overflows (k, j)
				then
					Result := True
				end
			end
		rescue
			if not must_retry then
				must_retry := True
				retry
			end
		end

feature -- Predicate

	sum_overflows (i, j: STS_INTEGER): BOOLEAN
			-- Does the sum `i' + `j' overflow?
		do
			if i > zero then
				if j > zero then
					Result := i + j < i
				end
			elseif i < zero then
				if j < zero then
					Result := i + j > i
				end
			end
		ensure
			positive_sum: if i > zero then
					if j > zero then
						Result = (i + j < i)
					else
						not Result
					end
				else
					Result implies True
				end
			negative_sum: if i < zero then
					if j < zero then
						Result = (i + j > i)
					else
						not Result
					end
				else
					Result implies True
				end
			no_overflow: if i <= zero then
					if i >= zero then
						not Result
					elseif j >= zero then
						not Result
					elseif i + j <= i then
						not Result
					else
						not Result implies True
					end
				elseif j <= zero then
					if i >= zero then
						not Result
					elseif j >= zero then
						not Result
					elseif i + j <= i then
						not Result
					else
						not Result implies True
					end
				elseif i + j >= i then
					if i >= zero then
						not Result
					elseif j >= zero then
						not Result
					elseif i + j <= i then
						not Result
					else
						not Result implies True
					end
				else
					not Result implies True
				end
		end

	subtraction_overflows (i, j: STS_INTEGER): BOOLEAN
			-- Does the subtraction  `i' - `j' overflow?
		do
			if i < zero then
				if j > zero then
					Result := i - j > i
				end
			elseif i > zero then
				if j < zero then
					Result := i - j < i
				end
			end
		ensure
			negative_subtraction: if i < zero then
					if j > zero then
						Result = (i - j > i)
					else
						not Result
					end
				else
					Result implies True
				end
			positive_subtraction: if i > zero then
					if j < zero then
						Result = (i - j < i)
					else
						not Result
					end
				else
					Result implies True
				end
			no_overflow: if i >= zero then
					if i <= zero then
						not Result
					elseif j >= zero then
						not Result
					elseif i - j >= i then
						not Result
					else
						not Result implies True
					end
				elseif j <= zero then
					if i <= zero then
						not Result
					elseif j >= zero then
						not Result
					elseif i - j >= i then
						not Result
					else
						not Result implies True
					end
				elseif i - j <= i then
					if i <= zero then
						not Result
					elseif j >= zero then
						not Result
					elseif i - j >= i then
						not Result
					else
						not Result implies True
					end
				else
					not Result implies True
				end
		end

	product_overflows (i, j: STS_INTEGER): BOOLEAN
			-- Does the product `i' * `j' overflow?
		do
			Result := (i * j).value /= i.value * j.value
		ensure
			definition: Result = ((i * j).value /= i.value * j.value)
		end

	quotient_overflows (i, j: STS_INTEGER): BOOLEAN
			-- Does the quotient `i' // `j' overflow?
		do
			Result := (i // j).value /= i.value // j.value
		ensure
			definition: Result = ((i // j).value /= i.value // j.value)
		end

note
	copyright: "Copyright (c) 2012-2019, Rosivaldo Fernandes Alves"
	license: "[
		Eiffel Forum License v2
		(see http://www.eiffel.com/licensing/forum.txt)
		]"
	source: "https://launchpad.net/set-theory"

end
