﻿note
	description: "Element that models a real number"
	author: "Rosivaldo Fernandes Alves"
	date: "$Date$"
	revision: "$Revision$"

deferred class
	STS_REAL

inherit

	STS_COMPLEX
		rename
				-- Status report
			divisible as complex_divisible,
				-- Comparison
			equals as complex_equals,
			unequals as complex_unequals,
				-- Basic operations
			plus as complex_plus,
			minus as complex_minus,
			product as complex_product,
			quotient as complex_quotient
		undefine
				-- Comparison
			is_equal
		redefine
				-- Basic operations
			opposite
		select
				-- Status report
			complex_divisible,
				-- Basic operations
			complex_plus,
			complex_minus,
			complex_product,
			complex_quotient
		end

	STS_NUMBER
		undefine
				-- Comparison
			is_equal
		end

	COMPARABLE
		rename
				-- Comparison
			three_way_comparison as native_three_way_comparison
		redefine -- todo: properties, tests...
				-- Comparison
			is_less_equal,
			is_greater,
			is_greater_equal,
			native_three_way_comparison,
			min,
			max
		end

feature -- Primitive

	value: REAL
			-- Native value of this real number
		deferred
		end

feature -- Access

	sign: like integer_anchor
			-- Sign of this number as an integer value (0, -1 or 1)
		do
			Result := three_way_comparison (zero)
		ensure
			three_way: Result ≍ three_way_comparison (zero)
		end

	zero: like natural_anchor
			-- <Precursor>
		deferred
		end

	one: like natural_anchor
			-- <Precursor>
		deferred
		end

feature -- Quality

--	is_invertible: BOOLEAN
--			-- Does this number have a multiplicative `inverse' (AKA `reciprocal')?
--		do
--			Result := Current ≭ zero
--		ensure
--			definition: Result = (Current ≭ zero)
--		end

feature -- Status report

	divisible (other: STS_REAL): BOOLEAN
			-- May current number be divided by `other`?
		do
			Result := not other.value.is_nan and value.divisible (other.value)
		ensure then
			definition: Result = (not other.value.is_nan and value.divisible (other.value))
		end

	exponentiable (other: NUMERIC): BOOLEAN
			-- <Precursor>
		do
			Result := value.exponentiable (other)
		ensure then
			definition: Result = value.exponentiable (other)
		end

feature -- Comparison

	equals alias "≍" (r: STS_REAL): BOOLEAN
			-- Do this real number and `r' have the same numeric value?
		do
			Result := same_value (value, r.value)
		ensure
			definition: Result = same_value (value, r.value)
		end

	unequals alias "≭" (r: STS_REAL): BOOLEAN
			-- Does not this real number equal `r'?
		do
			Result := not (Current ≍ r)
		ensure
			definition: Result = not (Current ≍ r)
		end

	same_value (v1, v2: REAL): BOOLEAN
			-- Do `v1' and `v2' represent the same numeric value?
		do
			if v1.is_nan then
				Result := v2.is_nan
			else
				Result := v1 = v2
			end
		ensure
			class
			not_a_number: v1.is_nan implies Result = v2.is_nan
			number: not v1.is_nan implies Result = (v1 = v2)
		end

	is_less alias "<" (other: STS_REAL): BOOLEAN
			-- <Precursor>
		do
			Result := value < other.value
		ensure then
			definition: Result = (value < other.value)
		end

	is_less_equal alias "<=" (other: STS_REAL): BOOLEAN
			-- <Precursor>
		do
			Result := not (other < Current)
		end

	is_greater alias ">" (other: STS_REAL): BOOLEAN
			-- <Precursor>
		do
			Result := other < Current
		end

	is_greater_equal alias ">=" (other: STS_REAL): BOOLEAN
			-- <Precursor>
		do
			Result := not (Current < other)
		end

	native_three_way_comparison (other: STS_REAL): INTEGER
			-- <Precursor>
		do
			if Current < other then
				Result := -1
			elseif other < Current then
				Result := 1
			end
		end

	three_way_comparison (other: STS_REAL): like integer_anchor
			-- If current object equal to `other`, 0;
			-- if smaller, -1; if greater, 1
		do
			if Current < other then
				Result := - one
			elseif other < Current then
				Result := one
			else
				Result := zero
			end
		ensure
			equal_zero: (Result ≍ zero) = (Current ≍ other)
			smaller_negative: (Result ≍ - one) = (Current < other)
			greater_positive: (Result ≍ one) = (Current > other)
		end

	min (other: STS_REAL): STS_REAL
			-- <Precursor>
		do
			if Current <= other then
				Result := Current
			else
				Result := other
			end
		end

	max (other: STS_REAL): STS_REAL
			-- <Precursor>
		do
			if Current >= other then
				Result := Current
			else
				Result := other
			end
		end

feature -- Basic operations

	abs: like real_anchor
			-- Absolute value of this number
		deferred
--			if Current < zero then
--				Result := - Current
--			else
--				Result := Current
--			end
		ensure
			definition: same_value (Result.value, normalized_value (value.abs))
		end

	plus alias "+" (other: STS_REAL): like real_anchor
			-- Sum with `other` (commutative).
		deferred
		ensure then
			definition: same_value (Result.value, normalized_value (value + other.value))
		end

--	identity alias "+": like Current
--			-- <Precursor>
--		do
--			Result := Current
--		ensure then
--			definition: same_value (Result.value, value)
--		end

	minus alias "-" (other: STS_REAL): like real_anchor
			-- Result of subtracting `other`
		deferred
		ensure then
			definition: same_value (Result.value, normalized_value (value - other.value))
		end

	opposite alias "-": like real_anchor
			-- <Precursor>
		do
			Result := zero.real_minus (Current)
		ensure then
			definition: same_value (Result.value, normalized_value (- value))
		end

	product alias "*" (other: STS_REAL): like real_anchor
			-- Product by `other`
		deferred
		ensure then
			definition: same_value (Result.value, normalized_value (value * other.value))
		end

	quotient alias "/" (other: STS_REAL): like real_anchor
			-- Division by `other`
		deferred
		ensure then
			good_divisor: value.divisible (other.value) -- divisible (other)
			definition: same_value (Result.value, normalized_value (value / other.value))
		end

	reciprocal,
	inverse alias "⁻¹": like real_anchor
			-- The multiplicative inverse (AKA reciprocal) of this bumber, i.e. `one' divided by this number.
--		require
--			is_invertible: is_invertible
		do
				check
					good_divisor: one.real_divisible (Current) -- is_invertible
				end
			Result := one.real_quotient (Current)
		ensure then
			good_divisor: ({REAL} 1.0).divisible (value) -- one.divisible (Current)
			definition: same_value (Result.value, normalized_value (1/ value))
		end

	power alias "^" (other: STS_REAL): like real_anchor
			-- Current real number to the power `other'
		deferred
		ensure
			definition: same_value (Result.value, normalized_value ((value ^ other.value).truncated_to_real))
		end

feature -- Factory

	normalized_value (v: REAL): REAL
			-- `v' adapted to represent the `value' of this real number
		deferred
		ensure
			stable: same_value (normalized_value (Result), Result)
		end

feature -- Anchor

--	real_anchor: STS_REAL
--			-- Anchor for real numbers
--		do
--			Result := Current
--		end

	integer_anchor: STS_INTEGER
			-- Anchor for integer numbers
		do
			Result := sign
		end

--	natural_anchor: STS_NATURAL
--			-- Anchor for natural numbers
--		do
--			Result := zero
--		end

invariant
	is_wholly_real: imaginary ≍ zero
	normalized: same_value (value, normalized_value (value))

note
	copyright: "Copyright (c) 2012-2019, Rosivaldo Fernandes Alves"
	license: "[
		Eiffel Forum License v2
		(see http://www.eiffel.com/licensing/forum.txt)
		]"
	source: "https://launchpad.net/set-theory"
end
